/*
 *  LibRoadnav
 *  MapControl.h
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// This file contains the map control class itself.
///
//////////////////////////////////////////////////////////////////////////////

#ifndef MAPCONTROL_H
#define MAPCONTROL_H

#include <wx/wx.h>

#include <map>
#include <vector>

#include "MapMarker.h"
#include "MapTrack.h"
#include "Point.h"
#include "MapControlData.h"
#include "MapAppearanceSettings.h"
#include "Constants.h"

DECLARE_EVENT_TYPE(wxEVT_SAFE_REFRESH, -1)

#define TIMER_LEFTMOUSEHOLD		3

#define UPDATED_CENTER			1
#define UPDATED_DETAILLEVEL		2
#define UPDATED_RASTERBKGD		4
#define UPDATED_DECORATION		8
#define UPDATED_MARKERS			16
#define UPDATED_ANGLE			32
#define UPDATED_3D				64
#define UPDATED_THEME			128
#define UPDATED_DATA			256
#define UPDATED_TRACKS			512

#define TILE_CACHE_SIZE			128

typedef bool (* tClickCallback)(MapControl * pMapCtl, void * pArg, bool bRightClick, wxPoint ptClicked);

struct SClickCallback
{
	tClickCallback m_pCallback;
	void * m_pArg;
	wxPoint m_ptTopLeft;
	wxPoint m_ptBottomRight;
};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Class used to store raster map tiles for faster drawing.
///
//////////////////////////////////////////////////////////////////////////////
class MapCacheEntry
{
	public:
		MapCacheEntry(int iWidth, int iHeight, int iX, int iY, int iDetailLevel, double fAngle, bool bAntiAliased);
		~MapCacheEntry();
		
		/// Detail level (zoom) for this tile.
		int m_iDetailLevel;
		
		/// Orientation of this tile.
		double m_fAngle;
		
		/// X coordinate of this tile.
		int m_iX;
		
		/// Y coordinate of this tile.
		int m_iY;
		
		/// wxBitmap containing raster image.
		wxBitmap m_bmp;
		
		/// Device context used to blit raster image.
		wxMemoryDC * m_pdc;
};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief A wxControl that draws a map
///
//////////////////////////////////////////////////////////////////////////////
class MapControl : public wxControl
{
	public:
		//////////////////////////////////////////////////////////////////////////////
		/// @name Constructors
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		MapControl();
		MapControl(IMapControlData * pMapControlData, const MapAppearanceSettings * pMapAppearanceSettings, wxWindow * parent, wxWindowID id, wxPoint pt, wxSize sz);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Destructor/clean up
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		~MapControl();
		/// @}
	
		//////////////////////////////////////////////////////////////////////////////
		/// @name Scrolling
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void MoveCenterUp(double fScreens = 0.25);
		void MoveCenterLeft(double fScreens = 0.25);
		void MoveCenterRight(double fScreens = 0.25);
		void MoveCenterDown(double fScreens = 0.25);

		void SetCenterCoordinates(Point pt);
		Point GetCenterCoordinates();
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name 3D mode
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void Enable3DMode(bool bEnabled);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Rotation
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void SetAngle(double fAngle);
		double GetAngle();
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Detail level adjustments
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		wxInt16 GetDetailLevel();
		wxInt16 GetDetailLevelCount();
		void SetDetailLevel(wxInt16 iDetailLevel);
		void SetBestDetailLevel(double scale);
		void MoreDetail();
		void LessDetail();

		const MapAppearanceSettings* GetDetailSettings();
	
		void AutoScale();

		void SetDrawingScaleFactor(double fDrawingScaleFactor);
		double GetDrawingScaleFactor();
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Markers
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		int GetMarkerCount();
		MapMarker & GetMarker(int iIndex = 0);
		MapMarker & GetMarker(const wxString & strName);
		int SetMarker(const MapMarker & cMarker, bool fEnsureOnScreen = true);
		void ClearMarkers();
		void ClearMarker(wxString strName);
		/// @}
			
		//////////////////////////////////////////////////////////////////////////////
		/// @name Map features
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void ShowCompass(bool bShow);
		void ShowScale(bool bShow);
		void ShowAerialPhotos(bool bShow, bool bDownload);
		void SetAntialiasing(bool bEnabled);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Tracks
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void ClearTracks();
		void ClearTrack(const wxString & strName);
		void SetTrack(const MapTrack & cTrack);
		MapTrack * GetTrack(const wxString & strName);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Coordinate functions
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		wxPoint MapLongLatToScreen(const Point & pt);
		Point MapScreenToLongLat(const wxPoint & ptScreen);

		bool IsOnScreen(const Point & pt, bool bQuick = true);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Misc
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void DrawAll(wxDC * realdc, wxSize szMap, bool bDirectlyDraw = false, bool bScreen = false);
		void MapUpdated(int iUpdatedFlags);
		void AddClickCallback(const wxPoint & ptTopLeft, const wxPoint & ptBottomRight, tClickCallback pCallback, void * pArg);

		IMapControlData * GetMapControlData();

		/// @}

		int m_iControlWidth;
		int m_iControlHeight;

protected:
		void Init(IMapControlData * pMapControlData,const MapAppearanceSettings* pMapAppearanceSettings);

	private:
		//////////////////////////////////////////////////////////////////////////////
		/// @name Event handlers
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		void OnPaint(wxPaintEvent& event);
		void OnEraseBackground(wxEraseEvent& event);
		void OnLeftUp(wxMouseEvent& event);
		void OnLeftDown(wxMouseEvent& event);
		void OnRightUp(wxMouseEvent& event);
		void OnMouseMove(wxMouseEvent& event);
		void OnMouseWheel(wxMouseEvent& event);
		void OnSafeRefresh(wxCommandEvent & event);
		void OnTimerLeftMouseHold(wxTimerEvent& event);
		/// @}

		//////////////////////////////////////////////////////////////////////////////
		/// @name Drawing functions
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		MapCacheEntry * CreateMapCacheEntry(int iX, int iY, int iDetailLevel, double fAngle, Point ptStepX, Point ptStepY, bool bScreen, bool bRender);

#ifdef FILLED_IN_WATER_SUPPORT
		void DrawMapWaterBoundaries(wxDC * dc, bool bScreen);
		void DrawMapWater(wxDC * dc, bool bScreen);
#endif

		void DrawMapTerraserver(wxDC * dc, bool bScreen);
		void DrawMapGeneral(wxDC * dc, bool bScreen);
		void DrawMapCompass(wxDC * dc, bool bScreen);
		void DrawMapKey(wxDC * dc, bool bScreen);
		int DrawMap(wxDC * dc, bool bScreen);
		void DrawBorder(wxDC * dc, bool bScreen);
		void DrawTracks(wxDC * dc, bool bScreen);
		/// @}
	
		void PerformAutoscaling();
		
		void UpdateTopLeftAndBottomRight();
		void ClearMapCache();
		
		void UpdateCrossHair();

		/// Current detail level
		wxInt16 m_iDetailLevel;

		/// Center long/lat coordinates of the map
		Point m_ptCenter;

		/// Angle in degrees to rotate map (positive = ccw rotation)
		double m_fMapAngle;

		/// Vector of markers
		std::vector<MapMarker> m_vsMarkers;
		
		/// Drawing scale factor. Setting this greater than 1 will cause
		/// everything to appear bigger. It should be used to compensate for
		/// devices with a high DPI. It will not change the detail level.
		double m_fDrawingScaleFactor;

		/// Bitmap containing the last map drawn
		
		/// Bitmap containing a picture of the control the last time it was
		/// drawn.
		wxBitmap m_bmpOldEverything;
		wxMemoryDC * m_pdcOldEverything;
		
		/// Long/lat coordinates of top left corner of the control
		Point m_ptTopLeft;
		
		/// Long/lat coordinates of the bottom right corner of the control
		Point m_ptBottomRight;

		Point m_ptViewTopLeft;
		Point m_ptViewBottomRight;

#ifdef FILLED_IN_WATER_SUPPORT
		/// Boolean - should water be filled in? Very slow on certain platforms.
		bool m_bFillInWater;
#endif

		/// Background brush
		wxBrush m_brBackground;
		
		/// Boolean - should the compass be visible?
		bool m_bShowCompass;
		
		/// Boolean - should the scale be visible?
		bool m_bShowScale;

		/// Boolean - should the aerial photos be visible?
		bool m_bShowAerialPhoto;

		/// Boolean - should the aerial photos be downloaded if not cached?
		bool m_bDownloadAerialPhoto;

		/// Boolean - should lines be enabled? true = enables lines specified by
		/// detail settings. false = disable all lines.
		bool m_bShowLines;
		
		/// Boolean - should labels be enabled? true = enable labels specified by
		/// detail settings. false = disable all labels.
		bool m_bShowLabels;

		double m_f3DHorizon;
		
		/// Boolean - is 3D mode on?
		bool m_b3D;

		/// Boolean - should the map be automatically scaled next time it's
		/// drawn? This will zoom out until all markers are visible.
		bool m_bAutoScale;
		
		/// Maximum distance a label may be drawn from the line
		/// it's referring to. 0 = always draw label next to line.
		int m_iMaxLineToLabelLength;

		/// Vector of tracks to draw
		std::vector<MapTrack> m_vcTracks;

		/// Cursor to use for this control
		wxCursor m_curCrossHair;

		/// true = use antialiasing, false = no antialiasing
		bool m_bAntiAlias;

		/// Map data associated with this control.
		IMapControlData * m_pMapControlData;
		
		/// Appearance settings associated with this control.
		const MapAppearanceSettings* m_pMapAppearanceSettings;

		/// Time data was last updated on disk. If IMapControlData reports a newer
		/// value, then refresh what's on the screen.
		wxDateTime m_tDataLastUpdated;
		
		/// Mutex locked whenever map settings are being changed, or when
		/// the control is being painted.
		wxMutex m_mxSettings;

		//////////////////////////////////////////////////////////////////////////////
		/// @name Left click handling
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		/// Indicates if the mouse is currentingly dragging the map.
		bool m_bLeftMouseMovingMap;
		
		/// Indicates if the left mouse click was already handled by
		/// the right click handler or the dragging handler.
		bool m_bLeftMouseHandled;
		
		/// Coordinates of where the left button was pressed.
		Point m_ptCenterWhenLeftMouseDown;
		
		/// Timer that goes off 2 sec after the left button was pressed.
		wxTimer * m_pTimerLeftMouseHold;
		
		/// Last left mouse down event.
		wxMouseEvent m_evLeftMouseHold;
		/// @}
		
		/// Callbacks associated with this control
		std::vector<SClickCallback> m_vClickCallbacks;
		
		bool m_bSendMapUpdatedNotification;
		
		//////////////////////////////////////////////////////////////////////////////
		/// @name Tile cache variables
		//////////////////////////////////////////////////////////////////////////////
		/// @{
		
		/// Array of MapCachEntry containing the cache.
		MapCacheEntry * m_parMapCache[TILE_CACHE_SIZE];
		
		/// Points to the next map cache entry that will be purged.
		int m_iMapCacheOldest;
		
		/// Size of each tile in pixels.
		wxSize m_szCacheBitmaps;
		/// @}
		
		bool m_bLastDrawWasComplete;

	private:
		void SafeRefresh();

		DECLARE_EVENT_TABLE()
		DECLARE_DYNAMIC_CLASS(MapControl);
};

#endif
