/*
 *  LibRoadnav
 *  MapControlDataEntry.h
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains IMapControlDataEntry - interface to access individual map
/// records.
///
//////////////////////////////////////////////////////////////////////////////

#ifndef MAPCONTROLDATAENTRY_H
#define MAPCONTROLDATAENTRY_H

#include "Point.h"
#include "UniqueID.h"
#include "Serialization.h"
#include "Constants.h"

typedef UniqueID tRecordID;

enum ERecordType
{
	RecordTypeDefault = 0,
	RecordTypeOneWaySmallRoad,
	RecordTypeOneWayLargeRoad,
	RecordTypeOneWayHighway,
	RecordTypeTwoWaySmallRoad,
	RecordTypeTwoWayLargeRoad,
	RecordTypeTwoWayHighway,
	RecordTypeRailroad,
	RecordTypeWater,
	RecordTypeInvisibleLandBoundary,
	RecordTypeInvisibleWaterBoundary,
	RecordTypeInvisibleMiscBoundary,
	RecordTypeLandmark,
	RecordTypePhysicalFeature,
	RecordTypeTownOrCity,
	RecordTypeCountyBoundary,
	RecordTypeStateBoundary
};

enum ECountryCode
{
	CountryCodeUnknown = 0,
	CountryCodeUnitedStates = 1,
	CountryCodeCanada = 2
};

#define	FIRST_RECORD_TYPE		RecordTypeDefault
#define LAST_RECORD_TYPE		RecordTypeStateBoundary

#define FIRST_RECORD_TYPE_ROAD	RecordTypeOneWaySmallRoad
#define LAST_RECORD_TYPE_ROAD	RecordTypeTwoWayHighway

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Represents a range of street addresses, associated with a
/// IMapControlDataEntry object
///
//////////////////////////////////////////////////////////////////////////////
struct SAddressRangeZip
{
	/// Starting street address number
	wxUint16 iFromAddr;

	/// Ending street address number
	wxUint16 iToAddr;

	/// Zip code for street
	wxUint32 iZip;

	/// true = refers to left hand side of street, false = right hand side
	bool bOnLeft;
};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Interface used by MapControl to access individual map records
///
//////////////////////////////////////////////////////////////////////////////
class IMapControlDataEntry : public Serialization
{
	public:
		virtual ~IMapControlDataEntry();

		virtual bool IsRoad() const = 0;
		virtual double CostFactor() const = 0;
		virtual wxString SmartName() const = 0;
		virtual bool IsOneWay() const = 0;
		virtual double TimeFactor() const = 0;
		virtual double RecordDistance() const = 0;
		virtual bool IsSameRoad(IMapControlDataEntry * psRec) const = 0;
		virtual wxString NameAndType(int iEntry = 0, bool bLongType = false) const = 0;
		virtual double Distance(const Point & ptSearch, int * piBestSegment = NULL) const = 0;
		virtual bool ContainsPoint(const Point & ptCheck) const = 0;
		virtual bool IntersectsRect(const Point & ptTopLeft, const Point & ptBottomRight) const = 0;
		virtual bool IsRegion() const = 0;
		virtual wxInt32 CountyCode() const = 0;
		virtual wxString Dump() const = 0;

		virtual void Clear() = 0;
	
	public:
		/// Unique ID number assigned to each record
		tRecordID idRecord;


		/// Array of feature names
		char ** pszFeatureNames;

		/// Array of feature types
		char ** pszFeatureTypes;

		/// Number of feature names/types
		wxUint16 nFeatureNames;

		/// Most common name for a feature		
		char * pszSmartName;

		/// What type of feature this is
		ERecordType eType;

		
		/// Array of street number ranges and zip codes
		struct SAddressRangeZip * psAddressZipRanges;

		/// Number of entries in psAddressZipRanges
		wxUint16 nAddressZipRanges;
		

		/// Array of coordinates that constitute this feature
		Point * psCoordinates;

		/// Number of coordinates in psCoordinates
		wxUint16 nCoordinates;
		
		ECountryCode eCountryCode;
	
#ifdef FILLED_IN_WATER_SUPPORT
		bool bWaterL;
		bool bWaterR;
	
		Point ptWaterL;
		Point ptWaterR; 
#endif
};

#endif
