/*
 *  LibRoadnav
 *  Address.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains the Address class, a class that contains a US address
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <wx/wx.h>

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include "Address.h"
#include "MapRepresentations.h"
#include "MapState.h"
#include "MapAbbreviations.h"

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - zero out everything
///
//////////////////////////////////////////////////////////////////////////////
Address::Address()
{
	m_iStreetNumber = 0;
	m_iZipCode = 0;
	m_ptCoordinates.Set(0, 0);
	m_iCountyCode = -1;
	m_idRecord.m_iLow = 0;
	m_idRecord.m_iHigh = 0;
	m_eRecordType = RecordTypeDefault;
	m_bValid = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Format this address nicely into a string.
///
/// \param bShowCoordinates  - true = append coordinates at end, false = just
///                            include the street, city, state, and zip.
///
/// \param bOneLine          - true = separate address into lines, false = 
///                            print everything onto one line.
///
/// \param bStreetOnly       - true = return only the street address,
///                            false = include city, state, and zip.
///
/// \return Address formatted as requested
///
//////////////////////////////////////////////////////////////////////////////
wxString Address::FormatAddress(bool bShowCoordinates, bool bOneLine, bool bStreetOnly)
{
	wxString rtn;
	wxString strType;
	wxString strLineBreak = wxT("\n");
	
	if (!m_bValid)
		return wxT("");

	if (bOneLine)
		strLineBreak = wxT(", ");

	if (this)
	{
		wxString strNumber;
		wxString strType;

		////////////////////////////////////////////////////////////////
		// convert the street number into a string, or blank for 0
		////////////////////////////////////////////////////////////////
		strNumber = m_iStreetNumber ? wxString::Format(wxT("%d "), m_iStreetNumber) : wxT("");

		////////////////////////////////////////////////////////////////
		// expand the type abbreviation
		////////////////////////////////////////////////////////////////
		strType = wxT(" ") + GetFullName(m_strStreetType);

		////////////////////////////////////////////////////////////////
		// if the abbreviation can't be expanded, then just use the abbreviation
		////////////////////////////////////////////////////////////////
		if (strType == wxT(" "))
			strType = (wxString) wxT(" ") + m_strStreetType;
		
		////////////////////////////////////////////////////////////////
		// still just contains a space? turn it into an empty string then
		////////////////////////////////////////////////////////////////
		if (strType == wxT(" "))
			strType = wxT("");

		////////////////////////////////////////////////////////////////
		// combine street number, name, and type
		////////////////////////////////////////////////////////////////
		rtn = strNumber + m_strStreetName + strType;
		
		if (bStreetOnly)
			return rtn;
		
		if (rtn != wxT(""))
			rtn += strLineBreak;
	
		////////////////////////////////////////////////////////////////
		// city/state/zip line
		////////////////////////////////////////////////////////////////
		if (m_strCityName != wxT(""))
			rtn += m_strCityName + wxT(", ") + StateAbbreviationByName(m_strStateName);
		else if (m_strStateName != wxT(""))
			rtn += StateAbbreviationByName(m_strStateName);
		
		if (m_iZipCode)
			rtn += wxString::Format(wxT(" %05d"), m_iZipCode);

		////////////////////////////////////////////////////////////////
		// coordinates, if bShowCoordinates
		////////////////////////////////////////////////////////////////
		if (bShowCoordinates)
		{
			rtn += strLineBreak + wxT("(") + m_ptCoordinates.FormatPoint() + wxT(")");
		}
	}
	else
	{
		rtn = wxT("");
	}
	return rtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert this address into a string suitable for storage in the
/// registry
///
/// \return Address in form suitable for storage in registry
///
//////////////////////////////////////////////////////////////////////////////
wxString Address::ToString()
{
	return wxString::Format(wxT("%d"), m_iStreetNumber) + wxT("\t") +
			m_strStreetName + wxT("\t") +
			m_strStreetType + wxT("\t") +
			m_strCityName + wxT("\t") +
			m_strStateName + wxT("\t") +
			wxString::Format(wxT("%d"), m_iZipCode) + wxT("\t") +
			m_ptCoordinates.ToString() + wxT("\t") +
			wxString::Format(wxT("%d"), m_iCountyCode) + wxT("\t") + 
			wxString::Format(wxT("%d"), m_bValid) + wxT("\t") + 
			m_idRecord.ToString() + wxT("\t") + 
			wxString::Format(wxT("%d"), (int) m_eRecordType);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert a string generated by Address::ToString() back into this
/// Address object
///
/// \param strValue			String previously generated by Address::ToString()
///
//////////////////////////////////////////////////////////////////////////////
void Address::FromString(wxString strValue)
{
	long lTmp;

	strValue.BeforeFirst('\t').ToLong(&lTmp);
	m_iStreetNumber = lTmp;
	strValue = strValue.AfterFirst('\t');

	m_strStreetName = strValue.BeforeFirst('\t');
	strValue = strValue.AfterFirst('\t');

	m_strStreetType = strValue.BeforeFirst('\t');
	strValue = strValue.AfterFirst('\t');

	m_strCityName = strValue.BeforeFirst('\t');
	strValue = strValue.AfterFirst('\t');

	m_strStateName = strValue.BeforeFirst('\t');
	strValue = strValue.AfterFirst('\t');

	strValue.BeforeFirst('\t').ToLong(&lTmp);
	m_iZipCode = lTmp;
	strValue = strValue.AfterFirst('\t');

	m_ptCoordinates.FromString(strValue.BeforeFirst('\t'));
	strValue = strValue.AfterFirst('\t');

	strValue.BeforeFirst('\t').ToLong(&lTmp);
	m_iCountyCode = lTmp;
	strValue = strValue.AfterFirst('\t');

	strValue.BeforeFirst('\t').ToLong(&lTmp);
	m_bValid = lTmp ? true : false;
	strValue = strValue.AfterFirst('\t');

	m_idRecord.FromString(strValue.BeforeFirst('\t'));
	strValue = strValue.AfterFirst('\t');

	strValue.BeforeFirst('\t').ToLong(&lTmp);
	m_eRecordType = (ERecordType) lTmp;
	strValue = strValue.AfterFirst('\t');
}
