/*
 *  LibRoadnav
 *  Angle.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <sstream>
#include <iomanip>
#include <wx/tokenzr.h>

#include "Angle.h"
#include "UnitConversion.h"
#include "MapSupport.h"
#include "Constants.h"


bool operator == (const Angle& a,const Angle& b)
{
   return a.AsDouble() == b.AsDouble();
}

bool operator != (const Angle& a,const Angle& b)
{
   return !operator==(a,b);
}

bool operator  > (const Angle& a,const Angle& b)
{
   return (a.AsDouble() > b.AsDouble());
}

bool operator  < (const Angle& a,const Angle& b)
{
   return (a.AsDouble() < b.AsDouble());
}

bool operator >= (const Angle& a,const Angle& b)
{
   return (a.AsDouble() > b.AsDouble());
}

bool operator <= (const Angle& a,const Angle& b)
{
   return (a.AsDouble() < b.AsDouble());
}

////////////////////////// PUBLIC     ///////////////////////////////////////

//======================== LIFECYCLE  =======================================
Angle::Angle(double angle)
{
   m_Angle = angle;
}

Angle::Angle(short deg, unsigned short min, double sec)
{
   m_Angle = DMSToDecimal( deg, min, sec );
}

Angle::Angle(unsigned short deg, unsigned short min, double sec, char dir)
{
   m_Angle = (toupper(dir) == 'L' ? -1 : 1) * DMSToDecimal(deg,min,sec);
}

Angle::Angle(const Angle& rOther)
{
   MakeCopy(rOther);
}

Angle::~Angle()
{
}

//======================== OPERATORS  =======================================
Angle& Angle::operator= (const Angle& rOther)
{
   if( this != &rOther )
   {
      MakeAssignment(rOther);
   }

   return *this;
}

Angle& Angle::operator = (double angle)
{
   m_Angle = angle;
   return *this;
}

Angle Angle::operator+(const Angle& rOther) const
{
   return Angle( m_Angle + rOther.AsDouble() );
}

Angle Angle::operator-(const Angle& rOther) const
{
   return Angle( m_Angle - rOther.AsDouble() );
}

Angle Angle::operator*(double scalar) const
{
   return Angle( m_Angle * scalar );
}

Angle Angle::operator/(double scalar) const
{
   return Angle( m_Angle / scalar );
}

Angle& Angle::operator+=(const Angle& rOther)
{
   m_Angle += rOther.AsDouble();
   return *this;
}

Angle& Angle::operator-=(const Angle& rOther)
{
   m_Angle -= rOther.AsDouble();
   return *this;
}

Angle& Angle::operator*=(double scalar)
{
   m_Angle *= scalar;
   return *this;
}

Angle& Angle::operator/=(double scalar)
{
   m_Angle /= scalar;
   return *this;
}

Angle Angle::operator-() const
{
   return Angle( -m_Angle );
}

//======================== OPERATIONS =======================================
void Angle::Normalize()
{
   do
   {
      if ( m_Angle < 0 )
         m_Angle += 360;

      if ( m_Angle >= 360 )
         m_Angle -= 360;

   } while ( m_Angle < 0 || 360 < m_Angle );
}

//======================== ACCESS     =======================================
void Angle::SetAngle(double angle)
{
   m_Angle = angle;
}

void Angle::SetAngle(short deg,unsigned short min,double sec)
{
   m_Angle = DMSToDecimal( deg, min, sec );
}

double Angle::AsDouble() const
{
   return m_Angle;
}

//////////////////////////////////////////////////////////////////////////////
/// \brief Convert the angle to a formatted string
///
/// bSigned			- if true, format negative numbers as positive
///
/// bFull			- true = always show minutes and seconds
///					  false = only show minutes and seconds if non-zero
///
/// bSymbols		- true = show degree, minute, and seconds symbols
///					  false = just the numbers
//////////////////////////////////////////////////////////////////////////////
wxString Angle::AsString(bool bSigned, bool bFull, bool bSymbols) const
{
   short deg;
   unsigned short min;
   double sec;
   char dir;

   dir = (m_Angle < 0) ? 'L' : 'R';
   DecimalToDMS( m_Angle, deg, min, sec );

   std::stringstream s;
   s << std::setw(2) << std::setfill('0') << (bSigned ? deg : abs(deg));

   if ( bSymbols )
      s << "";

   if ( bFull || (min != 0 || sec != 0 ) )
   {
      s << " " << std::setw(2) << std::setfill('0') << min << (bSymbols ? "' " : " ")
        << std::setw(4) << std::setfill('0') << std::showpoint << std::fixed << std::setprecision(1) << sec;
      if ( bSymbols ) 
         s << "\"";
   }

   if ( !bSigned )
     s << " " << dir;

   return wxString(s.str().c_str(), *wxConvCurrent);
}

//////////////////////////////////////////////////////////////////////////////
/// \brief [Static] Translate a formatted string back into pCAngle 
//////////////////////////////////////////////////////////////////////////////
bool Angle::AngleFromString(const char* str,Angle* pCAngle)
{
   if ( pCAngle == 0 )
      return false;

   // parse into space-delimited tokens
   wxStringTokenizer tokenizer(wxString(str, *wxConvCurrent), wxT(" \t"));

   int nParts = tokenizer.CountTokens();
   if (nParts < 1 || 4 < nParts)
      return false;  // string is empty or has too many parts

   if ( nParts == 1 )
   {
      // String is of the format [+|-]ddd.ddd[R|L]
      // check for R or L
      wxString stmp = tokenizer.GetNextToken();
      int endloc = stmp.Length()-1;
      char end = stmp.GetChar(endloc);
      double rlfactor = 1.;
      if (end=='R' || end=='r' || end=='L' || end=='l')
      {
         // can't begin with a - or + sign
         if ( stmp.GetChar(0) == '+' || stmp.GetChar(0) == '-' )
            return false;

         if (end=='L' || end=='l')
            rlfactor = -1;

         stmp.Remove(endloc);
      }

      double angle;
      if ( !stmp.ToDouble(&angle) )
         return false;

      angle *=rlfactor;

      *pCAngle = angle;
   }
   else if ( nParts == 2 )
   {
      // String is of the format ddd.ddd [L|R]

      wxString stmp = tokenizer.GetNextToken();

      // can't begin with +|- sign
      if ( stmp.GetChar(0) == '+' || stmp.GetChar(0) == '-' )
         return false;

      double angle;
      if ( !stmp.ToDouble(&angle) )
         return false;

      stmp = tokenizer.GetNextToken();
      if (stmp==wxT("L") || stmp==wxT("l") || stmp==wxT("R") || stmp==wxT("r"))
      {
         if (stmp==wxT("L") || stmp==wxT("l"))
            angle *= -1;
      }
      else 
         return false;

      *pCAngle = angle;
   }
   else if ( nParts == 3 || nParts == 4)
   {
      // String is of the format [+|-]ddd mm ss.s or ddd mm ss.s [L|R]
	   short deg;
	   unsigned short min;
	   double sec;

      wxString stmp = tokenizer.GetNextToken();

      // deal with [L|R]
      double rlfactor = 1;
      if (nParts==4)
      {
         // can't begin with +|- sign
         if ( stmp.GetChar(0) == '+' || stmp.GetChar(0) == '-' )
            return false;

         std::string rl = static_cast<const char *>(tokenizer.GetNextToken().mb_str(*wxConvCurrent));
         if (rl=="L" || rl=="l" || rl=="R" || rl=="r")
         {
            if (rl=="L" || rl=="l")
               rlfactor=-1;
         }
         else
            return false;
      }

      // ddd
      long ltmp;
      if ( !stmp.ToLong(&ltmp) )
         return false;

      deg = ltmp;

      // mm
      stmp = tokenizer.GetNextToken();
      if ( !stmp.ToLong(&ltmp) )
         return false;

      if (ltmp<0 || 60<ltmp)
         return false;

      min = ltmp;

      // sss.s
      stmp = tokenizer.GetNextToken();
      if ( !stmp.ToDouble(&sec) )
         return false;

      if (sec < 0 || 60 < sec)
         return false;

      double angle = DMSToDecimal( deg, min, sec ) * rlfactor;
      *pCAngle = angle;
   }

   return true;
}

//======================== INQUIRY    =======================================

////////////////////////// PROTECTED  ///////////////////////////////////////

//======================== LIFECYCLE  =======================================
//======================== OPERATORS  =======================================
//======================== OPERATIONS =======================================
void Angle::MakeCopy(const Angle& rOther)
{
   m_Angle = rOther.AsDouble();
}

void Angle::MakeAssignment(const Angle& rOther)
{
   MakeCopy( rOther );
}
