/*
 *  LibRoadnav
 *  Debug.cpp  
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Debugging functions
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>
#include <wx/datetime.h>
#include <wx/filename.h>
#include <wx/utils.h>

#include "Debug.h"

#ifdef LIBROADNAV_DEBUG

/// Controls if debug messages are logged to a file.
static bool g_bDebugging = false;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Records a log entry. Shouldn't be called directly.
///
/// \param strType				Section of code generating log message. Will
///								be used in the future for filtering.
///
/// \param strFormat			printf style format for log message
///
/// \param pszFilename			Filename of code generating log message.
///
/// \param iLineNumber			Line number of code generating log message.
///
/// Use the LibRoadnavDebug0, LibRoadnavDebug1, etc. macros to log messages
/// rather than calling this function directly.
///
/// wxLogDebug is always called if LIBROADNAV_DEBUG is defined, and
/// the log file is written if g_bDebugging is true.
///
//////////////////////////////////////////////////////////////////////////////
void LibRoadnavDebug(wxString strType, wxString strFormat, char * pszFilename, int iLineNumber, ...)
{
	va_list valArgs;
	wxString strMsg;
	wxString strPrefix;
	wxDateTime dtNow;
	FILE * hFile;
	int iFilenameOffset;
	int iPos;
	
	dtNow = wxDateTime::Now();
	
	va_start(valArgs, iLineNumber);
	
	strMsg.PrintfV(strFormat, valArgs);
	
	iFilenameOffset = 0;
	for (iPos = 0; pszFilename[iPos]; iPos++)
		if (pszFilename[iPos] == '/' || pszFilename[iPos] == '\\')
			iFilenameOffset = iPos + 1;
	
	strPrefix = wxString::Format(wxT("%30hs:%-5d %04d-%02d-%02d %02d:%02d:%02d.%03d - "),
				pszFilename + iFilenameOffset,
				iLineNumber,
				dtNow.GetYear(),
				dtNow.GetMonth() + 1,
				dtNow.GetDay(),
				dtNow.GetHour(),
				dtNow.GetMinute(),
				dtNow.GetSecond(),
				dtNow.GetMillisecond());

	wxLogDebug(wxT("%s%s"), strPrefix.c_str(), strMsg.c_str());

	if (!g_bDebugging)
		return;
				
	hFile = fopen("roadnav-debug.log", "a+b");
	
	if (hFile)
	{
		fputs((strPrefix + strMsg + wxT("\r\n")).mb_str(wxConvUTF8), hFile);
		fclose(hFile);
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Activates logging debug messages to a file.
///
//////////////////////////////////////////////////////////////////////////////
void LibRoadnavDebugEnable()
{
	int iScrW;
	int iScrH;
	int iScrD;
	
	g_bDebugging = true;
	
	LibRoadnavDebug0(wxT("Debug"), wxT("Debugging enabled for Roadnav ") wxT(VERSION) wxT(" compiled on ") wxT(__DATE__) wxT(" ") wxT(__TIME__) );
	LibRoadnavDebug1(wxT("Debug"), wxT("Running on %s"), wxGetOsDescription().c_str());

#if defined(__WINDOWS__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for Windows OS"));
#endif
#if defined(__WXCOCOA__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for Cocoa"));
#endif
#if defined(__WXWINCE__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for WinCE"));
#endif
#if defined(__WXGTK12__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for GTK+ 1.2"));
#endif
#if defined(__WXGTK20__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for GTK+ 2.0"));
#endif
#if defined(__WXMAC_CLASSIC__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for MacOS Classic"));
#endif
#if defined(__WXMAC_CARBON__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for MacOS Carbon"));
#endif
#if defined(__WXMAC_OSX__)
	LibRoadnavDebug0(wxT("Debug"), wxT("Compiled for MacOS OS X"));
#endif
	LibRoadnavDebug1(wxT("Debug"), wxT("Compiled against %s"), wxVERSION_STRING);

	wxDisplaySize(&iScrW, &iScrH);
	iScrD = wxDisplayDepth();
	LibRoadnavDebug3(wxT("Debug"), wxT("Display: %d x %d x %d"), iScrW, iScrH, iScrD);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Deactivates (default) logging debug messages to a file.
///
//////////////////////////////////////////////////////////////////////////////
void LibRoadnavDebugDisable()
{
	g_bDebugging = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the filename of the debug log
///
//////////////////////////////////////////////////////////////////////////////
wxString LibRoadnavDebugGetFilename()
{
	wxFileName fn;
	fn.SetFullName(wxT("roadnav-debug.log"));
	fn.Normalize();
	return fn.GetFullPath();
}

#endif // #if LIBROADNAV_DEBUG
