/*
 *  LibRoadnav
 *  Direction.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/tokenzr.h>
#include <sstream>
#include <iomanip>

#include "Direction.h"
#include "Angle.h"
#include "Constants.h"
#include "MapSupport.h"
#include "UnitConversion.h"

const Direction operator+(const Direction& dir,const Angle& angle)
{
   Direction d(dir);
   d += angle;
   return d;
}

const Direction operator-(const Direction& dir,const Angle& angle)
{
   Direction d(dir);
   d -= angle;
   return d;
}

const Angle operator-(const Direction& dir1,const Direction& dir2)
{
   Angle angle( dir1.AsDouble() - dir2.AsDouble() );
   angle.Normalize();
   return angle;
}


////////////////////////// PUBLIC     ///////////////////////////////////////

//======================== LIFECYCLE  =======================================
Direction::Direction()
{
   m_Direction = 0;
}

Direction::Direction(char n,unsigned short deg,unsigned short min,double sec,char e)
{
   m_Direction = Direction::BearingToDBearing( n, deg, min, sec, e );
}

Direction::Direction(short deg,unsigned short min,double sec)
{
   m_Direction = Direction::AzimuthToDAzimuth( deg, min, sec );
}

Direction::Direction(const Angle& rAzimuth)
{
   m_Direction = rAzimuth.AsDouble();
}

Direction::Direction(double direction)
{
   m_Direction = direction;
   Normalize();
}

Direction::Direction(const Direction& rOther)
{
   MakeCopy(rOther);
}

Direction::~Direction()
{
}

//======================== OPERATORS  =======================================
Direction& Direction::operator= (const Direction& rOther)
{
   if( this != &rOther )
   {
      MakeAssignment(rOther);
   }

   return *this;
}

Direction& Direction::operator = (double direction)
{
   m_Direction = direction;
   Normalize();
   return *this;
}

bool Direction::operator == (const Direction& rOther) const
{
   return (m_Direction == rOther.AsDouble() );
}

bool Direction::operator != (const Direction& rOther) const
{
   return !(*this == rOther);
}

bool Direction::operator < (const Direction& rOther) const
{
   return m_Direction < rOther.AsDouble();
}

Direction& Direction::operator+=(const Angle& angle)
{
   return Increment( angle );
}

Direction& Direction::operator-=(const Angle& angle)
{
   return Increment( -angle );
}


Direction& Direction::Increment(const Angle& angle)
{
   m_Direction += angle.AsDouble();
   Normalize();
   return *this;
}

Direction Direction::IncrementBy(const Angle& angle) const
{
   Direction dir( *this );
   dir += angle;
   return dir;
}

//======================== OPERATIONS =======================================
//======================== ACCESS     =======================================
double Direction::AsDouble() const
{
   return m_Direction;
}

wxString Direction::AsBearingString(bool bFull,bool bSymbols) const
{
   char n,e;
   unsigned short deg;
   unsigned short min;
   double sec;

   std::stringstream s;
   Direction::DBearingToBearing( m_Direction, n, deg, min, sec, e );
   short prec = 1;
   s << std::setw(1) << n << " " 
     << std::setw(2) << std::setfill('0') << deg;

#if defined _MSC_VER
   if ( bSymbols )
      s << "";
#endif

   if ( bFull || ( min != 0 || sec != 0 ) )
   {
      s << " " << std::setw(2) << std::setfill('0') << min;
      s << ( bSymbols ? "' " : " ");
      s << std::setw(4) << std::setfill('0') << std::showpoint << std::fixed << std::setprecision(prec) << sec;
      s << ( bSymbols ? "\"" : "");
   }

   s << " " << std::setw(1) << e << std::ends;
   return wxString(s.str().c_str(), *wxConvCurrent);
}

wxString Direction::AsAzimuthString(bool bSymbols) const
{
   wxString str;
   AsDmsString( str, bSymbols );
   return str;
}

//======================== INQUIRY    =======================================

////////////////////////// PROTECTED  ///////////////////////////////////////

//======================== LIFECYCLE  =======================================
//======================== OPERATORS  =======================================
//======================== OPERATIONS =======================================
void Direction::MakeCopy(const Direction& rOther)
{
   m_Direction = rOther.AsDouble();
}

void Direction::MakeAssignment(const Direction& rOther)
{
   MakeCopy( rOther );
}

//======================== ACCESS     =======================================
//======================== INQUIRY    =======================================

////////////////////////// PRIVATE    ///////////////////////////////////////

//======================== LIFECYCLE  =======================================
//======================== OPERATORS  =======================================
//======================== OPERATIONS =======================================
void Direction::Normalize()
{
   do
   {
      if ( m_Direction < 0 )
         m_Direction += 360.;

      if ( m_Direction > 360. )
         m_Direction -= 360.;
   } while ( m_Direction < 0 || 360. < m_Direction );
}

void Direction::DBearingToBearing(double dir,char& n,unsigned short& deg,unsigned short& min,double& sec,char& e)
{
   while ( dir < 0 )
      dir += 360.;

   while ( 360. < dir )
      dir -= 360.;

   n = ( (0 <= dir && dir <= 90.) || (270. <= dir && dir <= 360.) ) ? 'N' : 'S';
   e = ( 0 <= dir && dir <= 180. ) ? 'E' : 'W';

   if ( 90 <= dir && dir <= 180 )
      dir = 180 - dir;
   else if ( 90 <= dir && dir < 270 )
      dir -= 180;
   else if ( 270 <= dir && dir <= 360. )
      dir = 360. - dir;

   short temp_deg;
   DecimalToDMS( dir, temp_deg, min, sec );
   deg = temp_deg;
}

bool Direction::BearingFromString(const char* str,Direction* pDir)
{
   if ( pDir == 0 )
      return false;

   char cY;
   char cX;
   unsigned short deg=0;
   unsigned short min=0;
   double sec=0;

   wxString strDir( wxString(str, *wxConvCurrent) );

   // Trim the whitespace off both ends
   std::string::size_type last_leading_space   = strDir.find_first_not_of(wxT(" "));
   if ( last_leading_space == std::string::npos )
      return false; // string is all spaces
   std::string::size_type cChar;
   cChar = last_leading_space;  // number of characters to remove
   strDir.erase( 0, cChar );

   std::string::size_type first_trailing_space = strDir.find_last_not_of(wxT(" "));
   cChar = strDir.length() - first_trailing_space - 1;
   strDir.erase( first_trailing_space+1, cChar );

   if ( strDir.size() == 0 )
      return false; // string is empty

   // Lets start breaking apart the string
   cY = toupper( strDir.GetChar(0) );
   if ( cY != 'N' && cY != 'S' )
      return false;

   cX = toupper( strDir[strDir.size()-1] );
   if ( cX != 'E' && cX != 'W' )
      return false;

   strDir.erase( strDir.size()-1, 1 );
   strDir.erase( 0, 1 );

   wxStringTokenizer tokenizer(strDir, wxT(" \t"));

   int size = tokenizer.CountTokens();
   if (size < 1 || 3 < size)
      return false;

   // Get the degrees
   wxString stmp = tokenizer.GetNextToken();
   long ltmp;
   if ( !stmp.ToLong(&ltmp) )
      return false;

   if (ltmp < 0 || 90 < ltmp) 
      return false;

   deg = ltmp;

   // minutes
   if (1 < size)
   {
      stmp = tokenizer.GetNextToken();
      if (!stmp.ToLong(&ltmp))
         return false;

      if (ltmp < 0 || 60 < ltmp) 
         return false;

      min = ltmp;
   }

   // seconds
   if (size==3)
   {
      stmp = tokenizer.GetNextToken();
      if ( !stmp.ToDouble(&sec) )
         return false;

      if (sec < 0.0 || 60.0 < sec) 
         return false;
   }

   // check that total is not more than 90 degrees
   double ang = deg + (double)min/60.0 + sec/3600.0;
   if (90.0 < ang)
      return false;

   *pDir = Direction( cY, deg, min, sec, cX );
   return true;
}

double Direction::BearingToDBearing(char n,unsigned short deg,unsigned short min,double sec,char e)
{
   double angle = 0;
   double sign = 1;
	n = toupper(n);
	e = toupper(e);

	// Convert component parts to a decimal angle (radians)
   if( n == 'N' && e == 'E')
   {
      angle = 0.;
      sign  = 1.;
   }
   else if (n == 'S' && e == 'W')
   {
      angle = 180;
      sign  = 1.;
   }
   else if (n == 'S' && e == 'E')
   {
      angle = 180;
      sign  = -1.;
   }
   else if (n == 'N' && e == 'W')
   {
      angle = 360.;
      sign  = -1;
   }
   else
      wxASSERT(false); // should never get here

   angle += sign * DMSToDecimal(deg,min,sec);

   if ( angle == 360. )
      angle = 0;

   return angle;
}

double Direction::AzimuthToDAzimuth(short deg,unsigned short min,double sec)
{
   return DMSToDecimal(deg, min, sec);;
}

void Direction::DAzimuthToAzimuth(double angle,short& deg,unsigned short& min,double& sec)
{
   DecimalToDMS(angle, deg, min, sec);
}

void Direction::AsDmsString(wxString& str,bool bSymbols) const
{
   short deg;
   unsigned short min;
   double sec;
   std::stringstream s;
   DecimalToDMS( m_Direction, deg, min, sec );
   s << std::setw(2) << deg;
   s << (bSymbols ? "" : " ");
   s << std::setw(2) << std::setfill('0') << min;
   s << (bSymbols ? "' " : " ");
   s << std::setw(4) << std::showpoint << std::setprecision(1) << sec;
   if ( bSymbols )
      s << "\""; 
   s << std::ends;
   str = wxString(s.str().c_str(), *wxConvCurrent);
}
