/*
 *  LibRoadnav
 *  DownloadFiles.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Synchronous file download functions
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <wx/wx.h>
#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>

#include "httpbuilder.h"

#include "DownloadFiles.h"
#include "ProgressDialog.h"
#include "MapSupport.h"
#include "MapRepresentations.h"

using namespace std;

/// Directory where downloads are cached. If blank, caching is disabled.
static wxString g_strDownloadCacheDirectory;

/// Hostname of proxy to use. Blank for no proxy.
static wxString g_strProxyHost;

/// Proxy port. 0 for no proxy.
static int g_iProxyPort = 0;

/// Username to authenticate with proxy. Blank for no username/password auth.
static wxString g_strProxyUsername;

/// Password to authenticate with proxy. Blank for no username/password auth.
static wxString g_strProxyPassword;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor for SyncDownloadEntry - zero out everything
///
/////////////////////////////////////////////////////////////////////////////
SyncDownloadEntry::SyncDownloadEntry()
{
	m_bDownloadedOkay = false;
	m_iSocketTimeout = -1;
	m_iRetryAttempts = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor for SyncDownloadEntry - nothing needs to be done
///
/////////////////////////////////////////////////////////////////////////////
SyncDownloadEntry::	~SyncDownloadEntry()
{
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert a URL into a legal filename
///
/// \param strURL			URL to mangle
///
/// \return					Mangled URL
///
/////////////////////////////////////////////////////////////////////////////
wxString MangleURLtoFilename(wxString strURL)
{
	wxString strRtn;

	strRtn = strURL;

	strRtn.Replace(wxT(":"), wxT(""));
	strRtn.Replace(wxT("/"), wxT("."));
	strRtn.Replace(wxT("\\"), wxT("."));
	strRtn.Replace(wxT(" "), wxT("_"));
	strRtn.Replace(wxT("?"), wxT("_"));
	strRtn.Replace(wxT("*"), wxT("_"));
	strRtn.Replace(wxT("@"), wxT(""));
	strRtn.Replace(wxT("$"), wxT(""));
	strRtn.Replace(wxT("&"), wxT(""));

	return strRtn;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the username/password to authenticate with the proxy.
///
/// \param strProxyUsername		Username to use.
/// \param strProxyPassword		Password to authenticate with.
///
/////////////////////////////////////////////////////////////////////////////
void SetProxyUsernamePassword(wxString strProxyUsername, wxString strProxyPassword)
{
	g_strProxyUsername = strProxyUsername;
	g_strProxyPassword = strProxyPassword;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the hostname of the proxy
///
/// \param strProxyHost		Hostname of proxy. Empty string to disable proxy.	
///
/////////////////////////////////////////////////////////////////////////////
void SetProxyHost(wxString strProxyHost)
{
	g_strProxyHost = strProxyHost;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the hostname of the proxy
///
/// \return					Hostname of proxy. Empty string for no proxy.
///
/////////////////////////////////////////////////////////////////////////////
wxString GetProxyHost()
{
	return g_strProxyHost;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the port number of the proxy
///
/// \param iProxyPort		Port number of proxy. 0 to disable proxy.
///
/////////////////////////////////////////////////////////////////////////////
void SetProxyPort(int iProxyPort)
{
	g_iProxyPort = iProxyPort;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Get the port number of the proxy
///
/// \return					Port number of proxy. 0 for no proxy.
///
/////////////////////////////////////////////////////////////////////////////
int GetProxyPort()
{
	return g_iProxyPort;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the proxy hostname, port number, username, and password
/// from the URL strProxy
///
/////////////////////////////////////////////////////////////////////////////
void SetProxyFromURL(const wxString & strProxy)
{
	wxString strRemaining;
	int iTmp;

	wxString strUsername = wxT("");
	wxString strPassword = wxT("");
	wxString strHost = wxT("");
	long lPort = 80;
	
	strRemaining = strProxy;
	iTmp = strRemaining.Find(wxT("//"));
	if (iTmp >= 0)
		strRemaining = strRemaining.Mid(iTmp + 2); // Eliminate everything before and including the //
		
	iTmp = strRemaining.Find(wxT("@"));
	if (iTmp >= 0)
	{
		// Username present
		strUsername = strRemaining.Left(iTmp);
		strRemaining = strRemaining.Mid(iTmp + 1);
		
		iTmp = strUsername.Find(wxT(":"));
		if (iTmp >= 0)
		{
			// Password present
			strPassword = strUsername.Mid(iTmp + 1);
			strUsername = strUsername.Left(iTmp);
		}
	}

	iTmp = strRemaining.Find(wxT("/"));
	if (iTmp >= 0)
	{
		// Trailing slash present
		strHost = strRemaining.Left(iTmp);
		strRemaining = strRemaining.Mid(iTmp + 1);
	}
	else
	{
		// No trailing slash
		strHost = strRemaining;
		strRemaining = wxT("");
	}
	
	iTmp = strHost.Find(wxT(":"));
	if (iTmp >= 0)
	{
		// Port number present
		strHost.Mid(iTmp + 1).ToLong(&lPort);
		strHost = strHost.Left(iTmp);
	}
	
	SetProxyHost(strHost);
	SetProxyPort(lPort);
	SetProxyUsernamePassword(strUsername, strPassword);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the path of the download cache directory
///
/// \return					Directory that contains a cache of all
///							downloaded files. Empty string if cache directory
///							is disabled.
///
/////////////////////////////////////////////////////////////////////////////
wxString GetDownloadCacheDirectory()
{
	return g_strDownloadCacheDirectory;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the download cache directory
///
/// \param strPath			New directory to cache downloads in. Empty string
///							to disable caching.
///
/////////////////////////////////////////////////////////////////////////////
void SetDownloadCacheDirectory(wxString strPath)
{
	g_strDownloadCacheDirectory = strPath;

	if (!wxDirExists(g_strDownloadCacheDirectory))
		wxMkdir(g_strDownloadCacheDirectory);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Empties out the cache directory
///
/////////////////////////////////////////////////////////////////////////////
void ClearDownloadCache()
{
	wxString strPath;
	
	strPath = GetDownloadCacheDirectory();
	
	if (strPath == wxT(""))
		return;
		
	RemoveDirectoryAndContents(strPath);
}

bool IsValidURL(wxString strURL, wxString strUsername, wxString strPassword)
{
	// set URL
	wxHTTPBuilder cURL;
	wxProxySettings settingsProxy;
	wxInputStream * pIn;
	
	cURL.InitContentTypes();

	// set proxy if necessary
	if (g_strProxyHost != wxT("") && g_iProxyPort != 0)
	{
		settingsProxy.SetUseProxy(true);
		
		settingsProxy.m_strProxyHostname = g_strProxyHost;
		settingsProxy.m_nProxyPort = g_iProxyPort;
		settingsProxy.m_strProxyUsername = g_strProxyUsername;
		settingsProxy.m_strProxyPassword = g_strProxyPassword;
		settingsProxy.m_bRequiresAuth = (g_strProxyUsername != wxT("") || g_strProxyPassword != wxT(""));
		
		cURL.SetProxySettings(settingsProxy);
	}
	
	// set up authentication information
	if (strUsername != wxT(""))
	{
	    cURL.Authenticate(strUsername, strPassword);
	}
	
	pIn = cURL.GetInputStream(strURL);
	
	if (!pIn)
		return false;
		
	delete pIn;
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download a set of files, described by vFiles
///
/// \param vFiles			Vector of SyncDownloadEntry objects. One object
///							per file to download.
///
/// \param bProgressDialog	Indicates if a progress dialog should be shown.
///
/// \param pdlgProgress		Progress dialog to use. NULL to automatically
///							create a progress dialog.
///
/// \param iProgressStart	Starting position in progress dialog.
///
/// \param iProgressTicks	Added to iProgressStart to obtain the ending
///							position in the dialog.
///
/// \return					Zero on success. Non-zero on errors.
///
/////////////////////////////////////////////////////////////////////////////
int DownloadFiles(vector<SyncDownloadEntry> & vFiles, bool bProgressDialog, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	unsigned int iFile;
	int nErrors;
	int iRealProgressStart;
	int iRealProgressTicks;
	wxString strErrorMsgs;

	iRealProgressStart = 0;
	iRealProgressTicks = 1;
	nErrors = 0;

	ProgressDialog * dlgProgress = NULL;
	
	if (bProgressDialog)
	{
		if (pdlgProgress)
		{
			dlgProgress = pdlgProgress;
			iRealProgressStart = iProgressStart;
			iRealProgressTicks = iProgressTicks;
		}
		else
		{
			dlgProgress = new ProgressDialog(NULL, wxT("Downloading..."), 100);
			dlgProgress->Show();

			iRealProgressStart = 0;
			iRealProgressTicks = 100;
		}
	}

	for (iFile = 0; iFile < vFiles.size(); iFile++)
	{
		char szBuffer[8192];
		int nBuffer;
		FILE * hOut;
		wxInputStream * pIn;
		int iFileLength;
		int iFilePos;
		wxFileName fnCache;
		bool bCopiedFromCache;

		fnCache.AssignDir(GetDownloadCacheDirectory());
		fnCache.SetFullName(MangleURLtoFilename(vFiles[iFile].m_strURL));

		// attempt to copy the file from the cache
		bCopiedFromCache = false;
		if (wxFileExists(fnCache.GetFullPath()))
			bCopiedFromCache = wxCopyFile(fnCache.GetFullPath(), vFiles[iFile].m_strOutputFilename);

		if (bCopiedFromCache)
		{
			vFiles[iFile].m_bDownloadedOkay = true;
		}
		else
		{
			// if the file cannot be copied from the cache, then download it

			// update progress bar text
			if (bProgressDialog)
			{
				dlgProgress->SetLabel(vFiles[iFile].m_strProgressBarName);

				double fPercentage = ((double) iFile) / vFiles.size();
				dlgProgress->SetProgress((int) (iRealProgressStart + fPercentage * iRealProgressTicks));
			}
			
			// set URL
			wxHTTPBuilder cURL;
			wxProxySettings settingsProxy;
			
			cURL.InitContentTypes();

			// set proxy if necessary
			if (g_strProxyHost != wxT("") && g_iProxyPort != 0)
			{
				settingsProxy.SetUseProxy(true);
				
				settingsProxy.m_strProxyHostname = g_strProxyHost;
				settingsProxy.m_nProxyPort = g_iProxyPort;
				settingsProxy.m_strProxyUsername = g_strProxyUsername;
				settingsProxy.m_strProxyPassword = g_strProxyPassword;
				settingsProxy.m_bRequiresAuth = (g_strProxyUsername != wxT("") || g_strProxyPassword != wxT(""));
				
				cURL.SetProxySettings(settingsProxy);
			}
			
			// set up authentication information
			if (vFiles[iFile].m_strUsername != wxT(""))
			{
			    cURL.Authenticate(vFiles[iFile].m_strUsername, vFiles[iFile].m_strPassword);
			}
			
			// socket timeout
			if (vFiles[iFile].m_iSocketTimeout > 0)
				cURL.SetTimeout(vFiles[iFile].m_iSocketTimeout);

			int iAttemptsLeft = vFiles[iFile].m_iRetryAttempts + 1;
			pIn = NULL;		
			while (iAttemptsLeft-- && !pIn)
				pIn = cURL.GetInputStream(vFiles[iFile].m_strURL);

			if (!pIn)
			{
				wxString strError;
				
				switch (cURL.GetError())
				{
					case wxURL_NOERR:
						strError = wxT("No error");
						break;

					case wxURL_SNTXERR:	
						strError = wxT("Syntax error in the URL string");
						break;

					case wxURL_NOPROTO: 	
						strError = wxT("Found no protocol which can get this URL");
						break;

					case wxURL_NOHOST:	
						strError = wxT("An host name is required for this protocol");
						break;

					case wxURL_NOPATH:	
						strError = wxT("A path is required for this protocol");
						break;

					case wxURL_CONNERR:
						strError = wxT("Connection error");
						break;

					case wxURL_PROTOERR:
						strError = wxT("An error occurred during negotiation");
						break;

					default:
						strError = wxT("Unknown error");
						break;
				}

				strErrorMsgs += wxString::Format(wxT("Error reading %s (%s)\n"), vFiles[iFile].m_strURL.c_str(), strError.c_str());
				vFiles[iFile].m_bDownloadedOkay = false;
				nErrors++;
			}
			else
			{
				// download file
				hOut = fopen((vFiles[iFile].m_strOutputFilename + wxT(".tmp")).mb_str(*wxConvCurrent), "wb");
				if (!hOut)
				{
					strErrorMsgs += wxString::Format(wxT("Error opening %s for writing\n"), vFiles[iFile].m_strOutputFilename.c_str());

					vFiles[iFile].m_bDownloadedOkay = false;
					nErrors++;

					delete pIn;
				}
				else
				{					
					iFilePos = 0;
					iFileLength = pIn->GetSize();

					if (bProgressDialog && iFileLength >= 0)
					{
						dlgProgress->SetLabel(wxString::Format(wxT("%s (%s)"), vFiles[iFile].m_strProgressBarName.c_str(), FormatFileSize(iFileLength).c_str()));
					}					

					// if the web server doesn't report the content length, then
					// make up a big length so the progress bar doesn't go crazy
					if (iFileLength < 0)
						iFileLength = 1000000000;
						
					pIn->Read(szBuffer, sizeof(szBuffer));
					nBuffer = pIn->LastRead();
					
					while (nBuffer > 0)
					{
						iFilePos += nBuffer;
						fwrite(szBuffer, 1, nBuffer, hOut);

						// need to use floating point numbers, otherwise large counties will
						// cause an overflow
						if (bProgressDialog)
						{
							double fPercentage = (iFile * 100 + ((double) iFilePos) * 100.0 / iFileLength) / (vFiles.size() * 100);
							dlgProgress->SetProgress((int) (iRealProgressStart + fPercentage * iRealProgressTicks));
						}

						pIn->Read(szBuffer, sizeof(szBuffer));
						nBuffer = pIn->LastRead();
					}
					fclose(hOut);
					
					delete pIn;

					wxRemoveFile(vFiles[iFile].m_strOutputFilename);
					wxRenameFile(vFiles[iFile].m_strOutputFilename + wxT(".tmp"), vFiles[iFile].m_strOutputFilename);

					// cache download
					if (GetDownloadCacheDirectory() != wxT(""))
						wxCopyFile(vFiles[iFile].m_strOutputFilename, fnCache.GetFullPath());

					// mark file as downloaded
					vFiles[iFile].m_bDownloadedOkay = true;
				}
			}
		}
	}
	
	if (bProgressDialog)
	{
		if (!pdlgProgress)
		{
			// hide progress dialog
			dlgProgress->Show(FALSE);
			delete dlgProgress;
		}
	}

	if (nErrors)
		::wxMessageBox(strErrorMsgs, wxT("Error"), wxOK);

	return nErrors;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download a single file, described by sFile
///
/// \param sFile			Describes the file to be downloaded.
///
/// \param bProgressDialog	Indicates if a progress dialog should be shown.
///
/// \param pdlgProgress		Progress dialog to use. NULL to automatically
///							create a progress dialog.
///
/// \param iProgressStart	Starting position in progress dialog.
///
/// \param iProgressTicks	Added to iProgressStart to obtain the ending
///							position in the dialog.
///
/// \return					Zero on success. Non-zero on errors.
///
/////////////////////////////////////////////////////////////////////////////
int DownloadFile(SyncDownloadEntry & sFile, bool bProgressDialog, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	vector<SyncDownloadEntry> vFiles;
	int iRtn;

	vFiles.push_back(sFile);

	iRtn = DownloadFiles(vFiles, bProgressDialog, pdlgProgress, iProgressStart, iProgressTicks);

	sFile = vFiles[0];

	return iRtn;
}

