/*
 *  LibRoadnav
 *  DownloadThread.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains a thread that performs downloads asynchronously
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>

#include "DownloadThread.h"
#include "MapSupport.h"

/// List of files to download asynchronously
std::list<SDownloadEntry> g_vDownloadList;

/// Mutex used to protect g_vDownloadList
wxMutex g_mtxDownloadList;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief This function adds a file to the download queue.
///
/// \param strFilename				Filename where file should be stored.
///
/// \param strURL					URL to download
///
/// \param iPostDownloadAction		Flags representing what actions should be
///									taken once all downloads are complete.
///
/// Currently, only used to download Terraserver tiles.
///
//////////////////////////////////////////////////////////////////////////////
void QueueDownload(wxString strFilename, wxString strURL, int iPostDownloadAction)
{
	SDownloadEntry sEntry;

	sEntry.strFilename = strFilename;
	sEntry.strURL = strURL;
	sEntry.iPostDownloadAction = iPostDownloadAction;

	wxMutexLocker lock(g_mtxDownloadList);

	std::list<SDownloadEntry>::iterator i;

	for (i = g_vDownloadList.begin(); i != g_vDownloadList.end(); i++)
	{
		if (
				(*i).strURL == sEntry.strURL &&
				(*i).strFilename == sEntry.strFilename &&
				(*i).iPostDownloadAction == sEntry.iPostDownloadAction
			)
		{
			// entry is already queued
			return;
		}
	}

	g_vDownloadList.push_back(sEntry);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download thread constructor
///
//////////////////////////////////////////////////////////////////////////////
DownloadThread::DownloadThread(wxWindow * pParent, int idRedrawNotification) : wxThread(wxTHREAD_JOINABLE)
{
	m_pParent = pParent;
	m_idRedrawNotification = idRedrawNotification;
}


//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Download worker thread - all of the async downloads happen here.
///
//////////////////////////////////////////////////////////////////////////////
void * DownloadThread::Entry()
{
	int iQueuedActions = 0;
	int iErrors = 0;

	while (!TestDestroy())
	{
		SDownloadEntry sEntry;
		bool bDone;

		bDone = true;

		g_mtxDownloadList.Lock();

		// remove an entry if the list is non-empty
		if (g_vDownloadList.size() > 0)
		{
			sEntry = g_vDownloadList.back();
			g_vDownloadList.pop_back();
			bDone = false;
		}

		g_mtxDownloadList.Unlock();

		if (!bDone)
		{
			// process entry
			if (GetDataPath(sEntry.strFilename, sEntry.strURL, false, false) == wxT(""))
			{
				iErrors++;
				
				if (iErrors > 3)
				{
					int iCount;
					
					// burst of errors occurred - throttle for one minute
					for (iCount = 0; iCount < 60 && !TestDestroy(); iCount++)
						wxThread::Sleep(1000);
				}
			}
			else
			{
				iErrors = 0;
			}
			iQueuedActions |= sEntry.iPostDownloadAction;
		}
		else
		{
			// check if we're supposed to refresh the map control when we're done downloading
			if (iQueuedActions & DL_ACTION_REFRESH)
			{
				// yes .. ensure parent is set
				if (m_pParent)
				{
					// okay, send refresh event
					wxCommandEvent evCmd(wxEVT_COMMAND_MENU_SELECTED, m_idRedrawNotification);
					m_pParent->AddPendingEvent(evCmd);
				}
				
			}

			iQueuedActions = 0;

			// list was empty - sleep for a bit
			wxThread::Sleep(1000);
		}
	}

	return NULL;
}
