/*
 *  LibRoadnav
 *  Map.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains some initialization code for the map code.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>
#include <wx/filename.h>

#include "MapControlData_Tiles.h"
#include "Map.h"
#include "MapState.h"
#include "MapCounty.h"
#include "MapZip.h"
#include "MapAbbreviations.h"
#include "MapSupport.h"
#include "Constants.h"
#include "URLs.h"
#include "MapControlDataImporter_StateBoundaries.h"
#include "MapControlDataImporter_CountyBoundaries.h"
#include "MapControlDataImporter_TigerLine.h"
#include "MapControlDataImporter_GNISDECI.h"
#include "ProgressDialog.h"
#include "Debug.h"
#include "StringTable.h"

using namespace std;

#if !wxCHECK_VERSION(2, 6, 2)
#error Requires at least version 2.6.2 of wxWidgets
#endif

// Version number .. increment whenever an incompatible file format change is made.
// This will cause the old set of compiled maps to be wiped out.
#define DATA_VERSION		"0017"

StringTable * g_pcStringTable = NULL;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Reload the specified map data
///
//////////////////////////////////////////////////////////////////////////////
void ReloadMapData(IMapControlData * pcRecords, vector<string> vIDs)
{
	vector<string>::iterator i;
	int iCount;
	int iProgressPosition;
	
	iCount = vIDs.size();
	
	ProgressDialog dlgProgress(NULL, wxT("Downloading..."), iCount * 100);
	dlgProgress.Show();
	iProgressPosition = 0;

	for (i = vIDs.begin(); i != vIDs.end(); i++)
	{
		if (*i == "StateBoundaries")
		{
			MapControlDataImporter_StateBoundaries cSB;
			
			if (!MapControlDataImporter_StateBoundaries::IsLoaded())
				cSB.Download(pcRecords, &dlgProgress, iProgressPosition, 99);
			
			iProgressPosition += 100;
		}
		else if (*i == "CountyBoundaries")
		{
			MapControlDataImporter_CountyBoundaries cCB;
			
			if (!MapControlDataImporter_CountyBoundaries::IsLoaded())
				cCB.Download(pcRecords, &dlgProgress, iProgressPosition, 99);
				
			iProgressPosition += 100;
		}
		else if (!strncmp((*i).c_str(), "TigerLine-", 10))
		{
			int iCountyCode;
			MapControlDataImporter_TigerLine diTigerLine;
			
			iCountyCode = atoi((*i).c_str() + 10);
			
			wxASSERT(iCountyCode);

			if (!diTigerLine.IsCountyLoaded(iCountyCode))
				diTigerLine.DownloadCounty(pcRecords, iCountyCode, &dlgProgress, iProgressPosition, 99);
				
			iProgressPosition += 100;
		}
		else if (!strncmp((*i).c_str(), "GNIS-", 5))
		{
			char szStateAbbrev[3];
			MapControlDataImporter_GNISDECI diGNIS;
			
			strncpy(szStateAbbrev, (*i).c_str() + 5, 2);
			szStateAbbrev[2] = 0;
			
			if (!diGNIS.IsGNISStateLoaded(wxString(szStateAbbrev, *wxConvCurrent)))
				diGNIS.DownloadGNISState(pcRecords, wxString(szStateAbbrev, *wxConvCurrent), &dlgProgress, iProgressPosition, 99);

			iProgressPosition += 100;
		}
		else
		{
			wxASSERT(0);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Erases all compiled map data
///
//////////////////////////////////////////////////////////////////////////////
void ClearCompiledMapData()
{
	RemoveDirectoryAndContents(GetDataPath());
}

bool NeedToEraseData()
{
	FILE * hFile;
	wxFileName fnVersion;
	char szDataVersion[16] = "";

	fnVersion.AssignDir(GetDataPath());
	fnVersion.SetFullName(wxT("version.txt"));

	hFile = fopen(fnVersion.GetFullPath().mb_str(*wxConvCurrent), "rb");

	if (hFile)
	{
		// obtain version from version.txt
		fgets(szDataVersion, sizeof(szDataVersion) / sizeof(char), hFile);
		fclose(hFile);

		// strip CRs and LFs
		while (strlen(szDataVersion) && (szDataVersion[strlen(szDataVersion) - 1] == '\r' || szDataVersion[strlen(szDataVersion) - 1] == '\n'))
			szDataVersion[strlen(szDataVersion) - 1] = 0;
	}

	if (!strcmp(szDataVersion, DATA_VERSION))
		return false;
	
	return true;
}

bool IsFreshInstall()
{
	FILE * hFile;
	wxFileName fnVersion;

	fnVersion.AssignDir(GetDataPath());
	fnVersion.SetFullName(wxT("version.txt"));

	hFile = fopen(fnVersion.GetFullPath().mb_str(*wxConvCurrent), "rb");

	if (hFile)
	{
		fclose(hFile);
		return false;
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Check if the data files belong to an old version of Roadnav, and if so
/// erase them all and start over again.
///
/// \return true if existing data was wiped out, or false if either this is
///         a fresh install, or the data was not wiped out.
///
//////////////////////////////////////////////////////////////////////////////
bool EraseOldData()
{
	FILE * hFile;
	wxFileName fnVersion;

	fnVersion.AssignDir(GetDataPath());
	fnVersion.SetFullName(wxT("version.txt"));

	// the data we're using is out of data .. wipe it out and start again
	ClearCompiledMapData();

	// this will cause the data directory to be recreated
	GetDataPath();
	
	hFile = fopen(fnVersion.GetFullPath().mb_str(*wxConvCurrent), "wb");

	wxASSERT(hFile);

	if (hFile)
	{
		fprintf(hFile, "%s\r\n", DATA_VERSION);
		fclose(hFile);
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the various map related files
///
//////////////////////////////////////////////////////////////////////////////
int MapInit()
{
	wxFileName fnStringTable;
	MapControlData_Tiles cRecords;
	bool bNeedReload = false;
	vector<string> vIntegratedData;
	
	LibRoadnavDebug0(wxT("LibRoadnav"), wxT("LibRoadnav initializing"));

	if (NeedToEraseData())
	{
		if (!IsFreshInstall())
		{
			switch(::wxMessageBox(	wxT("This version of Roadnav uses a new and more efficient file format to represent map data. Roadnav can optionally automatically download and recompile the maps you had previously loaded in the old format. This may take a while. Alternately, you can start fresh without any maps and download maps as needed. Do you wish to download and recompile all of the maps you had previously downloaded? Choose yes to download, no to start fresh, or cancel to close Roadnav without modifying the data."), 
									wxT("New Map File Format"),
									wxYES_NO | wxCANCEL | wxICON_QUESTION))
			{
				case wxYES:
					vIntegratedData = GetIntegratedDataList();
					bNeedReload = true;
					break;
					
				case wxNO:
					break;
					
				case wxCANCEL:
					return 1;
			}
		}
		EraseOldData();
	}

	fnStringTable.AssignDir(GetDataPath());
	fnStringTable.SetFullName(wxT("stringtable.bin"));
	g_pcStringTable = new StringTable(fnStringTable.GetFullPath());
	
	LoadStateInformation(GetDataPath(wxT("states.tdf"), GetURL(wxT("States"))));
	LoadZipInformation(GetDataPath(wxT("zips.tdf"), GetURL(wxT("Zip Codes"))));
	LoadCountyInformation(GetDataPath(wxT("counties.txt"), GetURL(wxT("Counties"))));
	LoadAbbreviationInformation(GetDataPath(wxT("abbreviations.txt"), GetURL(wxT("Abbreviations"))));

	if (bNeedReload)
		ReloadMapData(&cRecords, vIntegratedData);

	if (!MapControlDataImporter_StateBoundaries::IsLoaded() || 
		!MapControlDataImporter_CountyBoundaries::IsLoaded())
	{
		MapControlDataImporter_StateBoundaries cSB;
		MapControlDataImporter_CountyBoundaries cCB;

		ProgressDialog dlgProgress(NULL, wxT("Downloading..."), 200);
		dlgProgress.Show();

		cSB.Download(&cRecords, &dlgProgress, 0, 100);
		cCB.Download(&cRecords, &dlgProgress, 100, 100);
	}

	return 0;
}

void MapCleanup()
{
	delete g_pcStringTable;
	g_pcStringTable = NULL;
}
