/*
 *  LibRoadnav
 *  MapAbbreviations.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// This file contains functions that translate feature types into their 
/// four letter abbreviation and abbreviations back into fully written out 
/// feature types.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <vector>
#include <map>

#include <wx/wx.h>

#include "MapSupport.h"
#include "Constants.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
/// Abbreviation maps
//////////////////////////////////////////////////////////////////////////////
static map<wxString,wxString> mapAbbreviationToFullName;
static map<wxString,wxString> mapNameToAbbreviation;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load abbreviation information from a file.
///
/// \param strFilename			Filename to read abbreviations from
///
/// \return						true on error, false on success
///
/// File Format is:
/// Long Name <tabs> Long abbreviation <tabs> 4 letter abbreviation <tabs> junk
///
//////////////////////////////////////////////////////////////////////////////
bool LoadAbbreviationInformation(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (abbreviations)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		wxString strName = wxT("");
		wxString strLongAbbrev = wxT("");
		wxString strShortAbbrev = wxT("");
		wxString strTranslation = wxT("");
		bool bValid = true;

		char * pszToken;
		
		// strip off CR/LF
		while (strlen(szLine) && (szLine[strlen(szLine) - 1] == '\r' || szLine[strlen(szLine) - 1] == '\n'))
			szLine[strlen(szLine) - 1] = 0;
	
		pszToken = strtok(szLine, "\t");
		
		if (pszToken)
		{
			strName = wxString(pszToken, *wxConvCurrent);
			strName.Trim();
			pszToken = strtok(NULL, "\t");
		}
		
		if (pszToken)
		{
			strLongAbbrev = wxString(pszToken, *wxConvCurrent);
			strLongAbbrev.Trim();
			pszToken = strtok(NULL, "\t");
		}
		
		if (pszToken)
		{
			strShortAbbrev = wxString(pszToken, *wxConvCurrent);
			strShortAbbrev.Trim();
			pszToken = strtok(NULL, "\t");
		}
		else
		{
			bValid = false;
		}
		
		if (pszToken)
		{
			strTranslation = wxString(pszToken, *wxConvCurrent);
			strTranslation.Trim();
			pszToken = strtok(NULL, "\t");
		}
		
		
		if (bValid)
		{
			if (strShortAbbrev == wxT("-"))
				strShortAbbrev = strLongAbbrev;
				
			mapAbbreviationToFullName[strShortAbbrev.Lower()] = strName;
			mapAbbreviationToFullName[strLongAbbrev.Lower()] = strName;
			
			mapNameToAbbreviation[strLongAbbrev.Lower()] = strShortAbbrev;
			mapNameToAbbreviation[strName.Lower()] = strShortAbbrev;
		}
		
		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the four letter abbreviation for something.
///
/// \return				Returns "" if it's not listed.
///
//////////////////////////////////////////////////////////////////////////////
wxString GetAbbreviation(const wxString & strName)
{
	if (mapNameToAbbreviation.find(strName.Lower()) != mapNameToAbbreviation.end())
		return mapNameToAbbreviation[strName.Lower()];
	
	return wxT("");
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Find what an abbreviation stands for.
///
/// \param strName		Abbreviation to look up.
///
/// \return				Returns "" if it's not listed.
///
//////////////////////////////////////////////////////////////////////////////
wxString GetFullName(const wxString & strName)
{
	if (mapAbbreviationToFullName.find(strName.Lower()) != mapAbbreviationToFullName.end())
		return mapAbbreviationToFullName[strName.Lower()];
	
	return wxT("");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determine if the supplied string is a feature type or not (drive, road,
/// lane, etc.)
///
/// \param strName		Abbreviation/long name to look up.
///
/// \return				true if abbreviation/long name found, false otherwise.
///
//////////////////////////////////////////////////////////////////////////////
bool IsFeatureType(const wxString & strName)
{
	if (GetAbbreviation(strName) != wxT("") || GetFullName(strName) != wxT(""))
		return true;
	
	return false;
}
