/*
 *  LibRoadnav
 *  MapAppearanceSettings.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// \brief Contains the MapAppearanceSettings class - a class representing the 
/// MapControl appearance related settings that are tied to a particular 
/// detail (zoom) level.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/wx.h>

#include "MapAppearanceSettings.h"
#include "MapSupport.h"
#include "MapRepresentations.h"

#define MAPINDEX(dl,rt) ((dl+1)*100 + rt)
#define MIN_DETAIL_LEVEL	0
#define MAX_DETAIL_LEVEL	14

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - call LoadDefaults()
///
//////////////////////////////////////////////////////////////////////////////
MapAppearanceSettings::MapAppearanceSettings()
{
	LoadDefaults();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the default visibility/style/font settings for each record
/// type at each detail level.
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::LoadDefaults()
{
	ClearDetailLevels();

	ERecordType eRT;
	wxUint16 iDetailLevel;

	std::vector<double> scales;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
		AddDetailLevel( 65536. / 1e6 * ::pow(0.5, iDetailLevel+1) );

	eRT = RecordTypeDefault;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 11 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 11 ? true : false;
		settings.m_iFontSize = 8;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypeTwoWaySmallRoad;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 8 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 10 ? true : false;
		settings.m_iFontSize = 12;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}
		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}
		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 18;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 16;
			settings.m_iFontSize = 7;
		}
		if (iDetailLevel == 10)
		{
			settings.m_iFontSize = 6;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 1;
		}
	}

	eRT = RecordTypeOneWaySmallRoad;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 8 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 10 ? true : false;
		settings.m_iFontSize = 12;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}
		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}
		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 18;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 16;
			settings.m_iFontSize = 7;
		}
		if (iDetailLevel == 10)
		{
			settings.m_iFontSize = 6;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 1;
		}
	}

	eRT = RecordTypeTwoWayLargeRoad;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 4;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 7 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 7 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}
		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}
		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}
		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}
		if (iDetailLevel == 10)
		{
			settings.m_iWidth = 12;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 9)
		{
			settings.m_iWidth = 10;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 5;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 7)
		{
			settings.m_iWidth = 2;
			settings.m_iFontSize = 8;
		}
	}

	eRT = RecordTypeOneWayLargeRoad;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 4;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 7 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 7 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}
		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}
		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}
		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}
		if (iDetailLevel == 10)
		{
			settings.m_iWidth = 12;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 9)
		{
			settings.m_iWidth = 10;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 5;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 7)
		{
			settings.m_iWidth = 2;
			settings.m_iFontSize = 8;
		}
	}

	eRT = RecordTypeTwoWayHighway;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 8;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 2 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 3 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}

		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}

		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}

		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}

		if (iDetailLevel == 9)
		{
			settings.m_iWidth = 10;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 5;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 7)
		{
			settings.m_iWidth = 4;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 6)
		{
			settings.m_iWidth = 4;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel <= 5)
		{
			settings.m_iWidth = 2;
			settings.m_iFontSize = 6;
		}
	}

	eRT = RecordTypeOneWayHighway;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 8;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 2 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 3 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 14)
		{
			settings.m_iWidth = 44;
			settings.m_iFontSize = 20;
		}

		if (iDetailLevel == 13)
		{
			settings.m_iWidth = 36;
			settings.m_iFontSize = 14;
		}

		if (iDetailLevel == 12)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}

		if (iDetailLevel == 11)
		{
			settings.m_iWidth = 24;
			settings.m_iFontSize = 12;
		}

		if (iDetailLevel == 9)
		{
			settings.m_iWidth = 10;
			settings.m_iFontSize = 10;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iWidth = 5;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 7)
		{
			settings.m_iWidth = 4;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel == 6)
		{
			settings.m_iWidth = 4;
			settings.m_iFontSize = 8;
		}
		if (iDetailLevel <= 5)
		{
			settings.m_iWidth = 2;
			settings.m_iFontSize = 6;
		}
	}

	eRT = RecordTypeRailroad;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 9 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 10 ? true : false;
		settings.m_iFontSize = 10;
		settings.m_iArrowSpacing = 0;
		settings.m_bRailroadStyle = true;
	}

	eRT = RecordTypeWater;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 8 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 10 ? true : false;
		settings.m_iFontSize = 10;
		settings.m_iArrowSpacing = 0;
	}
	eRT = RecordTypeInvisibleLandBoundary;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = iDetailLevel >= 9 ? 3 : 1;
		settings.m_iStyle = wxSOLID;
//		settings.m_bLineVisible = iDetailLevel >= 1 ? true : false;
		settings.m_bLineVisible = iDetailLevel >= 4 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 7 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypeInvisibleWaterBoundary;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 2;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 8 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 10 ? true : false;
		settings.m_iFontSize = 10;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypeInvisibleMiscBoundary;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = iDetailLevel >= 9 ? 3 : 1;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 12 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 12 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypeLandmark;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 3;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 12 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 12 ? true : false;
		settings.m_iFontSize = 10;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypePhysicalFeature;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = 6;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 6 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = iDetailLevel >= 8 ? true : false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel == 9)
		{
			settings.m_iFontSize = 12;
			settings.m_iWidth = 5;
		}
		if (iDetailLevel == 8)
		{
			settings.m_iFontSize = 8;
			settings.m_iWidth = 4;
		}
		if (iDetailLevel == 7)
		{
			settings.m_iWidth = 2;
		}
		if (iDetailLevel == 6)
		{
			settings.m_iWidth = 1;
		}
	}

	eRT = RecordTypeTownOrCity;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);
		wxColour clr;

		settings.m_iWidth = 2;
		settings.m_iStyle = wxTRANSPARENT;
		settings.m_bLineVisible = (iDetailLevel >= 3 && iDetailLevel <= 9) ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = (iDetailLevel >= 5 && iDetailLevel <= 9) ? true : false;
/*		
		if (iDetailLevel >= 8)
			clr.Set(192, 0, 0);
		else
			clr.Set(160, 160, 160);
		
		settings.m_clrLine = clr;
		settings.m_clrText = clr;
*/
		settings.m_iFontSize = 14;
		settings.m_iArrowSpacing = 0;

		if (iDetailLevel < 6)
		{
			settings.m_iFontSize = 6;
		}
		else if (iDetailLevel < 7)
		{
			settings.m_iFontSize = 8;
		}
		else if (iDetailLevel < 8)
		{
			settings.m_iFontSize = 10;
		}
		else if (iDetailLevel < 9)
		{
			settings.m_iFontSize = 14;
		}
	}

	eRT = RecordTypeStateBoundary;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = iDetailLevel >= 9 ? 3 : 1;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = true;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;
	}

	eRT = RecordTypeCountyBoundary;
	for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
	{
		DetailSettings& settings = GetDetailSettings(iDetailLevel,eRT);

		settings.m_iWidth = iDetailLevel >= 9 ? 3 : 1;
		settings.m_iStyle = wxSOLID;
		settings.m_bLineVisible = iDetailLevel >= 3 ? true : false;
		settings.m_bFastLine = false;
		settings.m_bTextVisible = false;
		settings.m_iFontSize = 16;
		settings.m_iArrowSpacing = 0;
	}

	SetMiscColor(MiscColorBackground,				ParseColor(wxT("#FFFFFF")) );
	SetMiscColor(MiscColorForeground,				ParseColor(wxT("#000000")) );
	SetMiscColor(MiscColorMarkerBackground,			ParseColor(wxT("#FFFF00")) );
	SetMiscColor(MiscColorMarkerForeground,			ParseColor(wxT("#000000")) );

	SetColor( RecordTypeDefault,					ParseColor(wxT("#C0C0C0")) );
	SetColor( RecordTypeOneWaySmallRoad,			ParseColor(wxT("#808080")) );
	SetColor( RecordTypeTwoWaySmallRoad,			ParseColor(wxT("#808080")) );
	SetColor( RecordTypeOneWayLargeRoad,			ParseColor(wxT("#827000")) );
	SetColor( RecordTypeTwoWayLargeRoad,			ParseColor(wxT("#827000")) );
	SetColor( RecordTypeOneWayHighway,				ParseColor(wxT("#B68628")) );
	SetColor( RecordTypeTwoWayHighway,				ParseColor(wxT("#B68628")) );
	SetColor( RecordTypeRailroad,					ParseColor(wxT("#808080")) );
	SetColor( RecordTypeWater,						ParseColor(wxT("#0000ff")) );
	SetColor( RecordTypeInvisibleLandBoundary,		ParseColor(wxT("#A0A0A0")) );
	SetColor( RecordTypeInvisibleWaterBoundary,		ParseColor(wxT("#C0C0FF")) );
	SetColor( RecordTypeInvisibleMiscBoundary,		ParseColor(wxT("#C0C0FF")) );
	SetColor( RecordTypeLandmark,					ParseColor(wxT("#2FAF41")) );
	SetColor( RecordTypePhysicalFeature,			ParseColor(wxT("#000000")) );
	SetColor( RecordTypeTownOrCity,					ParseColor(wxT("#808080")) );
	SetColor( RecordTypeStateBoundary,				ParseColor(wxT("#008000")) );
	SetColor( RecordTypeCountyBoundary,				ParseColor(wxT("#f0f0f0")) );
}

//////////////////////////////////////////////////////////////////////////////
///
/// Loads the color information from the registry/config file
///
//////////////////////////////////////////////////////////////////////////////
bool MapAppearanceSettings::Load(wxConfigBase* pConfig)
{
	int iRT;
	int iDL;
	int iMC;
	int nDetailLevels;
	
	nDetailLevels = pConfig->Read(wxT("NumberOfDetailLevels"), (long) 0);
	
	// abort if there's a bogus number of detail levels (probably indicates we're upgrading from an old version of Roadnav)
	if (nDetailLevels < 1)
		return true;
	
	m_vDetailLevelToScale.clear();
	
	for (iDL = 0; iDL < nDetailLevels; iDL++)
	{
		double fVal = 0;
		
		pConfig->Read(wxString::Format(wxT("DetailLevelToScale%d"), iDL), &fVal);
		
		if (fVal == 0)
			return true;
		
		m_vDetailLevelToScale.push_back(fVal);
	}

	for (iMC = 0; iMC <= LAST_MISCCOLOR; iMC++)
	{
		m_mapMiscColors[(EMiscColor) iMC] = ParseColor(pConfig->Read(wxString::Format(wxT("m%d"), iMC), wxT("")));
	}
	
	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		int iDetailLevel;
		
		for (iDetailLevel = 0; iDetailLevel < GetDetailLevelCount(); iDetailLevel++)
		{
			pConfig->SetPath(wxString::Format(wxT("%d,%d"), iRT, iDetailLevel));
			
			if (GetDetailSettings(iDetailLevel, (ERecordType) iRT).Load(pConfig))
			{
				pConfig->SetPath(wxT(".."));
				return true;
			}

			pConfig->SetPath(wxT(".."));
		}
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Saves the color information to the registry/config file
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::Save(wxConfigBase* pConfig) const
{
	int iRT;
	int iDL;
	int iMC;
	
	pConfig->Write(wxT("NumberOfDetailLevels"), GetDetailLevelCount());

	for (iDL = 0; iDL < GetDetailLevelCount(); iDL++)
	{
		pConfig->Write(wxString::Format(wxT("DetailLevelToScale%d"), iDL), m_vDetailLevelToScale[iDL]);
	}

	for (iMC = 0; iMC <= LAST_MISCCOLOR; iMC++)
	{
		std::map<EMiscColor, wxColour>::const_iterator iFind;
		
		iFind = m_mapMiscColors.find((EMiscColor) iMC);
		
		if (iFind != m_mapMiscColors.end())
		{
			pConfig->Write(wxString::Format(wxT("m%d"), iMC), FormatColor((*iFind).second));
		}
	}
	
	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		int iDetailLevel;
		
		for (iDetailLevel = 0; iDetailLevel < GetDetailLevelCount(); iDetailLevel++)
		{
			pConfig->SetPath(wxString::Format(wxT("%d,%d"), iRT, iDetailLevel));
			
			GetDetailSettings(iDetailLevel, (ERecordType) iRT).Save(pConfig);

			pConfig->SetPath(wxT(".."));
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the line color for the specified record type and 
/// detail level.
///
/// \param	eRecordType		Record type of the color setting.
///
/// \param	clrColor		New color for specified record type.
///
/// \param	iDetailLevel	Detail level to update, or -1 for all detail
///							levels.
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::SetLineColor(ERecordType eRecordType, const wxColour & clrColor, int iDetailLevel)
{
	if (iDetailLevel >= 0)
	{
		GetDetailSettings(iDetailLevel, eRecordType).m_clrLine = clrColor;
	}
	else
	{	
		int iIndexDetailLevel;

		for (iIndexDetailLevel = 0; iIndexDetailLevel < GetDetailLevelCount(); iIndexDetailLevel++)
		{
			GetDetailSettings(iIndexDetailLevel, eRecordType).m_clrLine = clrColor;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the text color for the specified record type and 
/// detail level.
///
/// \param	eRecordType		Record type of the color setting.
///
/// \param	clrColor		New color for specified record type.
///
/// \param	iDetailLevel	Detail level to update, or -1 for all detail
///							levels.
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::SetTextColor(ERecordType eRecordType, const wxColour & clrColor, int iDetailLevel)
{
	if (iDetailLevel >= 0)
	{
		GetDetailSettings(iDetailLevel, eRecordType).m_clrText = clrColor;
	}
	else
	{	
		int iIndexDetailLevel;

		for (iIndexDetailLevel = 0; iIndexDetailLevel < GetDetailLevelCount(); iIndexDetailLevel++)
		{
			GetDetailSettings(iIndexDetailLevel, eRecordType).m_clrText = clrColor;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the text and line color for the specified record type and 
/// detail level.
///
/// \param	eRecordType		Record type of the color setting.
///
/// \param	clrColor		New color for specified record type.
///
/// \param	iDetailLevel	Detail level to update, or -1 for all detail
///							levels.
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::SetColor(ERecordType eRecordType, const wxColour & clrColor, int iDetailLevel)
{
	SetLineColor(eRecordType, clrColor, iDetailLevel);
	SetTextColor(eRecordType, clrColor, iDetailLevel);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Wipe out the contents of this object
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::ClearDetailLevels()
{
	m_mapDetailSettings.clear();
	m_vDetailLevelToScale.clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Add the specified scale to the list of detail levels in this object.
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::AddDetailLevel(double scale)
{
	wxInt16 iDetailLevel;
	wxInt16 iRecordType;
	
	m_vDetailLevelToScale.push_back(scale);

	iDetailLevel = m_vDetailLevelToScale.size() - 1;
	
	for (iRecordType = FIRST_RECORD_TYPE; iRecordType <= LAST_RECORD_TYPE; iRecordType++)
	{
		DetailSettings dsSettings;
		wxInt16 iMapIndex;
		
		iMapIndex = MAPINDEX(iDetailLevel, iRecordType);
		m_mapDetailSettings[iMapIndex] = dsSettings;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the number of detail levels contained in this object.
///
//////////////////////////////////////////////////////////////////////////////
wxInt16 MapAppearanceSettings::GetDetailLevelCount() const
{
   return m_vDetailLevelToScale.size();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the (const) settings for a particular detail level and record
/// type.
///
//////////////////////////////////////////////////////////////////////////////
const DetailSettings& MapAppearanceSettings::GetDetailSettings(wxInt16 iDetailLevel, ERecordType eRT) const
{
   wxInt16 iMapIndex;
   std::map<wxInt16,DetailSettings>::const_iterator found;
   
   iMapIndex = MAPINDEX(iDetailLevel, eRT);
   
   found = m_mapDetailSettings.find(iMapIndex);
   
   wxASSERT( found != m_mapDetailSettings.end() );
   
   return (*found).second;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the (editable) settings for a particular detail level 
/// and record type.
///
//////////////////////////////////////////////////////////////////////////////
DetailSettings & MapAppearanceSettings::GetDetailSettings(wxInt16 iDetailLevel, ERecordType eRT)
{
   wxInt16 iMapIndex;
   std::map<wxInt16,DetailSettings>::iterator found;
   
   iMapIndex = MAPINDEX(iDetailLevel, eRT);
   
   found = m_mapDetailSettings.find(iMapIndex);
   
   wxASSERT( found != m_mapDetailSettings.end() );
   
   return (*found).second;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert a detail level to the corresponding map scale
///
//////////////////////////////////////////////////////////////////////////////
double MapAppearanceSettings::GetScale(wxInt16 iDetailLevel) const
{
   return m_vDetailLevelToScale[iDetailLevel];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the detai level with the largest scale less than the provided scale
///
//////////////////////////////////////////////////////////////////////////////
wxInt16 MapAppearanceSettings::GetBestDetailLevel(double scale) const
{
	std::vector<double>::const_iterator	iter;
	wxInt16	iDetailLevel = 0;

	for	( iter = m_vDetailLevelToScale.begin();	iter !=	m_vDetailLevelToScale.end(); iter++	)
	{
		double s = *iter;
		if ( s < scale )
		{
			iDetailLevel = iter	- m_vDetailLevelToScale.begin();
			break;
		}
	}

	if ( iter == m_vDetailLevelToScale.end() )
	{
		iDetailLevel = m_vDetailLevelToScale.size()-1;
	}

   return iDetailLevel;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set a color that's not tied to a particular record type or detail level
/// (e.g. background color)
///
//////////////////////////////////////////////////////////////////////////////
void MapAppearanceSettings::SetMiscColor(EMiscColor eIndex, const wxColour & clrColor)
{
	m_mapMiscColors[eIndex] = clrColor;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a color that's not tied to a particular record type or
/// detail level (e.g. background color).
///
//////////////////////////////////////////////////////////////////////////////
const wxColour & MapAppearanceSettings::GetMiscColor(EMiscColor eIndex) const
{
	std::map<EMiscColor, wxColour>::const_iterator i;

	i = m_mapMiscColors.find(eIndex);

	wxASSERT(i != m_mapMiscColors.end());

	return (*i).second;
}

void MapAppearanceSettings::FixTextColorInsideLines()
{
	int iRT;
	wxColour clrBackground;
	
	clrBackground = GetMiscColor(MiscColorBackground);
	
	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		int iDetailLevel;
		ERecordType eRT = (ERecordType) iRT;
		
		for (iDetailLevel = MIN_DETAIL_LEVEL; iDetailLevel <= MAX_DETAIL_LEVEL; iDetailLevel++)
		{
			DetailSettings& settings = GetDetailSettings(iDetailLevel, eRT);

			if (settings.m_iWidth >= settings.m_iFontSize * 2)
				settings.m_clrText = clrBackground;
		}
	}
}
