/*
 *  LibRoadnav
 *  MapControl.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// This file contains the map control class itself.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>
#include <wx/tipwin.h>
#include <wx/image.h>
#include <wx/utils.h>
#include <wx/thread.h>

#include "MapControl.h"
#include "MapDrawing.h"
#include "MapRepresentations.h"
#include "MapLookup.h"
#include "MapSupport.h"
#include "MapState.h"
#include "Terraserver.h"
#include "wxAntiAliasedDC.h"
#include "Debug.h"

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <map>

#if defined(__DARWIN__)
#define BROKEN_DRAWELLIPTICARC
#endif

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - create m_pdc based on bAntiAliased, and copy in
/// supplied arguments.
///
/// \param iWidth				Tile width
/// \param iHeight				Tile height
/// \param iX					Tile X coordinate (index)
/// \param iY					Tile Y coordinate (index)
/// \param iDetailLevel			Detail level (zoom) for this tile
/// \param fAngle				Orientation of tile
/// \param bAntiAliased			Indicates if this tile will be antialiased
///
//////////////////////////////////////////////////////////////////////////////
MapCacheEntry::MapCacheEntry(int iWidth, int iHeight, int iX, int iY, int iDetailLevel, double fAngle, bool bAntiAliased) : m_bmp(iWidth, iHeight)
{
	if (bAntiAliased)
		m_pdc = new wxAntiAliasedDC<wxMemoryDC>();
	else
		m_pdc = new wxMemoryDC();
		
	m_pdc->SelectObject(m_bmp);

	m_iDetailLevel = iDetailLevel;
	m_fAngle = fAngle;

	m_iX = iX;
	m_iY = iY;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - frees m_pdc
///
//////////////////////////////////////////////////////////////////////////////
MapCacheEntry::~MapCacheEntry()
{
	delete m_pdc;
}


DEFINE_EVENT_TYPE(wxEVT_SAFE_REFRESH)

//////////////////////////////////////////////////////////////////////////////
///
/// MapControl events
///
//////////////////////////////////////////////////////////////////////////////
BEGIN_EVENT_TABLE(MapControl, wxControl)
	EVT_PAINT(MapControl::OnPaint)
	EVT_ERASE_BACKGROUND(MapControl::OnEraseBackground)
	EVT_LEFT_DOWN(MapControl::OnLeftDown)
	EVT_LEFT_UP(MapControl::OnLeftUp)
	EVT_RIGHT_UP(MapControl::OnRightUp)
	EVT_MOTION(MapControl::OnMouseMove)
	EVT_MOUSEWHEEL(MapControl::OnMouseWheel)
	EVT_COMMAND(0, wxEVT_SAFE_REFRESH, MapControl::OnSafeRefresh)
	EVT_TIMER(TIMER_LEFTMOUSEHOLD, MapControl::OnTimerLeftMouseHold)
END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS(MapControl, wxControl);

//////////////////////////////////////////////////////////////////////////////
///
/// \brief MapControl constructor - call Init()
///
//////////////////////////////////////////////////////////////////////////////
MapControl::MapControl() : wxControl(), m_curCrossHair(wxCURSOR_CROSS), m_mxSettings(wxMUTEX_RECURSIVE)
{
	Init(NULL,NULL);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief MapControl constructor - call Init and Create
///
//////////////////////////////////////////////////////////////////////////////
MapControl::MapControl(IMapControlData * pMapControlData,const MapAppearanceSettings* pMapAppearanceSettings, wxWindow* parent, wxWindowID id, wxPoint pt, wxSize sz) : m_curCrossHair(wxCURSOR_CROSS), m_mxSettings(wxMUTEX_RECURSIVE)
{
	Init(pMapControlData,pMapAppearanceSettings);
	Create(
				parent, 
				id, 
				pt, 
				sz, 
#if wxCHECK_VERSION(2, 5, 0)
				wxNO_BORDER | wxFULL_REPAINT_ON_RESIZE
#else
				0
#endif
				);

	UpdateCrossHair();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief MapControl initialization. Sets up various variables to default settings.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::Init(IMapControlData * pMapControlData,const MapAppearanceSettings* pMapAppearanceSettings)
{
	int i;
	
	m_pMapControlData = pMapControlData;
	m_pMapAppearanceSettings = pMapAppearanceSettings;
	
	m_iMapCacheOldest = 0;
	
	for (i = 0; i < TILE_CACHE_SIZE; i++)
		m_parMapCache[i] = NULL;

	m_f3DHorizon = -100;
	m_b3D = false;

	m_pdcOldEverything = new wxMemoryDC();
	m_bmpOldEverything.Create(1, 1);
	SetAntialiasing(true);

	m_fDrawingScaleFactor = 1;
	m_iDetailLevel = 0;
	m_iControlWidth = 0;
	m_iControlHeight = 0;
#ifdef FILLED_IN_WATER_SUPPORT
	m_bFillInWater = true;
#endif
	m_bShowAerialPhoto = false;
	m_bShowCompass = true;
	m_bShowScale = true;
	m_bShowLines = true;
	m_bShowLabels = true;
	m_iMaxLineToLabelLength = 0;

	m_ptCenter.Set(0, 0);
	m_fMapAngle = 0;

	m_bAutoScale = false;
	
	m_tDataLastUpdated = wxDateTime((time_t) 0);
	m_bLeftMouseHandled = false;
	m_bLeftMouseMovingMap = false;
	m_pTimerLeftMouseHold = new wxTimer(this, TIMER_LEFTMOUSEHOLD);
	
	m_bSendMapUpdatedNotification = false;
	
	int iCacheWidth;
	int iCacheHeight;
	
	iCacheWidth = wxSystemSettings::GetMetric(wxSYS_SCREEN_X) / 2;
	iCacheHeight = wxSystemSettings::GetMetric(wxSYS_SCREEN_Y) / 2;
	
	if (iCacheWidth < 640)
		iCacheWidth = 640;
		
	if (iCacheHeight < 480)
		iCacheHeight = 480;
		
	m_szCacheBitmaps.Set(iCacheWidth, iCacheHeight);
	
	m_bLastDrawWasComplete = true;

	SetDetailLevel(12);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - free memory
///
//////////////////////////////////////////////////////////////////////////////
MapControl::~MapControl()
{
	delete m_pTimerLeftMouseHold;
	delete m_pdcOldEverything;
	ClearMapCache();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Called to erase the background of the control. Our function doesn't
/// do anything - OnPaint will draw everything
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnEraseBackground(wxEraseEvent& event)
{
	// don't do anything .. OnPaint will draw everything
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Left mouse button pressed somewhere over the control. Note the 
/// time.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnLeftDown(wxMouseEvent& event)
{
	m_bLeftMouseHandled = false;
	m_bLeftMouseMovingMap = false;
	m_evLeftMouseHold = event;

	// remember what the mouse was over when the mouse button was pressed
	m_mxSettings.Lock();
	m_ptCenterWhenLeftMouseDown = m_ptCenter;
	m_mxSettings.Unlock();

	// timer to convert a held down left click into a right click
	m_pTimerLeftMouseHold->Start(2000, TRUE);

	event.Skip();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Left mouse button released somewhere over the control. Recenter
/// the map on the click.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnLeftUp(wxMouseEvent& event)
{
	Point pt;

	m_pTimerLeftMouseHold->Stop();

	// check if left button was held down for more than 2 sec or if the map was dragged
	if (m_bLeftMouseHandled)
		return;
		
	// okay, left mouse button was pressed for less than 2 sec and the map wasn't dragged
	
	// check if there's a callback for this
	std::vector<SClickCallback>::iterator iClickCallback;
	
	for (iClickCallback = m_vClickCallbacks.begin(); iClickCallback != m_vClickCallbacks.end(); iClickCallback++)
	{
		// check if region is correct
		if (event.m_x >= (*iClickCallback).m_ptTopLeft.x &&
			event.m_y >= (*iClickCallback).m_ptTopLeft.y &&
			event.m_x <= (*iClickCallback).m_ptBottomRight.x &&
			event.m_y <= (*iClickCallback).m_ptBottomRight.y)
		{
			// check if callback wants to handle it
			if ((*iClickCallback).m_pCallback(this, (*iClickCallback).m_pArg, false, wxPoint(event.m_x, event.m_y)))
			{
				// handled - return
				return;
			}
		}
	}

	// limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_ptCenter = MapScreenToLongLat(wxPoint(event.m_x, event.m_y));
	}

	MapUpdated(UPDATED_CENTER);
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Right mouse button released somewhere over the control. Pop up a 
/// tooltip indicating the closest address.
///
/// Also sends an evEVT_COMMAND_RIGHT_CLICK to the parent window with
/// the county code in the integer portion of the message.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnRightUp(wxMouseEvent& event)
{
	wxString str;
	//wxTipWindow * pTip;
	Point pt;
	Address sAddr;
	int iCountyCode;
	int iRT;
	bool bVisibility[LAST_RECORD_TYPE + 1];
	vector<MapMarker>::iterator iMarker;

	// check if there's a callback for this
	std::vector<SClickCallback>::iterator iClickCallback;
	
	for (iClickCallback = m_vClickCallbacks.begin(); iClickCallback != m_vClickCallbacks.end(); iClickCallback++)
	{
		// check if region is correct
		if (event.m_x >= (*iClickCallback).m_ptTopLeft.x &&
			event.m_y >= (*iClickCallback).m_ptTopLeft.y &&
			event.m_x <= (*iClickCallback).m_ptBottomRight.x &&
			event.m_y <= (*iClickCallback).m_ptBottomRight.y)
		{
			// check if callback wants to handle it
			if ((*iClickCallback).m_pCallback(this, (*iClickCallback).m_pArg, true, wxPoint(event.m_x, event.m_y)))
			{
				// handled - return
				return;
			}
		}
	}

	// no callback registered for this click
	pt = MapScreenToLongLat(wxPoint(event.m_x, event.m_y));

	// figure out what county was clicked
	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		bVisibility[iRT] = ((ERecordType) iRT) == RecordTypeCountyBoundary;
	}

	sAddr = FindCoordinates(pt, bVisibility, true);

	// get it loaded
	iCountyCode = sAddr.m_iCountyCode;
	if (iCountyCode >= 0)
	{
		wxCommandEvent ev(wxEVT_COMMAND_RIGHT_CLICK, GetId());

		ev.SetInt(iCountyCode);

		GetParent()->GetEventHandler()->ProcessEvent(ev);
	}

	// now actually locate the address
	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		bVisibility[iRT] = m_pMapAppearanceSettings->GetDetailSettings(m_iDetailLevel, (ERecordType) iRT).m_bLineVisible;
	}

	/*str = sAddr.FormatAddress();

	int x, y;
	x = event.m_x;
	y = event.m_y;
	ClientToScreen(&x, &y);
	
	int iMargin;
	iMargin = 50;
	pTip = new wxTipWindow(this, str, 400);
	pTip->SetBoundingRect(wxRect(x - iMargin, y - iMargin, iMargin * 2, iMargin * 2));*/

	double fBestDistance = INFINITY;
	vector<MapMarker>::iterator iBestMarker = m_vsMarkers.end();
	
	for (iMarker = m_vsMarkers.begin(); iMarker != m_vsMarkers.end(); iMarker++)
	{
		// only consider invisible markers
		if (!(*iMarker).m_bVisible &&
			(*iMarker).m_strName != wxT("GPS Position"))
		{
			// ignore off screen invisible markers
			if (IsOnScreen((*iMarker).m_cAddress.m_ptCoordinates, true))
			{
				double fDistance;
				
				fDistance = ((*iMarker).m_cAddress.m_ptCoordinates - pt).abs();
				
				if (fDistance < fBestDistance)
				{
					fBestDistance = fDistance;
					iBestMarker = iMarker;
				}
			}
		}
	}

	double scale = m_pMapAppearanceSettings->GetScale(m_iDetailLevel);
	double fPixelDistance;

	if (m_b3D)
		scale /= 5;

	fPixelDistance = fBestDistance / (scale / m_fDrawingScaleFactor);
	
	if (iBestMarker != m_vsMarkers.end() && 
		fPixelDistance < m_iControlWidth / 5)
	{
		(*iBestMarker).m_bVisible = true;

		MapUpdated(UPDATED_MARKERS);
	}
	else
	{
		// couldn't find a nearby invisible marker .. create one
		sAddr = FindCoordinates(pt, bVisibility);
		
#ifdef LIBROADNAV_DEBUG
		IMapControlDataEntry * pcRec;
		
		m_pMapControlData->LoadRegion(pt, pt, bVisibility);
		pcRec = m_pMapControlData->GetRecordByID(sAddr.m_idRecord);
		wxASSERT(pcRec);
		if (pcRec)
			LibRoadnavDebug1(wxT("MapControl"), wxT("Right clicked on:\n%s------------\n"), pcRec->Dump().c_str());
#endif
		SetMarker(MapMarker(sAddr.FormatAddress(false, false, true), &sAddr));
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Mouse wheel moved - change detail level.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnMouseWheel(wxMouseEvent& event)
{
	if (event.GetWheelRotation() < 0)
		MoreDetail();
	if (event.GetWheelRotation() > 0)
		LessDetail();
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Control needs to be repainted - repaint it.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnPaint(wxPaintEvent& event)
{
	wxMutexLocker mlLock(m_mxSettings);
	wxPaintDC dc(this);
		
	int w, h;
	dc.GetSize(&w, &h);
	
	DrawAll(&dc, wxSize(w, h), false, true);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Show/hide the aerial photos of the area
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ShowAerialPhotos(bool bShow, bool bDownload)
{
	if (m_bShowAerialPhoto == bShow && m_bDownloadAerialPhoto == bDownload)
		return;

	// limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_bShowAerialPhoto = bShow;
		m_bDownloadAerialPhoto = bDownload;
	}

	if (m_bShowAerialPhoto)
	{
		// limit scope of locker
		{
			wxMutexLocker mlLock(m_mxSettings);
			m_fMapAngle = 0;
		}
		Enable3DMode(false);
	}
	
	MapUpdated(UPDATED_RASTERBKGD);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Show/hide the compass in the top left
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ShowCompass(bool bShow)
{
	if (m_bShowCompass == bShow)
		return;
		
	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_bShowCompass = bShow;
	}
	
	MapUpdated(UPDATED_DECORATION);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Show/hide the scale indicator in the bottom right
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ShowScale(bool bShow)
{
	if (m_bShowScale == bShow)
		return;
	
	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_bShowScale = bShow;
	}
	
	MapUpdated(UPDATED_DECORATION);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the number of detail levels.
///
//////////////////////////////////////////////////////////////////////////////
wxInt16 MapControl::GetDetailLevelCount()
{
	return m_pMapAppearanceSettings->GetDetailLevelCount();
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the detail level. 
///
/// Varies from MIN_DETAIL_LEVEL to MAX_DETAIL_LEVEL.
/// MIN_DETAIL_LEVEL provides the least detail and is zoomed out the farthest.
/// MAX_DETAIL_LEVEL provides the maximum detail and is obnoxiously zoomed in.
///
//////////////////////////////////////////////////////////////////////////////
wxInt16 MapControl::GetDetailLevel()
{
	return m_iDetailLevel;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the detail level.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetDetailLevel(wxInt16 iDetailLevel)
{
	wxInt16 iNewLevel;
	wxUint16 nDetailLevels;
	
	iNewLevel = iDetailLevel;
	nDetailLevels = m_pMapAppearanceSettings->GetDetailLevelCount();

	if (iNewLevel < 0)
		iNewLevel = 0;
	
	if (iNewLevel >= nDetailLevels)
		iNewLevel = nDetailLevels - 1;

	if ( iNewLevel != m_iDetailLevel )
	{
		// Limit scope of locker
		{
			wxMutexLocker mlLock(m_mxSettings);
			m_iDetailLevel = iNewLevel;
		}
		
		MapUpdated(UPDATED_DETAILLEVEL);
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Pick the highest detail level with a scaling factor less than
/// the supplied scaling factor
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetBestDetailLevel(double scale)
{
   wxInt16 iDetailLevel = m_pMapAppearanceSettings->GetBestDetailLevel(scale);
   SetDetailLevel(iDetailLevel);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Increase the detail level (zoom in)
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MoreDetail()
{
   SetDetailLevel(GetDetailLevel() + 1);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Decrease the detail level (zoom out)
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::LessDetail()
{
   SetDetailLevel(GetDetailLevel() - 1);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Add a track to the map (e.g. plot out a route)
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetTrack(const MapTrack & cTrack)
{
	vector<MapTrack>::iterator i;
	bool bFound = false;

	for (i = m_vcTracks.begin(); i != m_vcTracks.end(); i++)
	{
		if (i->m_strName == cTrack.m_strName)
		{
			*i = cTrack;
			bFound = true;
		}
	}

	if (!bFound)
		m_vcTracks.push_back(cTrack);
	
	MapUpdated(UPDATED_TRACKS);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Fetch a specific track by name
///
//////////////////////////////////////////////////////////////////////////////
MapTrack * MapControl::GetTrack(const wxString & strName)
{
	vector<MapTrack>::iterator i;

	for (i = m_vcTracks.begin(); i != m_vcTracks.end(); i++)
	{
		if (i->m_strName == strName)
		{
			return &(*i);
		}
	}

	return NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clear all tracks from the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ClearTracks()
{
	m_vcTracks.clear();

	MapUpdated(UPDATED_TRACKS);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clear all tracks from the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ClearTrack(const wxString & strName)
{
	vector<MapTrack>::iterator iter;
	
	for (iter = m_vcTracks.begin(); iter != m_vcTracks.end(); iter++)
	{
		if (iter->m_strName == strName)
		{
			m_vcTracks.erase(iter);

			MapUpdated(UPDATED_TRACKS);
			
			return;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Autoscale the map so the tracks are completely visible
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::AutoScale()
{
	m_bAutoScale = true;

	SafeRefresh();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Change the center coordinates of the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetCenterCoordinates(Point pt)
{
	if (m_ptCenter == pt)
		return;
		
	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_ptCenter = pt;
	}

	MapUpdated(UPDATED_CENTER);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the center coordinates of the map
///
//////////////////////////////////////////////////////////////////////////////
Point MapControl::GetCenterCoordinates()
{
	return m_ptCenter;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Move the map to the left
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MoveCenterLeft(double fScreens)
{
	if (m_b3D)
		SetCenterCoordinates(MapScreenToLongLat(wxPoint((int) (m_iControlWidth / 2 - fScreens * m_iControlWidth), m_iControlHeight)));
	else
		SetCenterCoordinates(MapScreenToLongLat(wxPoint((int) (m_iControlWidth / 2 - fScreens * m_iControlWidth), m_iControlHeight / 2)));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Move the map to the right
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MoveCenterRight(double fScreens)
{
	if (m_b3D)
		SetCenterCoordinates(MapScreenToLongLat(wxPoint((int) (m_iControlWidth / 2 + fScreens * m_iControlWidth), m_iControlHeight)));
	else
		SetCenterCoordinates(MapScreenToLongLat(wxPoint((int) (m_iControlWidth / 2 + fScreens * m_iControlWidth), m_iControlHeight / 2)));
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Move the map up
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MoveCenterUp(double fScreens)
{
	if (m_b3D)
		SetCenterCoordinates(MapScreenToLongLat(wxPoint(m_iControlWidth / 2, (int) (m_iControlHeight - fScreens * m_iControlHeight))));
	else
		SetCenterCoordinates(MapScreenToLongLat(wxPoint(m_iControlWidth / 2, (int) (m_iControlHeight / 2 - fScreens * m_iControlHeight))));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Move the map down
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MoveCenterDown(double fScreens)
{
	if (m_b3D)
	{
		// down is off screen in this case, so calculate it as the opposite of moving up
		SetCenterCoordinates(m_ptCenter + (m_ptCenter - MapScreenToLongLat(wxPoint(m_iControlWidth / 2, (int) (m_iControlHeight - fScreens * m_iControlHeight)))));
	}
	else
		SetCenterCoordinates(MapScreenToLongLat(wxPoint(m_iControlWidth / 2, (int) (m_iControlHeight / 2 + fScreens * m_iControlHeight))));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the coordinates of the marker
///
//////////////////////////////////////////////////////////////////////////////
MapMarker & MapControl::GetMarker(int iIndex)
{
	return m_vsMarkers[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns how many markers are in use
///
//////////////////////////////////////////////////////////////////////////////
int MapControl::GetMarkerCount()
{
	return m_vsMarkers.size();
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clears all markers
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ClearMarkers()
{
	m_vsMarkers.clear();
	
	MapUpdated(UPDATED_MARKERS);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clears a marker by name
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ClearMarker(wxString strName)
{
	vector<MapMarker>::iterator iter;
	
	for (iter = m_vsMarkers.begin(); iter != m_vsMarkers.end(); iter++)
	{
		if (iter->m_strName == strName)
		{
			m_vsMarkers.erase(iter);

			MapUpdated(UPDATED_MARKERS);
			
			return;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Gets a marker by name
///
//////////////////////////////////////////////////////////////////////////////
MapMarker & MapControl::GetMarker(const wxString & strName)
{
	vector<MapMarker>::iterator iter;
	static MapMarker cNull;
	
	for (iter = m_vsMarkers.begin(); iter != m_vsMarkers.end(); iter++)
	{
		if (iter->m_strName == strName)
		{
			return *iter;
		}
	}
	
	return cNull;
}

/////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets a marker
///
/// If a marker with the specified name exists, it will be updated. Otherwise
/// a new marker will be created.
///
//////////////////////////////////////////////////////////////////////////////
int MapControl::SetMarker(const MapMarker & cMarker, bool fEnsureOnScreen)
{
	int i;
	vector<MapMarker>::iterator iMarker;
	
	if (!IsOnScreen(cMarker.m_cAddress.m_ptCoordinates, false) && fEnsureOnScreen)
	{
		SetCenterCoordinates(cMarker.m_cAddress.m_ptCoordinates);
	}
	
	for (iMarker = m_vsMarkers.begin(), i = 0; iMarker != m_vsMarkers.end(); iMarker++, i++)
	{
		if ((*iMarker).m_strName == cMarker.m_strName)
		{
			*iMarker = cMarker;
			break;
		}
	}
	if (iMarker == m_vsMarkers.end())
		m_vsMarkers.push_back(cMarker);
	
	MapUpdated(UPDATED_MARKERS);

	SafeRefresh();
	
	return i;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Map has been updated in some way. Do a complete redraw.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::MapUpdated(int iUpdatedFlags)
{
	if (iUpdatedFlags & UPDATED_CENTER)
	{
		// no need to do anything
	}
	
	if (iUpdatedFlags & UPDATED_DETAILLEVEL)
	{
		// no need to do anything
	}
	
	if (iUpdatedFlags & UPDATED_RASTERBKGD)
	{
		ClearMapCache();
	}
	
	if (iUpdatedFlags & UPDATED_DECORATION)
	{
		// no need to do anything
	}
	
	if (iUpdatedFlags & UPDATED_MARKERS)
	{
		// no need to do anything
	}
	
	if (iUpdatedFlags & UPDATED_ANGLE)
	{
		// no need to do anything
	}
	
	if (iUpdatedFlags & UPDATED_3D)
	{
		// no need to do anything
	}

	if (iUpdatedFlags & UPDATED_THEME)
	{
		ClearMapCache();
		UpdateCrossHair();
	}

	if (iUpdatedFlags & UPDATED_DATA)
	{
		ClearMapCache();
	}
	
	if (iUpdatedFlags & UPDATED_TRACKS)
	{
		// no need to do anything
	}
	
	m_ptTopLeft = MapScreenToLongLat(wxPoint(0, 0));
	m_ptBottomRight = MapScreenToLongLat(wxPoint(m_iControlWidth - 1, m_iControlHeight - 1));

	m_bSendMapUpdatedNotification = true;

	SafeRefresh();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert screen coordinates to longitude/latitude.
///
/// This function may not use m_ptTopLeft or m_ptBottomRight
/// since it is called to initialize them
///
//////////////////////////////////////////////////////////////////////////////
Point MapControl::MapScreenToLongLat(const wxPoint & ptScreen)
{
	Point sRtn;
	wxPoint ptScreenModRotated;
	wxPoint ptScreenMod(ptScreen);
	
	double scale = m_pMapAppearanceSettings->GetScale(m_iDetailLevel);

	if (m_b3D)
	{
		double fDistFromCenterX = ptScreenMod.x - m_iControlWidth / 2; // X distance from center of screen
		double fDistFromHorizon = ptScreenMod.y - m_f3DHorizon; // Y distance from horizon
		double fVisibleRange = m_iControlHeight - m_f3DHorizon; // distance from bottom of screen to horizon

		scale /= 5;

		ptScreenMod.x = (int) floor(fDistFromCenterX / ( fDistFromHorizon / fVisibleRange ) + m_iControlWidth / 2); // unsqueeze the X axis

		double fDistFromBottom = m_iControlHeight - ptScreenMod.y; // distance from bottom of screen
		double fD = fDistFromBottom / ( 1.0 - fDistFromBottom / fVisibleRange); // inverse Y squeezing formula

		ptScreenMod.y = (int) floor(m_iControlHeight / 2 - fD); // center on screen
	}

	if (m_fMapAngle != 0)
		ptScreenModRotated = RotatePoint(ptScreenMod, wxPoint(m_iControlWidth / 2, m_iControlHeight / 2), m_fMapAngle);
	else
		ptScreenModRotated = ptScreenMod;

	sRtn.SetLat(m_ptCenter.GetLat() + -1 * (ptScreenModRotated.y - m_iControlHeight / 2) * scale / m_fDrawingScaleFactor);
	sRtn.SetLong(m_ptCenter.GetLong() + (ptScreenModRotated.x - m_iControlWidth / 2) * scale / m_fDrawingScaleFactor);

	return sRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert longitude/latitude to screen coordinates
///
//////////////////////////////////////////////////////////////////////////////
wxPoint MapControl::MapLongLatToScreen(const Point & pt)
{
	wxPoint ptRtn;
	double scale = m_pMapAppearanceSettings->GetScale(m_iDetailLevel);

	if (m_b3D)
	{
		scale /= 5;
	}

	ptRtn.x = (int) floor(m_iControlWidth / 2 + (pt.GetLong() - m_ptCenter.GetLong()) / (scale / m_fDrawingScaleFactor));
	ptRtn.y = (int) floor(m_iControlHeight / 2 - (pt.GetLat() - m_ptCenter.GetLat()) / (scale / m_fDrawingScaleFactor));

	if (m_fMapAngle != 0)
		ptRtn = RotatePoint(ptRtn, wxPoint(m_iControlWidth / 2, m_iControlHeight / 2), -m_fMapAngle);

	if (m_b3D)
	{
		double fDistFromCenterY = m_iControlHeight / 2 - ptRtn.y; // how far away is this point along the Y axis
		double fVisibleRange = m_iControlHeight - m_f3DHorizon; // how far from the bottom of the screen is the horizon

		ptRtn.y = (int) floor(m_iControlHeight - fDistFromCenterY / ( fabs(fDistFromCenterY / fVisibleRange) + 1 )); // make the Y coordinate converge on the horizon as the distance from the center goes to infinity

		double fDistFromCenterX = ptRtn.x - m_iControlWidth / 2; // X distance from center of the screen
		double fDistFromHorizon = ptRtn.y - m_f3DHorizon; // distance from the horizon after adjusting for perspective

		ptRtn.x = (int) floor(fDistFromCenterX * ( fDistFromHorizon / fVisibleRange ) + m_iControlWidth / 2); // squeeze the X axis, make it a point at the horizon and normal sized at the bottom of the screen
	}

	return ptRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determine if a certain longitude/latitude coordinate is on the screen
///
/// \param pt				Coordinates to check.
///
/// \param bQuick			true = faster, less accurate check - sometimes
///							returns false positives. false = slower, more
///							accurate check - no false negatives/positives.
///
/// \return					If the specified coordinates are visible.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControl::IsOnScreen(const Point & pt, bool bQuick)
{
	if (bQuick)
	{
		if (pt.GetLongRaw() >= m_ptTopLeft.GetLongRaw() &&
			pt.GetLongRaw() <= m_ptBottomRight.GetLongRaw() &&
			pt.GetLatRaw() >= m_ptBottomRight.GetLatRaw() &&
			pt.GetLatRaw() <= m_ptTopLeft.GetLatRaw())
			return true;
	}
	else
	{
		wxPoint ptScreen;

		ptScreen = MapLongLatToScreen(pt);

		if (ptScreen.x < 0 || ptScreen.y < 0 || ptScreen.x >= m_iControlWidth || ptScreen.y >= m_iControlHeight)
			return false;

		return true;
	}

	return false; 
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw only the edges of the water
///
//////////////////////////////////////////////////////////////////////////////
#ifdef FILLED_IN_WATER_SUPPORT
void MapControl::DrawMapWaterBoundaries(wxDC * dc, bool bScreen)
{
	if (!m_bFillInWater)
		return;

	int iRec;
	// draw the water boundaries
	for (iRec = 0; iRec < m_pMapControlData->Count(); iRec++)
	{
		IMapControlDataEntry * psRec = m_pMapControlData->GetRecord(iRec);
		bool bVisible = false;

		{
			int iSeg;

			for (iSeg = 0; iSeg < psRec->nCoordinates; iSeg++)
			{
				if (IsOnScreen(psRec->psCoordinates[iSeg]))
					bVisible = true;
			}
		}

		if (bVisible)
		{
			if ((psRec->eType == RecordTypeWater || (!psRec->bWaterL && !psRec->bWaterR)) && psRec->eType != RecordTypeInvisibleWaterBoundary && psRec->eType != RecordTypeInvisibleLandBoundary && psRec->eType != RecordTypeInvisibleMiscBoundary)
			{
				wxPoint * parPoints;
				wxColour clrLine = wxColour(254, 254, 254);
				int iWidth = 1;
				int iStyle = wxSOLID;
				int iSeg;

				parPoints = new wxPoint [psRec->nCoordinates];

				for (iSeg = 0; iSeg < psRec->nCoordinates; iSeg++)
					parPoints[iSeg] = MapLongLatToScreen(psRec->psCoordinates[iSeg]);

				for (iSeg = 0; iSeg < psRec->nCoordinates - 1; iSeg++)
				{
					DrawLine(dc, 
								parPoints[iSeg].x, 
								parPoints[iSeg].y, 
								parPoints[iSeg + 1].x, 
								parPoints[iSeg + 1].y,
								(int) floor(iWidth * m_fDrawingScaleFactor),
								clrLine,
								iStyle);
				}

				delete parPoints;

			}

		}
	}
}
#endif

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Fill in the water polygons
///
/// Seems very slow under X, but okay under Windows. Not terribly reliable.
///
//////////////////////////////////////////////////////////////////////////////
#ifdef FILLED_IN_WATER_SUPPORT
void MapControl::DrawMapWater(wxDC * dc, bool bScreen)
{
	if (!m_bFillInWater)
		return;
	
	int iRec;
	
	// fill in water
	for (iRec = 0; iRec < m_pMapControlData->Count(); iRec++)
	{
		IMapControlDataEntry * psRec = m_pMapControlData->GetRecord(iRec);

		if (psRec->bWaterL)
		{
			Point ptLongLat = psRec->ptWaterL;

			if (IsOnScreen(ptLongLat))
			{
				wxPoint ptScreen = MapLongLatToScreen(ptLongLat); 
				wxBrush br(wxColour(192, 192, 255), wxSOLID);
				dc->SetBrush(br);
				dc->FloodFill(ptScreen.x, ptScreen.y, wxColour(255, 255, 255));
				dc->SetBrush(wxNullBrush);
			}
		}
		if (psRec->bWaterR)
		{
			Point ptLongLat = psRec->ptWaterR;

			if (IsOnScreen(ptLongLat))
			{
				wxPoint ptScreen = MapLongLatToScreen(ptLongLat); 
				wxBrush br(wxColour(192, 192, 255), wxSOLID);
				dc->SetBrush(br);
				dc->FloodFill(ptScreen.x, ptScreen.y, wxColour(255, 255, 255));
				dc->SetBrush(wxNullBrush);
			}
		}
		
		/*if (psRec->bWaterL)
		{
			if (psRec->ContainsPoint(psRec->ptWaterL) || 1)
			{
				wxPoint * parPoints;
				int iPoint;
				
				parPoints = new wxPoint[psRec->nCoordinates];
				
				for (iPoint = 0; iPoint < psRec->nCoordinates; iPoint++)
				{
					parPoints[iPoint] = MapLongLatToScreen(psRec->psCoordinates[iPoint]);
				}
				
				wxBrush br(wxColour(192, 192, 255), wxSOLID);
				dc->SetBrush(br);
				dc->DrawPolygon(psRec->nCoordinates, parPoints);
				dc->SetBrush(wxNullBrush);
				
				delete parPoints;
			}
		}
		if (psRec->bWaterR)
		{
			if (psRec->ContainsPoint(psRec->ptWaterR) || 1)
			{
				wxPoint * parPoints;
				int iPoint;
				
				parPoints = new wxPoint[psRec->nCoordinates];
				
				for (iPoint = 0; iPoint < psRec->nCoordinates; iPoint++)
				{
					parPoints[iPoint] = MapLongLatToScreen(psRec->psCoordinates[iPoint]);
				}
				
				wxBrush br(wxColour(192, 192, 255), wxSOLID);
				dc->SetBrush(br);
				dc->DrawPolygon(psRec->nCoordinates, parPoints);
				dc->SetBrush(wxNullBrush);
				
				delete parPoints;
			}
		}*/
	}
}
#endif

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Structure used by DrawMapGeneral to organize the labels to be drawn
///
//////////////////////////////////////////////////////////////////////////////
struct SLabelInfo
{
	/// starting coordinate of segment
	Point ptStart; 
	
	/// ending coordinate of segment
	Point ptEnd; 
	
	/// ptStart and ptEnd must be within fMinLabelSpacing/3 of ptOrigin. If a line segment is found farther than 
	/// fMinLabelSpacing/1 of ptOrigin, a new SLabelInfo will be created.
	Point ptOrigin; 
	
	/// length of segment
	double fDistance; 
	
	/// detail settings for this label
	DetailSettings pMap; 
	
	/// indicates if this SLabelInfo has been drawn yet.
	bool bDrawn; 
};

typedef vector<SLabelInfo> VSLabelInfo;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw all features of the map, except for water.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawMapGeneral(wxDC * dc, bool bScreen)
{
	map<wxString,VSLabelInfo> mapLabels;
	map<wxString,VSLabelInfo>::iterator itLabels;
	int iRec;

	double scale = m_pMapAppearanceSettings->GetScale(m_iDetailLevel);

	if (m_b3D)
		scale /= 5;

	// minimum spacing between redundant street labels/markers
	double fMinLabelSpacing = Distance(m_ptViewTopLeft, m_ptViewBottomRight) / 6;
	
	// lower limit
	if (fMinLabelSpacing < 0.01)
		fMinLabelSpacing = 0.01;

	mapLabels.clear();

	for (iRec = 0; iRec < m_pMapControlData->Count(); iRec++)
	{
		IMapControlDataEntry * psRec = m_pMapControlData->GetRecord(iRec);
		bool bVisible = false;

		bVisible = psRec->IntersectsRect(m_ptTopLeft, m_ptBottomRight);

		if (bVisible)
		{
			const DetailSettings& pMap = m_pMapAppearanceSettings->GetDetailSettings(m_iDetailLevel,psRec->eType);
			
			if (pMap.m_bLineVisible && m_bShowLines)
			{
				wxPoint * parPoints;
				wxColour clrLine = pMap.m_clrLine;
				int iWidth = pMap.m_iWidth;
				int iStyle = pMap.m_iStyle;
				int iArrowSpacing = pMap.m_iArrowSpacing;
				bool bRailroadStyle = pMap.m_bRailroadStyle;
				int iSeg;

				// this doesn't seem to work at all, so don't bother
				/*{
					int iLeftZip = 0;
					int iRightZip = 0;
					
					int i;
					
					for (i = 0; i < psRec->nAddressZipRanges; i++)
					{
						if (psRec->psAddressZipRanges[i].iZip)
						{
							if (psRec->psAddressZipRanges[i].bOnLeft)
								iLeftZip = psRec->psAddressZipRanges[i].iZip;
							else
								iRightZip = psRec->psAddressZipRanges[i].iZip;
						}
						
						if (iLeftZip && iRightZip && StateByZip(iLeftZip) != StateByZip(iRightZip))
							iWidth *= 4;
					}
				}*/

				
				parPoints = new wxPoint [psRec->nCoordinates];

				for (iSeg = 0; iSeg < psRec->nCoordinates; iSeg++)
				{
					parPoints[iSeg] = MapLongLatToScreen(psRec->psCoordinates[iSeg]);
				}
				
				if (psRec->nCoordinates == 1)
				{
					// a place
					wxPen penThis(clrLine, 1, iStyle);
					wxBrush brThis(clrLine, iStyle);
					
					dc->SetPen(penThis);
					dc->SetBrush(brThis);
					
					dc->DrawCircle(parPoints[0].x, parPoints[0].y, (int) floor(iWidth * m_fDrawingScaleFactor));
					
					dc->SetPen(wxNullPen);
					dc->SetBrush(wxNullBrush);
				}
				else
				{
					// a polygon
					if (pMap.m_bFastLine)
					{
						DrawLine(dc, 
									parPoints[0].x, 
									parPoints[0].y, 
									parPoints[psRec->nCoordinates - 1].x, 
									parPoints[psRec->nCoordinates - 1].y,
									(int) floor(iWidth * m_fDrawingScaleFactor),
									clrLine,
									iStyle,
									(int) floor(iArrowSpacing * m_fDrawingScaleFactor),
									bRailroadStyle ? iWidth * 2 : 0,
									bRailroadStyle ? iWidth * 6 : 0
						);
					}
					else
					{
						for (iSeg = 0; iSeg < psRec->nCoordinates - 1; iSeg++)
						{
							DrawLine(dc, 
										parPoints[iSeg].x, 
										parPoints[iSeg].y, 
										parPoints[iSeg + 1].x, 
										parPoints[iSeg + 1].y,
										(int) floor(iWidth * m_fDrawingScaleFactor),
										clrLine,
										iStyle,
										(int) floor(iArrowSpacing * m_fDrawingScaleFactor),
										bRailroadStyle ? iWidth * 2 : 0,
										bRailroadStyle ? iWidth * 6 : 0
							);
						}
					}
				}

				delete [] parPoints;

			}

			if (pMap.m_bTextVisible && m_bShowLabels)
			{
				wxString strFullName;
				int iOffset;

				strFullName = psRec->SmartName();
				
				if (strFullName != wxT("") && strFullName != wxT(" "))
				{
					if (psRec->nCoordinates > 1)
					{
						for (iOffset = 0; iOffset < psRec->nCoordinates - 1; iOffset++)
						{
							if (IsOnScreen(psRec->psCoordinates[iOffset]) && IsOnScreen(psRec->psCoordinates[iOffset + 1], false))
							{
								vector<SLabelInfo>::iterator iLabelEntry;
								vector<SLabelInfo> & vLabels = mapLabels[strFullName];
								SLabelInfo * pMatchingLabel;
								
								for (iLabelEntry = vLabels.begin(); iLabelEntry != vLabels.end(); iLabelEntry++)
								{
									if (Distance((*iLabelEntry).ptOrigin, psRec->psCoordinates[iOffset]) < fMinLabelSpacing)
										break;
								}
								
								if (iLabelEntry == vLabels.end())
								{
									SLabelInfo sBlank;
									
									sBlank.bDrawn = false;
									sBlank.fDistance = 0;
									sBlank.ptOrigin = psRec->psCoordinates[iOffset];
									
									vLabels.push_back(sBlank);
									
									pMatchingLabel = &vLabels[vLabels.size() - 1];
								}
								else
								{
									pMatchingLabel = &(*iLabelEntry);
								}
								
								// try to put the text on the longest segment we can find - it will look best that way
								if (Distance(psRec->psCoordinates[iOffset], psRec->psCoordinates[iOffset + 1]) >= pMatchingLabel->fDistance &&
									Distance(pMatchingLabel->ptOrigin, psRec->psCoordinates[iOffset]) < fMinLabelSpacing / 3)
								{
									pMatchingLabel->fDistance = Distance(psRec->psCoordinates[0], psRec->psCoordinates[psRec->nCoordinates - 1]);
									pMatchingLabel->ptStart = psRec->psCoordinates[iOffset];
									pMatchingLabel->ptEnd = psRec->psCoordinates[iOffset + 1];
									pMatchingLabel->pMap = pMap;
									pMatchingLabel->bDrawn = false;
								}
							}
						}
					}
					else
					{
						// psRec->nCoordinates == 1, landmark
						SLabelInfo sInfo;
						
						sInfo.fDistance = 0;
						sInfo.ptStart = psRec->psCoordinates[0];
						sInfo.ptEnd = psRec->psCoordinates[0];
						sInfo.pMap = pMap;
						sInfo.bDrawn = false;

						mapLabels[strFullName].clear();
						mapLabels[strFullName].push_back(sInfo);
					}
				}
			}
		}
	}

	vector<wxRect> vTextRects;

	//////////////////////////////////////////////////////////////////////////
	// draw the highway symbols
	//////////////////////////////////////////////////////////////////////////
	for (itLabels = mapLabels.begin(); itLabels != mapLabels.end(); itLabels++)
	{
		VSLabelInfo::iterator iLabelEntry;
		wxString strText;

		strText = (*itLabels).first;
		
		for (iLabelEntry = (*itLabels).second.begin(); iLabelEntry != (*itLabels).second.end(); iLabelEntry++)
		{
			wxColour clrText;
			int iFontSize;
			wxCoord cFontHeight, cFontWidth;
			wxPoint p1, p2, p;
			Point ptStart, ptEnd;
			bool bDrawn;
			
			bDrawn = true;
			
			clrText = (*iLabelEntry).pMap.m_clrText;
			iFontSize = (*iLabelEntry).pMap.m_iFontSize;
			
			if (iFontSize < (*iLabelEntry).pMap.m_iWidth)
				iFontSize = (*iLabelEntry).pMap.m_iWidth;

			wxFont * pFont = wxFont::New((int) (iFontSize * m_fDrawingScaleFactor), (wxFontFamily) wxSWISS, wxFONTFLAG_ANTIALIASED);
			dc->SetFont(*pFont);
			dc->GetTextExtent(strText, &cFontWidth, &cFontHeight);

			ptStart = (*iLabelEntry).ptStart;
			ptEnd = (*iLabelEntry).ptEnd;

			p1 = MapLongLatToScreen(ptStart);
			p2 = MapLongLatToScreen(ptEnd);

			p.x = (p1.x + p2.x) / 2;
			p.y = (p1.y + p2.y) / 2;
			
			if (strText.Left(2) == wxT("I-"))
			{
				// Interstate
				// route number inside a blue shield with a red hat
				
				wxString strRouteNumber;
				wxPen penOutline;
				double fEllipseScaling = 1.3;
				int iEllipseWidth;
				int iEllipseHeight;
				
				strRouteNumber = strText.Mid(2);
				dc->GetTextExtent(strRouteNumber, &cFontWidth, &cFontHeight);
				
				iEllipseWidth = (int) (cFontWidth * fEllipseScaling);
				iEllipseHeight = (int) (cFontHeight * fEllipseScaling);
				if (iEllipseWidth < iEllipseHeight * 1.1)
					iEllipseWidth = (int) (iEllipseHeight * 1.1);
					
				if (p.x - iEllipseWidth * 2 > 0 &&
					p.y - iEllipseHeight * 2 > 0 &&
					p.x + iEllipseWidth * 2 < m_iControlWidth &&
					p.y + iEllipseHeight * 2 < m_iControlHeight)
				{
#ifdef BROKEN_DRAWELLIPTICARC
					wxPoint arPoints[3];
#endif
					
					penOutline = *wxBLACK_PEN;
					dc->SetPen(penOutline);

					dc->SetTextForeground(*wxWHITE);
					
					dc->SetBrush(*wxBLUE_BRUSH);
					dc->SetBackground(*wxBLUE_BRUSH);
					/*dc->DrawEllipse(	p.x - iEllipseWidth / 2,
										p.y - iEllipseHeight / 2,
										iEllipseWidth,
										iEllipseHeight);*/
					
					dc->DrawEllipticArc(	(int) (p.x - iEllipseWidth * 1.5),
											(int) (p.y - iEllipseHeight * 1.5 - iEllipseHeight * 0.1),
											(int) (iEllipseWidth * 2),
											(int) (iEllipseHeight * 2.4),
											300,
											360
											);
					dc->DrawEllipticArc(	(int) (p.x - iEllipseWidth * 0.5),
											(int) (p.y - iEllipseHeight * 1.5 - iEllipseHeight * 0.1),
											(int) (iEllipseWidth * 2),
											(int) (iEllipseHeight * 2.4),
											180,
											240
											);

#ifdef BROKEN_DRAWELLIPTICARC
					wxPen penBlue(*wxBLUE, 4);
					dc->SetBrush(*wxBLUE_BRUSH);
					dc->SetPen(penBlue);
					/* Workaround for a bug in wxDC::DrawEllipticArc on OSX */
					arPoints[0] = wxPoint((p.x - iEllipseWidth * 0.5), (p.y - iEllipseHeight * 0.45));
					arPoints[1] = wxPoint((p.x + iEllipseWidth * 0.5), (p.y - iEllipseHeight * 0.45));
					arPoints[2] = wxPoint((p.x), (p.y + iEllipseHeight * 0.6));
					dc->DrawPolygon(3, arPoints);
#endif

					dc->SetBrush(*wxRED_BRUSH);
					dc->SetPen(penOutline);
					dc->DrawRectangle(	
										p.x - iEllipseWidth / 2,
										(int) (p.y - iEllipseHeight / 2 - iEllipseHeight * 0.1),
										iEllipseWidth,
										iEllipseHeight / 4
										);
					dc->DrawText(		strRouteNumber,
										p.x - cFontWidth / 2,
										p.y - cFontHeight / 2	);
					
					dc->SetPen(wxNullPen);
					dc->SetBrush(wxNullBrush);
				
					dc->SetBackground(m_brBackground);
					
					vTextRects.push_back(wxRect(p.x - iEllipseWidth * 2, p.y - iEllipseHeight * 2, iEllipseWidth * 4, iEllipseHeight * 3));
				}
			}
			else if (strText.Left(3) == wxT("US-"))
			{
				// US highway
				// route number inside a white shield
				
				wxString strRouteNumber;
				wxPen penOutline(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground), (int) m_fDrawingScaleFactor, wxSOLID);
				wxBrush brCircleFill;
				double fSymbolScaling = 1.2;
				int iSymbolWidth;
				int iSymbolHeight;
				
				strRouteNumber = strText.Mid(3);
				dc->GetTextExtent(strRouteNumber, &cFontWidth, &cFontHeight);
				
				iSymbolWidth = (int) (cFontWidth * fSymbolScaling);
				iSymbolHeight = (int) (cFontHeight * fSymbolScaling);
				if (iSymbolWidth < iSymbolHeight * 1.1)
					iSymbolWidth = (int) (iSymbolHeight * 1.1);
					
				if (p.x - iSymbolWidth > 0 &&
					p.y - iSymbolHeight > 0 &&
					p.x + iSymbolWidth < m_iControlWidth &&
					p.y + iSymbolHeight < m_iControlHeight)
				{
					dc->SetPen(penOutline);

					brCircleFill = m_brBackground;
					dc->SetBrush(brCircleFill);
					
					dc->SetTextForeground(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground));
					
					dc->DrawRoundedRectangle(	p.x - iSymbolWidth / 2,
										p.y - iSymbolHeight / 2,
										iSymbolWidth,
										iSymbolHeight,
										iSymbolHeight / 8);
										
					dc->SetBackground(brCircleFill);
					dc->DrawText(		strRouteNumber,
										p.x - cFontWidth / 2,
										p.y - cFontHeight / 2	);
					
					dc->SetPen(wxNullPen);
					dc->SetBrush(wxNullBrush);
					dc->SetBackground(m_brBackground);

					vTextRects.push_back(wxRect(p.x - iSymbolWidth / 2, p.y - iSymbolHeight / 2, iSymbolWidth, iSymbolHeight));
				}
			}
			else if (strText.Length() >= 3 && strText.GetChar(2) == '-' && (StateNameByAbbreviation(strText.Left(2)) != wxT("") || strText.Left(2) == wxT("SR")))
			{
				// state route
				// route number in a white circle
				
				wxString strRouteNumber;
				wxPen penOutline(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground), (int) m_fDrawingScaleFactor, wxSOLID);
				wxBrush brCircleFill;
				double fEllipseScaling = 1.2;
				int iEllipseWidth;
				int iEllipseHeight;
				
				strRouteNumber = strText.Mid(3);
				dc->GetTextExtent(strRouteNumber, &cFontWidth, &cFontHeight);
				
				iEllipseWidth = (int) (cFontWidth * fEllipseScaling);
				iEllipseHeight = (int) (cFontHeight * fEllipseScaling);
				if (iEllipseWidth < iEllipseHeight * 1.1)
					iEllipseWidth = (int) (iEllipseHeight * 1.1);
				
				if (p.x - iEllipseWidth > 0 &&
					p.y - iEllipseHeight > 0 &&
					p.x + iEllipseWidth < m_iControlWidth &&
					p.y + iEllipseHeight < m_iControlHeight)
				{
					dc->SetPen(penOutline);

					brCircleFill = m_brBackground;
					dc->SetBrush(brCircleFill);
					
					dc->SetTextForeground(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground));
					
					dc->DrawEllipse(	p.x - iEllipseWidth / 2,
										p.y - iEllipseHeight / 2,
										iEllipseWidth,
										iEllipseHeight);
										
					dc->SetBackground(brCircleFill);
					dc->DrawText(		strRouteNumber,
										p.x - cFontWidth / 2,
										p.y - cFontHeight / 2	);
					
					dc->SetPen(wxNullPen);
					dc->SetBrush(wxNullBrush);
					dc->SetBackground(m_brBackground);

					vTextRects.push_back(wxRect(p.x - iEllipseWidth / 2, p.y - iEllipseHeight / 2, iEllipseWidth, iEllipseHeight));
				}
			}
			else
			{
				bDrawn = false;
			}
			
			(*iLabelEntry).bDrawn = bDrawn;
			
			dc->SetFont(wxNullFont);
			delete pFont;
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// draw the text now
	//////////////////////////////////////////////////////////////////////////
	for (itLabels = mapLabels.begin(); itLabels != mapLabels.end(); itLabels++)
	{
		VSLabelInfo::iterator iLabelEntry;
		wxString strText;

		strText = (*itLabels).first;
		
		for (iLabelEntry = (*itLabels).second.begin(); iLabelEntry != (*itLabels).second.end(); iLabelEntry++)
		{
			wxColour clrText;
			int iFontSize;
			wxCoord cFontHeight, cFontWidth;
			wxPoint p1, p2, p;
			Point ptStart, ptEnd;
			double fAngle;
			
			clrText = (*iLabelEntry).pMap.m_clrText;
			iFontSize = (*iLabelEntry).pMap.m_iFontSize;

			wxFont * pFont = wxFont::New((int) (iFontSize * m_fDrawingScaleFactor), (wxFontFamily) wxSWISS, wxFONTFLAG_ANTIALIASED | wxFONTFLAG_BOLD);
			dc->SetFont(*pFont);
			dc->GetTextExtent(strText, &cFontWidth, &cFontHeight);

			ptStart = (*iLabelEntry).ptStart;
			ptEnd = (*iLabelEntry).ptEnd;

			p1 = MapLongLatToScreen(ptStart);
			p2 = MapLongLatToScreen(ptEnd);

			p.x = (p1.x + p2.x) / 2;
			p.y = (p1.y + p2.y) / 2;
			
			if (!(*iLabelEntry).bDrawn)
			{
				int iTextOffsetY = 0;
				bool bTextInsideLine = false;
				
				if (p1 != p2)
					fAngle = ::atan2((double) (p2 - p1).x, (double) (p2 - p1).y) - PI / 2;
				else
					fAngle = 0;
		
				if (fAngle > PI / 2)
					fAngle -= PI;
		
				if (fAngle < -PI / 2)
					fAngle += PI;
		
				//////////////////////////////////////////////////////////////////////
				// center the text on this point
				//////////////////////////////////////////////////////////////////////
				p.x -= (int) ( cos(fAngle) * cFontWidth / 2 );
				p.y += (int) ( sin(fAngle) * cFontWidth / 2 );
		
				bTextInsideLine = cFontHeight < (*iLabelEntry).pMap.m_iWidth * m_fDrawingScaleFactor;
				if (bTextInsideLine)
				{
					// shift the line up half a line height so its vertical center 
					// passes through the line's center.
					// if this isn't done, then its top will pass through the
					// line's center.
					iTextOffsetY = -1 * cFontHeight / 2; 
				}
				else
				{
					// slide the line down half a line width so it's
					// always right on the edge of the line
					iTextOffsetY = (int) ((*iLabelEntry).pMap.m_iWidth * m_fDrawingScaleFactor / 2);
				}
	
				p.x += (int) ( sin(fAngle) * iTextOffsetY );
				p.y += (int) ( cos(fAngle) * iTextOffsetY );

				wxRect rectThis;
				
				vector<wxRect>::iterator i;
				bool bCollision;
				
				int iOffsetX = 0;
				int iOffsetY = 0;
				
				bCollision = true;
				
				while (bCollision && iOffsetX <= m_iMaxLineToLabelLength && iOffsetY <= m_iMaxLineToLabelLength)
				{
					bCollision = false;
					
					// coordinates of the top left corner of the text
					wxPoint ptTextTopLeft;
					wxPoint pt;
					
					// these points form the top-left and bottom-right corners
					// of a rectangle bounding the text we're trying to draw
					wxPoint ptBoundTopLeft;
					wxPoint ptBoundBottomRight;
					
					ptTextTopLeft = p;
					ptTextTopLeft.x += iOffsetX;
					ptTextTopLeft.y += iOffsetY;

					ptBoundTopLeft = ptTextTopLeft;
					ptBoundBottomRight = ptTextTopLeft;

					// top right edge of text
					
					pt.x = (int) (ptTextTopLeft.x + cos(fAngle) * cFontWidth);
					pt.y = (int) (ptTextTopLeft.y - sin(fAngle) * cFontWidth);
					
					if (pt.x > ptBoundBottomRight.x)
						ptBoundBottomRight.x = pt.x;
					if (pt.x < ptBoundTopLeft.x)
						ptBoundTopLeft.x = pt.x;
					if (pt.y > ptBoundBottomRight.y)
						ptBoundBottomRight.y = pt.y;
					if (pt.y < ptBoundTopLeft.y)
						ptBoundTopLeft.y = pt.y;
					
					// bottom right edge of text

					pt.x = (int) (ptTextTopLeft.x + cos(fAngle) * cFontWidth + sin(fAngle) * cFontHeight);
					pt.y = (int) (ptTextTopLeft.y - sin(fAngle) * cFontWidth + cos(fAngle) * cFontHeight);
					
					if (pt.x > ptBoundBottomRight.x)
						ptBoundBottomRight.x = pt.x;
					if (pt.x < ptBoundTopLeft.x)
						ptBoundTopLeft.x = pt.x;
					if (pt.y > ptBoundBottomRight.y)
						ptBoundBottomRight.y = pt.y;
					if (pt.y < ptBoundTopLeft.y)
						ptBoundTopLeft.y = pt.y;
					
					// bottom left edge of text
						
					pt.x = (int) (ptTextTopLeft.x + sin(fAngle) * cFontHeight);
					pt.y = (int) (ptTextTopLeft.y + cos(fAngle) * cFontHeight);
					
					if (pt.x > ptBoundBottomRight.x)
						ptBoundBottomRight.x = pt.x;
					if (pt.x < ptBoundTopLeft.x)
						ptBoundTopLeft.x = pt.x;
					if (pt.y > ptBoundBottomRight.y)
						ptBoundBottomRight.y = pt.y;
					if (pt.y < ptBoundTopLeft.y)
						ptBoundTopLeft.y = pt.y;
					
					rectThis = wxRect(ptBoundTopLeft, ptBoundBottomRight);
					
					// look for other text occupying the same space
					for (i = vTextRects.begin(); i != vTextRects.end() && !bCollision; i++)
					{
						if ((*i).Intersects(rectThis))
							bCollision = true;
					}
					
					if (ptBoundTopLeft.x < 0 ||
						ptBoundTopLeft.y < 0 ||
						ptBoundBottomRight.x >= m_iControlWidth ||
						ptBoundBottomRight.y >= m_iControlHeight)
					{
						bCollision = true;
					}
					
					// was there a collision?
					if (bCollision)
					{
						if (iOffsetX == 0)
						{
							// shift the text right a bit so the dotted line from
							// the feature to the text always points down and right
							iOffsetX = (int) ( cos(fAngle) * cFontWidth / 2 );
							
							// also, shift it vertically so it lines up with the
							// dot/line
							iOffsetY = (int) ( -sin(fAngle) * cFontWidth / 2 );
						}
						
						// shift the text over a bit
						
						// it makes it a little easier to read if all the lines aren't parallel
						iOffsetX += cFontHeight * (-5 + rand() % 20) / 10 + 1;
						iOffsetY += cFontHeight + 1;
						
						// set the angle to horizontal
						fAngle = 0;
						
						// no longer inside line
						bTextInsideLine = false;
					}
				}
				
				// debugging code
				/*DrawLine(dc, rectThis.GetLeft(), rectThis.GetTop(), rectThis.GetRight(), rectThis.GetBottom(), 1, *wxRED, wxSOLID, 0);
				p.y += 50;
				DrawRotatedText(dc, wxString::Format("%.1f", fAngle * 180 / PI), p, 0, *wxRED);
				p.y -= 50;*/
				
				if (iOffsetX <= m_iMaxLineToLabelLength && iOffsetY <= m_iMaxLineToLabelLength)
				{
					p.x += iOffsetX;
					p.y += iOffsetY;

					vTextRects.push_back(rectThis);
						
					if (iOffsetX || iOffsetY)
					{
						// draw a dotted line from halfway along the feature line to the top left
						// corner of the label
						DrawLine(dc, (p1.x + p2.x) / 2, (p1.y + p2.y) / 2, p.x, p.y, 1, clrText, wxDOT);
					}

					if (bTextInsideLine)
					{
						// inside the line, the background color is the line color
						wxBrush brLine((*iLabelEntry).pMap.m_clrLine, wxSOLID);
						dc->SetBackground(brLine);
					}

					DrawRotatedText(dc, strText, p, fAngle * 180 / PI, clrText);

					if (bTextInsideLine)
					{
						// restore background color
						dc->SetBackground(m_brBackground);
					}
				}
			}

			dc->SetFont(wxNullFont);
			delete pFont;
		}
	}

}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw the compass in the top left corner of the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawMapCompass(wxDC * dc, bool bScreen)
{
	if (!m_bShowCompass)
		return;
	
	int iFontSize = 8;
	double fThisDrawingScaleFactor = m_fDrawingScaleFactor;
	
	wxFont fnt((int) (iFontSize * fThisDrawingScaleFactor), wxSWISS, wxNORMAL, wxNORMAL);
	dc->SetFont(fnt);

	wxString strNKeyLabel = wxT("N");
	wxString strWKeyLabel = wxT("W");

	wxCoord cNFontHeight, cNFontWidth;
	wxCoord cWFontHeight, cWFontWidth;

	dc->GetTextExtent(strNKeyLabel, &cNFontWidth, &cNFontHeight);
	dc->GetTextExtent(strWKeyLabel, &cWFontWidth, &cWFontHeight);

	int iCompassRadius = (int) (6 * fThisDrawingScaleFactor);
	int iCompassArrowLength = (int) (15 * fThisDrawingScaleFactor);
	int iArrowRadius = (int) (5 * fThisDrawingScaleFactor);
	
	int iCompassX = (int) (3 * fThisDrawingScaleFactor + iCompassArrowLength + cWFontWidth);
	int iCompassY = (int) (3 * fThisDrawingScaleFactor + iCompassArrowLength + cNFontHeight);

	wxPoint pts[8];
	// top left
	pts[0].x = (int) (iCompassX - iArrowRadius * ::sqrt((double) 2) / 2);
	pts[0].y = (int) (iCompassY - iArrowRadius * ::sqrt((double) 2) / 2);
	// top
	pts[1].x = (int) (iCompassX);
	pts[1].y = (int) (iCompassY - iCompassArrowLength);
	// top right
	pts[2].x = (int) (iCompassX + iArrowRadius * ::sqrt((double) 2) / 2);
	pts[2].y = (int) (iCompassY - iArrowRadius * ::sqrt((double) 2) / 2);
	// right
	pts[3].x = (int) (iCompassX + iCompassArrowLength);
	pts[3].y = (int) (iCompassY);
	// bottom right
	pts[4].x = (int) (iCompassX + iArrowRadius * ::sqrt((double) 2) / 2);
	pts[4].y = (int) (iCompassY + iArrowRadius * ::sqrt((double) 2) / 2);
	// bottom
	pts[5].x = (int) (iCompassX);
	pts[5].y = (int) (iCompassY + iCompassArrowLength);
	// bottom left
	pts[6].x = (int) (iCompassX - iArrowRadius * ::sqrt((double) 2) / 2);
	pts[6].y = (int) (iCompassY + iArrowRadius * ::sqrt((double) 2) / 2);
	// left
	pts[7].x = (int) (iCompassX - iCompassArrowLength);
	pts[7].y = (int) (iCompassY);

	int iPoint;

	for (iPoint = 0; iPoint < 8; iPoint++)
		pts[iPoint] = RotatePoint(pts[iPoint], wxPoint(iCompassX, iCompassY), -m_fMapAngle);

	dc->SetTextForeground(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground));

	wxPen penBlack(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground), (int) fThisDrawingScaleFactor, wxSOLID);
	dc->SetPen(penBlack);

	wxBrush brBackground(m_pMapAppearanceSettings->GetMiscColor(MiscColorBackground), wxSOLID);
	dc->SetBrush(brBackground);

	int iLabel;
	char * pszLabels[4] = {"N", "E", "S", "W"};	
	for (iLabel = 0; iLabel < 4; iLabel++)
	{
		char * pszLabel = pszLabels[iLabel];
		wxCoord cWidth, cHeight;
		double fAngle = iLabel * 90 - m_fMapAngle;

		dc->GetTextExtent(wxString(pszLabel, *wxConvCurrent), &cWidth, &cHeight);

		dc->DrawText(wxString(pszLabel, *wxConvCurrent), 
						(int) (iCompassX + (iCompassArrowLength + cHeight / 2) * sin(fAngle / 180 * PI) - cWidth / 2),
						(int) (iCompassY - (iCompassArrowLength + cHeight / 2) * cos(fAngle / 180 * PI) - cHeight / 2)
						);
	}

	dc->DrawPolygon(8, pts);
	dc->DrawCircle(iCompassX, iCompassY, iCompassRadius);

	dc->SetPen(wxNullPen);
	dc->SetBrush(wxNullBrush);
	dc->SetFont(wxNullFont);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw the key, indicating the scale of the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawMapKey(wxDC * dc, bool bScreen)
{
	if (!m_bShowScale)
		return;
	
	// draw key
	int iKeyLengthPixels;
	double fKeyLengthMiles;
	double fThisDrawingScaleFactor = m_fDrawingScaleFactor;
	
	iKeyLengthPixels = m_iControlWidth / 8;

	int iFontSize = 8;
	int iKeyW = (int) (iKeyLengthPixels);
	int iKeyH = (int) (5 * fThisDrawingScaleFactor);
	int iKeyX = (int) (m_iControlWidth - 5 * fThisDrawingScaleFactor - iKeyW);
	int iKeyY = (int) (m_iControlHeight - 5 * fThisDrawingScaleFactor - iKeyH);

	fKeyLengthMiles = Distance(Point(MapScreenToLongLat(wxPoint(iKeyX, iKeyY))),
								Point(MapScreenToLongLat(wxPoint(iKeyX + iKeyW, iKeyY))));

	wxString strLabel;
	strLabel = FormatDistance(fKeyLengthMiles);

	wxPen penBlack(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground), (int) fThisDrawingScaleFactor, wxSOLID);
	wxBrush brBackground(m_pMapAppearanceSettings->GetMiscColor(MiscColorBackground), wxSOLID);

	dc->SetPen(penBlack);
	dc->SetBrush(brBackground);
	dc->DrawRectangle(iKeyX, iKeyY, iKeyW, iKeyH);

/*	dc->SetBrush(*wxGREY_BRUSH);
	dc->DrawRectangle(iKeyX, iKeyY, iKeyW / 4, iKeyH);
	dc->DrawRectangle(iKeyX + iKeyW / 2, iKeyY, iKeyW / 4, iKeyH);*/

	dc->SetPen(wxNullPen);
	dc->SetBrush(wxNullBrush);

	wxFont fnt((int) (iFontSize * fThisDrawingScaleFactor), wxSWISS, wxNORMAL, wxNORMAL);
	dc->SetFont(fnt);
	dc->SetTextForeground(m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground));

	wxCoord cFontHeight, cFontWidth;
	dc->GetTextExtent(strLabel, &cFontWidth, &cFontHeight);

	dc->DrawText(strLabel, iKeyX + iKeyW - cFontWidth, iKeyY - cFontHeight);

	dc->SetFont(wxNullFont);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw any Terraserver tiles available
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawMapTerraserver(wxDC * dc, bool bScreen)
{
	int iScale;
	double fZoom = m_pMapAppearanceSettings->GetScale(m_iDetailLevel);
	double fLong;
	double fLat;
	char szFilename[256];

	Point ptCenterTL;
	Point ptCenterBR;
	double fLongWidth;
	double fLatHeight;

	iScale = (int) ((log(fZoom) / log((double) 2) + 27));

	if (iScale < 10)
		iScale = 10;

	if (iScale > 18)
		iScale = 18;

	FetchTerraserverTile(iScale, m_ptCenter, szFilename, &ptCenterTL, &ptCenterBR, m_bDownloadAerialPhoto);
	fLongWidth = ptCenterBR.GetLong() - ptCenterTL.GetLong();
	fLatHeight = ptCenterTL.GetLat() - ptCenterBR.GetLat();

	// eliminate the risk of moving one tile height/width over resulting in actually
	// moving over two tiles (happens due to UTM vs. orthographic projection issues)

	fLongWidth *= 0.9;
	fLatHeight *= 0.9;

	// go through each tile on the screen

	for (fLat = m_ptTopLeft.GetLat(); (fLat + fLatHeight) > m_ptBottomRight.GetLat(); fLat -= fLatHeight)
	{
		for (fLong = m_ptTopLeft.GetLong(); (fLong - fLongWidth) < m_ptBottomRight.GetLong(); fLong += fLongWidth)
		{
			Point ptTL;
			Point ptBR;
			char szTmp[128];
			wxImage imgTile;
			Point ptTileTL;
			Point ptTileBR;
			wxPoint cTileTL;
			wxPoint cTileBR;

			// fetch the tile we're really looking for
			FetchTerraserverTile(iScale, Point(fLong, fLat), szFilename, &ptTileTL, &ptTileBR, m_bDownloadAerialPhoto);

			// go one tile down, store it's upper edge coordinates
			FetchTerraserverTile(iScale, Point(fLong, fLat - fLatHeight), szTmp, &ptTL, &ptBR, m_bDownloadAerialPhoto);
			ptTileBR.SetLatRaw(ptTL.GetLatRaw());
			
			// make sure it's not the same as the upper edge of this tile
			if (ptTileBR.GetLatRaw() == ptTileTL.GetLatRaw())
			{
				FetchTerraserverTile(iScale, Point(fLong, fLat - fLatHeight * 2), szTmp, &ptTL, &ptBR, m_bDownloadAerialPhoto);
				ptTileBR.SetLatRaw(ptTL.GetLatRaw());
			}

			// go one tile right, store its left edge
			FetchTerraserverTile(iScale, Point(fLong + fLongWidth, fLat), szTmp, &ptTL, &ptBR, m_bDownloadAerialPhoto);
			ptTileBR.SetLongRaw(ptTL.GetLongRaw());

			// make sure it's not the same as the left edge of this tile
			if (ptTileBR.GetLongRaw() == ptTileTL.GetLongRaw())
			{
				FetchTerraserverTile(iScale, Point(fLong + fLongWidth * 2, fLat), szTmp, &ptTL, &ptBR, m_bDownloadAerialPhoto);
				ptTileBR.SetLongRaw(ptTL.GetLongRaw());
			}

			// what about one tile down and right?
			FetchTerraserverTile(iScale, Point(fLong + fLongWidth, fLat - fLatHeight), szTmp, &ptTL, &ptBR, m_bDownloadAerialPhoto);

			if (ptTL.GetLongRaw() > ptTileBR.GetLongRaw())
				ptTileBR.SetLongRaw(ptTL.GetLongRaw());
			if (ptTL.GetLatRaw() < ptTileBR.GetLatRaw())
				ptTileBR.SetLatRaw(ptTL.GetLatRaw());

			// make long/lat of tile edges to screen coordinates
			cTileTL = MapLongLatToScreen(ptTileTL);
			cTileBR = MapLongLatToScreen(ptTileBR);

			if (*szFilename && wxFileExists(wxString(szFilename, *wxConvCurrent)))
			{
				imgTile.LoadFile(wxString(szFilename, *wxConvCurrent));

				if (imgTile.Ok())
				{
					int iWidth;
					int iHeight;

					iWidth = cTileBR.x - cTileTL.x;
					iHeight = cTileBR.y - cTileTL.y;

					imgTile.Rescale(iWidth, iHeight);

					dc->DrawBitmap(imgTile, cTileTL.x, cTileTL.y, false);
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw all of the map items that don't change much.
///
//////////////////////////////////////////////////////////////////////////////
int MapControl::DrawMap(wxDC * dc, bool bScreen)
{
	bool bVisibility[LAST_RECORD_TYPE + 1];
	int iRT;
	Point ptStepX;
	Point ptStepY;

	for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
	{
		const DetailSettings& pMap = m_pMapAppearanceSettings->GetDetailSettings(m_iDetailLevel, (ERecordType) iRT);
		bVisibility[iRT] = pMap.m_bLineVisible || pMap.m_bTextVisible;
	}

	if (!m_b3D)
	{
		ptStepX = MapScreenToLongLat(wxPoint(m_szCacheBitmaps.GetWidth(), 0)) - MapScreenToLongLat(wxPoint(0, 0));
		ptStepY = MapScreenToLongLat(wxPoint(0, m_szCacheBitmaps.GetHeight())) - MapScreenToLongLat(wxPoint(0, 0));
		ptStepX.SetLong(fabs(ptStepX.GetLong()));
		ptStepX.SetLat(-fabs(ptStepX.GetLat()));
		ptStepY.SetLong(fabs(ptStepY.GetLong()));
		ptStepY.SetLat(-fabs(ptStepY.GetLat()));
	}
	else
	{
		ptStepX.SetRaw(0, 0);
		ptStepY.SetRaw(0, 0);
	}

	m_pMapControlData->LoadRegion(m_ptViewTopLeft - ptStepX - ptStepY, m_ptViewBottomRight + ptStepX + ptStepY, bVisibility);
	
	dc->SetBackground(m_brBackground);
	dc->Clear();
	dc->SetBrush(wxNullBrush);

	if (m_bShowAerialPhoto)
		DrawMapTerraserver(dc, bScreen);

#ifdef FILLED_IN_WATER_SUPPORT
	DrawMapWaterBoundaries(dc, bScreen);
	DrawMapWater(dc, bScreen);
#endif
	
	DrawMapGeneral(dc, bScreen);

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw a border around the map
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawBorder(wxDC * dc, bool bScreen)
{
	int iWidth;
	wxColour clrSegment;
	wxCoord w, h;

	w = m_iControlWidth;
	h = m_iControlHeight;

	iWidth = (int) (2 * m_fDrawingScaleFactor);
	clrSegment.Set(0, 0, 0);

	DrawLine(dc, 0, 0, w, 0, iWidth, clrSegment, wxSOLID);
	DrawLine(dc, w, 0, w, h, iWidth, clrSegment, wxSOLID);
	DrawLine(dc, w, h, 0, h, iWidth, clrSegment, wxSOLID);
	DrawLine(dc, 0, h, 0, 0, iWidth, clrSegment, wxSOLID);

	return;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw any tracks that are defined.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawTracks(wxDC * dc, bool bScreen)
{
	vector<MapTrack>::iterator i;

	for (i = m_vcTracks.begin(); i != m_vcTracks.end(); i++)
		i->Draw(this, dc, m_fDrawingScaleFactor);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief This function performs autoscaling - ensuring that the entire track is
/// visible on the screen.  If it's not, it keeps zooming out until it is.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::PerformAutoscaling()
{
	bool bAllVisible = false;
	while (!bAllVisible)
	{
		bAllVisible = true;

		unsigned int iTrack;
		for (iTrack = 0; iTrack < m_vcTracks.size(); iTrack++)
		{
			if (m_vcTracks[iTrack].m_bAffectsAutoscale)
			{
				vector<Point>::iterator iPoint;
				
				for (iPoint = m_vcTracks[iTrack].m_vptCoordinates.begin(); iPoint != m_vcTracks[iTrack].m_vptCoordinates.end(); iPoint++)
				{
					if (!IsOnScreen(*iPoint))
					{
						bAllVisible = false;
					}
				}
			}
		}
		
		if (GetDetailLevel() == 0)
			bAllVisible = true;

		if (!bAllVisible)
		{
			LessDetail();
			UpdateTopLeftAndBottomRight();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Update m_ptTopLeft and m_ptBottomRight to contain the NW and SE-most
/// coordinates, respectively, shown on the screen.
///
/// The rectangle enclosed by m_ptTopLeft and m_ptBottomRight may encompass
/// some area that's actually off-screen due to the orientation of the map.
/// However, all on screen points should be enclosed between m_ptTopLeft and
/// m_ptBottomRight no matter what.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::UpdateTopLeftAndBottomRight()
{
	wxPoint testpts[4] = {
							wxPoint(0, 0), 
							wxPoint(m_iControlWidth - 1, 0),
							wxPoint(0, m_iControlHeight - 1),
							wxPoint(m_iControlWidth - 1, m_iControlHeight - 1)
						};

	Point pt[4];
	int i;

	for (i = 0; i < 4; i++)
	{
		pt[i] = MapScreenToLongLat(testpts[i]);
	}

	m_ptTopLeft = pt[0];
	m_ptBottomRight = pt[0];

	for (i = 1; i < 4; i++)
	{
		if (m_ptTopLeft.GetLongRaw() > pt[i].GetLongRaw())
			m_ptTopLeft.SetLongRaw(pt[i].GetLongRaw());
		if (m_ptTopLeft.GetLatRaw() < pt[i].GetLatRaw())
			m_ptTopLeft.SetLatRaw(pt[i].GetLatRaw());

		if (m_ptBottomRight.GetLongRaw() < pt[i].GetLongRaw())
			m_ptBottomRight.SetLongRaw(pt[i].GetLongRaw());
		if (m_ptBottomRight.GetLatRaw() > pt[i].GetLatRaw())
			m_ptBottomRight.SetLatRaw(pt[i].GetLatRaw());
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Create the requested tile and insert it into the cache.
///
/// \param iX				X coordinate of tile
/// \param iY				Y coordinate of tile
/// \param iDetailLevel		Detail level (zoom) of tile
/// \param fAngle			Orientation of tile
/// \param ptStepX			When this is multiplied by iX, one obtains the
///							X coordinate in degrees of the tile.
/// \param ptStepY			When this is multiplied by iY, one obtains the
///							Y coordinate in degrees of the tile.
/// \param bScreen			True if drawing onto the screen. False for saving to file.
///
/// \return					MapCacheEntry for requested tile.
///
/// The center of the tile is determined by (iX + 0.5) * ptStepX + (iY + 0.5) * ptStepY.
///
/// Will return right away if tile already exists in cache.
///
//////////////////////////////////////////////////////////////////////////////
MapCacheEntry * MapControl::CreateMapCacheEntry(int iX, int iY, int iDetailLevel, double fAngle, Point ptStepX, Point ptStepY, bool bScreen, bool bRender)
{
	int iMapCacheEntry;
	
	// try to find tile in cache
	for (iMapCacheEntry = 0; iMapCacheEntry < TILE_CACHE_SIZE; iMapCacheEntry++)
	{
		if (m_parMapCache[iMapCacheEntry])
		{
			if (m_parMapCache[iMapCacheEntry]->m_iX == iX &&
					m_parMapCache[iMapCacheEntry]->m_iY == iY &&
					m_parMapCache[iMapCacheEntry]->m_iDetailLevel == m_iDetailLevel &&
					m_parMapCache[iMapCacheEntry]->m_fAngle == m_fMapAngle)
				return m_parMapCache[iMapCacheEntry];
		}
	}

	if (!bRender)
		return NULL;
	
	// need to create entry
	int iOriginalDetailLevel;
	double fOriginalAngle;
	Point ptSliceTL;
	Point ptSliceBR;
	Point ptOriginalCenter;
	MapCacheEntry * pcMapCacheEntry;
	
	// remember settings		
	iOriginalDetailLevel = m_iDetailLevel;
	fOriginalAngle = m_fMapAngle;
	ptOriginalCenter = m_ptCenter;
	
	// apply new settings
	m_iDetailLevel = iOriginalDetailLevel;
	m_fMapAngle = fAngle;
	
	// these are the coordinates at the TL and BR corner of the
	// slice. They aren't expanded out to include rotation effects
	// like m_ptTopLeft
	ptSliceTL = ptStepX * iX + ptStepY * iY;
	ptSliceBR = ptSliceTL + ptStepX + ptStepY;

	// calculate center of tile
	m_ptCenter = ptSliceTL + ptStepX / 2 + ptStepY / 2;
			
	// set width and height of tile in pixels
	m_iControlWidth = m_szCacheBitmaps.GetWidth();
	m_iControlHeight = m_szCacheBitmaps.GetHeight();

	// calculate coordinates of top-left and bottom-right corner
	UpdateTopLeftAndBottomRight();

	// otherwise, create a new entry and draw the map onto it
	pcMapCacheEntry = new MapCacheEntry(m_iControlWidth, m_iControlHeight, iX, iY, m_iDetailLevel, m_fMapAngle, m_bAntiAlias);
	
	DrawMap(pcMapCacheEntry->m_pdc, bScreen);
	//DrawBorder(pcMapCacheEntry->m_pdc, bScreen);

	if (m_parMapCache[m_iMapCacheOldest])
		delete m_parMapCache[m_iMapCacheOldest];
	m_parMapCache[m_iMapCacheOldest] = pcMapCacheEntry;
	m_iMapCacheOldest = (m_iMapCacheOldest + 1) % TILE_CACHE_SIZE;
	
	// restore old settings
	m_iDetailLevel = iOriginalDetailLevel;
	m_fMapAngle = fOriginalAngle;
	m_ptCenter = ptOriginalCenter;
	
	return pcMapCacheEntry;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw everything
///
/// If bDirectlyDraw == false, everything is drawn on memdc, 
/// scaled down by iAntiAliasingLevel, and copied to realdc. fDPIFactor
/// scales the size of everything in the final product, so if fDPIFactor > 1,
/// text, lines, and everything else will take up more pixels. This should be
/// used if the DPI is not 72 on realdc.
///
/// If bDirectlyDraw == true, then everything is directly drawn onto the
/// realdc device context. iAntiAliasingLevel will be ignored in that case.
/// fDPIFactor will still be used.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::DrawAll(wxDC * realdc, wxSize szMap, bool bDirectlyDraw, bool bScreen)
{
	wxDC * pDC;
	
	static wxAntiAliasedDC<wxMemoryDC> * pAA = NULL;

	if (m_tDataLastUpdated != GetMapControlData()->GetLastUpdated())
	{
		ClearMapCache();
		m_tDataLastUpdated = GetMapControlData()->GetLastUpdated();
	}
	
	if (bDirectlyDraw)
		pDC = realdc;
	else
		pDC = m_pdcOldEverything;
	
	// Start out by calculating how big the control is and
	// what will fit on the screen
	
	m_iControlWidth = szMap.x;
	m_iControlHeight = szMap.y;
	UpdateTopLeftAndBottomRight();

	m_ptViewTopLeft = m_ptTopLeft;
	m_ptViewBottomRight = m_ptBottomRight;
	
	if (m_bAutoScale)
	{
		PerformAutoscaling();
		m_bAutoScale = false;

		UpdateTopLeftAndBottomRight();

		m_ptViewTopLeft = m_ptTopLeft;
		m_ptViewBottomRight = m_ptBottomRight;
	}

	m_brBackground = wxBrush(m_pMapAppearanceSettings->GetMiscColor(MiscColorBackground), wxSOLID);

	if (!bDirectlyDraw)
	{
		if (m_bmpOldEverything.GetWidth() != m_iControlWidth || m_bmpOldEverything.GetHeight() != m_iControlHeight)
		{
			wxBitmap bmpEmpty(1, 1);
			m_pdcOldEverything->SelectObject(bmpEmpty);
			m_bmpOldEverything.Create(m_iControlWidth, m_iControlHeight);
			m_pdcOldEverything->SelectObject(m_bmpOldEverything);
		}
		else
		{
			if (m_bAntiAlias)
			{
				if ( pAA == NULL )
				{
					pAA = wxDynamicCast(m_pdcOldEverything, wxAntiAliasedDC<wxMemoryDC>);
				}
				
				// purge pixel cache
				if (pAA)
					pAA->Setup(m_iControlWidth, m_iControlHeight);
			}
		}
	}

	if (m_b3D || bDirectlyDraw)
	{
		DrawMap(pDC, bScreen);
	}
	else
	{
		Point ptStepX;
		Point ptStepY;
		int iStartX;
		int iEndX;
		int iStartY;
		int iEndY;
		int iX;
		int iY;
		double fX;
		double fY;
		Point ptScrTL;
		Point ptScrBR;
		wxPoint ptTopLeftSlice;
		int iTilesRendered = 0;
		int iTilesNeedingRendering = 0;
		Point ptCenterSave;
		
		m_iControlWidth = 0;
		m_iControlHeight = 0;
		
		// Divide the world into a set of square tiles, rotated by m_fMapAngle with the origin at
		// 0N 0E. ptStepX and ptStepY are width and height of each tile. They're two dimensional
		// since the tiles may not run parallel to the equator if m_fMapAngle != 0.
		
		// Use a fixed m_ptCenter since having m_ptCenter change causes ptStepX and ptStepY
		// to change slightly as well due to rounding
		ptCenterSave = m_ptCenter;
		m_ptCenter.Set(0, 0);
		ptStepX = MapScreenToLongLat(wxPoint(m_szCacheBitmaps.GetWidth(), 0)) - MapScreenToLongLat(wxPoint(0, 0));
		ptStepY = MapScreenToLongLat(wxPoint(0, m_szCacheBitmaps.GetHeight())) - MapScreenToLongLat(wxPoint(0, 0));
		m_ptCenter = ptCenterSave;

		m_iControlWidth = szMap.x;
		m_iControlHeight = szMap.y;
	
		// determine screen boundary
		ptScrTL = MapScreenToLongLat(wxPoint(0, 0));
		ptScrBR = MapScreenToLongLat(wxPoint(szMap.x, szMap.y));
		
		// solve for top left coordinates
		SolveLinearEquationSet(ptStepX.GetLong(), ptStepY.GetLong(), ptScrTL.GetLong(), 
								ptStepX.GetLat(), ptStepY.GetLat(), ptScrTL.GetLat(),
								fX, fY);
		
		iStartX = (int) floor(fX);
		iStartY = (int) floor(fY);
		
		// solve for bottom right coordinates
		SolveLinearEquationSet(ptStepX.GetLong(), ptStepY.GetLong(), ptScrBR.GetLong(), 
								ptStepX.GetLat(), ptStepY.GetLat(), ptScrBR.GetLat(),
								fX, fY);
		
		iEndX = (int) floor(fX);
		iEndY = (int) floor(fY);
		
		wxASSERT(iEndX >= iStartX);
		wxASSERT(iEndY >= iStartY);
				
		m_iControlWidth = szMap.x;
		m_iControlHeight = szMap.y;
		ptTopLeftSlice = MapLongLatToScreen(ptStepX * iStartX + ptStepY * iStartY);
		
		// go through each tile that's on screen
		for (iX = iStartX; iX <= iEndX; iX++)
		{
			for (iY = iStartY; iY <= iEndY; iY++)
			{
				MapCacheEntry * pcMapCacheEntry;

				pcMapCacheEntry = CreateMapCacheEntry(iX, iY, m_iDetailLevel, m_fMapAngle, ptStepX, ptStepY, bScreen, false);

				if (!pcMapCacheEntry)
				{
					if (iTilesRendered < 1 && !m_bLastDrawWasComplete)
					{
						pcMapCacheEntry = CreateMapCacheEntry(iX, iY, m_iDetailLevel, m_fMapAngle, ptStepX, ptStepY, bScreen, true);
						iTilesRendered++;
					}
					else
					{
						iTilesNeedingRendering++;
					}
				}
				
				if (pcMapCacheEntry)
				{
					// copy the tile onto the appropriate place on the DC we're using
					// (typically off screen)
					pDC->Blit(	ptTopLeftSlice.x + (iX - iStartX) * m_szCacheBitmaps.GetWidth(),
								ptTopLeftSlice.y + (iY - iStartY) * m_szCacheBitmaps.GetHeight(),
								m_szCacheBitmaps.GetWidth(),
								m_szCacheBitmaps.GetHeight(),
								pcMapCacheEntry->m_pdc,
								0,
								0);
				}
				else
				{
					// nothing available yet
					wxColour clrBackground = m_pMapAppearanceSettings->GetMiscColor(MiscColorBackground);
					wxPen penBackground(clrBackground, 1, wxSOLID);
					wxBrush brBackground(clrBackground);
					
					// fill in with background color
					pDC->SetBrush(brBackground);
					pDC->SetPen(penBackground);
					
					pDC->DrawRectangle(	ptTopLeftSlice.x + (iX - iStartX) * m_szCacheBitmaps.GetWidth(),
										ptTopLeftSlice.y + (iY - iStartY) * m_szCacheBitmaps.GetHeight(),
										m_szCacheBitmaps.GetWidth(),
										m_szCacheBitmaps.GetHeight());
				}
			} // for (iY 
		} // for (iX
		
		if (iTilesNeedingRendering)
		{
			m_bLastDrawWasComplete = false;
			SafeRefresh();
		}
		else
		{
			m_bLastDrawWasComplete = true;
		}
	} // if (m_b3D) else

	// reset so we can draw tracks and markers correctly
	m_iControlWidth = szMap.x;
	m_iControlHeight = szMap.y;
	m_ptTopLeft = m_ptViewTopLeft;
	m_ptBottomRight = m_ptViewBottomRight;

	if (bScreen)
		m_vClickCallbacks.clear();

	pDC->SetBackground(m_brBackground);
	DrawTracks(pDC, bScreen);

	vector<MapMarker>::iterator iMarker;
	for (iMarker = m_vsMarkers.begin(); iMarker != m_vsMarkers.end(); iMarker++)
		(*iMarker).Draw(this, pDC, m_fDrawingScaleFactor, bScreen);

	if (!bDirectlyDraw)
	{
		// Blit everything onto the screen
		realdc->Blit(0, 0,
							szMap.x,
							szMap.y,
							m_pdcOldEverything,
							0,
							0
							);
	}

	if (m_bSendMapUpdatedNotification)
	{
		wxCommandEvent ev(wxEVT_COMMAND_BUTTON_CLICKED, GetId());
		GetParent()->GetEventHandler()->AddPendingEvent(ev);
		m_bSendMapUpdatedNotification = false;
	}

	DrawMapCompass(realdc, bScreen);
	DrawMapKey(realdc, bScreen);
	DrawBorder(realdc, bScreen);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the drawing scale factor - the magnification factor of everything
/// in the control - lines, fonts, symbols
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetDrawingScaleFactor(double fDrawingScaleFactor)
{
	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_fDrawingScaleFactor = fDrawingScaleFactor;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtains the current drawing scale factor
///
//////////////////////////////////////////////////////////////////////////////
double MapControl::GetDrawingScaleFactor()
{
	return m_fDrawingScaleFactor;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtains a pointer to the MapAppearanceSettings object - the container
/// for settings tied to a specific detail level
///
//////////////////////////////////////////////////////////////////////////////
const MapAppearanceSettings* MapControl::GetDetailSettings()
{
	return m_pMapAppearanceSettings;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Enable or disable antialiasing
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetAntialiasing(bool bEnabled)
{
	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_bAntiAlias = bEnabled;
	}
	
	delete m_pdcOldEverything;

	if (m_bAntiAlias)
		m_pdcOldEverything = new wxAntiAliasedDC<wxMemoryDC>();
	else
		m_pdcOldEverything = new wxMemoryDC();
		
	m_pdcOldEverything->SelectObject(m_bmpOldEverything);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtains a pointer to the IMapControlData object - the container
/// for all of the map features.
///
//////////////////////////////////////////////////////////////////////////////
IMapControlData * MapControl::GetMapControlData()
{
	return m_pMapControlData;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Trigger a refresh of the entire control in a thread safe manner
///
/// (calling Refresh() in a thread other than the main thread causes the
/// UI to hang under GTK)
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SafeRefresh()
{
	wxCommandEvent ev(wxEVT_SAFE_REFRESH, 0);

	AddPendingEvent(ev);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Triggered in the main thread by a SafeRefresh() call from any
/// thread.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnSafeRefresh(wxCommandEvent & event)
{
#if wxCHECK_VERSION(2, 5, 0)
	wxMilliSleep(1);
#else
	wxUsleep(1);
#endif
	Refresh();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtains the heading that faces up
///
/// 0 = north faces up
/// 90 = east faces up
/// 180 = south faces up
/// 270 = west faces up
///
/// Arbitrary angles can be used.
///
//////////////////////////////////////////////////////////////////////////////
double MapControl::GetAngle()
{
	return m_fMapAngle;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the heading that faces up
///
/// \param fAngle		New map orientation.
///
/// 0 = north faces up
/// 90 = east faces up
/// 180 = south faces up
/// 270 = west faces up
///
/// Arbitrary angles can be used.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::SetAngle(double fAngle)
{
	// The angle cannot be set if aerial photos are on.
	if (m_bShowAerialPhoto)
		return;

	if (fabs(m_fMapAngle - fAngle) > 0.01)
	{
		// Limit scope of locker
		{
			wxMutexLocker mlLock(m_mxSettings);
			m_fMapAngle = fAngle;

			while (m_fMapAngle >= 360)
				m_fMapAngle -= 360;

			while (m_fMapAngle < 0)
				m_fMapAngle += 360;
		}

		MapUpdated(UPDATED_ANGLE);
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Enable or disables the 3D perspective mode.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::Enable3DMode(bool bEnabled)
{
	if (m_b3D == bEnabled)
		return;

	// Limit scope of locker
	{
		wxMutexLocker mlLock(m_mxSettings);
		m_b3D = bEnabled;
	}

	if (m_b3D)
		ShowAerialPhotos(false, false);

	MapUpdated(UPDATED_3D);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Called after the left mouse button is held down for 2 seconds.
///
/// Triggers the right mouse click code.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnTimerLeftMouseHold(wxTimerEvent& event)
{
	// mark as handled
	m_bLeftMouseHandled = true;

	OnRightUp(m_evLeftMouseHold);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets up a callback for mouse clicks
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::AddClickCallback(const wxPoint & ptTopLeft, const wxPoint & ptBottomRight, tClickCallback pCallback, void * pArg)
{
	SClickCallback sCCB;
	
	sCCB.m_ptTopLeft = ptTopLeft;
	sCCB.m_ptBottomRight = ptBottomRight;
	sCCB.m_pCallback = pCallback;
	sCCB.m_pArg = pArg;
	
	m_vClickCallbacks.push_back(sCCB);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clear the cache of map tiles (e.g. due to theme change)
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::ClearMapCache()
{
	int iMapCacheEntry;
	
	for (iMapCacheEntry = 0; iMapCacheEntry < TILE_CACHE_SIZE; iMapCacheEntry++)
	{
		if (m_parMapCache[iMapCacheEntry])
			delete m_parMapCache[iMapCacheEntry];

		m_parMapCache[iMapCacheEntry] = NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Mouse has moved - if left button is down, then drag the map.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::OnMouseMove(wxMouseEvent& event)
{
	wxPoint ptDistance;

	if (!event.LeftIsDown())
	{
		event.Skip();
		return;
	}

	ptDistance = event.GetPosition() - m_evLeftMouseHold.GetPosition();
	
	if (!m_bLeftMouseMovingMap)
	{
		// haven't moved far enough to be considered a drag yet
		int iDistance;
		
		// how far have we gone (squared)?
		iDistance = ptDistance.x * ptDistance.x + ptDistance.y * ptDistance.y;

		if (iDistance > 400) // 20 pixels
		{
			// moved far enough to count as a drag
			m_pTimerLeftMouseHold->Stop();
			m_bLeftMouseHandled = true;
			m_bLeftMouseMovingMap = true;
		}
		else
		{
			// not enough
			event.Skip();
			return;
		}
	}

	// recenter accordingly
	m_mxSettings.Lock();
	m_ptCenter = m_ptCenterWhenLeftMouseDown + MapScreenToLongLat(m_evLeftMouseHold.GetPosition()) - MapScreenToLongLat(event.GetPosition());
	m_mxSettings.Unlock();

	MapUpdated(UPDATED_CENTER);

	event.Skip();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief This constructs the cross hair cursor based on the current 
/// foreground color.
///
//////////////////////////////////////////////////////////////////////////////
void MapControl::UpdateCrossHair()
{
#ifdef __WINDOWS__
	wxColour clrBackground = m_pMapAppearanceSettings->GetMiscColor(MiscColorBackground);
	wxColour clrForeground = m_pMapAppearanceSettings->GetMiscColor(MiscColorForeground);
	wxBitmap bmpCursor(32, 32);

	{
		wxMemoryDC dcCursor;
		wxPen penForeground(clrForeground, 1, wxSOLID);
		wxPen penBackground(clrBackground, 1, wxSOLID);
		wxBrush brBackground(clrBackground);
		
		dcCursor.SelectObject(bmpCursor);
		
		// fill in background
		dcCursor.SetBrush(brBackground);
		dcCursor.SetPen(penBackground);
		dcCursor.DrawRectangle(0, 0, 32, 32);
		
		// draw cross hairs
		dcCursor.SetPen(penForeground);
		dcCursor.DrawLine(15, 0, 15, 31);
		dcCursor.DrawLine(0, 15, 31, 15);
	}
	
	wxImage imgCursor = bmpCursor.ConvertToImage();
	imgCursor.SetMask(true);
	imgCursor.SetMaskColour(clrBackground.Red(), clrBackground.Green(), clrBackground.Blue());
	imgCursor.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_X, 15);
	imgCursor.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_Y, 15);

	m_curCrossHair = wxCursor(imgCursor);
#else
	m_curCrossHair = *wxCROSS_CURSOR;
#endif

	SetCursor(m_curCrossHair);
}
