/*
 *  LibRoadnav
 *  MapRecordEntry.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains MapControlDataEntry_Std - a class for a single TIGER/Line type 1 
/// feature or equivalent in the database.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "Map.h"
#include "MapSupport.h"
#include "MapZip.h"
#include "MapState.h"
#include "MapAbbreviations.h"
#include "Portability.h"
#include "StringTable.h"
#include "Debug.h"

#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (x > y ? x : y)

double MapControlDataEntry_Std::m_fPreferSmallRoads = 1.00;
double MapControlDataEntry_Std::m_fPreferLargeRoads = 1.44;
double MapControlDataEntry_Std::m_fPreferHighways = 1.54;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Tries to come up with an abbreviated name for a record
///
//////////////////////////////////////////////////////////////////////////////
wxString MapControlDataEntry_Std::SmartName() const
{
	// AF string table is in UTF8, so do conversion for correct display
	return wxString(pszSmartName, wxConvUTF8);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the feature name and type
///
//////////////////////////////////////////////////////////////////////////////
wxString MapControlDataEntry_Std::NameAndType(int iEntry, bool bLongType) const
{
	if (iEntry < nFeatureNames)
	{
		wxString strType = wxT("");
	
		if (bLongType)
			strType = GetFullName(wxString(pszFeatureTypes[iEntry], *wxConvCurrent));
		
		if (strType == wxT(""))
			strType = wxString(pszFeatureTypes[iEntry], *wxConvCurrent);
		
		if (strType != wxT(""))
			strType = wxT(" ") + strType;
		
		return wxString(pszFeatureNames[iEntry], *wxConvCurrent) + strType;
	}
	
	return wxT("");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines if a record is a road or not
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::IsRoad() const
{
	if (eType >= FIRST_RECORD_TYPE_ROAD && eType <= LAST_RECORD_TYPE_ROAD)
		return true;

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Cost factor for each record.
/// 
/// The cost factor defines the penalty per distance for going down a certain
/// road. Roads with smaller penalties are more likely to be taken than roads
/// with large penalties.
///
/// This can be tweaked to avoid small roads, or to avoid highways, etc.
///
//////////////////////////////////////////////////////////////////////////////
double MapControlDataEntry_Std::CostFactor() const
{
	switch (eType)
	{
		case RecordTypeOneWaySmallRoad:
		case RecordTypeTwoWaySmallRoad:
			return 60. / 25 / m_fPreferSmallRoads;

		case RecordTypeOneWayLargeRoad:
		case RecordTypeTwoWayLargeRoad:
			return 60. / 45 / m_fPreferLargeRoads;

		case RecordTypeOneWayHighway:
		case RecordTypeTwoWayHighway:
			return 60. / 65 / m_fPreferHighways;
		
		default:
		{
		}
	}
	return INFINITY;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if a record is a one-way street
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::IsOneWay() const
{
	switch (eType)
	{
		case RecordTypeOneWaySmallRoad:
		case RecordTypeOneWayLargeRoad:
		case RecordTypeOneWayHighway:
			return true;

		default:
		{
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// How long does it take to go one mile along a type of road?.
///
/// This is used to determine how long each leg of the trip takes.
/// The return value should be related to the speed limit on the road.
///
//////////////////////////////////////////////////////////////////////////////
double MapControlDataEntry_Std::TimeFactor() const
{
	switch (eType)
	{
		case RecordTypeOneWaySmallRoad:
		case RecordTypeTwoWaySmallRoad:
			return 60. / 35;

		case RecordTypeOneWayLargeRoad:
		case RecordTypeTwoWayLargeRoad:
			return 60. / 45;

		case RecordTypeOneWayHighway:
		case RecordTypeTwoWayHighway:
			return 60. / 65;

		default:
		{
		}
	}
	return INFINITY;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determine the distance covered by a single record entry
///
//////////////////////////////////////////////////////////////////////////////
double MapControlDataEntry_Std::RecordDistance() const
{
	int iSeg;
	double fDistance = 0;

	for (iSeg = 1; iSeg < nCoordinates; iSeg++)
		fDistance += ::Distance(psCoordinates[iSeg - 1], psCoordinates[iSeg]);

	return fDistance;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines if two records are part of the same road.
///
/// It does this by checking if they share any common names.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::IsSameRoad(IMapControlDataEntry * psRec) const
{
	int i, j;

	for (i = 0; i < nFeatureNames; i++)
	{
		if (!CaseInsensitiveStrCmp(pszFeatureTypes[i], "Cir"))
			return false;
	}

	for (i = 0; i < psRec->nFeatureNames; i++)
	{
		if (!CaseInsensitiveStrCmp(psRec->pszFeatureTypes[i], "Cir"))
			return false;
	}

	for (i = 0; i < nFeatureNames; i++)
	{
		for (j = 0; j < psRec->nFeatureNames; j++)
		{
			if (!CaseInsensitiveStrCmp(pszFeatureNames[i], psRec->pszFeatureNames[j]) &&
				!CaseInsensitiveStrCmp(pszFeatureTypes[i], psRec->pszFeatureTypes[j]))
			{
				return true;	
			}
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compare the coordinates of this record to the coordinates of another
/// record for sorting purposes
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::operator == (const MapControlDataEntry_Std & x) const
{
	return psCoordinates[0] == x.psCoordinates[0];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compare the coordinates of this record to the coordinates of another
/// record for sorting purposes
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::operator < (const MapControlDataEntry_Std & x) const
{
	return psCoordinates[0] < x.psCoordinates[0];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compare the coordinates of this record to the coordinates of another
/// record for sorting purposes
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::operator > (const MapControlDataEntry_Std & x) const
{
	return psCoordinates[0] > x.psCoordinates[0];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compare the coordinates of this record to the coordinates of another
/// record for sorting purposes
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::operator <= (const MapControlDataEntry_Std & x) const
{
	return psCoordinates[0] <= x.psCoordinates[0];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compare the coordinates of this record to the coordinates of another
/// record for sorting purposes
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::operator >= (const MapControlDataEntry_Std & x) const
{
	return psCoordinates[0] >= x.psCoordinates[0];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Shortest distance from this record to some point, optionally returning
/// which segment of this record was used for the calculation too.
///
//////////////////////////////////////////////////////////////////////////////
double MapControlDataEntry_Std::Distance(const Point & ptSearch, int * piBestSegment) const
{
	double fBestDistance = INFINITY;

	if (piBestSegment)
		*piBestSegment = -1;

	if (nCoordinates > 1 && psCoordinates[0] != psCoordinates[1])
	{
		int iSeg;
		double fBestEstDistance = INFINITY;
		for (iSeg = 0; iSeg < nCoordinates - 1; iSeg++)
		{
			double x0 = ptSearch.GetLong();
			double y0 = ptSearch.GetLat();
			double x1 = psCoordinates[iSeg].GetLong();
			double y1 = psCoordinates[iSeg].GetLat();
			double x2 = psCoordinates[iSeg + 1].GetLong();
			double y2 = psCoordinates[iSeg + 1].GetLat();
			double x3 = 0;
			double y3 = 0;

			if (y2 != y1 && x2 != x1)
			{
				double slope0 = (y2 - y1) / (x2 - x1);
				double intercept0 = y1 - slope0 * x1;

				// perpendicular line to previous line, passing through search point
				double slope1 = -1 / slope0;
				double intercept1 = y0 - slope1 * x0;

				// here we have the distance from the target coordinate
				// to segment iSeg of record iRec, assuming the segment extended
				// infinitely in both directions
				SolveLinearEquationSet(-1 * slope0, 1, intercept0, -1 * slope1, 1, intercept1, x3, y3);
			}
			else
			{
				// TODO: implement
			}
			Point ptProjected(x3, y3);

			double fEstDistance = (ptProjected - ptSearch).abs();
			double fRealDistance = ::Distance(ptProjected, ptSearch);

			// here we've got the distance from the endpoints to the target
			double h1 = (psCoordinates[iSeg] - ptSearch).abs();
			double h2 = (psCoordinates[iSeg + 1] - ptSearch).abs();

			// is the projected point between x1 and x2?
			if ((x2 > x1 && x2 > x3 && x3 > x1) || (x1 > x2 && x1 > x3 && x3 > x2))
			{
				// yes
				//wxASSERT(h1 >= fDistance);
				//wxASSERT(h2 >= fDistance);

				// sometimes occurs since earth is not flat?
				if (h1 < fEstDistance)
				{
					fEstDistance = h1;
					fRealDistance = ::Distance(ptSearch, psCoordinates[iSeg]);
				}
				if (h2 < fEstDistance)
				{
					fEstDistance = h2;
					fRealDistance = ::Distance(ptSearch, psCoordinates[iSeg + 1]);
				}
			}
			else
			{
				// no
				if (h1 > h2)
				{
					fEstDistance = h2;
					fRealDistance = ::Distance(ptSearch, psCoordinates[iSeg + 1]);
				}
				else
				{
					fEstDistance = h1;
					fRealDistance = ::Distance(ptSearch, psCoordinates[iSeg]);
				}
			}

			if (fEstDistance < fBestEstDistance)
			{
				fBestEstDistance = fEstDistance;
				fBestDistance = fRealDistance;

				if (piBestSegment)
					*piBestSegment = iSeg;
			}
		}
	}
	else // otherwise, nCoordinates == 1
	{
		fBestDistance = ::Distance(psCoordinates[0], ptSearch);

		if (piBestSegment)
			*piBestSegment = 0;
	}

	return fBestDistance;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Saves this record to the supplied file handle
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataEntry_Std::Serialize(FILE * hFile, bool bSave, void * pvArg)
{
	int iEntry;
	Point * pptRef = (Point *) pvArg;

	SerializeItem(hFile, bSave, &idRecord, NULL);

	SerializeItem(hFile, bSave, &nFeatureNames);

	if (!bSave)
	{
		pszFeatureNames = new char * [nFeatureNames];
		pszFeatureTypes = new char * [nFeatureNames];
	}

	for (iEntry = 0; iEntry < nFeatureNames; iEntry++)
	{
		if (bSave)
		{
			wxUint32 uiTableEntry;
			
			uiTableEntry = g_pcStringTable->GetEntryIndex(pszFeatureNames[iEntry], true);
			SerializeItem(hFile, bSave, &uiTableEntry);

			uiTableEntry = g_pcStringTable->GetEntryIndex(pszFeatureTypes[iEntry], true);
			SerializeItem(hFile, bSave, &uiTableEntry);
		}
		else
		{
			wxUint32 uiTableEntry;
			
			SerializeItem(hFile, bSave, &uiTableEntry);
			pszFeatureNames[iEntry] = g_pcStringTable->GetEntryPointer(uiTableEntry);
			
			SerializeItem(hFile, bSave, &uiTableEntry);
			pszFeatureTypes[iEntry] = g_pcStringTable->GetEntryPointer(uiTableEntry);
		}
	}

	if (bSave)
		fwrite(&eType, sizeof(eType), 1, hFile);
	else
		fread(&eType, sizeof(eType), 1, hFile);

	SerializeItem(hFile, bSave, &nAddressZipRanges);

	if (bSave)
	{
		fwrite(psAddressZipRanges, sizeof(SAddressRangeZip), nAddressZipRanges, hFile);
	}
	else
	{
		psAddressZipRanges = new SAddressRangeZip [nAddressZipRanges];
		fread(psAddressZipRanges, sizeof(SAddressRangeZip), nAddressZipRanges, hFile);
	}

	SerializeItem(hFile, bSave, &nCoordinates);

	if (!bSave)
		psCoordinates = new Point[nCoordinates];

	wxASSERT(pptRef);
	wxASSERT(nCoordinates > 0);
	SerializePointArray(hFile, bSave, psCoordinates, nCoordinates, *pptRef);
	*pptRef = psCoordinates[0];

	if (bSave)
	{
		wxUint32 uiTableEntry;
		
		uiTableEntry = g_pcStringTable->GetEntryIndex(pszSmartName, true);
		SerializeItem(hFile, bSave, &uiTableEntry);
	}
	else
	{
		wxUint32 uiTableEntry;
		
		SerializeItem(hFile, bSave, &uiTableEntry);
		pszSmartName = g_pcStringTable->GetEntryPointer(uiTableEntry);
	}

	if (bSave)
		fwrite(&eCountryCode, sizeof(eCountryCode), 1, hFile);
	else
		fread(&eCountryCode, sizeof(eCountryCode), 1, hFile);

#ifdef FILLED_IN_WATER_SUPPORT
	SerializeItem(hFile, bSave, &bWaterL);
	SerializeItem(hFile, bSave, &bWaterR);
	
	SerializeItem(hFile, bSave, &ptWaterL);
	SerializeItem(hFile, bSave, &ptWaterR); 
#endif
	
	return;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Release all memory used by this object
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataEntry_Std::Clear()
{
	delete [] pszFeatureNames;
	delete [] pszFeatureTypes;

	pszFeatureNames = NULL;
	pszFeatureTypes = NULL;
	nFeatureNames = 0;

	delete [] psAddressZipRanges;

	psAddressZipRanges = NULL;
	nAddressZipRanges = 0;

	delete [] psCoordinates;
	
	psCoordinates = NULL;
	nCoordinates = 0;

	pszSmartName = NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines if a point is contained within the region of this
/// record.
///
/// Only makes sense for state and county (polygon) records.
///
/// \return			true if ptCheck is inside the region, or false if it is
///					outside.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::ContainsPoint(const Point & ptCheck) const
{
	int iCrossings = 0;
	int iCoordinate;
	double fLongIntersection;

	for (iCoordinate = 0; iCoordinate < nCoordinates; iCoordinate++) 
	{
		Point pt1, pt2;

		pt1 = psCoordinates[iCoordinate];
		pt2 = psCoordinates[(iCoordinate + 1) % nCoordinates];

		if (
			ptCheck.GetLatRaw() > MIN(pt1.GetLatRaw(), pt2.GetLatRaw()) && 
			ptCheck.GetLatRaw() <= MAX(pt1.GetLatRaw(), pt2.GetLatRaw()) &&
			ptCheck.GetLongRaw() <= MAX(pt1.GetLongRaw(), pt2.GetLongRaw()) && 
			pt1.GetLatRaw() != pt2.GetLatRaw()
			)
		{
			fLongIntersection = (ptCheck.GetLat() - pt1.GetLat()) / (pt2.GetLat() - pt1.GetLat()) * (pt2.GetLong() - pt1.GetLong()) + pt1.GetLong();

			if (pt1.GetLongRaw() == pt2.GetLongRaw() || ptCheck.GetLong() <= fLongIntersection)
				iCrossings++;
		}
	}

	if (iCrossings % 2 == 0)
		return false;

	return true;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if this record represents a region.
///
/// \return			true if this record represents a region, or false if this
///					record represents a point/line.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::IsRegion() const
{
	return eType == RecordTypeCountyBoundary || eType == RecordTypeStateBoundary;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the 5 digit county + state code associated with this record.
///
/// Only applicable to county boundary records.
///
/// \return			5 digit county code for this record, or -1 on error.
///
//////////////////////////////////////////////////////////////////////////////
wxInt32 MapControlDataEntry_Std::CountyCode() const
{
	if (eType != RecordTypeCountyBoundary)
		return -1;

	if (nFeatureNames < 1)
		return -1;

	if (!isdigit(*pszFeatureNames[0]))
		return -1;

	return atoi(pszFeatureNames[0]);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Adjust the routing algorithm to prefer smaller or larger roads
///
/// \param fPreferSmallRoads	Multiplier giving extra weight to small roads
///
/// \param fPreferLargeRoads	Multiplier giving extra weight to large roads
///
/// \param fPreferHighways		Multiplier giving extra weight to highways
/// 
/// When a parameter is set to 1.0, the default weight is used. When the
/// parameter is between 0 and 1.0, that type of road is avoided. Greater
/// than 1.0 causes the routing algorithm to prefer that type of road.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataEntry_Std::SetRoutingPreferences(double fPreferSmallRoads, double fPreferLargeRoads, double fPreferHighways)
{
	m_fPreferSmallRoads = fPreferSmallRoads;
	m_fPreferLargeRoads = fPreferLargeRoads;
	m_fPreferHighways = fPreferHighways;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the bias towards or against small roads, large roads,
/// and highways.
///
/// \param pfPreferSmallRoads		Receives bias towards/against small roads.
///									Greater = prefer, smaller = avoid.
/// \param pfPreferSmallRoads		Receives bias towards/against large roads.
///									Greater = prefer, smaller = avoid.
/// \param pfPreferSmallRoads		Receives bias towards/against highways.
///									Greater = prefer, smaller = avoid.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataEntry_Std::GetRoutingPreferences(double * pfPreferSmallRoads, double * pfPreferLargeRoads, double * pfPreferHighways)
{
	if (pfPreferSmallRoads)
		*pfPreferSmallRoads = m_fPreferSmallRoads;
		
	if (pfPreferLargeRoads)
		*pfPreferLargeRoads = m_fPreferLargeRoads;
		
	if (pfPreferHighways)
		*pfPreferHighways = m_fPreferHighways;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines if this record passes through the specified rectangle.
///
/// \param ptTopLeft			Top left coordinates of rectangle to test.
/// \param ptBottomRight		Bottom right coordinates of rectangle to test.
/// \return						If intersects rectangle defined by ptTopLeft
///								and ptBottomRight.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataEntry_Std::IntersectsRect(const Point & ptTopLeft, const Point & ptBottomRight) const
{
	int iSeg;
	int iLastVertical = 0;
	int iLastHorizontal = 0;
	Point ptTopLeftChk(ptTopLeft);
	Point ptBottomRightChk(ptBottomRight);
	
	if (ptTopLeftChk.GetLongRaw() > ptBottomRightChk.GetLongRaw())
	{
		wxInt32 iTmp;
		
		iTmp = ptTopLeftChk.GetLongRaw();
		ptTopLeftChk.SetLongRaw(ptBottomRightChk.GetLongRaw());
		ptBottomRightChk.SetLongRaw(iTmp);
	}

	if (ptTopLeftChk.GetLatRaw() > ptBottomRightChk.GetLatRaw())
	{
		wxInt32 iTmp;
		
		iTmp = ptTopLeftChk.GetLatRaw();
		ptTopLeftChk.SetLatRaw(ptBottomRightChk.GetLatRaw());
		ptBottomRightChk.SetLatRaw(iTmp);
	}

	for (iSeg = 0; iSeg < nCoordinates; iSeg++)
	{
		int iThisVertical;
		int iThisHorizontal;
		
		if (psCoordinates[iSeg].GetLongRaw() < ptTopLeftChk.GetLongRaw())
			iThisHorizontal = -1;
		else if (psCoordinates[iSeg].GetLongRaw() > ptBottomRightChk.GetLongRaw())
			iThisHorizontal = 1;
		else
			iThisHorizontal = 0;
			
		
		if (psCoordinates[iSeg].GetLatRaw() < ptTopLeftChk.GetLatRaw())
			iThisVertical = -1;
		else if (psCoordinates[iSeg].GetLatRaw() > ptBottomRightChk.GetLatRaw())
			iThisVertical = 1;
		else
			iThisVertical = 0;
			
		if (iThisHorizontal == 0 && iThisVertical == 0)
			return true;
		
		if (iSeg != 0)
		{
			if (iThisHorizontal != iLastHorizontal &&
				iThisVertical != iLastVertical)
				return true;
			
			if (iThisHorizontal * iLastHorizontal == -1 &&
				iThisVertical == 0)
				return true;

			if (iThisVertical * iLastVertical == -1 &&
				iThisHorizontal == 0)
				return true;
			
		}
			
		iLastHorizontal = iThisHorizontal;
		iLastVertical = iThisVertical;
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Dumps the contents of this record into a string.
///
/// \return						Contents of this record.
///
//////////////////////////////////////////////////////////////////////////////
wxString MapControlDataEntry_Std::Dump() const
{
	wxString strRtn;
	int i;
	
	strRtn += wxT("Record ID: ") + idRecord.ToString() + wxT("\n");

	for (i = 0; i < nFeatureNames; i++)
	{
		strRtn += wxString::Format(wxT("Name #%d: "), i + 1);
		strRtn += wxString(pszFeatureNames[i], *wxConvCurrent);
		strRtn += wxT(" / ");
		strRtn += wxString(pszFeatureTypes[i], *wxConvCurrent);
		strRtn += wxT("\n");
	}

	strRtn += wxT("Smart Name: ") + wxString(pszSmartName, *wxConvCurrent) + wxT("\n");

	strRtn += wxString::Format(wxT("Record Type: %d\n"), (int) eType);

	strRtn += wxString::Format(wxT("Country Code: %d\n"), (int) eCountryCode);
	
	for (i = 0; i < nAddressZipRanges; i++)
	{
		strRtn += wxString::Format(wxT("Address Range #%d: %d - %d (%s), %d\n"), 
					i + 1,
					psAddressZipRanges[i].iFromAddr,
					psAddressZipRanges[i].iToAddr,
					psAddressZipRanges[i].bOnLeft ? wxT("left") : wxT("right"),
					psAddressZipRanges[i].iZip);
	}

	for (i = 0; i < nCoordinates; i++)
		strRtn += psCoordinates[i].ToString() + wxT("\n");
	
#ifdef FILLED_IN_WATER_SUPPORT
	strRtn += wxString::Format(wxT("Water on left at %s: %s\n"), ptWaterL.ToString(), bWaterL ? wxT("yes") : wxT("no"));
	strRtn += wxString::Format(wxT("Water on right at %s: %s\n"), ptWaterR.ToString(), bWaterR ? wxT("yes") : wxT("no"));
#endif

	return strRtn;	
}
