/*
 *  LibRoadnav
 *  MapControlDataImporter_CountyBoundaries.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Imports data defining the boundaries of the states in the US
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>

#include "Map.h"
#include "Constants.h"
#include "MapControlDataImporter_CountyBoundaries.h"
#include "MapSupport.h"
#include "MapAbbreviations.h"
#include "MapState.h"
#include "MapCounty.h"
#include "Portability.h"
#include "MapZip.h"
#include "Zip.h"
#include "URLs.h"
#include "UniqueIDGenerator.h"
#include "DownloadFiles.h"
#include "StringTable.h"

#include <map>

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor, optionally loads data from the specified directory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_CountyBoundaries::MapControlDataImporter_CountyBoundaries(char * pszSource) : m_psRecords(NULL), m_nRecords(0)
{
	if (pszSource)
		Load(wxString(pszSource, *wxConvCurrent));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_CountyBoundaries::~MapControlDataImporter_CountyBoundaries()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_CountyBoundaries::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;

	m_psRecords = NULL;
	
	m_nRecords = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database. This roughly corresponds
/// to the number of Type 1 records in the Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_CountyBoundaries::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlDataImporter_CountyBoundaries::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

bool MapControlDataImporter_CountyBoundaries::LoadIndex(wxString strBaseName)
{
	FILE * hFile;
	char szLine[2048];
	wxUint16 iEntry;

	m_mapIndex.clear();

	hFile = fopen(strBaseName.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (CountyBoundaries)!"), strBaseName.c_str()), wxT("Error"), wxOK);
		return true;
	}

	iEntry = 0;
	fgets(szLine, sizeof(szLine), hFile);

	while (!feof(hFile))
	{
		TrimWhitespace(szLine);

		if (*szLine == '"')
		{
			// remove trailing "
			szLine[strlen(szLine) - 1] = 0;

			m_mapIndex[iEntry].push_back(szLine + 1);
		}
		else
		{
			iEntry = atoi(szLine);
		}

		fgets(szLine, sizeof(szLine), hFile);
	}
	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Loads the data from the specified file into this object
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_CountyBoundaries::Load(wxString strBaseName)
{
	FILE * hFile;
	char szLine[2048];
	UniqueIDGenerator idGen;
	vector<int> vSegments;
	int iSegments;
	int iRecord;

	hFile = fopen(strBaseName.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (CountyBoundaries)!"), strBaseName.c_str()), wxT("Error"), wxOK);
		return true;
	}

	iSegments = 0;
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		char * pszToken1;
		char * pszToken2;
		char * pszToken3;

		pszToken1 = strtok(szLine, " \t\r\n");
		
		if (pszToken1)
			pszToken2 = strtok(NULL, " \t\r\n");
		else
			pszToken2 = NULL;

		if (pszToken2)
			pszToken3 = strtok(NULL, " \t\r\n");
		else
			pszToken3 = NULL;

		if (pszToken3 || (!pszToken2 && pszToken1 && atoi(pszToken1)))
		{
			// new ID
			iSegments = 0;
		}
		else if (pszToken2)
		{
			// polygon coordinates
			iSegments++;
		}
		else if (pszToken1)
		{
			if (iSegments)
			{
				vSegments.push_back(iSegments);
				iSegments = 0;
			}
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	m_psRecords = new MapControlDataEntry_Std[vSegments.size()];
	m_nRecords = vSegments.size();

	iRecord = 0;
	iSegments = 0;
	fseek(hFile, 0, SEEK_SET);
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		char * pszToken1;
		char * pszToken2;
		char * pszToken3;

		pszToken1 = strtok(szLine, " \t\r\n");
		
		if (pszToken1)
			pszToken2 = strtok(NULL, " \t\r\n");
		else
			pszToken2 = NULL;

		if (pszToken2)
			pszToken3 = strtok(NULL, " \t\r\n");
		else
			pszToken3 = NULL;

		if (pszToken3 || (!pszToken2 && pszToken1 && atoi(pszToken1)))
		{
			// new ID - remember ID, discard coordinates
			string strName;
			wxUint16 iID;

			iID = atoi(pszToken1);

			if (m_mapIndex.find(iID) != m_mapIndex.end())
				strName = m_mapIndex[iID][0] + m_mapIndex[iID][1] + " - " + m_mapIndex[iID][2];

			m_psRecords[iRecord].eType = RecordTypeCountyBoundary;
			m_psRecords[iRecord].nFeatureNames = 1;

			m_psRecords[iRecord].pszFeatureNames = new char * [1];
			m_psRecords[iRecord].pszFeatureNames[0] = g_pcStringTable->GetEntryPointer(strName.c_str(), true);

			m_psRecords[iRecord].pszFeatureTypes = new char * [1];
			m_psRecords[iRecord].pszFeatureTypes[0] = g_pcStringTable->GetEntryPointer("", true);
			
			m_psRecords[iRecord].pszSmartName = g_pcStringTable->GetEntryPointer(strName.c_str(), true);

			m_psRecords[iRecord].nAddressZipRanges = 0;
			m_psRecords[iRecord].psAddressZipRanges = NULL;

			m_psRecords[iRecord].eCountryCode = CountryCodeUnitedStates;

#ifdef FILLED_IN_WATER_SUPPORT
			m_psRecords[iRecord].bWaterL = false;
			m_psRecords[iRecord].bWaterR = false;
#endif

			m_psRecords[iRecord].psCoordinates = new Point [vSegments[iRecord]];
			m_psRecords[iRecord].nCoordinates = vSegments[iRecord];

			m_psRecords[iRecord].idRecord = idGen.MakeID();

			iSegments = 0;
		}
		else if (pszToken2)
		{
			// polygon coordinates
			double fLong = 0;
			double fLat = 0;

			fLong = atof(pszToken1);
			fLat = atof(pszToken2);

			m_psRecords[iRecord].psCoordinates[iSegments].Set(fLong, fLat);

			iSegments++;
		}
		else if (pszToken1)
		{
			// END
			if (iSegments)
			{
				iRecord++;
				iSegments = 0;
			}
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;

}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Check if the state boundary information has already been loaded
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_CountyBoundaries::IsLoaded()
{
	return IsDataIntegrated(wxT("CountyBoundaries"));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download the county boundaries for the entire US
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_CountyBoundaries::Download(IMapControlData * pData, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	bool bRtn;
	
	bRtn = false;
	
	if (!IsLoaded())
	{
		SyncDownloadEntry sEntry;

		sEntry.m_strURL = wxString::Format(wxT("%sco99_d00_ascii.zip"), GetURL(wxT("County Boundary Root")).c_str());
		sEntry.m_strOutputFilename = wxFileName::CreateTempFileName(wxT("CountyBoundaries"));
		sEntry.m_strProgressBarName = wxT("County Boundaries");

		if (DownloadFile(sEntry, true, pdlgProgress, iProgressStart, iProgressTicks / 2))
			bRtn = true;

		if (sEntry.m_bDownloadedOkay)
		{
			// Unzip
			Unzip(sEntry.m_strOutputFilename, GetDataPath());

			// remove zip file
			wxRemoveFile(sEntry.m_strOutputFilename);

			// figure out filename of unzipped file
			wxFileName fnInput;
			fnInput.AssignDir(GetDataPath());

			// process it
			fnInput.SetFullName(wxT("co99_d00a.dat"));
			if (!LoadIndex(fnInput.GetFullPath()))
			{
				fnInput.SetFullName(wxT("co99_d00.dat"));
				if (!Load(fnInput.GetFullPath()))
				{
					pData->Save(this, pdlgProgress, iProgressStart + iProgressTicks / 2, iProgressTicks / 2);
					MarkDataIntegrated(wxT("CountyBoundaries"));
				}
			}

			Clear();

			// remove the txt file 
			wxRemoveFile(fnInput.GetFullPath().c_str());

			fnInput.SetFullName(wxT("co99_d00a.dat"));
			wxRemoveFile(fnInput.GetFullPath().c_str());
		}
		else
		{
			bRtn = true;
		}
	}

	Clear();

	return bRtn;
}
