/*
 *  LibRoadnav
 *  MapControlDataImporter_GNISDECI.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// \brief Imports GNIS data - various landmarks (Delimited Format Files using pipe symbol)
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>
#include <wx/tokenzr.h>

#include "Map.h"
#include "Constants.h"
#include "MapControlDataImporter_GNISDECI.h"
#include "MapSupport.h"
#include "MapAbbreviations.h"
#include "MapState.h"
#include "MapCounty.h"
#include "Portability.h"
#include "MapZip.h"
#include "Zip.h"
#include "URLs.h"
#include "UniqueIDGenerator.h"
#include "DownloadFiles.h"
#include "StringTable.h"

#include <map>

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor, optionally loads data from the specified directory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_GNISDECI::MapControlDataImporter_GNISDECI(char * pszSource) : m_psRecords(NULL), m_nRecords(0)
{
	if (pszSource)
		Load(wxString(pszSource, *wxConvCurrent));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_GNISDECI::~MapControlDataImporter_GNISDECI()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_GNISDECI::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;

	m_psRecords = NULL;
	
	m_nRecords = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database. This roughly corresponds
/// to the number of Type 1 records in the Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_GNISDECI::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlDataImporter_GNISDECI::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the GNIS data from the supplied filename
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_GNISDECI::Load(wxString strBaseName)
{
	FILE * hFile;
	char szLine[2048];
	int nNewRecords;
	MapControlDataEntry_Std * psOldSegments = NULL;
	UniqueIDGenerator idGen;
	
	hFile = fopen(strBaseName.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (GNISDECI)!"), strBaseName.c_str()), wxT("Error"), wxOK);
		return true;
	}

	nNewRecords = 0;

	// skip first line (column headers)
	fgets(szLine, sizeof(szLine), hFile);

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		nNewRecords++;
		fgets(szLine, sizeof(szLine), hFile);
	}

	if (m_nRecords)
	{
		psOldSegments = m_psRecords;
	}
	m_psRecords = new MapControlDataEntry_Std[m_nRecords + nNewRecords];
	if (m_nRecords)
	{
		memcpy(m_psRecords, psOldSegments, sizeof(MapControlDataEntry_Std) * m_nRecords);
		delete [] psOldSegments;
	}
	
	fseek(hFile, 0, SEEK_SET);

	// skip first line (column headers)
	fgets(szLine, sizeof(szLine), hFile);

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		char szFeatureName[128] = "";
		char szState[3] = "";
		char szType[16] = "";
		vector<wxUint32> vZips;
		char * pszStripSuffixes[] = {", town of", ", city of", " County"};
		ERecordType eSuffixTypes[] = {RecordTypeTownOrCity, RecordTypeTownOrCity, RecordTypeLandmark};
		unsigned int iSuffix;
		wxStringTokenizer tkzLine;

		tkzLine.SetString(wxString(szLine, *wxConvCurrent), wxT("|\t"), wxTOKEN_RET_EMPTY_ALL);
		
		// Skip Feature ID
		tkzLine.GetNextToken();
		
		// Feature name
		strncpy(szFeatureName, tkzLine.GetNextToken().mb_str(*wxConvCurrent), 127);
		szFeatureName[127] = 0;

		// Type
		strncpy(szType, tkzLine.GetNextToken().mb_str(*wxConvCurrent), 15);
		szType[15] = 0;
		
		if (!CaseInsensitiveStrCmp(szType, "civil"))
			m_psRecords[m_nRecords].eType = RecordTypeTownOrCity;
		else
			m_psRecords[m_nRecords].eType = RecordTypeLandmark;

		for (iSuffix = 0; iSuffix < sizeof(pszStripSuffixes) / sizeof(char *); iSuffix++)
		{
			if (strlen(szFeatureName) >= strlen(pszStripSuffixes[iSuffix]))
			{
				char * pszSuffix = szFeatureName + strlen(szFeatureName) - strlen(pszStripSuffixes[iSuffix]);
				
				if (!CaseInsensitiveStrCmp(pszSuffix, pszStripSuffixes[iSuffix]))
				{
					*pszSuffix = 0;
					
					m_psRecords[m_nRecords].eType = eSuffixTypes[iSuffix];
				}
			}
		}
	
		// State
		strncpy(szState, tkzLine.GetNextToken().mb_str(*wxConvCurrent), 2);
		szState[2] = 0;

		
		m_psRecords[m_nRecords].pszFeatureNames = new char * [1];
		m_psRecords[m_nRecords].pszFeatureNames[0] = g_pcStringTable->GetEntryPointer(szFeatureName, true);

		m_psRecords[m_nRecords].pszFeatureTypes = new char * [1];
		m_psRecords[m_nRecords].pszFeatureTypes[0] = g_pcStringTable->GetEntryPointer("", true);

		m_psRecords[m_nRecords].nFeatureNames = 1;

		m_psRecords[m_nRecords].pszSmartName = g_pcStringTable->GetEntryPointer(szFeatureName, true);

		m_psRecords[m_nRecords].nAddressZipRanges = 0;
		m_psRecords[m_nRecords].psAddressZipRanges = NULL;

		m_psRecords[m_nRecords].eCountryCode = CountryCodeUnitedStates;

		m_psRecords[m_nRecords].psCoordinates = new Point [1];
		
		m_psRecords[m_nRecords].idRecord = idGen.MakeID();
		
		double fLong = 0;
		double fLat = 0;
		
		// Skip County
		tkzLine.GetNextToken();
		
		// Skip State Code
		tkzLine.GetNextToken();
		
		// Skip County Code
		tkzLine.GetNextToken();
		
		// Skip Latitude in non-floating point format
		tkzLine.GetNextToken();

		// Skip Longitude in non-floating point format
		tkzLine.GetNextToken();

		// latitude
		fLat = atof(tkzLine.GetNextToken().mb_str(*wxConvCurrent));

		// longitude
		fLong = atof(tkzLine.GetNextToken().mb_str(*wxConvCurrent));
		
		m_psRecords[m_nRecords].psCoordinates[0].SetLong(fLong);
		m_psRecords[m_nRecords].psCoordinates[0].SetLat(fLat);
		m_psRecords[m_nRecords].nCoordinates = 1;

#ifdef FILLED_IN_WATER_SUPPORT
		m_psRecords[m_nRecords].bWaterL = false;
		m_psRecords[m_nRecords].bWaterR = false;
#endif

		vZips = ZipCodesByCityState(wxString(szFeatureName, *wxConvCurrent), wxString(szState, *wxConvCurrent));
		if (vZips.size())
		{
			m_psRecords[m_nRecords].psAddressZipRanges = new struct SAddressRangeZip [1];
			m_psRecords[m_nRecords].psAddressZipRanges[0].iFromAddr = 0;
			m_psRecords[m_nRecords].psAddressZipRanges[0].iToAddr = 0;
			m_psRecords[m_nRecords].psAddressZipRanges[0].iZip = vZips[0];
			m_psRecords[m_nRecords].nAddressZipRanges = 1;
		}

		m_nRecords++;

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is the GNIS landmark data for a particular state loaded? Returns true if
/// it is, and false otherwise.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_GNISDECI::IsGNISStateLoaded(wxString strState)
{
	wxString strStateAbbreviation;
	
	// take state provided
	strStateAbbreviation = strState;

	// convert to abbreviation, if necessary
	if (StateAbbreviationByName(strStateAbbreviation) != wxT(""))
		strStateAbbreviation = StateAbbreviationByName(strStateAbbreviation);

	// make upper case
	strStateAbbreviation = strStateAbbreviation.MakeUpper();

	// check
	return IsDataIntegrated(wxT("GNIS-") + strStateAbbreviation);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download the landmark data for a set of states. Takes a vector of
/// state abbreviations.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_GNISDECI::DownloadGNISStates(IMapControlData * pData, std::vector<wxString> vstrStates, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	unsigned int i;
	int iFile;
	vector<SyncDownloadEntry> vFiles;
	int iFilesProcessed;
	bool bRtn;
	
	bRtn = false;

	for (i = 0; i < vstrStates.size(); i++)
	{
		wxString strStateAbbreviation;
		
		strStateAbbreviation = vstrStates[i];
		
		if (StateAbbreviationByName(strStateAbbreviation) != wxT(""))
			strStateAbbreviation = StateAbbreviationByName(strStateAbbreviation);

		strStateAbbreviation = strStateAbbreviation.MakeUpper();

		if (!IsGNISStateLoaded(strStateAbbreviation))
		{
			SyncDownloadEntry sEntry;
			wxString strStateFull;

			strStateFull = StateNameByAbbreviation(strStateAbbreviation);

			sEntry.m_strProgressBarName = strStateFull;

			strStateFull.Replace(wxT(" "), wxT("_"));
			strStateFull = strStateFull.MakeUpper();

			sEntry.m_strURL = wxString::Format(wxT("%s%s_DECI.zip"), GetURL(wxT("GNIS DECI Root")).c_str(), strStateAbbreviation.c_str());
			sEntry.m_strOutputFilename = wxFileName::CreateTempFileName(strStateFull);

			vFiles.push_back(sEntry);
		}
	}

	if (DownloadFiles(vFiles, true, pdlgProgress, iProgressStart, iProgressTicks / 2))
		bRtn = true;

	for (i = 0, iFile = 0, iFilesProcessed = 0; i < vstrStates.size(); i++)
	{
		wxString strStateAbbreviation;
		
		strStateAbbreviation = vstrStates[i];
		
		if (StateAbbreviationByName(strStateAbbreviation) != wxT(""))
			strStateAbbreviation = StateAbbreviationByName(strStateAbbreviation);

		strStateAbbreviation = strStateAbbreviation.MakeUpper();

		if (!IsGNISStateLoaded(strStateAbbreviation))
		{
			if (vFiles[iFile].m_bDownloadedOkay)
			{
				wxFileName fnInput;
				wxString strStateFull;

				strStateFull = vFiles[iFile].m_strProgressBarName;

				strStateFull.Replace(wxT(" "), wxT("_"));
				strStateFull = strStateFull.MakeUpper();

				// unzip
				Unzip(vFiles[iFile].m_strOutputFilename, GetDataPath());

				// remove zip file
				wxRemoveFile(vFiles[iFile].m_strOutputFilename);

				// determine filename of file inside zip file
				fnInput.AssignDir(GetDataPath());
				fnInput.SetFullName(wxString::Format(wxT("%s_deci.txt"), strStateAbbreviation.c_str()));
				
				if (!wxFileExists(fnInput.GetFullPath()))
					fnInput.SetFullName(wxString::Format(wxT("%s_DECI.txt"), strStateAbbreviation.c_str()));

				if (!wxFileExists(fnInput.GetFullPath()))
					fnInput.SetFullName(wxString::Format(wxT("%s_DECI.TXT"), strStateAbbreviation.c_str()));

				// process data
				if (!Load(fnInput.GetFullPath()))
				{
					pData->Save(this, pdlgProgress, iProgressStart + iProgressTicks / 2 + iProgressTicks / 2 * iFilesProcessed / vFiles.size(), iProgressTicks / 2 / vFiles.size());
					MarkDataIntegrated(wxT("GNIS-") + strStateAbbreviation);
				}

				iFilesProcessed++;

				Clear();

				// remove the txt file 
				wxRemoveFile(fnInput.GetFullPath().c_str());
			}
			else
			{
				bRtn = true;
			}

			iFile++;
		}
	}

	Clear();

	return bRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download the landmark data for a particular state. Specify the state by
/// its abbreviation.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_GNISDECI::DownloadGNISState(IMapControlData * pData, wxString strStates, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	vector<wxString> v;
	v.push_back(strStates);
	
	return DownloadGNISStates(pData, v, pdlgProgress, iProgressStart, iProgressTicks);
}
