/*
 *  LibRoadnav
 *  MapControlDataImporter_GPX.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Imports GPX data.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>
#include <wx/tokenzr.h>

#include <list>

#include "tinyxml.h"

#include "Map.h"
#include "Constants.h"
#include "MapControlDataImporter_GPX.h"
#include "MapSupport.h"
#include "MapState.h"
#include "MapCounty.h"
#include "Portability.h"
#include "MapZip.h"
#include "Zip.h"
#include "URLs.h"
#include "UniqueIDGenerator.h"
#include "DownloadFiles.h"
#include "StringTable.h"
#include "MapMarker.h"
#include "MapLookup.h"

using namespace std;

#define	STREETNAME_SIZE		256
#define	STREETTYPE_SIZE		256

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_GPX::MapControlDataImporter_GPX() : m_psRecords(NULL), 
															m_nRecords(0), 
															m_pcMarkers(NULL),
															m_nMarkers(0)
{
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_GPX::~MapControlDataImporter_GPX()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_GPX::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;
	m_psRecords = NULL;
	
	m_nRecords = 0;

	if (m_pcMarkers)
		delete [] m_pcMarkers;
	m_pcMarkers = NULL;
	m_nMarkers = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_GPX::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of markers
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_GPX::MarkerCount()
{
	return m_nMarkers;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlDataImporter_GPX::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular marker by its index number
///
//////////////////////////////////////////////////////////////////////////////
MapMarker * MapControlDataImporter_GPX::GetMarker(wxUint32 iIndex)
{
	wxASSERT(iIndex < m_nMarkers);
	return &m_pcMarkers[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load data from a GPX file this object. Loads
/// entire file into memory, so can't be used on huge GPX files.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_GPX::Load(wxString strFilename, IMapControlData * pMapControlData, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks, Point * ptCenter, ERecordType eDefaultType)
{
	TiXmlDocument doc;
	TiXmlElement * xelRoot;
	TiXmlElement * xelTrk;
	TiXmlElement * xelWpt;
	UniqueIDGenerator idGen;
	int iRecord;
	wxFileName fn;
	wxString strDefaultName;

	fn.Assign(strFilename);
	strDefaultName = fn.GetName();
	strDefaultName = strDefaultName.Left(STREETNAME_SIZE - 5);

	pdlgProgress->SetLabel(wxT("Building maps"));

	Clear();

	doc.SetValue(strFilename.mb_str(*wxConvCurrent));
	doc.LoadFile();

	xelRoot = doc.RootElement();

	if (!xelRoot)
		return true;

	// count the tracks
	m_nRecords = 0;
	for (xelTrk = xelRoot->FirstChildElement("trk"); xelTrk; xelTrk = xelTrk->NextSiblingElement("trk"))
	{
		TiXmlElement * xelTrkSeg;

		for (xelTrkSeg = xelTrk->FirstChildElement("trkseg"); xelTrkSeg; xelTrkSeg = xelTrkSeg->NextSiblingElement("trkseg"))
			m_nRecords++;
	}

	// Set up memory
	m_psRecords = new MapControlDataEntry_Std[m_nRecords];
		
	// iterate through the tracks
	iRecord = 0;
	for (xelTrk = xelRoot->FirstChildElement("trk"); xelTrk; xelTrk = xelTrk->NextSiblingElement("trk"))
	{
		TiXmlElement * xelTrkSeg;
		ERecordType eType = eDefaultType;

		// Name of feature
		char szName[STREETNAME_SIZE + 1] = "";

		// Type of feature
		char szType[STREETTYPE_SIZE + 1] = "";

		TiXmlElement * xelName;
		TiXmlElement * xelType;

		// Default name
		strcpy(szName, strDefaultName.mb_str(*wxConvCurrent));
		if (iRecord)
			sprintf(szName + strlen(szName), " #%d", iRecord + 1);

		// Read name from GPX data
		xelName = xelTrk->FirstChildElement("name");
		if (xelName)
		{
			if (xelName->Value())
			{
				strncpy(szName, xelName->Value(), STREETNAME_SIZE);
				szName[STREETNAME_SIZE - 1] = 0;
				szName[STREETNAME_SIZE] = 0;
			}
		}	

		// Read type from GPX data
		xelType = xelTrk->FirstChildElement("type");
		if (xelType)
		{
			if (xelType->Value())
			{
				strncpy(szType, xelType->Value(), STREETTYPE_SIZE);
				szType[STREETTYPE_SIZE] = 0;
			}
		}	

		// extract segments
		for (xelTrkSeg = xelTrk->FirstChildElement("trkseg"); xelTrkSeg; xelTrkSeg = xelTrkSeg->NextSiblingElement("trkseg"))
		{
			TiXmlElement * xelTrkPt;
			vector<Point> vCoordinates;
			int iCoordinateIndex;
			vector<Point>::iterator iCoordinateIter;

			// extract individual points
			for (xelTrkPt = xelTrkSeg->FirstChildElement("trkpt"); xelTrkPt; xelTrkPt = xelTrkPt->NextSiblingElement("trkpt"))
			{
				const char * pszLat;
				const char * pszLon;
				
				pszLat = xelTrkPt->Attribute("lat");
				pszLon = xelTrkPt->Attribute("lon");
				
				wxASSERT(pszLat);
				wxASSERT(pszLon);
				
				if (pszLat && pszLon)
				{
					Point ptThis(atof(pszLon), atof(pszLat));
					vCoordinates.push_back(ptThis);
				}
			}

			// construct record
			m_psRecords[iRecord].eType = eType;
			m_psRecords[iRecord].nFeatureNames = 1;

			m_psRecords[iRecord].pszFeatureNames = new char * [1];
			m_psRecords[iRecord].pszFeatureNames[0] = g_pcStringTable->GetEntryPointer(szName, true);

			m_psRecords[iRecord].pszFeatureTypes = new char * [1];
			m_psRecords[iRecord].pszFeatureTypes[0] = g_pcStringTable->GetEntryPointer(szType, true);

			m_psRecords[iRecord].pszSmartName = g_pcStringTable->GetEntryPointer(szName, true);

			m_psRecords[iRecord].nAddressZipRanges = 0;
			m_psRecords[iRecord].psAddressZipRanges = NULL;

			m_psRecords[iRecord].eCountryCode = CountryCodeUnknown;

#ifdef FILLED_IN_WATER_SUPPORT
			m_psRecords[iRecord].bWaterL = false;
			m_psRecords[iRecord].bWaterR = false;
#endif

			m_psRecords[iRecord].psCoordinates = new Point [vCoordinates.size()];
			m_psRecords[iRecord].nCoordinates = vCoordinates.size();
			
			for (iCoordinateIndex = 0, iCoordinateIter = vCoordinates.begin();
					iCoordinateIter != vCoordinates.end();
					iCoordinateIter++, iCoordinateIndex++)
			{
				m_psRecords[iRecord].psCoordinates[iCoordinateIndex] = *iCoordinateIter;

				if (ptCenter)
					*ptCenter = *iCoordinateIter;
			}

			m_psRecords[iRecord].idRecord = idGen.MakeID();

			iRecord++;

			pdlgProgress->SetProgress((int) (iProgressStart + ((double) iProgressTicks) * iRecord / m_nRecords / 2));
		}
	}

	// count the waypoints
	m_nMarkers = 0;
	for (xelWpt = xelRoot->FirstChildElement("wpt"); xelWpt; xelWpt = xelWpt->NextSiblingElement("wpt"))
		m_nMarkers++;

	// Set up memory
	m_pcMarkers = new MapMarker[m_nMarkers];
		
	// iterate through the waypoints
	m_nMarkers = 0;
	for (xelWpt = xelRoot->FirstChildElement("wpt"); xelWpt; xelWpt = xelWpt->NextSiblingElement("wpt"))
	{
		const char * pszLat = NULL;
		const char * pszLon = NULL;
		const char * pszName = NULL;
		TiXmlElement * xelName = NULL;
		char szDefaultName[STREETNAME_SIZE + 1] = "";

		strcpy(szDefaultName, strDefaultName.mb_str(*wxConvCurrent));
		if (m_nMarkers)
			sprintf(szDefaultName + strlen(szDefaultName), " #%d", m_nMarkers + 1);
		pszName = szDefaultName;
		
		pszLat = xelWpt->Attribute("lat");
		pszLon = xelWpt->Attribute("lon");
		
		wxASSERT(pszLat);
		wxASSERT(pszLon);

		xelName = xelWpt->FirstChildElement("name");
		if (xelName)
		{
			if (xelName->Value())
				pszName = xelName->Value();
		}
		
		if (pszLat && pszLon && pszName)
		{
			MapMarker cMarker;
			Point ptThis;
			Address cAddress;
			wxString strName(pszName, *wxConvCurrent);
				
			ptThis.Set(atof(pszLon), atof(pszLat));
			cAddress = FindCoordinates(ptThis, AllVisibility());

			m_pcMarkers[m_nMarkers].Init(strName, &cAddress);
			m_nMarkers++;

			if (ptCenter)
				*ptCenter = ptThis;
		}
	}

	pMapControlData->Save(this, pdlgProgress, iProgressStart + iProgressTicks / 2, iProgressTicks / 2);

	return false;
}
