/*
 *  LibRoadnav
 *  MapControlDataImporter_OSM.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Imports OpenStreetMap data.
///
//////////////////////////////////////////////////////////////////////////////



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>
#include <wx/tokenzr.h>

#include <list>
#include <algorithm>


#include "tinyxml.h"

#include "Map.h"
#include "Constants.h"
#include "MapControlDataImporter_OSM.h"
#include "MapSupport.h"
#include "MapState.h"
#include "MapCounty.h"
#include "Portability.h"
#include "MapZip.h"
#include "Zip.h"
#include "URLs.h"
#include "UniqueIDGenerator.h"
#include "DownloadFiles.h"
#include "StringTable.h"

using namespace std;

#define	STREETNAME_SIZE		256

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_OSM::MapControlDataImporter_OSM() : m_psRecords(NULL), m_nRecords(0)
{
	m_fDownloadGranularity = 0.2;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_OSM::~MapControlDataImporter_OSM()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_OSM::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;
	m_psRecords = NULL;
	
	m_nRecords = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_OSM::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlDataImporter_OSM::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

struct SLineSegment
{
	Point m_ptFirst;
	Point m_ptSecond;
	
	bool m_bUsed;
};

typedef pair<int,int> tLineSegment;

class COSMNode
{
public:
	int nodeId;
	float f1;
	float f2;
	COSMNode()
	{
	}
	~COSMNode()
	{
	}
	bool operator<(const COSMNode& rhs) const{ return nodeId<rhs.nodeId; }

};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the requested attribute from the supplied XML data.
///
/// \param pszLine			Line of the XML file containing '... attribute="xyz"'
/// \param pszAttribute		Name of attribute to look for
/// \return					Value of requested attribute.
///
//////////////////////////////////////////////////////////////////////////////
static char * GetAttribute(char * pszLine, char * pszAttribute)
{
	int iOffset;
	static char szRtn[256];
	int iLenAttribute = strlen(pszAttribute);
	struct lconv * pLocaleInfo;

	pLocaleInfo = localeconv();
	wxASSERT(pLocaleInfo);
	
	for (iOffset = 0; pszLine[iOffset]; iOffset++)
	{
		if (!strncmp(pszLine + iOffset, pszAttribute, iLenAttribute) && 
				pszLine[iOffset + iLenAttribute] == '=' &&
				(pszLine[iOffset + iLenAttribute + 1] == '"' || pszLine[iOffset + iLenAttribute + 1] == '\'')
			)
		{
			int iPos;
			
			strncpy(szRtn, pszLine + iOffset + strlen(pszAttribute) + 2, 256);
			

			for (iPos = 0; szRtn[iPos] != pszLine[iOffset + iLenAttribute + 1] && iPos < 255; iPos++)
			{
				// XML file contains only dots as seperator. Let's convert them to current locale
				if (szRtn[iPos] == '.') 
					szRtn[iPos] = pLocaleInfo->decimal_point[0];
			}
			szRtn[iPos] = 0;
			
			return szRtn;
		}
	}
	
	return "";
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Converts a tag name/value into a LibRoadnav road type.
///
/// \param pszK				Attribute name/
/// \param pszV				Attribute value.
/// \param peType			Receives LibRoadnav road type.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_OSM::ClassifyType(const char * pszK, const char * pszV, ERecordType * peType)
{
	if (!strcmp(pszK, "highway"))
	{
		if (!strncmp(pszV, "motorway", 8))
			*peType = RecordTypeTwoWayHighway;
		else if (!strncmp(pszV, "trunk", 5))
			*peType = RecordTypeTwoWayHighway;
		else if (!strncmp(pszV, "primary", 7))
			*peType = RecordTypeTwoWayHighway;
		else if (!strncmp(pszV, "secondary", 9))
			*peType = RecordTypeTwoWayLargeRoad;
		else if (!strncmp(pszV, "minor", 5))
			*peType = RecordTypeTwoWaySmallRoad;
		// AF check
		else if (!strncmp(pszV, "tertiary", 5))
			*peType = RecordTypeTwoWaySmallRoad;

		else if (!strncmp(pszV, "unclassified", 12))
			*peType = RecordTypeTwoWaySmallRoad;
		else if (!strncmp(pszV, "track", 5))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "residential", 11))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "service", 7))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "bridleway", 9))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "cycleway", 8))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "footway", 7))
			*peType = RecordTypeDefault;
		else if (!strncmp(pszV, "steps", 5))
			*peType = RecordTypeDefault;
	}
	
	if (!strcmp(pszK, "railway"))
	{
		*peType = RecordTypeRailroad;
	}

	if (!strcmp(pszK, "waterway"))
	{
		*peType = RecordTypeWater;
	}

	if (!strcmp(pszK, "leisure"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "man_made"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "amenity"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "tourism"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "historic"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "natural"))
	{
		*peType = RecordTypeLandmark;
	}

	if (!strcmp(pszK, "boundary"))
	{
		// AF *peType = RecordTypeStateBoundary;
	}

	if (!strcmp(pszK, "place"))
	{
		if (!strncmp(pszV, "city", 4))
		{
			*peType = RecordTypeTownOrCity;
		}

		*peType = RecordTypeTownOrCity;
	}
	}

int CompareOsmNodes( const void* a, const void* b )
{
	COSMNode* oa = (COSMNode*)a;
	COSMNode* ob = (COSMNode*)b;

	if ( oa->nodeId >  ob->nodeId ) return 1;
	if ( oa->nodeId <  ob->nodeId ) return -1;

	return 0;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load data from an OSM XML API 3 file, directly into a
/// IMapControlData. Uses lots of shortcuts to keep memory consumption
/// reasonable while processing planet.osm.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_OSM::LoadDirect(wxString strFilename, IMapControlData * pMapControlData, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks, Point * ptCenter)
{
	UniqueIDGenerator idGen;
	vector<COSMNode> vOsmNodes;
	FILE * hFile = NULL;
	int nSegments = 0;
	char szLine[128  * 1024];
	int iMaxRecords = 50000;
	int nElements;
	int iElement;
	double fLongTotal = 0;
	double fLatTotal = 0;
	int nNodes = 0;

	if ( pdlgProgress != NULL )
	{
	pdlgProgress->SetLabel(wxT("Building maps"));
	}
	
	Clear();
	
	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");

	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	nElements = 0;
	fgets(szLine, sizeof(szLine) / sizeof(char), hFile);

	while (!feof(hFile))
	{
		char * pszStartLine;
		
		// eat up white spaces at start of the line
		for (pszStartLine = szLine; wxIsspace(*pszStartLine); pszStartLine++);

		if (!strncmp(pszStartLine, "<node", 5))
		{
			int iID;
			double fLong;
			double fLat;

			iID = atoi(GetAttribute(pszStartLine, "id"));
			fLong = atof(GetAttribute(pszStartLine, "lon"));
			fLat = atof(GetAttribute(pszStartLine, "lat"));
			
			fLongTotal += fLong;
			fLatTotal += fLat;
			nNodes++;

			COSMNode osmNode;
			osmNode.f1 = fLong;
			osmNode.f2 = fLat;
			osmNode.nodeId = iID;
			vOsmNodes.push_back( osmNode );
		}
		
		
		if (!strncmp(pszStartLine, "<way", 4))
		{
			nElements ++;
		}
			
		fgets(szLine, sizeof(szLine) / sizeof(char), hFile);
	}

	// sort the nodes, so we can perform a binaray search
	// when looking up de node references
	std::sort( vOsmNodes.begin(), vOsmNodes.end() );

	// nodes have been processed, so start at the beginning
	// of the file again
	fseek(hFile, 0, SEEK_SET);
	
	// Set up memory
	m_psRecords = new MapControlDataEntry_Std[iMaxRecords];
	m_nRecords = 0;
	
	iElement = 0;
	fgets(szLine, sizeof(szLine) / sizeof(char), hFile);
	while (!feof(hFile))
	{
		char * pszStartLine;
		bool bReadLine = true;
		
		for (pszStartLine = szLine; wxIsspace(*pszStartLine); pszStartLine++);
		
		if (!strncmp(pszStartLine, "<node", 5))
		{
		}
		else if (!strncmp(pszStartLine, "<way", 4))
		{
			// Name of feature
			
			char szName[STREETNAME_SIZE + 1] = {'\0',};
			char szNatRefName[STREETNAME_SIZE + 1] = {'\0',};
			char szRefName[STREETNAME_SIZE + 1] = {'\0',};
			bool isOneWayStreet = false;
			
			// Coordinates belonging to feature
			list<Point> lCoordinates;
			list<Point>::iterator iCoordinateIter;
			int iCoordinateIndex;
			vector<SLineSegment> vSegments;
			vector<SLineSegment>::iterator iSegment;
			
			// Feature type
			ERecordType eType = RecordTypeDefault;
			
			bReadLine = false;

			fgets(szLine, sizeof(szLine) / sizeof(char), hFile);
			
			for (pszStartLine = szLine; wxIsspace(*pszStartLine); pszStartLine++);
			
			while (!feof(hFile) && strncmp(pszStartLine, "<way", 4))
			{
				if (!strncmp(pszStartLine, "<nd", 3))
				{
					int iID;

					iID = atoi(GetAttribute(pszStartLine, "ref"));
						
					COSMNode* pbs = (COSMNode*) bsearch(
						&iID, 
						(void *)&vOsmNodes[0], 
						vOsmNodes.size(),
						sizeof( COSMNode ),
						(int (*)(const void*, const void*))CompareOsmNodes );
						
					if ( pbs != NULL )
						{
						lCoordinates.push_back( Point( pbs->f1, pbs->f2 ) );
						}
					}
				
				if (!strncmp(pszStartLine, "<tag", 4))
				{
					char pszK[256];
					char pszV[256];
					
					strcpy(pszK, GetAttribute(pszStartLine, "k"));
					strcpy(pszV, GetAttribute(pszStartLine, "v"));
					
					if (!strcmp(pszK, "name"))
					{
						strncpy(szName, pszV, STREETNAME_SIZE);
						szName[STREETNAME_SIZE] = 0;
					}
					if (!strcmp(pszK, "ref"))
					{
						strncpy( szRefName, pszV, STREETNAME_SIZE );
						szRefName[ STREETNAME_SIZE ] = 0;
					}					
					if (!strcmp(pszK, "nat_ref"))
					{
						strncpy( szNatRefName, pszV, STREETNAME_SIZE );
						szNatRefName[ STREETNAME_SIZE ] = 0;
					}					
					if (!strcmp(pszK, "oneway"))
					{
						if ( pszV[0] == '1' ) 
						{
							isOneWayStreet = true;
						}
						else
						{
							isOneWayStreet = false;
						}
					}
					
					ClassifyType(pszK, pszV, &eType);
				} // if (!strncmp(pszStartLine, "<tag", 4))
				
				fgets(szLine, sizeof(szLine) / sizeof(char), hFile);
				for (pszStartLine = szLine; wxIsspace(*pszStartLine); pszStartLine++);

			} // while (strncmp(pszStartLine, "<way", 4))
			
			if ( 0 == strlen( szName ) )
						{
				//if ( ( strlen( szNatRefName ) > 0 )  && ( strlen(  szRefName ) > 0 ) )
				//{
				//	sprintf( szName, "%s/%s", szNatRefName, szRefName );
				//} 
				if ( ( strlen( szNatRefName ) > 0 ) )
							{
					strcpy( szName, szNatRefName );
							}
				else if ( ( strlen( szRefName ) > 0 ) )
							{
					strcpy( szName, szRefName );
							}
			}
			else
							{
				if ( ( strlen( szNatRefName ) > 0 ) )
				{
					sprintf( szName, "%s (%s)", szName, szNatRefName );
							}
				else if ( ( strlen( szRefName ) > 0 ) )
							{
					sprintf( szName, "%s (%s)", szName, szRefName );
							}
						}
						

			// handle oneway streets
			if ( isOneWayStreet )
			{
				switch (eType )
				{
					case RecordTypeTwoWaySmallRoad: eType = RecordTypeOneWaySmallRoad; break;
					case RecordTypeTwoWayLargeRoad: eType = RecordTypeOneWayLargeRoad; break;
					case RecordTypeTwoWayHighway: eType = RecordTypeOneWayHighway; break;
				}
			}
			
		
			// construct record
			m_psRecords[m_nRecords].eType = eType;
			m_psRecords[m_nRecords].nFeatureNames = 1;

			m_psRecords[m_nRecords].pszFeatureNames = new char * [1];
			m_psRecords[m_nRecords].pszFeatureNames[0] = g_pcStringTable->GetEntryPointer(szName, true);

			m_psRecords[m_nRecords].pszFeatureTypes = new char * [1];
			m_psRecords[m_nRecords].pszFeatureTypes[0] = g_pcStringTable->GetEntryPointer("", true);

			m_psRecords[m_nRecords].pszSmartName = g_pcStringTable->GetEntryPointer(szName, true);

			m_psRecords[m_nRecords].nAddressZipRanges = 0;
			m_psRecords[m_nRecords].psAddressZipRanges = NULL;

			m_psRecords[m_nRecords].eCountryCode = CountryCodeUnknown;

#ifdef FILLED_IN_WATER_SUPPORT
			m_psRecords[m_nRecords].bWaterL = false;
			m_psRecords[m_nRecords].bWaterR = false;
#endif

			// test unicode stuff
			// 55.74 N, 37.62 E :55.74 N, 37.63 E
			// 55.70 N, 37.60 E :55.80 N, 37.70 E

			m_psRecords[m_nRecords].psCoordinates = new Point [lCoordinates.size()];
			m_psRecords[m_nRecords].nCoordinates = lCoordinates.size();
			
			for (iCoordinateIndex = 0, iCoordinateIter = lCoordinates.begin();
					iCoordinateIter != lCoordinates.end();
					iCoordinateIter++, iCoordinateIndex++)
			{
				m_psRecords[m_nRecords].psCoordinates[iCoordinateIndex] = *iCoordinateIter;
			}

			m_psRecords[m_nRecords].idRecord = idGen.MakeID();

			m_nRecords++;
			
			if (m_nRecords == iMaxRecords)
			{
				if ( pMapControlData != NULL )
				{
				pMapControlData->Save(this, pdlgProgress, (int) (iProgressStart + ((double) iProgressTicks) * iElement / nElements), 1);
				Clear();
				m_psRecords = new MapControlDataEntry_Std[iMaxRecords];
				m_nRecords = 0;
			}
			
			}
			
			iElement++;
		}

		// update progress information when possible
		if ( pdlgProgress != NULL )
		{
			pdlgProgress->SetProgress((int) (iProgressStart + ((double) iProgressTicks) * iElement / nElements));
		}
		
		if (bReadLine)
		{
			fgets(szLine, sizeof(szLine) / sizeof(char), hFile);
		}
	}

	// were done, close the file
	fclose(hFile);
	
	if ( pMapControlData != NULL )
	{
	pMapControlData->Save(this, pdlgProgress, (int) (iProgressStart + ((double) iProgressTicks) * iElement / nElements), 1);
	Clear();
	}
	
	if ( ptCenter != NULL )
	{
		if (nNodes)
			*ptCenter = Point(fLongTotal / nNodes, fLatTotal / nNodes);
		else
			ptCenter->SetInvalid();
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load data from an OSM XML API 3 file into this object. Loads
/// entire file into memory, so can't be used on huge OSM files.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_OSM::Load(wxString strFilename)
{
	return LoadDirect( strFilename, NULL, NULL, 0, 100, NULL );
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if specified point is loaded or not
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_OSM::IsLoaded(Point pt)
{
	int iLong;
	int iLat;
	
	iLong = (int) (pt.GetLong() / m_fDownloadGranularity);
	iLat = (int) (pt.GetLat() / m_fDownloadGranularity);
	
	return IsLoaded(
					Point(iLong * m_fDownloadGranularity, iLat * m_fDownloadGranularity), 
					Point((iLong + 1) * m_fDownloadGranularity, (iLat + 1) * m_fDownloadGranularity)
					);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if specified range is loaded or not. Must be
/// m_fDownloadGranularity large and aligned properly.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_OSM::IsLoaded(Point ptTL, Point ptBR)
{
	return IsDataIntegrated(wxString::Format(wxT("OSM-%.2f,%.2f-%.2f,%.2f"), ptTL.GetLong(), ptTL.GetLat(), ptBR.GetLong(), ptBR.GetLat()));
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Downloads a single m_fDownloadGranularity by m_fDownloadGranularity
/// slice of map.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_OSM::DownloadSlice(IMapControlData * pData, Point ptTL, Point ptBR, wxString strUsername, wxString strPassword, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	bool bRtn = false;
	
	if (!IsLoaded(ptTL, ptBR))
	{
		SyncDownloadEntry sEntry;

		sEntry.m_strURL = wxString::Format(wxT("%s?bbox=%.1f,%.1f,%.1f,%.1f"),
												GetURL(wxT("OpenStreetMaps")).c_str(),
												ptTL.GetLong() > ptBR.GetLong() ? ptBR.GetLong() : ptTL.GetLong(),
												ptTL.GetLat() > ptBR.GetLat() ? ptBR.GetLat() : ptTL.GetLat(),
												ptTL.GetLong() > ptBR.GetLong() ? ptTL.GetLong() : ptBR.GetLong(),
												ptTL.GetLat() > ptBR.GetLat() ? ptTL.GetLat() : ptBR.GetLat()
												);
		sEntry.m_strOutputFilename = wxFileName::CreateTempFileName(wxString::Format(wxT("%s_osm"), wxT("LibRoadnav")));
		sEntry.m_strProgressBarName = wxString::Format(wxT("OpenStreetMaps (%s - %s)"), ptTL.FormatPoint().c_str(), ptBR.FormatPoint().c_str());
		sEntry.m_strUsername = strUsername;
		sEntry.m_strPassword = strPassword;
		sEntry.m_iRetryAttempts = 5;

		if (DownloadFile(sEntry, true, pdlgProgress, iProgressStart, iProgressTicks / 2))
			bRtn = true;

		if (sEntry.m_bDownloadedOkay)
		{
			if (!Load(sEntry.m_strOutputFilename))
			{
				pData->Save(this, pdlgProgress, iProgressStart + iProgressTicks / 2, iProgressTicks / 2);
				
				MarkDataIntegrated(wxString::Format(wxT("OSM-%.2f,%.2f-%.2f,%.2f"), ptTL.GetLong(), ptTL.GetLat(), ptBR.GetLong(), ptBR.GetLat()));
			}
			
			wxRemoveFile(sEntry.m_strOutputFilename);
		}
	}

	Clear();

	return bRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Downloads a region of slices (if not already downloaded)
///
//////////////////////////////////////////////////////////////////////////////

bool MapControlDataImporter_OSM::DownloadRegion(IMapControlData * pData, Point ptTL, Point ptBR, wxString strUsername, wxString strPassword, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	int iTLLong;
	int iTLLat;
	int iBRLong;
	int iBRLat;
	int iLong;
	int iLat;
	int iSlices;
	int iDone;
	bool bRtn = false;
	
	// arrange ptTL and ptBR so ptTL has the numerically larger coordinates
	if (ptTL.GetLatRaw() > ptBR.GetLatRaw())
	{
		wxInt32 iTmp;
		iTmp = ptTL.GetLatRaw();
		ptTL.SetLatRaw(ptBR.GetLatRaw());
		ptBR.SetLatRaw(iTmp);
	}

	if (ptTL.GetLongRaw() > ptBR.GetLongRaw())
	{
		wxInt32 iTmp;
		iTmp = ptTL.GetLongRaw();
		ptTL.SetLongRaw(ptBR.GetLongRaw());
		ptBR.SetLongRaw(iTmp);
	}

	// reduce precision to m_fDownloadGranularity
	iTLLong = (int) (ptTL.GetLong() / m_fDownloadGranularity);
	iTLLat = (int) (ptTL.GetLat() / m_fDownloadGranularity);
	
	iBRLong = (int) (ptBR.GetLong() / m_fDownloadGranularity + 1);
	iBRLat = (int) (ptBR.GetLat() / m_fDownloadGranularity + 1);
	
	iDone = 0;
	iSlices = (iTLLong - iBRLong) * (iTLLat - iBRLat);
	
	// iterate through slices
	for (iLong = iTLLong; iLong < iBRLong && !bRtn; iLong++)
	{
		for (iLat = iTLLat; iLat < iBRLat && !bRtn; iLat++)
		{
			if (DownloadSlice(	pData, 
								Point(iLong * m_fDownloadGranularity, iLat * m_fDownloadGranularity), 
								Point((iLong + 1) * m_fDownloadGranularity, (iLat + 1) * m_fDownloadGranularity), 
								strUsername,
								strPassword,
								pdlgProgress, 
								iProgressStart + iProgressTicks * iDone / iSlices, 
								iProgressTicks / iSlices))
				bRtn = true;
			iDone++;
		}
	}
	
	return bRtn;
}

