/*
 *  LibRoadnav
 *  MapControlDataImporter_TigerLine.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Imports Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>
#include <wx/tokenzr.h>

#include "Map.h"
#include "Constants.h"
#include "MapControlDataImporter_TigerLine.h"
#include "MapSupport.h"
#include "MapState.h"
#include "MapCounty.h"
#include "Portability.h"
#include "MapZip.h"
#include "Zip.h"
#include "URLs.h"
#include "UniqueIDGenerator.h"
#include "DownloadFiles.h"
#include "MapAbbreviations.h"
#include "StringTable.h"

#include <map>

using namespace std;

#define	STREETNAME_SIZE		40
#define STREETTYPE_SIZE		10
#define CITY_SIZE			80
#define STATE_SIZE			50

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Translate "xxx (Alt)" into "xxxA"
///
//////////////////////////////////////////////////////////////////////////////
wxString TranslateAlt(const wxString & strIn)
{
	if (strIn.Right(6).CmpNoCase(wxT(" (Alt)")))
		return strIn;
		
	return strIn.Left(strIn.Length() - 6) + wxT("A");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor, optionally loads data from the specified directory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_TigerLine::MapControlDataImporter_TigerLine(char * pszSource) : m_psRecords(NULL), m_nRecords(0)
{
	if (pszSource)
		Load(wxString(pszSource, *wxConvCurrent));

	m_bCompactMaps = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlDataImporter_TigerLine::~MapControlDataImporter_TigerLine()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_TigerLine::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;
	m_psRecords = NULL;
	
	m_mapTLIDtoRecord.clear();
	m_mapPolyIDtoRecord.clear();
	m_mapPolyIDtoSide.clear();
	m_mapAdditionalNameIDtoRecord.clear();

	m_nRecords = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database. This roughly corresponds
/// to the number of Type 1 records in the Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlDataImporter_TigerLine::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlDataImporter_TigerLine::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Some Tiger/Line records don't have a zip code specified. Try to guess
/// the correct zip code.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_TigerLine::FixZipCodes()
{
	// fill in zip codes
	int iRec;
	for (iRec = 0; iRec < m_nRecords; iRec++)
	{
		MapControlDataEntry_Std * psRec = &m_psRecords[iRec];

		if (!psRec->nAddressZipRanges)
		{
			int iRec2;
			double fBestDistance;
			double fDeltaLong;
			double fDeltaLat;
			int iBestZip;

			fBestDistance = INFINITY;
			iBestZip = -1;
			fDeltaLong = 0;
			fDeltaLat = 0;
			
			// so, we've got a record that doesn't have a zip code
			// the optimal thing to do would be calculate the distance between
			// this record's coordinates and every other record's coordinates,
			// and find the closest record. That takes forever. Instead,
			// the records are already sorted according to longitude, so we're
			// going to start at the next record, and work our way out,
			// tracking the closest record until we've traveled so far away
			// along the longitutde axis that nothing from this point forward
			// could beat our best match
			for (iRec2 = iRec + 1; iRec2 < m_nRecords && (iBestZip == -1 || fDeltaLong < fBestDistance); iRec2++)
			{
				MapControlDataEntry_Std * psRec2;
				
				psRec2 = &m_psRecords[iRec2];
				
				if (psRec2->nAddressZipRanges && psRec2->psAddressZipRanges[0].iZip)
				{
					double fDistance;
				
					fDeltaLong = fabs(m_psRecords[iRec2].psCoordinates[0].GetLong() - m_psRecords[iRec].psCoordinates[0].GetLong());
					fDeltaLat = fabs(m_psRecords[iRec2].psCoordinates[0].GetLat() - m_psRecords[iRec].psCoordinates[0].GetLat());

					// for the sake of time, we're going to use a simpler distance
					// formula.
					fDistance = fDeltaLong + fDeltaLat;
					
					if (fDistance < fBestDistance)
					{
						fBestDistance = fDistance;
						iBestZip = psRec2->psAddressZipRanges[0].iZip;
					}
				}
			}
			psRec->psAddressZipRanges = new SAddressRangeZip[1];
			psRec->psAddressZipRanges[0].bOnLeft = true;
			psRec->psAddressZipRanges[0].iFromAddr = 0;
			psRec->psAddressZipRanges[0].iToAddr = 0;
			psRec->psAddressZipRanges[0].iZip = iBestZip;
			
			psRec->nAddressZipRanges = 1;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sort the records according to their longitutde coordinate using the
/// venerable quick sort algorithm.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_TigerLine::SortRecords(int left, int right)
{
	int l_hold, r_hold;
	int divide;
	MapControlDataEntry_Std pivot;
	
	if (left > right)
		return;
	
	l_hold = left;
	r_hold = right;
	pivot = m_psRecords[left];
	while (left < right)
	{
		while ((m_psRecords[right] >= pivot) && (left < right))
			right--;
		
		if (left != right)
		{
			m_psRecords[left] = m_psRecords[right];
			left++;
		}
		
		while ((m_psRecords[left] <= pivot) && (left < right))
			left++;
		
		if (left != right)
		{
			m_psRecords[right] = m_psRecords[left];
			right--;
		}
	}
	
	m_psRecords[left] = pivot;
	divide = left;
	
	left = l_hold;
	right = r_hold;
	
	if (left < divide)
		SortRecords(left, divide-1);
	if (right > divide)
		SortRecords(divide+1, right);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type 1 records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadType1(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];
	int nNewRecords;
	MapControlDataEntry_Std * psOldSegments = NULL;
	UniqueIDGenerator idGen;

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type 1)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	nNewRecords = 0;
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		nNewRecords++;
		fgets(szLine, sizeof(szLine), hFile);
	}

	if (m_nRecords)
	{
		psOldSegments = m_psRecords;
	}
	m_psRecords = new MapControlDataEntry_Std[m_nRecords + nNewRecords];
	if (m_nRecords)
	{
		memcpy(m_psRecords, psOldSegments, sizeof(MapControlDataEntry_Std) * m_nRecords);
		delete [] psOldSegments;
	}
	
	fseek(hFile, 0, SEEK_SET);
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iTLID = atoi(ExtractField(szLine, 6, 15));

		char szPrefix[4];
		char szSuffix[4];
		char szFeatureName[STREETNAME_SIZE + 1];
		char szFeatureBaseName[STREETNAME_SIZE + 1];
		char szFeatureType[STREETTYPE_SIZE + 1];
		char szCFCC[5];

		strcpy(szPrefix, ExtractField(szLine, 18, 19));
		strcpy(szSuffix, ExtractField(szLine, 54, 55));
		strcpy(szFeatureBaseName, ExtractField(szLine, 20, 49));
		
		if (*szPrefix)
			strcat(szPrefix, " ");
		if (*szSuffix)
			strcat(szSuffix, " ");
		
		strcpy(szFeatureName, szPrefix);
		strcat(szFeatureName, szSuffix);
		strcat(szFeatureName, szFeatureBaseName);
		
		m_psRecords[m_nRecords].pszFeatureNames = new char * [1];
		m_psRecords[m_nRecords].pszFeatureNames[0] = g_pcStringTable->GetEntryPointer(szFeatureName, true);

		strcpy(szFeatureType, ExtractField(szLine, 50, 53));
		wxString strFeatureType = GetAbbreviation(wxString(szFeatureType, *wxConvCurrent));
		wxASSERT(strFeatureType.Length() < (STREETTYPE_SIZE + 1));
		if (strFeatureType != wxT(""))
			strcpy(szFeatureType, strFeatureType.mb_str(*wxConvCurrent));
		
		m_psRecords[m_nRecords].pszFeatureTypes = new char * [1];
		m_psRecords[m_nRecords].pszFeatureTypes[0] = g_pcStringTable->GetEntryPointer(szFeatureType, true);

		m_psRecords[m_nRecords].nFeatureNames = 1;

		//m_psRecords[m_nRecords].iTLID = iTLID;

		strcpy(szCFCC, ExtractField(szLine, 56, 58));
		ERecordType eRT;
		eRT = RecordTypeDefault;
		int iCFCCCode = atoi(szCFCC + 1);
		switch(szCFCC[0])
		{
			case 'A': // roads
			case 'P': // provisional roads
				eRT = RecordTypeTwoWaySmallRoad; // default to small roads

				if (iCFCCCode >= 11 && iCFCCCode <= 14)
					eRT = RecordTypeTwoWayHighway;
				if (iCFCCCode >= 15 && iCFCCCode <= 18)
					// FIXME: one way highways are all messed up in the tiger/line files
					// don't bother counting them as one way
					eRT = RecordTypeTwoWayHighway;
				if (iCFCCCode >= 19 && iCFCCCode <= 24)
					eRT = RecordTypeTwoWayHighway;
				if (iCFCCCode >= 25 && iCFCCCode <= 28)
					// FIXME: ditto here
					eRT = RecordTypeTwoWayHighway;
				if (iCFCCCode >= 29 && iCFCCCode <= 29)
					eRT = RecordTypeTwoWayHighway;

				if (iCFCCCode >= 31 && iCFCCCode <= 34)
					eRT = RecordTypeTwoWayLargeRoad;
				if (iCFCCCode >= 35 && iCFCCCode <= 38)
					// FIXME: here too
					eRT = RecordTypeTwoWayLargeRoad;
				if (iCFCCCode >= 39 && iCFCCCode <= 39)
					eRT = RecordTypeTwoWayLargeRoad;

				if (iCFCCCode >= 45 && iCFCCCode <= 48)
					// FIXME: and here
					eRT = RecordTypeTwoWaySmallRoad;

				// traffic circle, seems to listed in the correct direction
				if (iCFCCCode >= 62 && iCFCCCode <= 62)
					eRT = RecordTypeOneWaySmallRoad;
				// ramp
				if (iCFCCCode >= 63 && iCFCCCode <= 63)
					eRT = RecordTypeTwoWaySmallRoad;

				break;

			case 'B': // railroads
				eRT = RecordTypeRailroad;
				break;

			case 'D': // landmark
				eRT = RecordTypeLandmark;
				break;

			case 'E': // physical feature
				eRT = RecordTypePhysicalFeature;
				break;

			case 'F': // county lines and such
				eRT = RecordTypeInvisibleMiscBoundary;

				if (iCFCCCode >= 10 && iCFCCCode <= 12)
					eRT = RecordTypeInvisibleLandBoundary;
				break;

			case 'H': // water
				eRT = RecordTypeWater;
				if (iCFCCCode >= 70 && iCFCCCode <= 79)
					eRT = RecordTypeInvisibleWaterBoundary;
				break;

		}

		m_psRecords[m_nRecords].eType = eRT;

		int iFL, iTL, iFR, iTR, iZipL, iZipR;

		m_psRecords[m_nRecords].nAddressZipRanges = 0;
		m_psRecords[m_nRecords].psAddressZipRanges = NULL;

		iFL = atoi(ExtractField(szLine, 59, 69));
		iTL = atoi(ExtractField(szLine, 70, 80));
		iFR = atoi(ExtractField(szLine, 81, 91));
		iTR = atoi(ExtractField(szLine, 92, 102));
		iZipL = atoi(ExtractField(szLine, 107, 111));
		iZipR = atoi(ExtractField(szLine, 112, 116));

		if (iFL || iTL || iZipL)
		{
			struct SAddressRangeZip * psOldAddressZipRanges;
			int i;

			psOldAddressZipRanges = m_psRecords[m_nRecords].psAddressZipRanges;
			m_psRecords[m_nRecords].psAddressZipRanges = new struct SAddressRangeZip [m_psRecords[m_nRecords].nAddressZipRanges + 1];

			for (i = 0; i < m_psRecords[m_nRecords].nAddressZipRanges; i++)
				m_psRecords[m_nRecords].psAddressZipRanges[i] = psOldAddressZipRanges[i];

			m_psRecords[m_nRecords].psAddressZipRanges[i].iFromAddr = iFL;
			m_psRecords[m_nRecords].psAddressZipRanges[i].iToAddr = iTL;
			m_psRecords[m_nRecords].psAddressZipRanges[i].iZip = iZipL;

			if (m_psRecords[m_nRecords].nAddressZipRanges)
				delete [] psOldAddressZipRanges;

			m_psRecords[m_nRecords].nAddressZipRanges++;
		}
		if (iFR || iTR || iZipR)
		{
			struct SAddressRangeZip * psOldAddressZipRanges;
			int i;

			psOldAddressZipRanges = m_psRecords[m_nRecords].psAddressZipRanges;
			m_psRecords[m_nRecords].psAddressZipRanges = new struct SAddressRangeZip [m_psRecords[m_nRecords].nAddressZipRanges + 1];

			for (i = 0; i < m_psRecords[m_nRecords].nAddressZipRanges; i++)
				m_psRecords[m_nRecords].psAddressZipRanges[i] = psOldAddressZipRanges[i];

			m_psRecords[m_nRecords].psAddressZipRanges[i].iFromAddr = iFR;
			m_psRecords[m_nRecords].psAddressZipRanges[i].iToAddr = iTR;
			m_psRecords[m_nRecords].psAddressZipRanges[i].iZip = iZipR;

			if (m_psRecords[m_nRecords].nAddressZipRanges)
				delete [] psOldAddressZipRanges;

			m_psRecords[m_nRecords].nAddressZipRanges++;
		}


		m_psRecords[m_nRecords].psCoordinates = new Point [2];
		m_psRecords[m_nRecords].psCoordinates[0].SetLong((double) atoi(ExtractField(szLine, 191, 200)) / 1000000.0);
		m_psRecords[m_nRecords].psCoordinates[0].SetLat((double) atoi(ExtractField(szLine, 201, 209)) / 1000000.0);
		m_psRecords[m_nRecords].psCoordinates[1].SetLong((double) atoi(ExtractField(szLine, 210, 219)) / 1000000.0);
		m_psRecords[m_nRecords].psCoordinates[1].SetLat((double) atoi(ExtractField(szLine, 220, 228)) / 1000000.0);
		m_psRecords[m_nRecords].nCoordinates = 2;

		m_psRecords[m_nRecords].eCountryCode = CountryCodeUnitedStates;

#ifdef FILLED_IN_WATER_SUPPORT		
		m_psRecords[m_nRecords].bWaterL = false;
		m_psRecords[m_nRecords].bWaterR = false;
#endif

		if (m_psRecords[m_nRecords].nAddressZipRanges)
		{
			if (	m_psRecords[m_nRecords].psCoordinates[0].abs() > 0.001 && 
					m_psRecords[m_nRecords].psAddressZipRanges[0].iZip > 0)
				MarkZipCodeCoordinates(m_psRecords[m_nRecords].psAddressZipRanges[0].iZip, m_psRecords[m_nRecords].psCoordinates[0]);
		}

		m_psRecords[m_nRecords].idRecord = idGen.MakeID();

		m_mapTLIDtoRecord[iTLID] = m_nRecords;

		m_nRecords++;

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	MarkZipCodeCoordinates(-1, Point(0, 0));

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type 2 records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadType2(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type 2)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iTLID = atoi(ExtractField(szLine, 6, 15));

		if (m_mapTLIDtoRecord.find(iTLID) != m_mapTLIDtoRecord.end())
		{
			int iRecord;
			Point sPoints[10];
			Point sEndPoint;
			int nPoints;
			bool bDone = false;
			int iPoint;

			iRecord = m_mapTLIDtoRecord[iTLID];
			
			for (nPoints = 0; nPoints < 10 && !bDone; nPoints++)
			{
				sPoints[nPoints].SetLong((double) atoi(ExtractField(szLine, 19 + 19 * nPoints, 28 + 19 * nPoints)) / 1000000.0);
				sPoints[nPoints].SetLat((double) atoi(ExtractField(szLine, 29 + 19 * nPoints, 37 + 19 * nPoints)) / 1000000.0);

				if (sPoints[nPoints].GetLongRaw() == 0 && sPoints[nPoints].GetLatRaw() == 0)
					bDone = true;
			}
			nPoints--;

			Point * psNewPoints;
			psNewPoints = new Point [m_psRecords[iRecord].nCoordinates + nPoints];
			memcpy(psNewPoints, m_psRecords[iRecord].psCoordinates, sizeof(Point) * m_psRecords[iRecord].nCoordinates);
			delete [] m_psRecords[iRecord].psCoordinates;
			m_psRecords[iRecord].psCoordinates = psNewPoints;

			sEndPoint = m_psRecords[iRecord].psCoordinates[m_psRecords[iRecord].nCoordinates - 1];

			for (iPoint = 0; iPoint < nPoints; iPoint++)
			{
				m_psRecords[iRecord].psCoordinates[m_psRecords[iRecord].nCoordinates - 1] = sPoints[iPoint];
				m_psRecords[iRecord].nCoordinates++;
			}
			m_psRecords[iRecord].psCoordinates[m_psRecords[iRecord].nCoordinates - 1] = sEndPoint;

		}
		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type 4 records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadType4(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type 4)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iTLID = atoi(ExtractField(szLine, 6, 15));

		if (m_mapTLIDtoRecord.find(iTLID) != m_mapTLIDtoRecord.end())
		{
			int iRecord;

			iRecord = m_mapTLIDtoRecord[iTLID];

			int nNewNames;

			for (nNewNames = 0; nNewNames < 5 && atoi(ExtractField(szLine, 19 + nNewNames * 8, 26 + nNewNames * 8)); nNewNames++)
			{
				int iID = atoi(ExtractField(szLine, 19 + nNewNames * 8, 26 + nNewNames * 8));
				m_mapAdditionalNameIDtoRecord[iID].push_back(iRecord);
			}
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type 5 records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadType5(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type 5)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		unsigned int iEntry;
		int iID = atoi(ExtractField(szLine, 11, 18));
		for (iEntry = 0; iEntry < m_mapAdditionalNameIDtoRecord[iID].size(); iEntry++)
		{
			int iRecord = m_mapAdditionalNameIDtoRecord[iID][iEntry];
			char ** pszOldFeatureNames;
			char ** pszOldFeatureTypes;
			int i;

			pszOldFeatureNames = m_psRecords[iRecord].pszFeatureNames;
			pszOldFeatureTypes = m_psRecords[iRecord].pszFeatureTypes;

			m_psRecords[iRecord].pszFeatureNames = new char * [m_psRecords[iRecord].nFeatureNames + 1];
			m_psRecords[iRecord].pszFeatureTypes = new char * [m_psRecords[iRecord].nFeatureNames + 1];

			for (i = 0; i < m_psRecords[iRecord].nFeatureNames; i++)
			{
				m_psRecords[iRecord].pszFeatureNames[i] = pszOldFeatureNames[i];
				m_psRecords[iRecord].pszFeatureTypes[i] = pszOldFeatureTypes[i];
			}

			char szFeatureName[STREETNAME_SIZE + 1];
			char szFeatureType[STREETTYPE_SIZE + 1];

			strcpy(szFeatureName, ExtractField(szLine, 21, 50));

			strcpy(szFeatureType, ExtractField(szLine, 51, 54));
			wxString strFeatureType = GetAbbreviation(wxString(szFeatureType, *wxConvCurrent));
			wxASSERT(strFeatureType.Length() < (STREETTYPE_SIZE + 1));
			if (strFeatureType != wxT(""))
				strcpy(szFeatureType, strFeatureType.mb_str(*wxConvCurrent));

			m_psRecords[iRecord].pszFeatureNames[i] = g_pcStringTable->GetEntryPointer(szFeatureName, true);
			m_psRecords[iRecord].pszFeatureTypes[i] = g_pcStringTable->GetEntryPointer(szFeatureType, true);

			m_psRecords[iRecord].nFeatureNames++;

			delete [] pszOldFeatureNames;
			delete [] pszOldFeatureTypes;
		}
		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type 6 records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadType6(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		// Some counties don't come with any type 6 records.
		return false;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iTLID = atoi(ExtractField(szLine, 6, 15));

		if (m_mapTLIDtoRecord.find(iTLID) != m_mapTLIDtoRecord.end())
		{
			int iRecord = m_mapTLIDtoRecord[iTLID];

			int iFL = atoi(ExtractField(szLine, 19, 29));
			int iTL = atoi(ExtractField(szLine, 30, 40));
			int iFR = atoi(ExtractField(szLine, 41, 51));
			int iTR = atoi(ExtractField(szLine, 52, 62));
			int iZipL = atoi(ExtractField(szLine, 67, 71));
			int iZipR = atoi(ExtractField(szLine, 72, 76));

			if (iFL || iTL || iZipL)
			{
				struct SAddressRangeZip * psOldAddressZipRanges;
				int i;

				psOldAddressZipRanges = m_psRecords[iRecord].psAddressZipRanges;
				m_psRecords[iRecord].psAddressZipRanges = new struct SAddressRangeZip [m_psRecords[iRecord].nAddressZipRanges + 1];

				for (i = 0; i < m_psRecords[iRecord].nAddressZipRanges; i++)
					m_psRecords[iRecord].psAddressZipRanges[i] = psOldAddressZipRanges[i];

				m_psRecords[iRecord].psAddressZipRanges[i].iFromAddr = iFL;
				m_psRecords[iRecord].psAddressZipRanges[i].iToAddr = iTL;
				m_psRecords[iRecord].psAddressZipRanges[i].iZip = iZipL;

				if (m_psRecords[iRecord].nAddressZipRanges)
					delete [] psOldAddressZipRanges;

				m_psRecords[iRecord].nAddressZipRanges++;
			}
			if (iFR || iTR || iZipR)
			{
				struct SAddressRangeZip * psOldAddressZipRanges;
				int i;

				psOldAddressZipRanges = m_psRecords[iRecord].psAddressZipRanges;
				m_psRecords[iRecord].psAddressZipRanges = new struct SAddressRangeZip [m_psRecords[iRecord].nAddressZipRanges + 1];

				for (i = 0; i < m_psRecords[iRecord].nAddressZipRanges; i++)
					m_psRecords[iRecord].psAddressZipRanges[i] = psOldAddressZipRanges[i];

				m_psRecords[iRecord].psAddressZipRanges[i].iFromAddr = iFR;
				m_psRecords[iRecord].psAddressZipRanges[i].iToAddr = iTR;
				m_psRecords[iRecord].psAddressZipRanges[i].iZip = iZipR;

				if (m_psRecords[iRecord].nAddressZipRanges)
					delete [] psOldAddressZipRanges;

				m_psRecords[iRecord].nAddressZipRanges++;
			}

		}
		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type I records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadTypeI(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type I)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iTLID = atoi(ExtractField(szLine, 11, 20));

		if (m_mapTLIDtoRecord.find(iTLID) != m_mapTLIDtoRecord.end())
		{
			int iRecord;

			iRecord = m_mapTLIDtoRecord[iTLID];

			int iPolyIDL = atoi(ExtractField(szLine, 46, 55));
			int iPolyIDR = atoi(ExtractField(szLine, 61, 70));

			m_mapPolyIDtoRecord[iPolyIDL] = iRecord;
			m_mapPolyIDtoRecord[iPolyIDR] = iRecord;

			m_mapPolyIDtoSide[iPolyIDL] = 0;
			m_mapPolyIDtoSide[iPolyIDR] = 1;
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load a Tiger/Line file containing type P records.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::LoadTypeP(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (type P)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		int iPolyID = atoi(ExtractField(szLine, 16, 25));

		if (m_mapPolyIDtoRecord.find(iPolyID) != m_mapPolyIDtoRecord.end())
		{
#ifdef FILLED_IN_WATER_SUPPORT			

			int iRecord;

			int iWater = atoi(ExtractField(szLine, 45, 45));
			double fLong = (double) atoi(ExtractField(szLine, 26, 35)) / 1000000.0;
			double fLat = (double) atoi(ExtractField(szLine, 36, 44)) / 1000000.0;

			iRecord = m_mapPolyIDtoRecord[iPolyID];

			if (m_mapPolyIDtoSide[iPolyID])
			{
				m_psRecords[iRecord].bWaterR = iWater;
				m_psRecords[iRecord].ptWaterR.Set(fLong, fLat);
			}
			else
			{
				m_psRecords[iRecord].bWaterL = iWater;
				m_psRecords[iRecord].ptWaterL.Set(fLong, fLat);
			}
#endif
			
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}



//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load all of the Tiger/Line files for a particular county.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::Load(wxString strBaseName)
{
	wxFileName fn = strBaseName;

	fn.SetExt(wxT("RT1"));
	if (LoadType1(fn.GetFullPath()))
		return true;

	if (!m_bCompactMaps)
	{
		fn.SetExt(wxT("RT2"));
		if (LoadType2(fn.GetFullPath()))
			return true;
	}

	fn.SetExt(wxT("RT4"));
	if (LoadType4(fn.GetFullPath()))
		return true;

	fn.SetExt(wxT("RT5"));
	if (LoadType5(fn.GetFullPath()))
		return true;

	fn.SetExt(wxT("RT6"));
	if (LoadType6(fn.GetFullPath()))
		return true;

	fn.SetExt(wxT("RTI"));
	if (LoadTypeI(fn.GetFullPath()))
		return true;

	fn.SetExt(wxT("RTP"));
	if (LoadTypeP(fn.GetFullPath()))
		return true;

	m_mapTLIDtoRecord.clear();
	m_mapPolyIDtoRecord.clear();
	m_mapPolyIDtoSide.clear();
	m_mapAdditionalNameIDtoRecord.clear();

	SortRecords(0, m_nRecords - 1);
	FixZipCodes();
	CalculateNameTypeInstances();
	CalculateSmartNames();
	m_mapFeatureNameTypeInstances.clear();
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is the mapping data for a particular county loaded? Returns true if it is,
/// and false otherwise.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::IsCountyLoaded(wxInt32 iStateCountyCode)
{
	return IsDataIntegrated(wxT("TigerLine-") + wxString::Format(wxT("%05d"), iStateCountyCode));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Downloads the mapping data for the selected counties. Takes an array
/// of integers, containing the state code + county code.
///
/// Returns false on success and true on error.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::DownloadCounties(IMapControlData * pData, std::vector<wxInt32> viStateCountyCodes, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	unsigned int i;
	int iFile;
	vector<SyncDownloadEntry> vFiles;
	int iFilesProcessed;
	wxString strTIGERLineRoot;
	wxString strTIGERLineURLs;
	wxStringTokenizer tkzTIGERLineURLs;
	bool bValidURL;
	bool bRtn;
	bool bLowerCaseTgr = false;
	
	bRtn = false;
	
	strTIGERLineURLs = GetURL(wxT("Tiger-Line Root"));
	
	tkzTIGERLineURLs.SetString(strTIGERLineURLs, wxT(","), wxTOKEN_STRTOK);
	
	bValidURL = false;
	
	while (!bValidURL)
	{
		if (!bLowerCaseTgr)
			strTIGERLineRoot = tkzTIGERLineURLs.GetNextToken();
			
		bLowerCaseTgr = !bLowerCaseTgr;
		
		if (strTIGERLineRoot == wxT(""))
		{
			// if we get to the end of the list and don't find a good URL, perhaps there's just no
			// internet connection, and we should use the cache instead.
			tkzTIGERLineURLs.SetString(strTIGERLineURLs, wxT(","), wxTOKEN_STRTOK);
			strTIGERLineRoot = tkzTIGERLineURLs.GetNextToken();
			bLowerCaseTgr = true;
			bValidURL = true;
		}
		
		// create URL object
		int iCombinedCode = viStateCountyCodes[0];
		int iStateCode = iCombinedCode / 1000;
		wxString strStateAbbrev = StateAbbreviationByCode(iStateCode);
		wxString strURL = wxString::Format(wxT("%s%s/%s%05d.%s"), 
													strTIGERLineRoot.c_str(), 
													strStateAbbrev.c_str(), 
													bLowerCaseTgr ? wxT("tgr") : wxT("TGR"),
													iCombinedCode,
													bLowerCaseTgr ? wxT("zip") : wxT("ZIP")
													);
		if (IsValidURL(strURL))
			bValidURL = true;
	}

	for (i = 0; i < viStateCountyCodes.size(); i++)
	{
		int iCombinedCode = viStateCountyCodes[i];

		if (!IsCountyLoaded(iCombinedCode))
		{
			SyncDownloadEntry sEntry;
			int iStateCode = iCombinedCode / 1000;
			wxString strStateAbbrev = StateAbbreviationByCode(iStateCode);

			sEntry.m_strURL = wxString::Format(wxT("%s%s/%s%05d.%s"), 
													strTIGERLineRoot.c_str(), 
													strStateAbbrev.c_str(), 
													bLowerCaseTgr ? wxT("tgr") : wxT("TGR"),
													iCombinedCode,
													bLowerCaseTgr ? wxT("zip") : wxT("ZIP")
													);
			sEntry.m_strOutputFilename = wxFileName::CreateTempFileName(wxString::Format(wxT("tgr%05d"), iCombinedCode));
			sEntry.m_strProgressBarName = CountyNameByCode(iCombinedCode);

			vFiles.push_back(sEntry);
		}
	}

	if (DownloadFiles(vFiles, true, pdlgProgress, iProgressStart, iProgressTicks / 2))
		bRtn = true;

	for (i = 0, iFile = 0, iFilesProcessed = 0; i < viStateCountyCodes.size(); i++)
	{
		int iCombinedCode = viStateCountyCodes[i];

		if (!IsCountyLoaded(iCombinedCode))
		{
			if (vFiles[iFile].m_bDownloadedOkay)
			{
				Unzip(vFiles[iFile].m_strOutputFilename, GetDataPath());
				
				wxRemoveFile(vFiles[iFile].m_strOutputFilename);

				wxFileName fnThis;

				fnThis.AssignDir(GetDataPath());
				fnThis.SetFullName(wxString::Format(wxT("TGR%05d"), iCombinedCode));
				
				if (!Load(fnThis.GetFullPath()))
				{
					pData->Save(this, pdlgProgress, iProgressStart + iProgressTicks / 2 + iProgressTicks / 2 * iFilesProcessed / vFiles.size(), iProgressTicks / 2 / vFiles.size());
					MarkDataIntegrated(wxT("TigerLine-") + fnThis.GetFullPath().Right(5));
				}
				
				iFilesProcessed++;

				Clear();

				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT1"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT2"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT4"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT5"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT1"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT6"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT7"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RT8"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTA"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTC"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTE"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTH"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTI"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTM"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTP"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTR"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTS"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTT"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".RTZ"));
				wxRemoveFile(fnThis.GetFullPath() + wxT(".MET"));
			}
			else
			{
				bRtn = true;
			}

			iFile++;
		}
	}

	Clear();

	return bRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Download the mapping data for a single county.
///
/// Returns false on success and true on error.
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlDataImporter_TigerLine::DownloadCounty(IMapControlData * pData, wxInt32 iStateCountyCode, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	vector<wxInt32> v;
	v.push_back(iStateCountyCode);
	return DownloadCounties(pData, v, pdlgProgress, iProgressStart, iProgressTicks);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Calculate the number of times each road name is used.
///
/// This is used in SmartName to find the most common name for a road.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlDataImporter_TigerLine::CalculateNameTypeInstances()
{
	int iRecord;
	
	m_mapFeatureNameTypeInstances.clear();
	
	for (iRecord = 0; iRecord < m_nRecords; iRecord++)
	{
		int iEntry;
		IMapControlDataEntry * psRec = &m_psRecords[iRecord];
		
		wxASSERT(psRec);
		
		for (iEntry = 0; iEntry < psRec->nFeatureNames; iEntry++)
		{
			wxString strNameType = psRec->NameAndType(iEntry);
			
			m_mapFeatureNameTypeInstances[strNameType]++;
		}
	}
}

wxString MapControlDataImporter_TigerLine::CalculateOneSmartName(int iRecord)
{
	IMapControlDataEntry * psRec = &m_psRecords[iRecord];
	wxString strMostCommonName;
	wxString strState;
	int iName;
	int iMostCommonCount = 0;
	
	if (psRec->nAddressZipRanges)
		strState = StateAbbreviationByZip(psRec->psAddressZipRanges[0].iZip);

	for (iName = 0; iName < psRec->nFeatureNames; iName++)
	{
		wxString strName(psRec->pszFeatureNames[iName], *wxConvCurrent);
		
		if (strName.Left(2) == wxT("I-") && strState != wxT(""))
			return TranslateAlt(strName);

		if (strName.Left(6) == wxT("Route ") && strState != wxT(""))
			return TranslateAlt(wxString::Format(wxT("%s-%s"), strState.c_str(), psRec->NameAndType(iName).c_str() + 6));

		if (strName.Left(12) == wxT("State Route "))
		{
			if (strState == wxT(""))
				return TranslateAlt(wxString::Format(wxT("SR-%s"), psRec->NameAndType(iName).c_str() + 12));

			return TranslateAlt(wxString::Format(wxT("%s-%s"), strState.c_str(), psRec->NameAndType(iName).c_str() + 12));
		}

		if (strName.Left(14) == wxT("State Highway "))
		{
			if (strState == wxT(""))
				return TranslateAlt(wxString::Format(wxT("SR-%s"), psRec->NameAndType(iName).c_str() + 14));
			
			return TranslateAlt(wxString::Format(wxT("%s-%s"), strState.c_str(), psRec->NameAndType(iName).c_str() + 14));
		}

		if (strName.Left(21) == wxT("United States Highway"))
			return TranslateAlt(wxString::Format(wxT("US-%s"), psRec->NameAndType(iName).c_str() + 22));
		if (strName.Left(10) == wxT("US Highway"))
			return TranslateAlt(wxString::Format(wxT("US-%s"), psRec->NameAndType(iName).c_str() + 11));
		if (strName.Left(6) == wxT("US Hwy"))
			return TranslateAlt(wxString::Format(wxT("US-%s"), psRec->NameAndType(iName).c_str() + 7));
	}

	for (iName = 0; iName < psRec->nFeatureNames; iName++)
	{
		wxString strNameType = psRec->NameAndType(iName);
		
		if (m_mapFeatureNameTypeInstances[strNameType] > iMostCommonCount)
		{
			iMostCommonCount = m_mapFeatureNameTypeInstances[strNameType];
			strMostCommonName = strNameType;
		}
	}
	return strMostCommonName;
}

void MapControlDataImporter_TigerLine::CalculateSmartNames()
{
	int iRecord;
	
	for (iRecord = 0; iRecord < m_nRecords; iRecord++)
	{
		wxString strSmartName = CalculateOneSmartName(iRecord);
		m_psRecords[iRecord].pszSmartName = g_pcStringTable->GetEntryPointer(strSmartName.mb_str(*wxConvCurrent), true);
	}
}

void MapControlDataImporter_TigerLine::SetCompactMapsMode(bool bCompactMaps)
{
	m_bCompactMaps = bCompactMaps;
}

