/*
 *  LibRoadnav
 *  MapControlData_Foundation.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Manages the Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>

#include "MapControlData_Foundation.h"
#include "MapRepresentations.h"
#include "MapSupport.h"
#include "Portability.h"
#include "MapZip.h"
#include "Debug.h"

#include <map>

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Class used internally to represent the map as a bunch of vertices
/// rather than as a bunch of edges.
///
//////////////////////////////////////////////////////////////////////////////
class MapVertex
{
	public:
		//////////////////////////////////////////////////////////////////////////////
		/// @name Constructors/destructors
		//////////////////////////////////////////////////////////////////////////////
		MapVertex();
		~MapVertex();
		
		//////////////////////////////////////////////////////////////////////////////
		/// @name Public functions
		//////////////////////////////////////////////////////////////////////////////
		void Clear();
		Point & GetCoordinates();
		void FindRecords(IMapControlData * pcMapControlData, Point & ptThis, std::map<Point,tVectorOfInts> & mapCoordinateToRecords);
		
		//////////////////////////////////////////////////////////////////////////////
		/// @name Information describing the associated records (edges)
		//////////////////////////////////////////////////////////////////////////////
		IMapControlDataEntry ** m_pcAssociatedRecords;
		int * m_piAssociatedRecordsCoordinate;
		int m_nAssociatedRecords;
		
		//////////////////////////////////////////////////////////////////////////////
		/// @name Routing information
		//////////////////////////////////////////////////////////////////////////////
		double m_fBestPathToThisCost;
		MapVertex * m_pcBestPathToThisPreviousVertex;
		IMapControlDataEntry * m_pcBestPathRecordUsed;
		
		double m_fBestCompletePathCost;
		
		//////////////////////////////////////////////////////////////////////////////
		/// @name Linked list variables
		//////////////////////////////////////////////////////////////////////////////
		MapVertex * m_pcNext;
};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - initialize everything to zero.
///
//////////////////////////////////////////////////////////////////////////////
MapVertex::MapVertex()
{
	m_pcAssociatedRecords = NULL;
	m_piAssociatedRecordsCoordinate = NULL;
	m_nAssociatedRecords = 0;
	
	m_fBestCompletePathCost = INFINITY;
	m_fBestPathToThisCost = INFINITY;
	m_pcBestPathToThisPreviousVertex = NULL;
	
	m_pcBestPathRecordUsed = NULL;
	
	m_pcNext = NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapVertex::~MapVertex()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Frees all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapVertex::Clear()
{
	if (m_pcAssociatedRecords)
	{
		delete m_pcAssociatedRecords;
		m_pcAssociatedRecords = NULL;
	}
	if (m_piAssociatedRecordsCoordinate)
	{
		delete m_piAssociatedRecordsCoordinate;
		m_piAssociatedRecordsCoordinate = NULL;
	}
	
	// deleting the head deletes the entire list
	while (m_pcNext)
	{
		MapVertex * pcNextNext;
		
		pcNextNext = m_pcNext->m_pcNext;
		m_pcNext->m_pcNext = NULL;
		
		delete m_pcNext;
		
		m_pcNext = pcNextNext;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the coordinates of this vertex
///
//////////////////////////////////////////////////////////////////////////////
Point & MapVertex::GetCoordinates()
{
	wxASSERT(m_nAssociatedRecords);
	return m_pcAssociatedRecords[0]->psCoordinates[m_piAssociatedRecordsCoordinate[0]];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the records associated with this vertex and fill
/// m_pcAssociatedRecords and m_piAssociatedRecordsCoordinate.
///
//////////////////////////////////////////////////////////////////////////////
void MapVertex::FindRecords(IMapControlData * pcMapControlData, Point & ptThis, map<Point,tVectorOfInts> & mapCoordinateToRecords)
{
	vector<int>::iterator iRecIter;
	int iRecIndex;
	
	map<Point,tVectorOfInts>::iterator iFind;

	// Free existing memory
	if (m_pcAssociatedRecords)
	{
		delete m_pcAssociatedRecords;
		m_pcAssociatedRecords = NULL;
	}
	
	if (m_piAssociatedRecordsCoordinate)
	{
		delete m_piAssociatedRecordsCoordinate;
		m_piAssociatedRecordsCoordinate = NULL;
	}
	
	// Get associated records from map
	iFind = mapCoordinateToRecords.find(ptThis);

	wxASSERT(iFind != mapCoordinateToRecords.end());
	
	vector<int> & vEntries = iFind->second;

	m_nAssociatedRecords = vEntries.size();
	
	wxASSERT(m_nAssociatedRecords);

	// allocate memory
	m_pcAssociatedRecords = new IMapControlDataEntry * [m_nAssociatedRecords];
	m_piAssociatedRecordsCoordinate = new int [m_nAssociatedRecords];
	
	// copy records to this MapVertex object
	for (iRecIter = vEntries.begin(), iRecIndex = 0; iRecIter != vEntries.end(); iRecIter++, iRecIndex++)
	{
		IMapControlDataEntry * pcRec;
		
		pcRec = pcMapControlData->GetRecord(*iRecIter);
		
		wxASSERT(pcRec);

		m_pcAssociatedRecords[iRecIndex] = pcRec;

		if (pcRec->psCoordinates[0] == ptThis)
			m_piAssociatedRecordsCoordinate[iRecIndex] = 0;
		else
			m_piAssociatedRecordsCoordinate[iRecIndex] = pcRec->nCoordinates - 1;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor
///
//////////////////////////////////////////////////////////////////////////////
MapControlData_Foundation::MapControlData_Foundation() : m_psRecords(NULL), m_nRecords(0)
{
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor, free all memory
///
//////////////////////////////////////////////////////////////////////////////
MapControlData_Foundation::~MapControlData_Foundation()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free all memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Foundation::Clear()
{
	int iRec;

	for (iRec = 0; iRec < m_nRecords; iRec++)
		m_psRecords[iRec].Clear();

	if (m_psRecords)
		delete [] m_psRecords;
	m_psRecords = NULL;
	
	ClearVertices();
	
	m_nRecords = 0;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Count of the number of records in the database. This roughly corresponds
/// to the number of Type 1 records in the Tiger/Line data.
///
//////////////////////////////////////////////////////////////////////////////
int MapControlData_Foundation::Count()
{
	return m_nRecords;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its index number
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlData_Foundation::GetRecord(wxUint32 iIndex)
{
	return &m_psRecords[iIndex];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve a particular record by its ID
///
//////////////////////////////////////////////////////////////////////////////
IMapControlDataEntry * MapControlData_Foundation::GetRecordByID(const tRecordID & idRec)
{
	int i;

	for (i = 0; i < m_nRecords; i++)
		if (m_psRecords[i].idRecord == idRec)
			return &m_psRecords[i];

	return NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Build a list of vertices that exist in the current record set.
///
/// This also determines the cost of moving from one vertex to another.
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Foundation::EnumerateVertices()
{
	// build a map that relates coordinates to record numbers
	int iRec;
	for (iRec = 0; iRec < m_nRecords; iRec++)
	{
		IMapControlDataEntry * psRec = &m_psRecords[iRec];

		// can't drive on rivers or railroad tracks
		if (psRec->IsRoad())
		{
			// add first coordinate
			m_mapCoordinateToRecords[psRec->psCoordinates[0]].push_back(iRec);
			
			// add last coordinate if there's more than one
			if (psRec->nCoordinates > 1)
				m_mapCoordinateToRecords[psRec->psCoordinates[psRec->nCoordinates - 1]].push_back(iRec);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clear the vertex list
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Foundation::ClearVertices()
{
	m_mapCoordinateToRecords.clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determine the shortest path between two records.
///
/// Returns a vector containing the records along the shortest path from
/// iStartRec to iEndRec.
///
/// Uses Dijkstra's algorithm.
///
//////////////////////////////////////////////////////////////////////////////
vector<tRecordID> MapControlData_Foundation::ShortestPath(const Point & ptStart, const Point & ptEnd, const tRecordID & idStart, const tRecordID & idEnd, bool bProgressDialog)
{
	ProgressDialog * pdlgProgress;
	double fProgress = 0;
	int iRound;
	bool bFoundSolution = false;
	double fTurnPenalty = 0.5;
	
	if (bProgressDialog)
		pdlgProgress = new ProgressDialog(NULL, wxT("Calculating route..."), 100);
	else
		pdlgProgress = NULL;
	
	if (pdlgProgress)
		pdlgProgress->Show();
	
	Clear();

	// load the records covering both endpoints
	int iRT;
	map<Point,MapVertex *> mapPointToMapVertex;

	LibRoadnavDebug0(wxT("Routing"), wxT("Loading region"));

	for (iRT = FIRST_RECORD_TYPE_ROAD; iRT <= LAST_RECORD_TYPE_ROAD; iRT++)
	{
		bool bVisibility[LAST_RECORD_TYPE + 1];
		int iRT2;
		double fRadius = INFINITY;

		// mark everything invisible, except this layer
		for (iRT2 = FIRST_RECORD_TYPE; iRT2 <= LAST_RECORD_TYPE; iRT2++)
			bVisibility[iRT2] = (iRT == iRT2);

		if (iRT == RecordTypeOneWaySmallRoad || iRT == RecordTypeTwoWaySmallRoad)
			fRadius = 0.1;

		if (iRT == RecordTypeOneWayLargeRoad || iRT == RecordTypeTwoWayLargeRoad)
			fRadius = 0.5;

		if (fabs((ptStart - ptEnd).GetLat()) < fRadius * 2 &&
			fabs((ptStart - ptEnd).GetLong()) < fRadius * 2)
		{
			Point ptTopLeft;
			Point ptBottomRight;
			
			if (ptStart.GetLongRaw() > ptEnd.GetLongRaw())
			{
				ptTopLeft.SetLongRaw(ptEnd.GetLongRaw());
				ptBottomRight.SetLongRaw(ptStart.GetLongRaw());
			}
			else
			{
				ptBottomRight.SetLongRaw(ptEnd.GetLongRaw());
				ptTopLeft.SetLongRaw(ptStart.GetLongRaw());
			}
			
			if (ptStart.GetLatRaw() > ptEnd.GetLatRaw())
			{
				ptTopLeft.SetLatRaw(ptEnd.GetLatRaw());
				ptBottomRight.SetLatRaw(ptStart.GetLatRaw());
			}
			else
			{
				ptBottomRight.SetLatRaw(ptEnd.GetLatRaw());
				ptTopLeft.SetLatRaw(ptStart.GetLatRaw());
			}
			
			LoadRegion(ptTopLeft, ptBottomRight, bVisibility, false);
		}
		else
		{
			LoadRegion(ptStart - Point(fRadius, fRadius), ptStart + Point(fRadius, fRadius), bVisibility, false);
			LoadRegion(ptEnd - Point(fRadius, fRadius), ptEnd + Point(fRadius, fRadius), bVisibility, false);
		}
	}

	MapVertex cOptimalVertexList;
	MapVertex cSubOptimalVertexList;
	Point ptEndThis;
	for (iRound = 0; iRound < 3 && !bFoundSolution; iRound++)
	{
		LibRoadnavDebug1(wxT("Routing"), wxT("Round %d"), iRound);

		if (iRound == 1 || iRound == 2)
		{
			bool bVisibility[LAST_RECORD_TYPE + 1];
			int iRT2;
			Point ptTopLeft;
			Point ptBottomRight;
			double fMargin = 0;
			
			if (iRound == 1)
				fMargin = 0.5;
				
			if (iRound == 2)
				fMargin = 2;
			
			if (ptStart.GetLongRaw() > ptEnd.GetLongRaw())
			{
				ptTopLeft.SetLongRaw(ptEnd.GetLongRaw());
				ptBottomRight.SetLongRaw(ptStart.GetLongRaw());
			}
			else
			{
				ptBottomRight.SetLongRaw(ptEnd.GetLongRaw());
				ptTopLeft.SetLongRaw(ptStart.GetLongRaw());
			}
			
			if (ptStart.GetLatRaw() > ptEnd.GetLatRaw())
			{
				ptTopLeft.SetLatRaw(ptEnd.GetLatRaw());
				ptBottomRight.SetLatRaw(ptStart.GetLatRaw());
			}
			else
			{
				ptBottomRight.SetLatRaw(ptEnd.GetLatRaw());
				ptTopLeft.SetLatRaw(ptStart.GetLatRaw());
			}

			for (iRT2 = FIRST_RECORD_TYPE; iRT2 <= LAST_RECORD_TYPE; iRT2++)
				bVisibility[iRT2] = (RecordTypeTwoWayLargeRoad == iRT2) || 
									(RecordTypeOneWayLargeRoad == iRT2) ||
									(RecordTypeTwoWayHighway == iRT2) || 
									(RecordTypeOneWayHighway == iRT2);
				
			LoadRegion(ptTopLeft - Point(fMargin, fMargin), ptBottomRight + Point(fMargin, fMargin), bVisibility, false);
		}
		
		LibRoadnavDebug0(wxT("Routing"), wxT("Enumerating vertices"));

		ClearVertices();
		mapPointToMapVertex.clear();
		cOptimalVertexList.Clear();
		cSubOptimalVertexList.Clear();

		EnumerateVertices();
		
		bool bInvalidIdEnd = false;
		IMapControlDataEntry * pcRecEnd = GetRecordByID(idEnd);
		
		if (pcRecEnd == NULL)
			bInvalidIdEnd = true;
		else if (pcRecEnd->Distance(ptEnd) > 0.001)
			bInvalidIdEnd = true;
			
		if (bInvalidIdEnd)
		{
			// FIXME: should be a more elegant solution than this for the problem
			// of record IDs from previous map compiles.
			pcRecEnd = GetRecordByID(FindCoordinates(ptEnd, OnlyRoadsVisibility()).m_idRecord);
			
			if (!pcRecEnd)
			{
				vector<tRecordID> vPath;
				LibRoadnavDebug0(wxT("Routing"), wxT("Can't find end coordinates"));

				if (pdlgProgress)
					delete pdlgProgress;

				return vPath;	
			}
		}

		ptEndThis = pcRecEnd->psCoordinates[0];
		
		//////////////////////////////////////////////////////////////////////////////
		// the first vertex is the starting point .. set that up
		//////////////////////////////////////////////////////////////////////////////

		cSubOptimalVertexList.m_pcNext = new MapVertex();

		cSubOptimalVertexList.m_pcNext->m_fBestPathToThisCost = 0;
		cSubOptimalVertexList.m_pcNext->m_pcBestPathToThisPreviousVertex = NULL;
		cSubOptimalVertexList.m_pcNext->m_fBestCompletePathCost = Distance(ptStart, ptEndThis);
		
		cSubOptimalVertexList.m_pcNext->FindRecords(this, GetRecordByID(idStart)->psCoordinates[0], m_mapCoordinateToRecords);

		mapPointToMapVertex[GetRecordByID(idStart)->psCoordinates[0]] = cSubOptimalVertexList.m_pcNext;

		LibRoadnavDebug0(wxT("Routing"), wxT("Beginning A* algorithm"));

		while (cSubOptimalVertexList.m_pcNext)
		{
			//////////////////////////////////////////////////////////////////////////////
			// locate the vertex in the suboptimal list with the lowest cost
			//////////////////////////////////////////////////////////////////////////////
			MapVertex * pcLowestCostVertexParent;
			MapVertex * pcThisVertexParent;

			pcLowestCostVertexParent = &cSubOptimalVertexList;
			for (pcThisVertexParent = cSubOptimalVertexList.m_pcNext; pcThisVertexParent->m_pcNext; pcThisVertexParent = pcThisVertexParent->m_pcNext)
				if (pcThisVertexParent->m_pcNext->m_fBestCompletePathCost < pcLowestCostVertexParent->m_pcNext->m_fBestCompletePathCost)
					pcLowestCostVertexParent = pcThisVertexParent;
			
			MapVertex * pcLowestCostVertex;

			pcLowestCostVertex = pcLowestCostVertexParent->m_pcNext;
			
			//////////////////////////////////////////////////////////////////////////////
			// remove the lowest cost vertex from the suboptimal list and put it in
			// the optimal list
			//////////////////////////////////////////////////////////////////////////////
		
			pcLowestCostVertexParent->m_pcNext = pcLowestCostVertex->m_pcNext;
			pcLowestCostVertex->m_pcNext = cOptimalVertexList.m_pcNext;
			cOptimalVertexList.m_pcNext = pcLowestCostVertex;
			
			//////////////////////////////////////////////////////////////////////////////
			// was the lowest cost vertex the ending point?
			// if so, we're done
			//////////////////////////////////////////////////////////////////////////////
			if (pcLowestCostVertex->GetCoordinates() == ptEndThis)
			{
				bFoundSolution = true;
				break;
			}
				
			//////////////////////////////////////////////////////////////////////////////
			// find the edges leading away from pcLowestCostVertex
			//////////////////////////////////////////////////////////////////////////////
			int iAssociatedRecord;

			for (iAssociatedRecord = 0; iAssociatedRecord < pcLowestCostVertex->m_nAssociatedRecords; iAssociatedRecord++)
			{
				int iAdjacentCoordinateNumber;
				IMapControlDataEntry * pcRec;
				
				pcRec = pcLowestCostVertex->m_pcAssociatedRecords[iAssociatedRecord];
				
				if (pcLowestCostVertex->m_piAssociatedRecordsCoordinate[iAssociatedRecord] == 0)
				{
					iAdjacentCoordinateNumber = pcRec->nCoordinates - 1;
				}
				else
				{
					if (pcRec->IsOneWay())
						iAdjacentCoordinateNumber = -1;
					else
						iAdjacentCoordinateNumber = 0;
				}
				
				// make sure we don't go the wrong way down a one way street
				if (iAdjacentCoordinateNumber >= 0)
				{
					MapVertex * pcAdjacentVertex;
					double fThisCostToHere;
					double fThisCompleteCost;
					
					// so we can get to pcRec->psCoordinates[iAdjacentCoordinateNumber] from pcLowestCostVertex
					// the cost is pcLowestCostVertex->m_fBestPathToThisCost + pcRec->CostFactor() * pcRec->RecordDistance()
					// we need to first find the corresponding MapVertex, then see if our cost is lower
					// than its cost. If so, we become the best path leading up to that vertex.
					
					Point & ptAdjacent = pcRec->psCoordinates[iAdjacentCoordinateNumber];

					pcAdjacentVertex = mapPointToMapVertex[ptAdjacent];
					
					if (!pcAdjacentVertex)
					{
						// vertex doesn't exist yet .. create it

						pcAdjacentVertex = new MapVertex();
						
						pcAdjacentVertex->m_pcNext = cSubOptimalVertexList.m_pcNext;
						cSubOptimalVertexList.m_pcNext = pcAdjacentVertex;
						
						pcAdjacentVertex->FindRecords(this, ptAdjacent, m_mapCoordinateToRecords);
						
						mapPointToMapVertex[ptAdjacent] = pcAdjacentVertex;
					}

					// calculate cost for this path
					fThisCostToHere = pcLowestCostVertex->m_fBestPathToThisCost + 
										pcRec->CostFactor() * Distance(pcRec->psCoordinates[0], pcRec->psCoordinates[pcRec->nCoordinates - 1]);
					
					// either pcLowestCostVertex->m_pcBestPathRecordUsed should be non-NULL,
					// i.e. we know what record was used to reach the lowest cost vertex,
					// or, the lowest cost vertex should be the starting vertex.
					wxASSERT(pcLowestCostVertex->m_pcBestPathRecordUsed || pcLowestCostVertex->m_fBestPathToThisCost < 1e-6);
					if (pcLowestCostVertex->m_pcBestPathRecordUsed)
						if (!pcLowestCostVertex->m_pcBestPathRecordUsed->IsSameRoad(pcRec))
							fThisCostToHere += fTurnPenalty;

					fThisCompleteCost = fThisCostToHere + Distance(ptAdjacent, ptEndThis);

					// if coming from this vertex is faster than the previous route, use this path						
					if (fThisCompleteCost < pcAdjacentVertex->m_fBestCompletePathCost && fThisCostToHere < pcAdjacentVertex->m_fBestPathToThisCost)
					{
						pcAdjacentVertex->m_fBestPathToThisCost = fThisCostToHere;
						pcAdjacentVertex->m_fBestCompletePathCost = fThisCompleteCost;
						
						pcAdjacentVertex->m_pcBestPathToThisPreviousVertex = pcLowestCostVertex;
						pcAdjacentVertex->m_pcBestPathRecordUsed = pcRec;

						MapVertex * pcSearchVertex;
						for (pcSearchVertex = &cOptimalVertexList; pcSearchVertex->m_pcNext; pcSearchVertex = pcSearchVertex->m_pcNext)
						{
							if (pcSearchVertex->m_pcNext == pcAdjacentVertex)
							{
								// remove from optimal list
								pcSearchVertex->m_pcNext = pcSearchVertex->m_pcNext->m_pcNext;
								
								// add to suboptimal
								pcAdjacentVertex->m_pcNext = cSubOptimalVertexList.m_pcNext;
								cSubOptimalVertexList.m_pcNext = pcAdjacentVertex;
								
								if (!pcSearchVertex->m_pcNext)
									break;
							}
						}
						
						if (pdlgProgress)
						{
							double fThisProgress;
							
							fThisProgress = fThisCostToHere / fThisCompleteCost * 100.0;
							
							if (fThisProgress > fProgress)
							{
								fProgress = fThisProgress;
								pdlgProgress->SetProgress((int) fProgress);
							}
						}
					}
				}
			}
		}
	}
	
	if (!bFoundSolution)
	{
		vector<tRecordID> vPath;
		LibRoadnavDebug0(wxT("Routing"), wxT("Couldn't find solution"));

		if (pdlgProgress)
			delete pdlgProgress;

		return vPath;	
	}

	LibRoadnavDebug0(wxT("Routing"), wxT("Backtracking"));

	// extract path from end to start	
	MapVertex * pcThisVertex;
	vector<tRecordID> vPathReversed;

	pcThisVertex = mapPointToMapVertex[ptEndThis];
	
	while (pcThisVertex)
	{
		if (pcThisVertex->m_pcBestPathToThisPreviousVertex)
			vPathReversed.push_back(pcThisVertex->m_pcBestPathRecordUsed->idRecord);
		pcThisVertex = pcThisVertex->m_pcBestPathToThisPreviousVertex;
	}

	LibRoadnavDebug0(wxT("Routing"), wxT("Generating forward solution"));
	
	// reverse path
	vector<tRecordID> vPath;
	int i;
	
	for (i = vPathReversed.size() - 1; i >= 0; i--)
		vPath.push_back(vPathReversed[i]);

	// delete progress dialog
	if (pdlgProgress)
		delete pdlgProgress;
	
	return vPath;	
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Takes a list of records, and returns a list of coordinates that the records
/// cover.
///
//////////////////////////////////////////////////////////////////////////////
vector<Point> MapControlData_Foundation::TranslateRecordsToCoordinates(const vector<tRecordID> & vRecords)
{
	unsigned int iLeg;
	vector<Point> vTrack;

	if (vRecords.size() < 2)
		return vTrack;

	for (iLeg = 0; iLeg < vRecords.size(); iLeg++)
	{
		int iSeg;
		IMapControlDataEntry * psRec = GetRecordByID(vRecords[iLeg]);
		bool bReversed = false;

		if (iLeg > 0)
		{
			IMapControlDataEntry * psPreviousRec = GetRecordByID(vRecords[iLeg - 1]);
			Point ptPreviousStart = psPreviousRec->psCoordinates[0];
			Point ptPreviousEnd = psPreviousRec->psCoordinates[psPreviousRec->nCoordinates - 1];
			Point ptThisEnd = psRec->psCoordinates[psRec->nCoordinates - 1];

			if (ptThisEnd == ptPreviousStart || ptThisEnd == ptPreviousEnd)
				bReversed = true;
		}
		else // iLeg == 0
		{
			IMapControlDataEntry * psNextRec = GetRecordByID(vRecords[iLeg + 1]);
			Point ptNextStart = psNextRec->psCoordinates[0];
			Point ptNextEnd = psNextRec->psCoordinates[psNextRec->nCoordinates - 1];
			Point ptThisStart = psRec->psCoordinates[0];

			if (ptThisStart == ptNextStart || ptThisStart == ptNextEnd)
				bReversed = true;
		}

		if (bReversed)
		{
			for (iSeg = psRec->nCoordinates - 1; iSeg >= 0; iSeg--)
				vTrack.push_back(psRec->psCoordinates[iSeg]);
		}
		else
		{
			for (iSeg = 0; iSeg < psRec->nCoordinates; iSeg++)
				vTrack.push_back(psRec->psCoordinates[iSeg]);
		}
	}
	return vTrack;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns how many record types exist.
///
//////////////////////////////////////////////////////////////////////////////
wxInt16 MapControlData_Foundation::GetNumRecordTypes()
{
	return LAST_RECORD_TYPE+1;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Maps a record type number to its name.
///
//////////////////////////////////////////////////////////////////////////////
wxString MapControlData_Foundation::GetRecordTypeDisplayName(wxInt16 iRecordType)
{
	wxString strName;
	ERecordType eRT = (ERecordType)iRecordType;
	switch( eRT )
	{
		case RecordTypeDefault:
			strName = wxT("Default");
		break;
		
		case RecordTypeOneWaySmallRoad:
			strName = wxT("One Way Small Road");
		break;
		
		case RecordTypeOneWayLargeRoad:
			strName = wxT("One Way Large Road");
		break;
		
		case RecordTypeOneWayHighway:
			strName = wxT("One Way Highway");
		break;
		
		case RecordTypeTwoWaySmallRoad:
			strName = wxT("Two Way Small Road");
		break;
		
		case RecordTypeTwoWayLargeRoad:
			strName = wxT("Two Way Large Road");
		break;
		
		case RecordTypeTwoWayHighway:
			strName = wxT("Two Way Highway");
		break;
		
		case RecordTypeRailroad:
			strName = wxT("Railroad");
		break;
		
		case RecordTypeWater:
			strName = wxT("Water");
		break;
		
		case RecordTypeInvisibleLandBoundary:
			strName = wxT("Invisible Land Boundary");
		break;
		
		case RecordTypeInvisibleWaterBoundary:
			strName = wxT("Invisible Water Boundary");
		break;
		
		case RecordTypeInvisibleMiscBoundary:
			strName = wxT("Invisible Misc Boundary");
		break;
		
		case RecordTypeLandmark:
			strName = wxT("Landmark");
		break;
		
		case RecordTypePhysicalFeature:
			strName = wxT("Physical Feature");
		break;
		
		case RecordTypeTownOrCity:
			strName = wxT("Town or City");
		break;
		
		case RecordTypeCountyBoundary:
			strName = wxT("County Boundary");
		break;
		
		case RecordTypeStateBoundary:
			strName = wxT("State Boundary");
		break;
	}
	return strName;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines some information about the intersection where two
/// records come together, like the angle between them, vertex #,
/// how many other roads are at the intersection, etc.
///
//////////////////////////////////////////////////////////////////////////////
vector<IMapControlDataEntry *> MapControlData_Foundation::IntersectionInformation(const tRecordID & iPreviousRec, const tRecordID & iRec, double * pfDeltaAngle, wxString * pstrDirectionChange)
{
	bool bThisReversed = false;
	bool bPreviousReversed = false;
	MapVertex cVertex;
	vector<IMapControlDataEntry *> vRtn;

	IMapControlDataEntry * psRec = GetRecordByID(iRec);
	IMapControlDataEntry * psPreviousRec = GetRecordByID(iPreviousRec);

	Point ptPreviousStart = psPreviousRec->psCoordinates[0];
	Point ptPreviousEnd = psPreviousRec->psCoordinates[psPreviousRec->nCoordinates - 1];

	Point ptThisStart = psRec->psCoordinates[0];
	Point ptThisEnd = psRec->psCoordinates[psRec->nCoordinates - 1];

	if (ptThisEnd == ptPreviousStart)
	{
		bThisReversed = true;
		bPreviousReversed = true;
	}
	if (ptThisEnd == ptPreviousEnd)
	{
		bThisReversed = true;
	}
	if (ptThisStart == ptPreviousStart)
	{
		bPreviousReversed = true;
	}

	if (bPreviousReversed)
	{
		ptPreviousStart = psPreviousRec->psCoordinates[1];
		ptPreviousEnd = psPreviousRec->psCoordinates[0];
	}
	else
	{
		ptPreviousStart = psPreviousRec->psCoordinates[psPreviousRec->nCoordinates - 2];
		ptPreviousEnd = psPreviousRec->psCoordinates[psPreviousRec->nCoordinates - 1];
	}
	if (bThisReversed)
	{
		ptThisStart = psRec->psCoordinates[psRec->nCoordinates - 1];
		ptThisEnd = psRec->psCoordinates[psRec->nCoordinates - 2];
	}
	else
	{
		ptThisStart = psRec->psCoordinates[0];
		ptThisEnd = psRec->psCoordinates[1];
	}


	// determine the angle between the previous and the current leg
	Point ptDeltaThis = ptThisEnd - ptThisStart;
	Point ptDeltaPrevious = ptPreviousEnd - ptPreviousStart;

	double fAngleThis = ptDeltaThis.angle();
	double fAnglePrevious = ptDeltaPrevious.angle();

	*pfDeltaAngle = fAngleThis - fAnglePrevious;

	if (*pfDeltaAngle > PI)
		*pfDeltaAngle = *pfDeltaAngle - 2 * PI;
	if (*pfDeltaAngle < -PI)
		*pfDeltaAngle = *pfDeltaAngle + 2 * PI;

	cVertex.FindRecords(this, ptThisStart, m_mapCoordinateToRecords);

	*pstrDirectionChange = wxT("continue");

	double fMinTurn = PI / 6;
	if (*pfDeltaAngle < -fMinTurn)
		*pstrDirectionChange = wxString::Format(wxT("turn %sright"), *pfDeltaAngle > (-PI / 4) ? wxT("slightly ") : wxT(""));
	else if (*pfDeltaAngle > fMinTurn)
		*pstrDirectionChange = wxString::Format(wxT("turn %sleft"), *pfDeltaAngle < (PI / 4) ? wxT("slightly ") : wxT(""));
	else
		*pstrDirectionChange = wxString::Format(wxT("go straight"));

	// if there are only two paths leaving this vertex, then we haven't
	// reached an intersection, rather the name of the road is changing
	if (cVertex.m_nAssociatedRecords == 2)
		*pstrDirectionChange = wxT("continue");

	int i;
	for (i = 0; i < cVertex.m_nAssociatedRecords; i++)
		vRtn.push_back(cVertex.m_pcAssociatedRecords[i]);
	
	return vRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Translate a list of record entries into human readable directions
///
//////////////////////////////////////////////////////////////////////////////
wxString MapControlData_Foundation::TranslateRecordsToDirections(const vector<tRecordID> & vRecords)
{
	wxString strRtn;
	unsigned int iLeg;
	double fTotalDistance = 0;
	double fDistance = 0;
	double fTotalTime = 0;
	double fTime = 0;

	for (iLeg = 0; iLeg < vRecords.size(); iLeg++)
	{
		tRecordID iRec = vRecords[iLeg];
		IMapControlDataEntry * psRec = GetRecordByID(iRec);
		wxString strIntersecting;
		bool bSharpTurn = false;
		wxString strDirectionChange;
		
		wxASSERT(psRec);

		if (iLeg > 0)
		{
			tRecordID iPreviousRec = vRecords[iLeg - 1];
			IMapControlDataEntry * psPreviousRec = GetRecordByID(iPreviousRec);

			double fDeltaAngle = 0;
			vector<IMapControlDataEntry *> vAssociatedRecords;
			
			wxASSERT(psPreviousRec);

			vAssociatedRecords = IntersectionInformation(iPreviousRec, iRec, &fDeltaAngle, &strDirectionChange);

			if (fabs(fDeltaAngle) > (60. / 180 * PI))
				bSharpTurn = true;

			unsigned int iPath;
			map<wxString,int> mapIntersectionMentioned;
			for (iPath = 0; iPath < vAssociatedRecords.size(); iPath++)
			{
				IMapControlDataEntry * psPeerRec = vAssociatedRecords[iPath];
				
				wxASSERT(psPeerRec);
				
				wxString strSmartName = psPeerRec->SmartName();
				if (!psPeerRec->IsSameRoad(psRec) && !psPeerRec->IsSameRoad(psPreviousRec) && !mapIntersectionMentioned[strSmartName])
				{
					strIntersecting += wxString::Format(wxT("%s/"), strSmartName.c_str());
					mapIntersectionMentioned[strSmartName] = 1;
				}
			}
			if (strIntersecting.Length())
				strIntersecting = strIntersecting.Left(strIntersecting.Length() - 1);

			fDistance += psPreviousRec->RecordDistance();
			fTime += psPreviousRec->TimeFactor() * psPreviousRec->RecordDistance();
		}

		wxString strNewStreet;
		wxString strNewStreetLocalName;
		wxString strNewStreetSmartName;

		strNewStreetSmartName = psRec->SmartName();
		strNewStreetLocalName = psRec->NameAndType();
		
		if (strNewStreetSmartName != strNewStreetLocalName)
			strNewStreet = strNewStreetLocalName + wxT(" (") + strNewStreetSmartName + wxT(")");
		else
			strNewStreet = strNewStreetSmartName;

		bool bSwitchedRoads;
		
		if (iLeg > 0)
			bSwitchedRoads = !psRec->IsSameRoad(GetRecordByID(vRecords[iLeg - 1]));
		else
			bSwitchedRoads = true;

		bool bWroteSomething = true;
		if (iLeg == 0)
		{
			strRtn += 
						FormatTime(fTime) + 
						wxT(" ") +
						FormatDistance(fTotalDistance + fDistance) + 
						wxT(" -") +
						wxT(" Start out on ") +
						strNewStreet;
			strRtn += wxT("\n");
		}
		else if (iLeg == vRecords.size() - 1)
		{
			strRtn += 
						FormatTime(fTotalTime + fTime) + 
						wxT(" ") +
						FormatDistance(fTotalDistance + fDistance) + 
						wxT(" -") +
						wxT(" Stop after ") +
						FormatDistance(fDistance) + 
						wxT(" at ") + 
						strNewStreet;
			strRtn += wxT("\n");
		}
		else if (bSwitchedRoads)
		{
			strRtn += 
						FormatTime(fTotalTime + fTime) + 
						wxT(" ") +
						FormatDistance(fTotalDistance + fDistance) + 
						wxT(" -") +
						wxT(" Go ") + 
						FormatDistance(fDistance) + 
						wxT(" then ") + 
						strDirectionChange + 
						wxT(" onto ") + 
						strNewStreet;
			
			if (strIntersecting != wxT(""))
				strRtn +=
						wxT(" (at intersection with ") + 
						strIntersecting + 
						wxT(")");
			
			strRtn += wxT("\n");
		}
		else if (bSharpTurn)
		{
			strRtn += 
						FormatTime(fTotalTime + fTime) + 
						wxT(" ") +
						FormatDistance(fTotalDistance + fDistance) + 
						wxT(" -") +
						wxT(" Go ") + 
						FormatDistance(fDistance) + 
						wxT(" then ") + 
						strDirectionChange + 
						wxT(" staying on ") + 
						strNewStreet;
			
			if (strIntersecting != wxT(""))
				strRtn +=
						wxT(" (at intersection with ") + 
						strIntersecting + 
						wxT(")");
			
			strRtn += wxT("\n");
		}
		else
		{
			bWroteSomething = false;
		}
		
		if (bWroteSomething)
		{
			fTotalDistance += fDistance;
			fTotalTime += fTime;
			fDistance = 0;
			fTime = 0;
		}
	}
	strRtn += wxString::Format(wxT("Total Distance: %s\n"), FormatDistance(fTotalDistance).c_str());
	strRtn += wxString::Format(wxT("Total Time: %s\n"), FormatTime(fTotalTime).c_str());

	return strRtn;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/*bool MapControlData_Foundation::LoadGazetteerPlaces(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];
	int nNewRecords;
	MapControlDataEntry_Std * psOldSegments = NULL;

	hFile = fopen(strFilename, "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format("Error opening %s (gazetteer)!", strFilename.c_str()), "Error", wxOK);
		return true;
	}

	nNewRecords = 0;
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		nNewRecords++;
		fgets(szLine, sizeof(szLine), hFile);
	}

	if (m_nRecords)
	{
		psOldSegments = m_psRecords;
	}
	m_psRecords = new MapControlDataEntry_Std[m_nRecords + nNewRecords];
	if (m_nRecords)
	{
		memcpy(m_psRecords, psOldSegments, sizeof(MapControlDataEntry_Std) * m_nRecords);
		delete psOldSegments;
	}
	
	fseek(hFile, 0, SEEK_SET);
	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		char szFeatureName[65];
		char szState[3];
		char * pszStripSuffixes[] = {" borough", " CDP", " village", " city", " town"};
		unsigned int iSuffix;
		vector<wxUint32> vZips;

		m_psRecords[m_nRecords].m_pParent = this;

		strcpy(szState, ExtractField(szLine, 1, 2));
		strcpy(szFeatureName, ExtractField(szLine, 10, 73));
		
		for (iSuffix = 0; iSuffix < sizeof(pszStripSuffixes) / sizeof(char *); iSuffix++)
		{
			if (strlen(szFeatureName) >= strlen(pszStripSuffixes[iSuffix]))
			{
				char * pszSuffix = szFeatureName + strlen(szFeatureName) - strlen(pszStripSuffixes[iSuffix]);
				
				if (!CaseInsensitiveStrCmp(pszSuffix, pszStripSuffixes[iSuffix]))
					*pszSuffix = 0;
			}
		}
		
		m_psRecords[m_nRecords].pszFeatureNames = new char * [1];
		m_psRecords[m_nRecords].pszFeatureNames[0] = new char [strlen(szFeatureName) + 1];
		strcpy(m_psRecords[m_nRecords].pszFeatureNames[0], szFeatureName);

		m_psRecords[m_nRecords].pszFeatureTypes = new char * [1];
		m_psRecords[m_nRecords].pszFeatureTypes[0] = new char [1];
		strcpy(m_psRecords[m_nRecords].pszFeatureTypes[0], "");

		m_psRecords[m_nRecords].nFeatureNames = 1;

		m_psRecords[m_nRecords].eType = RecordTypeTownOrCity;

		m_psRecords[m_nRecords].nAddressZipRanges = 0;
		m_psRecords[m_nRecords].psAddressZipRanges = NULL;

		m_psRecords[m_nRecords].psCoordinates = new Point [1];
		m_psRecords[m_nRecords].psCoordinates[0].GetLong() = atof(ExtractField(szLine, 154, 164));
		m_psRecords[m_nRecords].psCoordinates[0].GetLat() = atof(ExtractField(szLine, 144, 153));
		m_psRecords[m_nRecords].nCoordinates = 1;

#ifdef FILLED_IN_WATER_SUPPORT
		m_psRecords[m_nRecords].bWaterL = false;
		m_psRecords[m_nRecords].bWaterR = false;
#endif

		vZips = ZipCodesByCityState(szFeatureName, szState);
		if (vZips.size())
		{
			m_psRecords[m_nRecords].psAddressZipRanges = new struct SAddressRangeZip [1];
			m_psRecords[m_nRecords].psAddressZipRanges[0].iFromAddr = 0;
			m_psRecords[m_nRecords].psAddressZipRanges[0].iToAddr = 0;
			m_psRecords[m_nRecords].psAddressZipRanges[0].iZip = vZips[0];
			m_psRecords[m_nRecords].nAddressZipRanges = 1;
		}

		m_nRecords++;

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;

}*/

