/*
 *  LibRoadnav
 *  MapControlData_Tiles.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Stores records in a set of tile-layer files
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/utils.h>
#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/url.h>

#include "Debug.h"
#include "Map.h"
#include "MapControlData_Tiles.h"
#include "MapSupport.h"
#include "Portability.h"
#include "MapZip.h"
#include "ProgressDialog.h"
#include "StringTable.h"

#include <map>

using namespace std;

wxDateTime MapControlData_Tiles::m_tLastUpdated((time_t) 0);
wxMutex MapControlData_Tiles::m_mtxMaps;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor
///
//////////////////////////////////////////////////////////////////////////////
MapControlData_Tiles::MapControlData_Tiles() : MapControlData_Foundation()
{
	m_pvstrLoadedTiles = NULL;
	m_nLoadedTiles = 0;
	m_iLoadedTileSearchPos = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor
///
//////////////////////////////////////////////////////////////////////////////
MapControlData_Tiles::~MapControlData_Tiles()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Clear all of the records from memory
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Tiles::Clear()
{
	MapControlData_Foundation::Clear();

	if (m_pvstrLoadedTiles)
		delete [] m_pvstrLoadedTiles;
	m_pvstrLoadedTiles = NULL;
	m_nLoadedTiles = 0;
	
	m_mapIDtoRecordNumber.clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the size of a tile in degrees for a particular layer
/// (record type)
///
//////////////////////////////////////////////////////////////////////////////
double MapControlData_Tiles::TileSize(ERecordType eType)
{
	// Due to the tile naming convention, tile sizes must be multiples
	// of 0.1.
	switch (eType)
	{
		case RecordTypeStateBoundary:
			return 5;

		case RecordTypeCountyBoundary:
			return 2;

		case RecordTypeOneWaySmallRoad:
		case RecordTypeTwoWaySmallRoad:
			return 0.1;

		default:
			return 0.5;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the specified tile is loaded
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlData_Tiles::IsTileLoaded(const wxString & strName)
{
	int iPos;

	if (!m_pvstrLoadedTiles || !m_nLoadedTiles)
		return false;
		
	if (m_iLoadedTileSearchPos >= m_nLoadedTiles)
		m_iLoadedTileSearchPos = 0;
	
	iPos = m_iLoadedTileSearchPos;
	
	do
	{
		if (m_pvstrLoadedTiles[iPos] == strName)
		{
			m_iLoadedTileSearchPos = iPos;
			return true;
		}
		iPos++;
		
		if (iPos >= m_nLoadedTiles)
			iPos = 0;
	}
	while (iPos != m_iLoadedTileSearchPos);
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Mark the specified tile as loaded
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Tiles::MarkTileLoaded(const wxString & strName)
{
	if ((m_nLoadedTiles & (LOADEDTILE_BLOCK_SIZE - 1)) == 0)
	{
		wxString * pvstrNewLoadedTiles;
		int i;
		
		// need to grow array
		pvstrNewLoadedTiles = new wxString [m_nLoadedTiles + LOADEDTILE_BLOCK_SIZE];
		
		for (i = 0; i < m_nLoadedTiles; i++)
			pvstrNewLoadedTiles[i] = m_pvstrLoadedTiles[i];
			
		delete [] m_pvstrLoadedTiles;
		m_pvstrLoadedTiles = pvstrNewLoadedTiles;
	}
	
	m_pvstrLoadedTiles[m_nLoadedTiles++] = strName;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the rectangular region from ptTopLeft to ptBottomRight into
/// this object.
///
/// Only loads layers specified by pbVisibility
///
//////////////////////////////////////////////////////////////////////////////
void MapControlData_Tiles::LoadRegion(const Point & ptTopLeft, const Point & ptBottomRight, const bool * pbVisibility, bool bClearExisting)
{
	unsigned int iType;
	int nRecords;
	vector<wxString> vstrFilenamesToLoad;
	vector<wxString>::iterator iterFN;
	bool bMissingTiles;
	
	bMissingTiles = false;

	////////////////////////////////////////////////////////////////
	// Make a list of tiles to load
	////////////////////////////////////////////////////////////////
	for (iType = FIRST_RECORD_TYPE; iType <= LAST_RECORD_TYPE; iType++)
	{
		if (pbVisibility[iType])
		{
			int iLong;

			int iStartLong;
			int iEndLong;
			int iStartLat;
			int iEndLat;

			iStartLong = (int) (ptTopLeft.GetLong() / TileSize((ERecordType) iType));
			iEndLong = (int) (ptBottomRight.GetLong() / TileSize((ERecordType) iType));

			iStartLat = (int) (ptTopLeft.GetLat() / TileSize((ERecordType) iType));
			iEndLat = (int) (ptBottomRight.GetLat() / TileSize((ERecordType) iType));

			////////////////////////////////////////////////////////////////
			// arrange it so iStartLong < iEndLong and iStartLat < iEndLat
			////////////////////////////////////////////////////////////////
			if (iStartLong > iEndLong)
			{
				int iTmp;

				iTmp = iStartLong;
				iStartLong = iEndLong;
				iEndLong = iTmp;
			}

			if (iStartLat > iEndLat)
			{
				int iTmp;
	
				iTmp = iStartLat;
				iStartLat = iEndLat;
				iEndLat = iTmp;
			}

			for (iLong = iStartLong; iLong <= iEndLong; iLong++)
			{
				int iLat;
				for (iLat = iStartLat; iLat <= iEndLat; iLat++)
				{
					wxFileName fnThis;
					bool bIsLoaded;

					fnThis.AssignDir(GetDataPath());
					fnThis.AppendDir(wxString::Format(wxT("Lon%05d"), iLong * ((int)(TileSize((ERecordType) iType) * 10.0))));
					fnThis.SetFullName(wxString::Format(wxT("Lat%05dType%02d"), iLat * ((int)(TileSize((ERecordType) iType) * 10.0)), iType));
					
					bIsLoaded = IsTileLoaded(fnThis.GetFullPath());
					
					if (!bIsLoaded || bClearExisting)
						vstrFilenamesToLoad.push_back(fnThis.GetFullPath());

					if (!bIsLoaded)
						bMissingTiles = true;
				}
			}
		}
	}

	////////////////////////////////////////////////////////////////
	// If nothing is missing then don't bother clearing and
	// reloading
	////////////////////////////////////////////////////////////////

	if (!bMissingTiles && m_tLastUpdatedForLoadedData == m_tLastUpdated)
		return;

	////////////////////////////////////////////////////////////////
	// wipe out anything loaded
	////////////////////////////////////////////////////////////////

	if (bClearExisting)
		Clear();

	////////////////////////////////////////////////////////////////
	// Don't load data while saving data
	////////////////////////////////////////////////////////////////
	if (m_mtxMaps.TryLock() != wxMUTEX_NO_ERROR)
		return;

	////////////////////////////////////////////////////////////////
	// Count up the records
	////////////////////////////////////////////////////////////////
	nRecords = 0;
	for (iterFN = vstrFilenamesToLoad.begin(); iterFN != vstrFilenamesToLoad.end(); iterFN++)
	{
		FILE * hFile;

		if (wxFileExists(*iterFN + wxT(".dat")))
		{
			hFile = fopen((*iterFN + wxT(".dat")).mb_str(*wxConvCurrent), "rb");

			if (hFile)
			{
				wxUint32 nRecThisFile = 0;

				fread(&nRecThisFile, sizeof(wxUint32), 1, hFile);
				nRecords += nRecThisFile;

				fclose(hFile);
			}
		}
	}

	////////////////////////////////////////////////////////////////
	// Allocate memory
	////////////////////////////////////////////////////////////////
	if (m_nRecords + nRecords > 0)
	{
		if (m_nRecords)
		{
			MapControlDataEntry_Std * psRecordsNew;

			psRecordsNew = new MapControlDataEntry_Std [m_nRecords + nRecords + 1];

			memcpy(psRecordsNew, m_psRecords, sizeof(MapControlDataEntry_Std) * m_nRecords);

			delete [] m_psRecords;

			m_psRecords = psRecordsNew;
		}
		else
		{
			m_psRecords = new MapControlDataEntry_Std [nRecords];
		}
	}
	else
		m_psRecords = NULL;

	////////////////////////////////////////////////////////////////
	// Load up the tile-layers
	////////////////////////////////////////////////////////////////
	for (iterFN = vstrFilenamesToLoad.begin(); iterFN != vstrFilenamesToLoad.end(); iterFN++)
	{
		FILE * hFile;

		if (wxFileExists(*iterFN + wxT(".dat")))
		{
			hFile = fopen((*iterFN + wxT(".dat")).mb_str(*wxConvCurrent), "rb");

			if (hFile)
			{
				wxUint32 nRecThisFile = 0;
				wxUint32 iRec;
				Point ptRef;
				
				ptRef.SetInvalid();

				fread(&nRecThisFile, sizeof(wxUint32), 1, hFile);

				for (iRec = 0; iRec < nRecThisFile; iRec++)
				{
					bool bBelongsToMultipleFiles;
					
					m_psRecords[m_nRecords].Serialize(hFile, false, &ptRef);
					m_psRecords[m_nRecords].SerializeItem(hFile, false, &bBelongsToMultipleFiles);

					if (bBelongsToMultipleFiles)
					{
						if (m_mapIDtoRecordNumber.find(m_psRecords[m_nRecords].idRecord) == m_mapIDtoRecordNumber.end())
						{
							m_mapIDtoRecordNumber[m_psRecords[m_nRecords].idRecord] = m_nRecords;
							m_nRecords++;
						}
						else
						{
							m_psRecords[m_nRecords].Clear();
						}
					}
					else
					{
						m_nRecords++;
					}
				}

#ifdef LIBROADNAV_DEBUG
				// verify we're at the end of the file
				long lCurrentPos;
				long lEndPos;

				lCurrentPos = ftell(hFile);
				fseek(hFile, 0, SEEK_END);
				lEndPos = ftell(hFile);
				wxASSERT(lEndPos == lCurrentPos);
#endif
				fclose(hFile);
			}
		}
		MarkTileLoaded(*iterFN);
	}

	////////////////////////////////////////////////////////////////
	// Post-processing
	////////////////////////////////////////////////////////////////

	m_tLastUpdatedForLoadedData = m_tLastUpdated;
	
	m_mtxMaps.Unlock();
}

typedef pair<wxInt16,wxInt16> tPairwxInt16;

//////////////////////////////////////////////////////////////////////////////
///
/// Saves the data in a MapControlDataImporter object into tile-layers
///
//////////////////////////////////////////////////////////////////////////////
bool MapControlData_Tiles::Save(MapControlDataImporter * pImporter, ProgressDialog * pdlgProgress, int iProgressStart, int iProgressTicks)
{
	int i;
	int iStartRec;
	unsigned int iStartCoord;
	int iSearch;
	FILE * hFileDat = NULL;
	wxFileName fnThis;
	IMapControlDataEntry * psRec;
	wxString strBaseFilename;
	int iLong;
	int iLat;
	int iType;
	vector<int> * piLocs = NULL;
	int iCount;
	wxFileName fnDataPath;
	ProgressDialog * dlgRealProgress;
	int iRealProgressStart;
	int iRealProgressTicks;
	int iTotal;

	iCount = 0;
	iTotal = 0;

	if (pdlgProgress)
	{
		dlgRealProgress = pdlgProgress;
		iRealProgressStart = iProgressStart;
		iRealProgressTicks = iProgressTicks;
	}
	else
	{
		dlgRealProgress = new ProgressDialog(NULL, wxT("Processing..."), 100);
		iRealProgressStart = 0;
		iRealProgressTicks = 100;
	}

	dlgRealProgress->Show();
	dlgRealProgress->SetLabel(wxT("Building maps"));

	fnDataPath.AssignDir(GetDataPath());

	////////////////////////////////////////////////////////////////
	// determine filenames
	////////////////////////////////////////////////////////////////
	piLocs = new vector<int> [pImporter->Count()];

	for (i = 0; i < pImporter->Count(); i++)
	{
		int iCoordinate;
		map<tPairwxInt16,bool> bIncluded;

		psRec = pImporter->GetRecord(i);

		for (iCoordinate = 0; iCoordinate < psRec->nCoordinates; iCoordinate++)
		{
			wxInt16 iLong = (int) (psRec->psCoordinates[iCoordinate].GetLong() / TileSize(psRec->eType));
			wxInt16 iLat = (int) (psRec->psCoordinates[iCoordinate].GetLat() / TileSize(psRec->eType));
			tPairwxInt16 pCoord(iLong, iLat);

			if (!bIncluded[pCoord])
			{
				bIncluded[pCoord] = true;

				piLocs[i].push_back(iLong);
				piLocs[i].push_back(iLat);
				piLocs[i].push_back((int) (psRec->eType));

				iTotal++;
			}
		}
	}

	iStartRec = 0;
	iStartCoord = 0;

	wxMutexLocker mlMaps(m_mtxMaps);

	while (iStartRec < pImporter->Count())
	{
		wxUint32 nNewRecords;
		Point ptRef;
		
		ptRef.SetInvalid();

		////////////////////////////////////////////////////////////////
		// open files	
		////////////////////////////////////////////////////////////////
		iLong = piLocs[iStartRec][iStartCoord + 0];
		iLat = piLocs[iStartRec][iStartCoord + 1];
		iType = piLocs[iStartRec][iStartCoord + 2];

		wxASSERT(iType >= 0);

		fnThis = fnDataPath;
		fnThis.AppendDir(wxString::Format(wxT("Lon%05d"), iLong * ((int)(TileSize((ERecordType) iType) * 10.0))));
		fnThis.SetFullName(wxString::Format(wxT("Lat%05dType%02d"), iLat * ((int)(TileSize((ERecordType) iType) * 10.0)), iType));

		strBaseFilename = fnThis.GetFullPath();

		hFileDat = fopen((strBaseFilename + wxT(".dat")).mb_str(*wxConvCurrent), "a+b");

		if (!hFileDat)
		{
			wxFileName fnDir;

			fnDir = strBaseFilename;
			fnDir.SetFullName(wxT(""));

			if (!fnDir.DirExists())
				fnDir.Mkdir(0755);

			hFileDat = fopen((strBaseFilename + wxT(".dat")).mb_str(*wxConvCurrent), "a+b");
		}

		wxASSERT(hFileDat);

		////////////////////////////////////////////////////////////////
		// Make sure the count is at the beginning
		////////////////////////////////////////////////////////////////
		fseek(hFileDat, 0, SEEK_END);
		if (ftell(hFileDat) == 0)
		{
			wxUint32 nRecThisFile = 0;
			fwrite(&nRecThisFile, sizeof(wxUint32), 1, hFileDat);
		}

		////////////////////////////////////////////////////////////////
		// Find records that go into this file
		////////////////////////////////////////////////////////////////
		nNewRecords = 0;
		for (iSearch = iStartRec; iSearch < pImporter->Count(); iSearch++)
		{
			unsigned int iSearchCoord;

			for (iSearchCoord = 0; iSearchCoord < piLocs[iSearch].size(); iSearchCoord += 3)
			{
				if (iLong == piLocs[iSearch][iSearchCoord + 0] &&
					iLat == piLocs[iSearch][iSearchCoord + 1] &&
					iType == piLocs[iSearch][iSearchCoord + 2])
				{
					bool bBelongsToMultipleFiles = piLocs[iSearch].size() > 3;
					
					psRec = pImporter->GetRecord(iSearch);

					////////////////////////////////////////////////////////////////
					// write data
					////////////////////////////////////////////////////////////////
					psRec->Serialize(hFileDat, true, &ptRef);
					psRec->SerializeItem(hFileDat, true, &bBelongsToMultipleFiles);
					
					dlgRealProgress->SetProgress((int) (iRealProgressStart + ((double) iCount) * iRealProgressTicks / iTotal));
					iCount++;
					nNewRecords++;

					////////////////////////////////////////////////////////////////
					// mark as saved
					////////////////////////////////////////////////////////////////
					piLocs[iSearch][iSearchCoord + 2] = -1;
				}
			}
		}

		////////////////////////////////////////////////////////////////
		// Update count at beginning of file
		////////////////////////////////////////////////////////////////
		wxUint32 nRecThisFile = 0;

		fseek(hFileDat, 0, SEEK_SET);
		fread(&nRecThisFile, sizeof(wxUint32), 1, hFileDat);
		nRecThisFile += nNewRecords;

		fclose(hFileDat);
		hFileDat = fopen((strBaseFilename + wxT(".dat")).mb_str(*wxConvCurrent), "r+b");

		if (hFileDat)
		{
			fseek(hFileDat, 0, SEEK_SET);
			fwrite(&nRecThisFile, sizeof(wxUint32), 1, hFileDat);
		}

		////////////////////////////////////////////////////////////////
		// close file
		////////////////////////////////////////////////////////////////
		fclose(hFileDat);

		////////////////////////////////////////////////////////////////
		// find the next record to save
		////////////////////////////////////////////////////////////////
		bool bFound = false;

		for (iStartRec = 0; iStartRec < pImporter->Count() && !bFound; iStartRec++)
		{
			for (iStartCoord = 0; !bFound && iStartCoord < piLocs[iStartRec].size(); iStartCoord += 3)
			{
				if (piLocs[iStartRec][iStartCoord + 2] >= 0)
				{
					bFound = true;
					iStartCoord -= 3;
					iStartRec--;
				}
			}
		}
	}

	g_pcStringTable->Save();

	if (!pdlgProgress)
		delete dlgRealProgress;

	delete [] piLocs;

	m_tLastUpdated = wxDateTime::Now();

	return true;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the time the map data was last updated.
///
//////////////////////////////////////////////////////////////////////////////
wxDateTime MapControlData_Tiles::GetLastUpdated()
{
	return m_tLastUpdated;
}
