/*
 *  LibRoadnav
 *  MapCounty.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Maps counties to/from their FIPS code.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <vector>
#include <map>

#include <wx/wx.h>

#include "MapSupport.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// County associations
///
//////////////////////////////////////////////////////////////////////////////
static map<wxString,wxInt32> mapCountyNameToCode;
static map<wxInt32,wxString> mapCodeToCountyName;
static vector<wxString> vCountyNames;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load county information from a file.
///
/// Format:
/// Columns  1 -  2: State two letter abbreviation
/// Columns  3 -  7: FIPS state and county code
/// Columns  8 - 60: County name
///
//////////////////////////////////////////////////////////////////////////////
bool LoadCountyInformation(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (counties)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		// County name
		wxString strCounty;
		// State abbreviation
		wxString strState;
		// State + county
		wxString strCombined;
		// FIPS code
		wxInt32 iCode;

		// only process lines starting with a state code
		if (isdigit(szLine[0]))
		{
			int iInstanceNumber;
			
			// strip off CR/LF
			while (strlen(szLine) && (szLine[strlen(szLine) - 1] == '\r' || szLine[strlen(szLine) - 1] == '\n'))
				szLine[strlen(szLine) - 1] = 0;
			
			// combine state and county code
			iCode = atoi(ExtractField(szLine, 1, 2)) * 1000 + atoi(ExtractField(szLine, 8, 10));
			
			strCounty = wxString(szLine + 15, *wxConvCurrent);
			strState = strCounty.Right(2);
			strCounty = strCounty.Left(strCounty.Length() - 2);
			strCounty.Trim();
			
			strCombined = strState + wxT(" - ") + strCounty;
			
			iInstanceNumber = 2;
			while (mapCountyNameToCode.find(strCombined) != mapCountyNameToCode.end())
			{
				strCombined = strState + wxT(" - ") + strCounty + wxString::Format(wxT(" (%d)"), iInstanceNumber);
				iInstanceNumber++;
			}
			
			mapCountyNameToCode[strCombined] = iCode;
			mapCodeToCountyName[iCode] = strCombined;
			vCountyNames.push_back(strCombined);
		}

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a vector of all the county names
///
//////////////////////////////////////////////////////////////////////////////
vector<wxString> GetCountyNames()
{
	return vCountyNames;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a five digit state/county code given the name of the county
///
//////////////////////////////////////////////////////////////////////////////
wxInt32 CountyCodeByName(wxString strName)
{
	if (mapCountyNameToCode.find(strName) != mapCountyNameToCode.end())
		return mapCountyNameToCode[strName];

	return -1;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a county name given the five digit state/county code
///
//////////////////////////////////////////////////////////////////////////////
wxString CountyNameByCode(wxInt32 iCode)
{
	if (mapCodeToCountyName.find(iCode) != mapCodeToCountyName.end())
		return mapCodeToCountyName[iCode];

	return wxT("");
}
