/*
 *  LibRoadnav
 *  MapDrawing.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Drawing support functions for MapControl.cpp
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/wx.h>

#include "Constants.h"

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw a line from (sx, sy) to (dx, dy), iWidth pixels wide, 
/// color clrColor, using style iStyle.
///
/// If iArrowHeadSpacing is non-zero, draw arrow heads along the line too
///
//////////////////////////////////////////////////////////////////////////////
void DrawLine(wxDC * dc, int sx, int sy, int dx, int dy, int iWidth, const wxColour & clrColor, int iStyle, int iArrowHeadSpacing, int iRailroadWidth, int iRailroadSpacing)
{
	wxPen & pen = *wxThePenList->FindOrCreatePen(clrColor, iWidth, iStyle);

	dc->SetPen(pen);

	dc->DrawLine(sx, sy, dx, dy);

#if 0	
	if (iWidth >= 12)
	{
		double fAngle;
		int iOffset;
		int iColorOffset = 128;
		
		wxColour clrShaded(0, 0, 0);
							
		wxPen penShaded(clrShaded, 1, iStyle);
		
		fAngle = atan2((double) dy - sy, (double) dx - sx);
		
		dc->SetPen(penShaded);
		
		iOffset = iWidth / 2 - 2;

		dc->DrawLine(
					(int) (sx + cos(fAngle - PI / 2) * iOffset), 
					(int) (sy + sin(fAngle - PI / 2) * iOffset),
					(int) (dx + cos(fAngle - PI / 2) * iOffset), 
					(int) (dy + sin(fAngle - PI / 2) * iOffset)
					);

		dc->DrawLine(
					(int) (sx + cos(fAngle + PI / 2) * iOffset), 
					(int) (sy + sin(fAngle + PI / 2) * iOffset),
					(int) (dx + cos(fAngle + PI / 2) * iOffset), 
					(int) (dy + sin(fAngle + PI / 2) * iOffset)
					);
	}
#endif
	
	if (iArrowHeadSpacing)
	{
		int iLength;
		int nArrows;
		int iArrow;
		
		iLength = (int) sqrt(pow((double) dx - sx, 2) + pow((double) dy - sy, 2));
		nArrows = iLength / iArrowHeadSpacing;
		
		for (iArrow = 0; iArrow < nArrows; iArrow++)
		{
			wxPoint pt;
			double fAngle;
			
			fAngle = atan2((double) dy - sy, (double) dx - sx);
			
			pt.x = (int) (sx + ((double)(dx - sx) * iArrow / nArrows));
			pt.y = (int) (sy + ((double)(dy - sy) * iArrow / nArrows));

			pen.SetCap(wxCAP_BUTT);
			dc->SetPen(pen);

			dc->DrawLine(pt.x, 
						pt.y, 
						(int) (pt.x + cos(fAngle - PI / 8) * iWidth * 2), 
						(int) (pt.y + sin(fAngle - PI / 8) * iWidth * 2));
		}
	}
	
	if (iRailroadWidth && iRailroadSpacing)
	{
		int iLength;
		int nTracks;
		int iTrack;
		
		iLength = (int) sqrt(pow((double) dx - sx, 2) + pow((double) dy - sy, 2));
		nTracks = iLength / iRailroadSpacing;
		if (!nTracks)
			nTracks = 1;
		
		for (iTrack = 0; iTrack < nTracks; iTrack++)
		{
			wxPoint pt;
			double fAngle;
			
			fAngle = atan2((double) dy - sy, (double) dx - sx);
			
			pt.x = (int) (sx + ((double)(dx - sx) * iTrack / nTracks));
			pt.y = (int) (sy + ((double)(dy - sy) * iTrack / nTracks));

			dc->DrawLine(
						(int) (pt.x - cos(fAngle - PI / 2) * iRailroadWidth), 
						(int) (pt.y - sin(fAngle - PI / 2) * iRailroadWidth),
						(int) (pt.x + cos(fAngle - PI / 2) * iRailroadWidth), 
						(int) (pt.y + sin(fAngle - PI / 2) * iRailroadWidth));
		}
	}

	dc->SetPen(wxNullPen);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draws rotated text
///
//////////////////////////////////////////////////////////////////////////////
void DrawRotatedText(wxDC* pDC, const wxString & str, const wxPoint & p, double angle, const wxColour & clrText)
{
	if (!*str)
		return;
	
	pDC->SetTextForeground(clrText);
	pDC->DrawRotatedText(str, p.x, p.y, angle);
}
