/*
 *  LibRoadnav
 *  MapLookup.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Looks up addresses or finds the closest address to a set of coordinates.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <vector>
#include <list>

#include "Map.h"
#include "MapZip.h"
#include "MapLookup.h"
#include "MapSupport.h"
#include "MapAbbreviations.h"
#include "MapControlData_Tiles.h"
#include "MapState.h"
#include "MapRepresentations.h"
#include "StringTable.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the closest match to the supplied street name for the
/// supplied city, state, and map data.
///
/// \param strSearchStreet		Misspelled street name.
/// \param strSearchType		Misspelled street type.
/// \param strCity				Correctly spelled city name.
/// \param strState				Correctly spell state name or abbreviation.
/// \param pDataForced			Data to use, or NULL to autoload data.
/// \return						Recommended street name and type.
///
//////////////////////////////////////////////////////////////////////////////
wxString ClosestStreetName(const wxString & strSearchStreet, const wxString & strSearchType, const wxString & strCity, const wxString & strState, IMapControlData * pDataForced)
{
	int iRec;
	vector<wxUint32> vZips;
	IMapControlData * pData = pDataForced;
	wxString strSearchTypeAbbreviated;
	wxString strSearchStreetCondensed;
	wxString strBestMatch;
	double fBestMatch = -1;

	// abbreviate type into strSearchTypeAbbreviated
	strSearchTypeAbbreviated = GetAbbreviation(strSearchType);
	if (strSearchTypeAbbreviated == wxT(""))
		strSearchTypeAbbreviated = strSearchType;

	// look up zip codes
	if (strCity != wxT("") && strState != wxT(""))
		vZips = ZipCodesByCityState(strCity, strState);

	// condense street name
	strSearchStreetCondensed = strSearchStreet + strSearchTypeAbbreviated;
	strSearchStreetCondensed.Replace(wxT(" "), wxT(""));
	strSearchStreetCondensed.Replace(wxT("."), wxT(""));
	strSearchStreetCondensed.Replace(wxT(","), wxT(""));
	
	// go through records
	for (iRec = 0; iRec < pData->Count(); iRec++)
	{
		IMapControlDataEntry * psRec = pData->GetRecord(iRec);
		bool bMatch = true;
		
		if (strCity != wxT("") && strState != wxT("") && bMatch)
		{
			vector<wxUint32>::iterator iZip;
			int iRange;
			
			bMatch = false;
			
			for (iRange = 0; iRange < psRec->nAddressZipRanges && !bMatch; iRange++)
			{
				for (iZip = vZips.begin(); iZip != vZips.end() && !bMatch; iZip++)
				{
					if (psRec->psAddressZipRanges[iRange].iZip == *iZip)
						bMatch = true;
				}
			}
		}

		if (bMatch)
		{
			int i;
			for (i = 0; i < psRec->nFeatureNames; i++)
			{
				double fThisMatch;
				
				wxString strThisStreetNameType = wxString(psRec->pszFeatureNames[i], wxConvUTF8) + wxString(psRec->pszFeatureTypes[i], wxConvUTF8);
				// remove spaces
				strThisStreetNameType.Replace(wxT(" "), wxT(""));
				
				// and punctuation
				strThisStreetNameType.Replace(wxT("."), wxT(""));
				strThisStreetNameType.Replace(wxT(","), wxT(""));
				
				fThisMatch = WordSimilarity(strThisStreetNameType , strSearchStreetCondensed);
				
				if (fThisMatch > fBestMatch)
				{
					fBestMatch = fThisMatch;
					strBestMatch = wxString(psRec->pszFeatureNames[i], wxConvUTF8);
					
					if (*psRec->pszFeatureTypes[i])
						strBestMatch += wxT(" ") + wxString(psRec->pszFeatureTypes[i], wxConvUTF8);
				}
			}
		}
	}
	return strBestMatch;
}

void LoadCityStateIntoRecords(const wxString & strCity, const wxString & strState, const bool * pbVisibility, IMapControlData * pData, 	vector<wxUint32> & vZips)
{
	wxString strStateMod = strState;
	
	//////////////////////////////////////////////////////////////////////////
	// state .. convert to abbreviation if necessary
	//////////////////////////////////////////////////////////////////////////
	if (strStateMod.Length())
	{
		wxString strTmp;
		
		strTmp = StateAbbreviationByName(strStateMod);

		if (strTmp != wxT(""))
			strStateMod = strTmp;
			
		strStateMod = strStateMod.MakeUpper();
	}

	vZips = ZipCodesByCityState(strCity, strStateMod);

	if (vZips.size())
	{
		Point ptLoadTL, ptLoadBR;
		vector<wxUint32>::iterator iZip;
		bool bFirstEntry;
		
		// find a zip code that we have coordinates for
		bFirstEntry = true;
		for (iZip = vZips.begin(); iZip != vZips.end(); iZip++)
		{
			Point ptTmpTL, ptTmpBR;
			
			if (!GetZipCodeCoordinates(*iZip, &ptTmpTL, &ptTmpBR))
			{
				// check validity
				if (ptTmpTL.abs() > 0.001 && ptTmpBR.abs() > 0.001)
				{
					// update top left
					if (ptTmpTL.GetLatRaw() < ptLoadTL.GetLatRaw() || bFirstEntry)
						ptLoadTL.SetLatRaw(ptTmpTL.GetLatRaw());
					if (ptTmpTL.GetLongRaw() < ptLoadTL.GetLongRaw() || bFirstEntry)
						ptLoadTL.SetLongRaw(ptTmpTL.GetLongRaw());

					// update bottom right
					if (ptTmpBR.GetLatRaw() > ptLoadBR.GetLatRaw() || bFirstEntry)
						ptLoadBR.SetLatRaw(ptTmpBR.GetLatRaw());
					if (ptTmpBR.GetLongRaw() > ptLoadBR.GetLongRaw() || bFirstEntry)
						ptLoadBR.SetLongRaw(ptTmpBR.GetLongRaw());
						
					bFirstEntry = false;
				}
			}
		}
	
		// if at least one of the zip codes had coordinates, then load the region
		// otherwise, typically this means maps haven't been downloaded	
		if (ptLoadTL.abs() > 0.001 && ptLoadBR.abs() > 0.001)
			pData->LoadRegion(ptLoadTL, ptLoadBR, pbVisibility);
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Locate the record cooresponding to an address.
///
/// \param strSearchFullStreet		Street number, name, and type.
///									May be empty.
/// \param strCity					City name
/// \param strState					State name. May be in written out or abbreviated
///									form.
/// \param bVisibility				Indicates which record types should be
///									searched.
/// \param pDataForced				Map data to use, or NULL to automatically
///									load whatever data is needed.
/// \param pstrSuggestedStreet		May return a suggested street number, name, and
///									type if the address cannot be found.
/// \param pstrSuggestedCity		May return a suggested city name if the
///									city cannot be found.
/// \param pstrSuggestedState		May return a suggested state name if the
///									state cannot be found.
/// \return							Address object for the requested address.
///
//////////////////////////////////////////////////////////////////////////////
Address FindAddress(const wxString & strSearchFullStreet, const wxString & strCity, const wxString & strState, const bool * pbVisibility, IMapControlData * pDataForced, wxString * pstrErrorMsg, wxString * pstrSuggestedStreet, wxString * pstrSuggestedCity, wxString * pstrSuggestedState)
{
	long iStreetNumber = 0;
	wxString strStreetName;
	wxString strStreetType;
	Address sAddr;
	IMapControlData * pData;
	MapControlData_Tiles cRec;
	wxString strStateMod(strState);

	if (pstrSuggestedStreet)
		*pstrSuggestedStreet = wxT("");
	
	if (pstrSuggestedCity)
		*pstrSuggestedCity = wxT("");
	
	if (pstrSuggestedState)
		*pstrSuggestedState = wxT("");
	
	strStreetName = strSearchFullStreet;
	
	//////////////////////////////////////////////////////////////////////////
	// extract the street number
	//////////////////////////////////////////////////////////////////////////
	strStreetName.BeforeFirst(' ').ToLong(&iStreetNumber);
	
	if (iStreetNumber)
	{
		strStreetName = strStreetName.AfterFirst(' ');
	}

	//////////////////////////////////////////////////////////////////////////
	// street type
	//////////////////////////////////////////////////////////////////////////
	strStreetType = strStreetName.AfterLast(' ');
	
	// strip off any .
	strStreetType.Replace(wxT("."), wxT(""));
	
	if (IsFeatureType(strStreetType))
	{
		strStreetName = strStreetName.BeforeLast(' ');
	}
	else
	{
		strStreetType = wxT("");
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Handle north/south/east/west/northeast/southeast/southwest/northwest
	// prefix/suffix
	//////////////////////////////////////////////////////////////////////////
	wxString arstrDirectionAbbreviations[] = 
	{
		wxT("N"),		wxT("North"),
		wxT("S"),		wxT("South"),
		wxT("E"),		wxT("East"),
		wxT("W"),		wxT("West"),

		wxT("NE"),		wxT("Northeast"),
		wxT("SE"),		wxT("Southeast"),
		wxT("SW"),		wxT("Southwest"),
		wxT("NW"),		wxT("Northwest"),

		wxT("NE"),		wxT("North east"),
		wxT("SE"),		wxT("South east"),
		wxT("SW"),		wxT("South west"),
		wxT("NW"),		wxT("North west"),

		wxT("NE"),		wxT("North-east"),
		wxT("SE"),		wxT("South-east"),
		wxT("SW"),		wxT("South-west"),
		wxT("NW"),		wxT("North-west"),
		
		wxT(""),		wxT("")
	};
	int iDirectionAbbreviation;
	
	for (iDirectionAbbreviation = 0; arstrDirectionAbbreviations[iDirectionAbbreviation] != wxT(""); iDirectionAbbreviation += 2)
	{
		wxString strAbbreviation;
		wxString strWrittenOut;
		
		strAbbreviation = arstrDirectionAbbreviations[iDirectionAbbreviation];
		strWrittenOut = arstrDirectionAbbreviations[iDirectionAbbreviation + 1];

		// only replace if the street name is longer than the written out direction
		if (strStreetName.Length() > strWrittenOut.Length())
		{
			// check if prefix matches direction written out
			if (strStreetName.Left(strWrittenOut.Length()).Lower() == strWrittenOut.Lower())
			{
				// replace prefix with abbreviation
				strStreetName = strAbbreviation + strStreetName.Mid(strWrittenOut.Length());
			}

			// check if suffix matches direction written out
			if (strStreetName.Right(strWrittenOut.Length()).Lower() == strWrittenOut.Lower())
			{
				// replace suffix with abbreviation
				strStreetName = strAbbreviation + strStreetName.Left(strStreetName.Length() - strWrittenOut.Length());
			}
		}
	}
	

	//////////////////////////////////////////////////////////////////////////
	// city - no processing
	//////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////
	// state .. convert to abbreviation if necessary
	//////////////////////////////////////////////////////////////////////////
	if (strStateMod.Length())
	{
		wxString strTmp;
		
		strTmp = StateAbbreviationByName(strStateMod);

		if (strTmp != wxT(""))
			strStateMod = strTmp;
			
		strStateMod = strStateMod.MakeUpper();
	}
	
	vector<wxUint32> vZips;
	vZips = ZipCodesByCityState(strCity, strStateMod);

	//////////////////////////////////////////////////////////////////////////
	// load region
	//////////////////////////////////////////////////////////////////////////
	if (pDataForced)
		pData = pDataForced;
	else
		pData = &cRec;
		
	if (vZips.size() && !pDataForced)
	{
		Point ptLoadTL, ptLoadBR;
		vector<wxUint32>::iterator iZip;
		bool bFirstEntry;

		// find a zip code that we have coordinates for
		bFirstEntry = true;
		for (iZip = vZips.begin(); iZip != vZips.end(); iZip++)
		{
			Point ptTmpTL, ptTmpBR;
			
			if (!GetZipCodeCoordinates(*iZip, &ptTmpTL, &ptTmpBR))
			{
				// check validity
				if (ptTmpTL.abs() > 0.001 && ptTmpBR.abs() > 0.001)
				{
					// update top left
					if (ptTmpTL.GetLatRaw() < ptLoadTL.GetLatRaw() || bFirstEntry)
						ptLoadTL.SetLatRaw(ptTmpTL.GetLatRaw());
					if (ptTmpTL.GetLongRaw() < ptLoadTL.GetLongRaw() || bFirstEntry)
						ptLoadTL.SetLongRaw(ptTmpTL.GetLongRaw());

					// update bottom right
					if (ptTmpBR.GetLatRaw() > ptLoadBR.GetLatRaw() || bFirstEntry)
						ptLoadBR.SetLatRaw(ptTmpBR.GetLatRaw());
					if (ptTmpBR.GetLongRaw() > ptLoadBR.GetLongRaw() || bFirstEntry)
						ptLoadBR.SetLongRaw(ptTmpBR.GetLongRaw());
						
					bFirstEntry = false;
				}
			}
		}
	
		// if at least one of the zip codes had coordinates, then load the region
		// otherwise, typically this means maps haven't been downloaded	
		if (ptLoadTL.abs() > 0.001 && ptLoadBR.abs() > 0.001)
			pData->LoadRegion(ptLoadTL, ptLoadBR, pbVisibility);
	}



	sAddr = FindAddress(iStreetNumber, 
							strStreetName,
							strStreetType,
							strCity, 
							strStateMod,
							pbVisibility,
							pData);
	
	//////////////////////////////////////////////////////////////////////////
	// try out some variations if we can't find the address
	//////////////////////////////////////////////////////////////////////////
	if (!sAddr.m_bValid)
	{
		//////////////////////////////////////////////////////////////////////////////
		// ignore the street number
		//////////////////////////////////////////////////////////////////////////////
		sAddr = FindAddress(0, 
								strStreetName,
								strStreetType,
								strCity, 
								strStateMod,
								pbVisibility,
								pData);
	}
	
	if (!sAddr.m_bValid)
	{
		//////////////////////////////////////////////////////////////////////////////
		// perhaps the street name starts with a number?
		//////////////////////////////////////////////////////////////////////////////
		sAddr = FindAddress(0, 
								strSearchFullStreet.BeforeLast(' '),
								strStreetType,
								strCity, 
								strStateMod,
								pbVisibility,
								pData);
	}
	
	if (!sAddr.m_bValid)
	{
		//////////////////////////////////////////////////////////////////////////////
		// perhaps the street name start with a number?
		//////////////////////////////////////////////////////////////////////////////
		sAddr = FindAddress(0, 
								strSearchFullStreet,
								wxT(""),
								strCity, 
								strStateMod,
								pbVisibility,
								pData);
	}
	
	if (!sAddr.m_bValid && pstrErrorMsg)
	{
		if (StateNameByAbbreviation(strStateMod) == wxT(""))
		{
			wxString strRecommendation = ClosestStateName(strStateMod);

			*pstrErrorMsg = wxT("Unable to locate that state");
	
			if (strRecommendation != wxT(""))
			{
				*pstrErrorMsg += wxString::Format(wxT(". Perhaps you meant %s instead of %s?"),
					strRecommendation.c_str(),
					FormatName(strStateMod).c_str());

				if (pstrSuggestedState)
				{
					*pstrSuggestedState = strRecommendation;
				}
			}
		}
		else if (!vZips.size())
		{
			wxString strRecommendation = ClosestCityName(strCity, strStateMod);

			*pstrErrorMsg = wxT("Unable to locate that city");
	
			if (strRecommendation != wxT(""))
			{
				*pstrErrorMsg += wxString::Format(wxT(". Perhaps you meant %s instead of %s?"),
					strRecommendation.c_str(),
					FormatName(strCity).c_str());

				if (pstrSuggestedCity)
				{
					*pstrSuggestedCity = strRecommendation;
				}
			}
		}
		else
		{
			wxString strRecommendation = ClosestStreetName(strStreetName, strStreetType, strCity, strStateMod, pData);

			*pstrErrorMsg = wxT("Unable to locate that street address");
			
			if (strRecommendation != wxT(""))
			{
				if (iStreetNumber)
					strRecommendation = wxString::Format(wxT("%d "), (int) iStreetNumber) + strRecommendation;

				*pstrErrorMsg += wxString::Format(wxT(". Perhaps you meant %s instead of %s?"),
					strRecommendation.c_str(),
					FormatName(strSearchFullStreet).c_str());

				if (pstrSuggestedStreet)
				{
					*pstrSuggestedStreet = strRecommendation;
				}
			}
		}
	}
	
	return sAddr;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Locate the record cooresponding to an address.
///
/// \param iSearchNumber	Street number.
///							0 indicates any number.
/// \param strSearchStreet	Street name.
///							Empty string indicates any name.
/// \param strSearchType	Street type, may be abbreviated or written out.
///							Empty string indicates any type.
/// \param strCity			City name.
///							Empty string indicates any city.
/// \param strState			State name.
///							May be either written out or in abbreviation form.
///							Empty string indicates any state.
/// \param pbVisibility		Indicates which record types should be
///							searched.
/// \param pDataForced		Map data to use, or NULL to automatically
///							load whatever data is needed.
/// \return					Address object for the requested address.
///
//////////////////////////////////////////////////////////////////////////////
Address FindAddress(int iSearchNumber, const wxString & strSearchStreet, const wxString & strSearchType, const wxString & strCity, const wxString & strState, const bool * pbVisibility, IMapControlData * pDataForced)
{
	int iRec;
	vector<wxUint32> vZips;
	static Address sRtn;
	wxString strSearchTypeAbbreviated;
	MapControlData_Tiles cRec;
	IMapControlData * pData;
	wxString wxSearchStreet;
	bool bByCityStateOnly = false;

	// if it's not in the string table, then the address is
	// obviously invalid
	if (!g_pcStringTable->GetEntryPointer(strSearchStreet.mb_str(*wxConvCurrent), false) ||
		!g_pcStringTable->GetEntryPointer(strSearchType.mb_str(*wxConvCurrent), false))
	{
		Address cBlank;
		return cBlank;
	}

	if (pDataForced)
		pData = pDataForced;
	else
		pData = &cRec;
	
	strSearchTypeAbbreviated = GetAbbreviation(strSearchType);
	if (strSearchTypeAbbreviated == wxT(""))
		strSearchTypeAbbreviated = strSearchType;

	if (strCity != wxT("") && strState != wxT(""))
		vZips = ZipCodesByCityState(strCity, strState);
	
	if (strSearchStreet != wxT(""))
	{
		wxSearchStreet = strSearchStreet;
		// spaces don't matter, so get rid of them
		wxSearchStreet.Replace(wxT(" "), wxT(""));

		// same for punctuation
		wxSearchStreet.Replace(wxT("."), wxT(""));
		wxSearchStreet.Replace(wxT(","), wxT(""));
	}

	if (!iSearchNumber && strSearchStreet == wxT("") && strSearchType == wxT(""))
		bByCityStateOnly = true;
	
	if (vZips.size() && (!pDataForced || bByCityStateOnly))
	{
		Point ptLoadTL, ptLoadBR;
		vector<wxUint32>::iterator iZip;
		bool bFirstEntry;
		
		// find a zip code that we have coordinates for
		bFirstEntry = true;
		for (iZip = vZips.begin(); iZip != vZips.end(); iZip++)
		{
			Point ptTmpTL, ptTmpBR;
			
			if (!GetZipCodeCoordinates(*iZip, &ptTmpTL, &ptTmpBR))
			{
				// check validity
				if (ptTmpTL.abs() > 0.001 && ptTmpBR.abs() > 0.001)
				{
					// update top left
					if (ptTmpTL.GetLatRaw() < ptLoadTL.GetLatRaw() || bFirstEntry)
						ptLoadTL.SetLatRaw(ptTmpTL.GetLatRaw());
					if (ptTmpTL.GetLongRaw() < ptLoadTL.GetLongRaw() || bFirstEntry)
						ptLoadTL.SetLongRaw(ptTmpTL.GetLongRaw());

					// update bottom right
					if (ptTmpBR.GetLatRaw() > ptLoadBR.GetLatRaw() || bFirstEntry)
						ptLoadBR.SetLatRaw(ptTmpBR.GetLatRaw());
					if (ptTmpBR.GetLongRaw() > ptLoadBR.GetLongRaw() || bFirstEntry)
						ptLoadBR.SetLongRaw(ptTmpBR.GetLongRaw());
						
					bFirstEntry = false;
				}
			}
		}

		if (ptLoadTL.abs() > 0.001 && ptLoadBR.abs() > 0.001)
		{
			if (bByCityStateOnly)
			{
				// search by city/state only
				Point ptCenter = (ptLoadTL + ptLoadBR) / 2;
				
				return FindCoordinates(ptCenter, pbVisibility, false, true, pDataForced);
			}

			pData->LoadRegion(ptLoadTL, ptLoadBR, pbVisibility);
		}
	}

	for (iRec = 0; iRec < pData->Count(); iRec++)
	{
		IMapControlDataEntry * psRec = pData->GetRecord(iRec);
		
		bool bMatch = true;
		bool bOnLeft = false;
		bool bOnRight = false;
		double fPercentageAlongLeg = 0;
		int iAddressZipEntry = -1;

		if (iSearchNumber || (strCity != wxT("") && strState != wxT("")))
		{
			int iRange;
			for (iRange = 0; iRange < psRec->nAddressZipRanges && !bOnLeft && !bOnRight; iRange++)
			{
				int iLow = psRec->psAddressZipRanges[iRange].iFromAddr;
				int iHigh = psRec->psAddressZipRanges[iRange].iToAddr;
				bool bLeft = psRec->psAddressZipRanges[iRange].bOnLeft;
				int iSwapped = 0;

				if (iLow > iHigh)
				{
					int iTmp;
					iTmp = iLow;
					iLow = iHigh;
					iHigh = iTmp;
					iSwapped = 1;
				}

				if ((iLow <= iSearchNumber && iSearchNumber <= iHigh) || !iSearchNumber)
				{
					if (bLeft)
						bOnLeft = true;
					else
						bOnRight = true;

					fPercentageAlongLeg = (double) (iSearchNumber - iLow) / (iHigh - iLow + 0.001);

					if (iSwapped)
						fPercentageAlongLeg = 1 - fPercentageAlongLeg;
					
					if (!iSearchNumber)
						fPercentageAlongLeg = 0;
				}


				if ((bOnLeft || bOnRight) && (strCity != wxT("") || strState != wxT("")))
				{
					bool bZipOkay = false;
					unsigned int i;

					for(i = 0; i < vZips.size(); i++)
					{
						if (psRec->psAddressZipRanges[iRange].iZip == vZips[i])
							bZipOkay = true;
					}

					if (!bZipOkay)
					{
						bOnLeft = false;
						bOnRight = false;
					}
				}
				if (bOnLeft || bOnRight)
					iAddressZipEntry = iRange;
			}
		}
		else // we're not searching by street number, city, or zip
		{
			bOnLeft = true;
		}

		if (!bOnLeft && !bOnRight)
			bMatch = false;

		if (bMatch)
		{
			bool bMatchesName = false;
			int i;

			for (i = 0; i < psRec->nFeatureNames; i++)
			{
				wxString wxThisStreet = wxString(psRec->pszFeatureNames[i], wxConvUTF8);
				// remove spaces
				wxThisStreet.Replace(wxT(" "), wxT(""));
				
				// and punctuation
				wxThisStreet.Replace(wxT("."), wxT(""));
				wxThisStreet.Replace(wxT(","), wxT(""));
				
				if (
						( strSearchStreet == wxT("") || !wxThisStreet.CmpNoCase(wxSearchStreet))
						&& 
						( strSearchType == wxT("") || !strSearchTypeAbbreviated.CmpNoCase(wxString(psRec->pszFeatureTypes[i], wxConvUTF8)))
					)
					bMatchesName = true;
			}

			if (!bMatchesName)
				bMatch = false;
		}


		if (bMatch)
		{
			double fTotalDistance = 0;
			{
				int iSeg;
				for (iSeg = 0; iSeg < psRec->nCoordinates - 1; iSeg ++)
				{
					fTotalDistance += Distance(psRec->psCoordinates[iSeg], psRec->psCoordinates[iSeg + 1]);
				}
			}
			
			double fDesiredDistance = fTotalDistance * fPercentageAlongLeg;
			{
				double fSearchDistance = 0;
				int iSeg;
				for (iSeg = 0; iSeg < psRec->nCoordinates - 1; iSeg ++)
				{
					double fChange = Distance(psRec->psCoordinates[iSeg], psRec->psCoordinates[iSeg + 1]);

					if (fSearchDistance <= fDesiredDistance && fDesiredDistance < fSearchDistance + fChange)
					{
						int iZip = 0;
						
						if (iAddressZipEntry >= 0)
							iZip = psRec->psAddressZipRanges[iAddressZipEntry].iZip;

						sRtn.m_ptCoordinates.SetLong((psRec->psCoordinates[iSeg + 1].GetLong() - psRec->psCoordinates[iSeg].GetLong()) * (fDesiredDistance - fSearchDistance) / fChange + psRec->psCoordinates[iSeg].GetLong());
						sRtn.m_ptCoordinates.SetLat((psRec->psCoordinates[iSeg + 1].GetLat() - psRec->psCoordinates[iSeg].GetLat()) * (fDesiredDistance - fSearchDistance) / fChange + psRec->psCoordinates[iSeg].GetLat());

						sRtn.m_iStreetNumber = iSearchNumber;
						sRtn.m_strStreetName = wxString(psRec->pszFeatureNames[0], wxConvUTF8);
						sRtn.m_strStreetType = wxString(psRec->pszFeatureTypes[0], wxConvUTF8);
						
						sRtn.m_strCityName = CityByZip(iZip);
						sRtn.m_strStateName = StateByZip(iZip);
						sRtn.m_iZipCode = iZip;

						sRtn.m_idRecord = psRec->idRecord;

						sRtn.m_iCountyCode = psRec->CountyCode();
						
						sRtn.m_eRecordType = psRec->eType;

						sRtn.m_bValid = true;

						return sRtn;
					}

					fSearchDistance += fChange;
				}
			}
		}
	}
	Address cBlank;
	return cBlank;
}

struct SPtCacheEntry
{
	Point m_pt;
	bool m_bVisibility[LAST_RECORD_TYPE + 1];
	Address m_cAddress;
};

static list<SPtCacheEntry> g_ptCache;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the closest thing to a set of coordinates
///
//////////////////////////////////////////////////////////////////////////////
Address FindCoordinates(const Point & ptSearch, const bool * pbVisibility, bool bRegionMode, bool bUseCoordFromRec, IMapControlData * pDataForced)
{
	double fBestDistance = 1e99;
	int iBestRecord = -1;
	int iBestSegment = -1;
	Point ptBestCoordinates;

	list<SPtCacheEntry>::iterator iList;

	ptBestCoordinates = ptSearch;

	// check the cache for these coordinates first
	for (iList = g_ptCache.begin(); iList != g_ptCache.end(); iList++)
	{
		if (ptSearch == (*iList).m_pt && VisibilityEqual(pbVisibility, (*iList).m_bVisibility))
		{
			// found it
			SPtCacheEntry cEntry;
			
			cEntry = *iList;

			// pull entry to front
			g_ptCache.erase(iList);
			g_ptCache.push_front(cEntry);

			return cEntry.m_cAddress;
		}
	}

	// not in cache .. load up the appropriate section of the map
	MapControlData_Tiles cRec;
	IMapControlData * pData;
	
	if (pDataForced)
	{
		pData = pDataForced;
	}
	else
	{
		cRec.LoadRegion(ptSearch, ptSearch, pbVisibility);
		pData = &cRec;
	}

	if (bRegionMode)
	{
		// region mode - search within polygons
		int iRec;
		for (iRec = 0; iRec < pData->Count(); iRec++)
		{
			IMapControlDataEntry * psRec = pData->GetRecord(iRec);

			if (psRec->IsRegion())
			{
				if (pbVisibility[psRec->eType])
				{
					// is the search point inside?
					if (psRec->ContainsPoint(ptSearch))
						iBestRecord = iRec;
				}
			}
		}

		fBestDistance = 0;
		iBestSegment = 0;
		ptBestCoordinates = ptSearch;
	}
	else
	{
		// normal open polygon mode
		int iRec;
		for (iRec = 0; iRec < pData->Count(); iRec++)
		{
			IMapControlDataEntry * psRec = pData->GetRecord(iRec);
			
			if (pbVisibility[psRec->eType])
			{
				double fDistance;

				fDistance = psRec->Distance(ptSearch);

				if (fDistance < fBestDistance)
				{
					fBestDistance = fDistance;
					iBestRecord = iRec;
				}
			}
		}

		if (iBestRecord >= 0)
		{
			IMapControlDataEntry * psRec = pData->GetRecord(iBestRecord);
			psRec->Distance(ptSearch, &iBestSegment);
			ptBestCoordinates = psRec->psCoordinates[iBestSegment];
		}
	}

	// if no hits, return empty address
	if (iBestRecord < 0)
	{
		Address cBlank;
		return cBlank;
	}

	Address sRtn;
	IMapControlDataEntry * psRec = pData->GetRecord(iBestRecord);

	if (psRec->nAddressZipRanges)
		sRtn.m_iZipCode = psRec->psAddressZipRanges[0].iZip;
	else
		sRtn.m_iZipCode = 0;

	sRtn.m_idRecord = psRec->idRecord;
	sRtn.m_iCountyCode = psRec->CountyCode();

	sRtn.m_strStreetName = wxString(psRec->pszFeatureNames[0], wxConvUTF8);
	sRtn.m_strStreetType = wxString(psRec->pszFeatureTypes[0], wxConvUTF8);
	
	sRtn.m_strCityName = CityByZip(sRtn.m_iZipCode);
	sRtn.m_strStateName = StateByZip(sRtn.m_iZipCode);
	
	sRtn.m_eRecordType = psRec->eType;

	// calculating the street number will be the hardest

	double fTotalDistance = 0;
	{
		int iSeg;
		for (iSeg = 0; iSeg < psRec->nCoordinates - 1; iSeg++)
		{
			fTotalDistance += Distance(psRec->psCoordinates[iSeg], psRec->psCoordinates[iSeg + 1]);
		}
	}
	
	// calculate street number
	if (psRec->nAddressZipRanges && fTotalDistance)
	{
		double fDistanceUntilTargetSegment = 0;
		int iSeg;
		
		wxASSERT(iBestSegment < psRec->nCoordinates);
		
		for (iSeg = 0; iSeg < iBestSegment; iSeg++)
		{
			fDistanceUntilTargetSegment += Distance(psRec->psCoordinates[iSeg], psRec->psCoordinates[iSeg + 1]);
		}

		double h = Distance(psRec->psCoordinates[iBestSegment], ptSearch);
		
		fDistanceUntilTargetSegment += sqrt(h * h - fBestDistance * fBestDistance);

		sRtn.m_iStreetNumber = (int) (fDistanceUntilTargetSegment / fTotalDistance * (psRec->psAddressZipRanges[0].iToAddr - psRec->psAddressZipRanges[0].iFromAddr) + psRec->psAddressZipRanges[0].iFromAddr);
		
		if (sRtn.m_iStreetNumber < 1)
			sRtn.m_iStreetNumber = 0;
	}
	else
	{
		sRtn.m_iStreetNumber = 0;
	}

	if (bUseCoordFromRec)
	{
		wxASSERT(iBestSegment >= 0);
		wxASSERT(iBestSegment < psRec->nCoordinates);
		
		sRtn.m_ptCoordinates = psRec->psCoordinates[iBestSegment];
	}
	else
	{
		sRtn.m_ptCoordinates = ptSearch;
	}


	sRtn.m_bValid = true;

	// cache entry
	SPtCacheEntry cEntry;
	cEntry.m_pt = ptBestCoordinates;
	CopyVisibility(cEntry.m_bVisibility, pbVisibility);
	cEntry.m_cAddress = sRtn;
	g_ptCache.push_front(cEntry);

	// prune cache
	while (g_ptCache.size() > 50)
		g_ptCache.pop_back();

	return sRtn;
}

AutoComplete_Street::AutoComplete_Street()
{
	SetVisibility(OnlyRoadsVisibility());
}

void AutoComplete_Street::SetState(const wxString & strState)
{
	m_strState = strState;
}

void AutoComplete_Street::SetCity(const wxString & strCity)
{
	m_strCity = strCity;
}

void AutoComplete_Street::SetVisibility(const bool * pbVisibility)
{
	memcpy(m_bVisibility, pbVisibility, sizeof(m_bVisibility));
}

wxString AutoComplete_Street::ClosestMatch(const wxString & strInput)
{
	wxString strRtn;
	wxString strErrorMsg;
	vector<wxUint32> vZips;
	Address cAddr;
	static MapControlData_Tiles cRec;
	
	LoadCityStateIntoRecords(m_strCity, m_strState, OnlyRoadsVisibility(), &cRec, vZips);
	cAddr = FindAddress(strInput, m_strCity, m_strState, m_bVisibility, &cRec, &strErrorMsg, &strRtn, NULL, NULL);
	
	if (cAddr.m_bValid)
		return cAddr.FormatAddress(false, true, true);

	return strRtn;
}

int AutoComplete_Street::GetUpdateDelay()
{
	return 1;
}


