/*
 *  LibRoadnav
 *  MapMarker.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains the MapMarker class - a class to store information about
/// (GPS, current address, etc.) markers on the map.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/wx.h>
#include <wx/image.h>

#include <vector>

#include "MapMarker.h"
#include "MapControl.h"
#include "MapLookup.h"
#include "MapRepresentations.h"

using namespace std;

bool CloseMarkerCallback(MapControl * pMap, void * pvMarker, bool bRightClick, wxPoint ptClicked)
{
	if (bRightClick)
	{
		MapMarker * pMarker = (MapMarker *) pvMarker;

		pMarker->m_bVisible = false;

		pMap->SetMarker(*pMarker, false);
		
		return true;
	}
	else
	{
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Constructor
///
//////////////////////////////////////////////////////////////////////////////
MapMarker::MapMarker()
{
	m_fHeading = -1;
	m_bCalloutBox = false;
	m_bSmallLabel = true;
	m_bArrow = true;
	m_bVisible = true;
	m_bScaleIcon = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Constructor - initialize marker from provided arguments
///
//////////////////////////////////////////////////////////////////////////////
MapMarker::MapMarker(const wxString & strName, Address * psAddr, double fHeading, bool bCalloutBox, bool bSmallLabel, bool bArrow, const wxString & strIcon, bool bScaleIcon, bool bVisible)
{
	Init(strName, psAddr, fHeading, bCalloutBox, bSmallLabel, bArrow, strIcon, bScaleIcon, bVisible);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Initialize marker from provided arguments
///
/// Fills in m_strText according to provided address
/// 
//////////////////////////////////////////////////////////////////////////////
void MapMarker::Init(const wxString & strName, Address * psAddr, double fHeading, bool bCalloutBox, bool bSmallLabel, bool bArrow, const wxString & strIcon, bool bScaleIcon, bool bVisible)
{
	if (!psAddr)
		return;
	
	m_strName = strName;
	m_strText = wxT("[") + strName + wxT("]\n") + psAddr->FormatAddress(false);
	m_cAddress = *psAddr;
	m_fHeading = fHeading;

	m_bCalloutBox = bCalloutBox;
	m_bSmallLabel = bSmallLabel;
	m_strIcon = strIcon;
	m_bArrow = bArrow;
	m_bVisible = bVisible;
	
	m_bScaleIcon = bScaleIcon;
}

//////////////////////////////////////////////////////////////////////////////
//
// Draw the marker
//
//////////////////////////////////////////////////////////////////////////////
int MapMarker::Draw(MapControl * pCtl, wxDC * dc, double fDrawingScaleFactor, bool bScreen)
{
	if (!m_bVisible)
		return 0;
		
	if (m_strIcon != wxT(""))
		DrawIcon(pCtl, dc, fDrawingScaleFactor, bScreen);

	if (m_bArrow)
		DrawArrow(pCtl, dc, fDrawingScaleFactor, bScreen);

	if (m_bCalloutBox)
		DrawCalloutBox(pCtl, dc, fDrawingScaleFactor, bScreen);

	if (m_bSmallLabel)
		DrawSmallLabel(pCtl, dc, fDrawingScaleFactor, bScreen);

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Draw the icon for the marker
//
//////////////////////////////////////////////////////////////////////////////
int MapMarker::DrawIcon(MapControl * pCtl, wxDC * dc, double fDrawingScaleFactor, bool bScreen)
{
	wxPoint ptTarget;
	wxBitmap bmpIcon;
	double fScale;
	int iReduction;
	double fRotate;

	if (!wxFileExists(m_strIcon))
		return 1;

	if (!bmpIcon.LoadFile(m_strIcon, wxBITMAP_TYPE_ANY))
		return 1;

	// figure out where the icon goes
	ptTarget = pCtl->MapLongLatToScreen(m_cAddress.m_ptCoordinates);

	// figure out the number of levels to zoom out	
	iReduction = pCtl->GetDetailLevelCount() - pCtl->GetDetailLevel() - 2;
	
	// put some limits on the zooming
	if (iReduction < 0)
		iReduction = 0;
		
	if (iReduction > 3)
		iReduction = 3;
	
	// don't zoom if zooming is disabled
	if (!m_bScaleIcon)
		iReduction = 0;

	// calculate scale
	fScale = fDrawingScaleFactor * pow(0.5, iReduction);
	
	// calculate rotation angle in radians
	fRotate = -1 * m_fHeading / 180 * PI;
	if (m_fHeading < 0)
		fRotate = 0;

	if (fScale != 1 || fRotate != 0)
	{
		wxImage imgScaled;
		
		// convert to wxImage
		imgScaled = bmpIcon.ConvertToImage();

		// scale, if necessary
		if (fScale != 1)
		{		
			imgScaled.Rescale((int) (bmpIcon.GetWidth() * fScale), (int) (bmpIcon.GetHeight() * fScale));
		}
		
		// rotate, if necessary
		if (fRotate != 0)
		{
			wxPoint ptRotate;

			ptRotate.x = imgScaled.GetWidth() / 2;
			ptRotate.y = imgScaled.GetHeight() / 2;
			
			imgScaled = imgScaled.Rotate(fRotate, ptRotate);
		}

		// convert back to wxBitmap
		bmpIcon = wxBitmap(imgScaled);
	}

	dc->DrawBitmap(	bmpIcon, 
					ptTarget.x - bmpIcon.GetWidth() / 2, 
					ptTarget.y - bmpIcon.GetHeight() / 2,
					true);
	
	if (bScreen)
		pCtl->AddClickCallback(	ptTarget - wxPoint(bmpIcon.GetWidth() / 2, bmpIcon.GetHeight() / 2), 
								ptTarget + wxPoint(bmpIcon.GetWidth() / 2, bmpIcon.GetHeight() / 2), 
								CloseMarkerCallback, 
								this);
	
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Draw the arrow for the marker
//
//////////////////////////////////////////////////////////////////////////////
int MapMarker::DrawArrow(MapControl * pCtl, wxDC * dc, double fDrawingScaleFactor, bool bScreen)
{
	if (m_fHeading >= 0)
	{
		wxColour clrArrow;
		int iArrowWidth;
		int iArrowLength;
		int iArrowHeadLength;
		wxPoint ptTarget;

		// settings
		clrArrow.Set(255, 0, 0);
		iArrowWidth = 3;
		iArrowLength = 12;
		iArrowHeadLength = 4;

		// DPI adjustment
		iArrowWidth = (int) (iArrowWidth * fDrawingScaleFactor);
		iArrowLength = (int) (iArrowLength * fDrawingScaleFactor);
		iArrowHeadLength = (int) (iArrowHeadLength * fDrawingScaleFactor);

		wxPen penArrow(clrArrow, iArrowWidth, wxSOLID);
		
		ptTarget = pCtl->MapLongLatToScreen(m_cAddress.m_ptCoordinates);

		dc->SetPen(penArrow);
		
		wxPoint ptTail;
		wxPoint ptHead;
		wxPoint ptHeadLeft;
		wxPoint ptHeadRight;

		double fAngle;

		fAngle = m_fHeading - 90 - pCtl->GetAngle();
		
		ptHead.x = (int) (ptTarget.x + iArrowLength * cos((fAngle) * PI / 180) / 2);
		ptHead.y = (int) (ptTarget.y + iArrowLength * sin((fAngle) * PI / 180) / 2);
		ptTail.x = (int) (ptTarget.x - iArrowLength * cos((fAngle)* PI / 180) / 2);
		ptTail.y = (int) (ptTarget.y - iArrowLength * sin((fAngle) * PI / 180) / 2);
		
		ptHeadLeft.x = (int) (ptHead.x + iArrowHeadLength * cos((fAngle + 135) * PI / 180));
		ptHeadLeft.y = (int) (ptHead.y + iArrowHeadLength * sin((fAngle + 135) * PI / 180));
		ptHeadRight.x = (int) (ptHead.x + iArrowHeadLength * cos((fAngle - 135) * PI / 180));
		ptHeadRight.y = (int) (ptHead.y + iArrowHeadLength * sin((fAngle - 135) * PI / 180));
		
		dc->DrawLine(ptHead.x, ptHead.y, ptTail.x, ptTail.y);
		dc->DrawLine(ptHead.x, ptHead.y, ptHeadLeft.x, ptHeadLeft.y);
		dc->DrawLine(ptHead.x, ptHead.y, ptHeadRight.x, ptHeadRight.y);
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Draw the callout box for the marker
//
//////////////////////////////////////////////////////////////////////////////
int MapMarker::DrawCalloutBox(MapControl * pCtl, wxDC * dc, double fDrawingScaleFactor, bool bScreen)
{
	wxColour clrLine;
	wxColour clrCallout;
	int iCalloutOffsetX;
	int iCalloutOffsetY;
	wxCoord iCalloutWidth;
	wxCoord iCalloutHeight;
	int iLineWidth;
	wxPoint ptTarget;
	wxString strText;
	vector<wxString> vLines;
	int iCalloutFontSize;
	int iTextMargin;
	int iLineHeight;
	int iLineSpacing;
	unsigned int iLine;
	wxColour clrText;
	wxBrush brOldBackground;

	// settings
	iLineSpacing = 1;
	strText = m_strText + wxT("\n(") + m_cAddress.m_ptCoordinates.FormatPoint() + wxT(")");
	iTextMargin = 5;
	iCalloutFontSize = 10;
	iCalloutOffsetX = 30;
	iCalloutOffsetY = 70;
	iLineWidth = 1;
	clrLine = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerForeground);
	clrCallout = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerBackground);
	clrText = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerForeground);
	
	// adjust for DPI
	iLineSpacing = (int) (iLineSpacing * fDrawingScaleFactor);
	iTextMargin = (int) (iTextMargin * fDrawingScaleFactor);
	iCalloutFontSize = (int) (iCalloutFontSize * fDrawingScaleFactor);
	iCalloutOffsetX = (int) (iCalloutOffsetX * fDrawingScaleFactor);
	iCalloutOffsetY = (int) (iCalloutOffsetY * fDrawingScaleFactor);
	iLineWidth = (int) (iLineWidth * fDrawingScaleFactor);
	
	// split the callout text into multiple lines since DrawText doesn't handle
	// CRs correctly
	{
		wxString strLine;
		
		strLine = strText.BeforeFirst('\n');
		strText = strText.AfterFirst('\n');
		
		while (strLine != wxT(""))
		{
			vLines.push_back(strLine);

			strLine = strText.BeforeFirst('\n');
			strText = strText.AfterFirst('\n');
		}
	}

	// figure out how big the callout box should be
	wxFont fnt(iCalloutFontSize, wxSWISS, wxNORMAL, wxNORMAL);
	dc->SetFont(fnt);
	
	{
		int x, y;
		
		iCalloutWidth = 0;
		iLineHeight = 0;
		
		for (iLine = 0; iLine < vLines.size(); iLine++)
		{
			dc->GetTextExtent(vLines[iLine], &x, &y);
			
			if (x > iCalloutWidth)
				iCalloutWidth = x;
			if (y > iLineHeight)
				iLineHeight = y;
		}
		
		iCalloutHeight = (iLineHeight + iLineSpacing) * vLines.size();
	}
	
	iCalloutWidth += 2 * iTextMargin;
	iCalloutHeight += 2 * iTextMargin;

	ptTarget = pCtl->MapLongLatToScreen(m_cAddress.m_ptCoordinates);
	
	wxPen penLine(clrLine, iLineWidth, wxSOLID);
	wxBrush brCallout(clrCallout, wxSOLID);

	dc->SetPen(penLine);
	dc->SetBrush(brCallout);
	
	brOldBackground = dc->GetBackground();
	dc->SetBackground(brCallout);

	dc->SetTextForeground(clrText);
	
	wxPoint ptCorners[6];
	
	// top left corner of box
	ptCorners[0].x = ptTarget.x + iCalloutOffsetX;
	ptCorners[0].y = ptTarget.y + iCalloutOffsetY;
	
	// bottom left corner of box
	ptCorners[1].x = ptTarget.x + iCalloutOffsetX;
	ptCorners[1].y = ptTarget.y + iCalloutOffsetY + iCalloutHeight;
	
	// bottom right corner of box
	ptCorners[2].x = ptTarget.x + iCalloutOffsetX + iCalloutWidth;
	ptCorners[2].y = ptTarget.y + iCalloutOffsetY + iCalloutHeight;
	
	// top right corner of box
	ptCorners[3].x = ptTarget.x + iCalloutOffsetX + iCalloutWidth;
	ptCorners[3].y = ptTarget.y + iCalloutOffsetY;

	// target
	ptCorners[4].x = ptTarget.x;
	ptCorners[4].y = ptTarget.y;

	// mid top of box
	ptCorners[5].x = ptTarget.x + iCalloutOffsetX + (int) (iCalloutWidth * 0.85);
	ptCorners[5].y = ptTarget.y + iCalloutOffsetY;

	dc->DrawPolygon(6, ptCorners);
	
	for (iLine = 0; iLine < vLines.size(); iLine++)
	{
		dc->DrawText(	vLines[iLine], 
						ptTarget.x + iCalloutOffsetX + iTextMargin, 
						ptTarget.y + iCalloutOffsetY + iTextMargin + (iLineHeight + iLineSpacing) * iLine); 
	}
	
	dc->SetPen(wxNullPen);
	dc->SetBrush(wxNullBrush);
	dc->SetFont(wxNullFont);
	dc->SetBackground(brOldBackground);

	if (bScreen)
		pCtl->AddClickCallback(ptCorners[0], ptCorners[2], CloseMarkerCallback, this);

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Draw the callout box for the marker
//
//////////////////////////////////////////////////////////////////////////////
int MapMarker::DrawSmallLabel(MapControl * pCtl, wxDC * dc, double fDrawingScaleFactor, bool bScreen)
{
	wxColour clrLine;
	wxColour clrCallout;
	wxColour clrText;
	wxCoord iCalloutWidth;
	wxCoord iCalloutHeight;
	wxPoint ptTarget;
	wxString strText;
	int iTextMargin;
	int iCalloutFontSize;
	int iLineWidth;
	wxBrush brOldBackground;

	// settings
	strText = m_strName;
	iTextMargin = 5;
	iLineWidth = 1;
	iCalloutFontSize = 10;
	clrLine = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerForeground);
	clrCallout = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerBackground);
	clrText = pCtl->GetDetailSettings()->GetMiscColor(MiscColorMarkerForeground);
	
	// adjust for DPI
	iTextMargin = (int) (iTextMargin * fDrawingScaleFactor);
	iCalloutFontSize = (int) (iCalloutFontSize * fDrawingScaleFactor);
	iLineWidth = (int) (iLineWidth * fDrawingScaleFactor);

	// figure out how big the callout box should be
	wxFont fnt(iCalloutFontSize, wxSWISS, wxNORMAL, wxNORMAL);
	dc->SetFont(fnt);

	dc->GetTextExtent(strText, &iCalloutWidth, &iCalloutHeight);
	
	iCalloutWidth += 2 * iTextMargin;
	iCalloutHeight += 2 * iTextMargin;

	ptTarget = pCtl->MapLongLatToScreen(m_cAddress.m_ptCoordinates);
	
	wxPen penLine(clrLine, iLineWidth, wxSOLID);
	wxBrush brCallout(clrCallout, wxSOLID);

	dc->SetPen(penLine);
	dc->SetBrush(brCallout);
	
	brOldBackground = dc->GetBackground();
	dc->SetBackground(brCallout);

	dc->SetTextForeground(clrText);

	dc->DrawRectangle(ptTarget.x, ptTarget.y, iCalloutWidth, iCalloutHeight);

	dc->DrawText(strText, ptTarget.x + iTextMargin, ptTarget.y + iTextMargin);

	dc->SetPen(wxNullPen);
	dc->SetBrush(wxNullBrush);
	dc->SetFont(wxNullFont);
	dc->SetBackground(brOldBackground);

	if (bScreen)
		pCtl->AddClickCallback(ptTarget, ptTarget + wxPoint(iCalloutWidth, iCalloutHeight), CloseMarkerCallback, this);

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Convert marker to a string that can be saved in the registry.
///
//////////////////////////////////////////////////////////////////////////////
wxString MapMarker::ToString()
{
	return	m_strName + wxT('\t') + 
			m_strText + wxT('\t') + 
			m_cAddress.ToString() + wxT('\001') + 
			wxString::Format(wxT("%f"), m_fHeading) + wxT("\t") + 
			wxString::Format(wxT("%d"), m_bCalloutBox) + wxT("\t") + 
			wxString::Format(wxT("%d"), m_bSmallLabel) + wxT("\t") + 
			wxString::Format(wxT("%d"), m_bArrow) + wxT("\t") + 
			m_strIcon + wxT("\t") +
			wxString::Format(wxT("%d"), m_bVisible) + wxT("\t") +
			wxString::Format(wxT("%d"), m_bScaleIcon);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Initialize this class from a string generated by MapMarker::ToString()
///
//////////////////////////////////////////////////////////////////////////////
int MapMarker::FromString(wxString strDesc)
{
	long lTmp;

	m_strName = strDesc.BeforeFirst('\t');
	strDesc = strDesc.AfterFirst('\t');

	m_strText = strDesc.BeforeFirst('\t');
	strDesc = strDesc.AfterFirst('\t');
	
	m_cAddress.FromString(strDesc.BeforeFirst('\001'));
	strDesc = strDesc.AfterFirst('\001');
	
	strDesc.BeforeFirst('\t').ToDouble(&m_fHeading);
	strDesc = strDesc.AfterFirst('\t');
	
	strDesc.BeforeFirst('\t').ToLong(&lTmp);
	m_bCalloutBox = (lTmp != 0);
	strDesc = strDesc.AfterFirst('\t');
	
	strDesc.BeforeFirst('\t').ToLong(&lTmp);
	m_bSmallLabel = (lTmp != 0);
	strDesc = strDesc.AfterFirst('\t');
	
	strDesc.BeforeFirst('\t').ToLong(&lTmp);
	m_bArrow = (lTmp != 0);
	strDesc = strDesc.AfterFirst('\t');

	m_strIcon = strDesc.BeforeFirst('\t');
	strDesc = strDesc.AfterFirst('\t');
	
	lTmp = 1;
	strDesc.BeforeFirst('\t').ToLong(&lTmp);
	m_bVisible = (lTmp != 0);
	strDesc = strDesc.AfterFirst('\t');
	
	lTmp = 0;
	strDesc.BeforeFirst('\t').ToLong(&lTmp);
	m_bScaleIcon = (lTmp != 0);
	strDesc = strDesc.AfterFirst('\t');
	
	return 0;
}
