/*
 *  LibRoadnav
 *  MapPolygon.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Implements the MapPolygon class - a class representing an open or
/// closed polygon.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>

#include <vector>

#include "MapPolygon.h"
#include "MapSupport.h"

using namespace std;

#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (x > y ? x : y)

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - initialized object with no vertices.
///
//////////////////////////////////////////////////////////////////////////////
MapPolygon::MapPolygon()
{
	m_pPoints = NULL;
	m_nPoints = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - initialize object with supplied vertices.
///
//////////////////////////////////////////////////////////////////////////////
MapPolygon::MapPolygon(Point * pptPoints, int nPoints)
{
	int i;

	if (nPoints == 0)
	{
		m_pPoints = NULL;
		m_nPoints = 0;
		return;
	}

	m_pPoints = new Point [nPoints] ();
	m_nPoints = nPoints;
	
	for (i = 0; i < nPoints; i++)
		m_pPoints[i] = pptPoints[i];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Copy constructor
///
//////////////////////////////////////////////////////////////////////////////
MapPolygon::MapPolygon(const MapPolygon & cCopy)
{
	m_pPoints = NULL;
	m_nPoints = 0;

	*this = cCopy;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - call Clear()
///
//////////////////////////////////////////////////////////////////////////////
MapPolygon::~MapPolygon()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Free memory occupied by object.
///
//////////////////////////////////////////////////////////////////////////////
void MapPolygon::Clear()
{
	if (m_pPoints)
		delete [] m_pPoints;

	m_pPoints = NULL;
	m_nPoints = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Assignment operator
///
//////////////////////////////////////////////////////////////////////////////
const MapPolygon & MapPolygon::operator =(const MapPolygon & cCopy)
{
	Clear();

	if (cCopy.m_nPoints)
	{
		int i;
		
		m_pPoints = new Point[cCopy.m_nPoints] ();
		m_nPoints = cCopy.m_nPoints;

		for (i = 0; i < m_nPoints; i++)
			m_pPoints[i] = cCopy.m_pPoints[i];
	}
	else
	{
		m_pPoints = NULL;
		m_nPoints = 0;
	}

	return *this;	
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the specified vertex of the polygon.
///
//////////////////////////////////////////////////////////////////////////////
Point MapPolygon::GetPoint(int n)
{
	if (n >= m_nPoints)
	{
		Point ptInvalid;
		
		ptInvalid.SetInvalid();
		
		return ptInvalid;
	}
		
	return m_pPoints[n];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the number of vertices in the polygon.
///
//////////////////////////////////////////////////////////////////////////////
int MapPolygon::GetSize()
{
	return m_nPoints;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the area occupied by the polygon.
///
//////////////////////////////////////////////////////////////////////////////
double MapPolygon::GetArea()
{
	int i;
	double fArea;
	
	fArea = 0;
	
	for (i = 0; i < GetSize(); i++)
	{
		Point pt1 = GetPoint(i);
		Point pt2 = GetPoint((i + 1) % GetSize());
		fArea += pt1.GetLong() * pt2.GetLat() - pt2.GetLong() * pt1.GetLat();
	}
	
	return fArea;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the first vertex.
///
//////////////////////////////////////////////////////////////////////////////
Point MapPolygon::GetBeginning()
{
	return GetPoint(0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieve the last vertex.
///
//////////////////////////////////////////////////////////////////////////////
Point MapPolygon::GetEnd()
{
	if (GetSize() == 0)
	{
		Point ptInvalid;
		
		ptInvalid.SetInvalid();
		
		return ptInvalid;
	}
	
	return GetPoint(GetSize() - 1);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the polygon contains any vertices.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::IsEmpty()
{
	return GetSize() == 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the polygon is closed.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::IsClosed()
{
	return Distance(GetBeginning(), GetEnd()) < 0.01;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Internal function used to allocate memory for more vertices.
///
//////////////////////////////////////////////////////////////////////////////
void MapPolygon::AllocatePoints(int nPointsNew)
{
	Point * pPointsOld;
	
	if (nPointsNew <= 0)
	{
		if (m_pPoints)
			delete [] m_pPoints;
		
		m_pPoints = NULL;
		m_nPoints = 0;
		
		return;
	}
	
	pPointsOld = m_pPoints;
	
	m_pPoints = new Point[nPointsNew] ();
	
	if (m_nPoints)
	{
		int i;
		int nCopy;
		
		if (nPointsNew > m_nPoints)
			nCopy = m_nPoints;
		else
			nCopy = nPointsNew;
		
		for (i = 0; i < nCopy; i++)
			m_pPoints[i] = pPointsOld[i];
	
		delete [] pPointsOld;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Attached the supplied polygon to this (open) polygon.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::Attach(MapPolygon cPolygon)
{
	bool bReverse;
	int i;
	
	bReverse = false;
	
	if (GetEnd() == cPolygon.GetBeginning())
	{
	}
	else if (GetEnd() == cPolygon.GetEnd())
	{
		bReverse = true;
	}
	else
		return true;

	AllocatePoints(m_nPoints + cPolygon.GetSize());

	for (i = 0; i < cPolygon.GetSize(); i++)
	{
		Point pt;
		
		pt = cPolygon.GetPoint(bReverse ? cPolygon.GetSize() - i - 1 : i);

		m_pPoints[m_nPoints] = pt;
		m_nPoints++;
			
		if (GetBeginning() == GetEnd())
			return false;
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if this polygon contains the supplied vertex.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::HasVertex(Point pt)
{
	int i;
	
	for (i = 0; i < GetSize(); i++)
		if (GetPoint(i) == pt)
			return true;
			
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if this polygon contains the supplied polygon.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::HasPolygon(MapPolygon cPolygon)
{
	int iThis;
	int iSearch;
	
	for (iThis = 0; iThis < GetSize() - cPolygon.GetSize() + 1; iThis++)
	{
		for (iSearch = 0; iSearch < cPolygon.GetSize() && GetPoint(iThis + iSearch) == cPolygon.GetPoint(iSearch); iSearch++);
		
		if (iSearch == cPolygon.GetSize())
			return true;

		for (iSearch = 0; iSearch < cPolygon.GetSize() && GetPoint(iThis + iSearch) == cPolygon.GetPoint(cPolygon.GetSize() - iSearch - 1); iSearch++);
		
		if (iSearch == cPolygon.GetSize())
			return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if this open polygon is adjacent to the supplied polygon.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::IsAdjacentTo(MapPolygon cPolygon)
{
	if (GetEnd() == cPolygon.GetBeginning() || GetEnd() == cPolygon.GetEnd())
	{
		if (HasPolygon(cPolygon))
			return false;

		return true;
	}
		
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the supplied point is inside this polygon.
///
//////////////////////////////////////////////////////////////////////////////
bool MapPolygon::ContainsPoint(Point ptCheck)
{
	int iCrossings = 0;
	int iCoordinate;
	double fLongIntersection;

	for (iCoordinate = 0; iCoordinate < m_nPoints; iCoordinate++) 
	{
		Point pt1, pt2;

		pt1 = m_pPoints[iCoordinate];
		pt2 = m_pPoints[(iCoordinate + 1) % m_nPoints];

		if (
			ptCheck.GetLatRaw() > MIN(pt1.GetLatRaw(), pt2.GetLatRaw()) && 
			ptCheck.GetLatRaw() <= MAX(pt1.GetLatRaw(), pt2.GetLatRaw()) &&
			ptCheck.GetLongRaw() <= MAX(pt1.GetLongRaw(), pt2.GetLongRaw()) && 
			pt1.GetLatRaw() != pt2.GetLatRaw()
			)
		{
			fLongIntersection = (ptCheck.GetLat() - pt1.GetLat()) / (pt2.GetLat() - pt1.GetLat()) * (pt2.GetLong() - pt1.GetLong()) + pt1.GetLong();

			if (pt1.GetLongRaw() == pt2.GetLongRaw() || ptCheck.GetLong() <= fLongIntersection)
				iCrossings++;
		}
	}

	if (iCrossings % 2 == 0)
		return false;

	return true;
}

MapPolygon CombinePolygons(vector<MapPolygon> vPolygons, Point ptInside, MapPolygon cChain)
{
	MapPolygon cNull;
	bool bFoundMore;

	if (cChain.IsEmpty())
		return cNull;

	if (cChain.IsClosed())
		return cChain;
	
	bFoundMore = true;
	while (bFoundMore)
	{
		int nMethods;
		
		bFoundMore = false;
		vector<MapPolygon> vOptions;
		
		vector<MapPolygon>::iterator iPoly;
		for (iPoly = vPolygons.begin(); iPoly != vPolygons.end(); iPoly++)
		{
			if (cChain.IsAdjacentTo(*iPoly))
			{
				vOptions.push_back(*iPoly);
			}
		}
		
		nMethods = vOptions.size();

		if (nMethods == 1)
		{
			bFoundMore = true;
			cChain.Attach(vOptions[0]);
			
			if (cChain.IsClosed())
				return cChain;
		}
		else if (nMethods > 1)
		{
			int iMethod;
			MapPolygon cBestSolution;
			
			for (iMethod = 0; iMethod < nMethods; iMethod++)
			{
				MapPolygon cRtn;
				MapPolygon cCombined;
				
				cCombined = cChain;
				cCombined.Attach(vOptions[iMethod]);
				
				if (cCombined.IsClosed())
					return cCombined;

				cRtn = CombinePolygons(vPolygons, ptInside, cCombined);
				
				if (!cRtn.IsEmpty())
				{
					if (cRtn.ContainsPoint(ptInside))
					{
						if (cRtn.GetArea() > cBestSolution.GetArea() || cBestSolution.IsEmpty())
						{
							bool bInvalid;
							vector<MapPolygon>::iterator iCheck;
							
							bInvalid = false;
							for (iCheck = vPolygons.begin(); iCheck != vPolygons.end(); iCheck++)
							{
								if (!cRtn.HasVertex((*iCheck).GetBeginning()) && cRtn.ContainsPoint((*iCheck).GetBeginning()))
									bInvalid = true;
							}
							
							if (!bInvalid)
								cBestSolution = cRtn;
						}
					}
				}
			}
			
			return cBestSolution;
		}
	}
	return cNull;
}

