/*
 *  LibRoadnav
 *  MapRepresentations.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Functions related to representing colors, positions, distances, times,
/// addresses, etc. in a string. It also contains functions to parse such
/// strings back into their original form.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>

#include "MapRepresentations.h"
#include "MapSupport.h"
#include "MapState.h"
#include "MapAbbreviations.h"
#include "UnitConversion.h"

//////////////////////////////////////////////////////////////////////////////
///
/// Convert a time in seconds into a string
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatTime(double fTime)
{
	int iHours;
	int iMinutes;
	int iSeconds;
	
	iSeconds = (int)(fTime * 60) % 60;
	iMinutes = ((int) fTime) % 60;
	iHours = (int) (fTime / 60);
	
	if (iHours > 0)
		return wxString::Format(wxT("%d:%02d:%02d"), iHours, iMinutes, iSeconds);
	else
		return wxString::Format(wxT("%d:%02d"), iMinutes, iSeconds);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Convert a file size from bytes to a more readable format (5 MB, 10 KB, etc.)
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatFileSize(wxUint32 uiSize)
{
	if (uiSize < 1024)
		return wxString::Format(wxT("%d bytes"), uiSize);
	if (uiSize < (10 * 1024))
		return wxString::Format(wxT("%.1f KB"), ((double) uiSize) / 1024.0);
	if (uiSize < (1024 * 1024))
		return wxString::Format(wxT("%d KB"), uiSize / 1024);
	if (uiSize < (10 * 1024 * 1024))
		return wxString::Format(wxT("%.1f MB"), ((double) uiSize) / 1024.0 / 1024.0);
	if (uiSize < (1024 * 1024 * 1024))
		return wxString::Format(wxT("%d MB"), uiSize / 1024 / 1024);

	return wxString::Format(wxT("%.1f GB"), ((double) uiSize) / 1024.0 / 1024.0 / 1024.0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Convert a distance in miles to a string with a unit tag in the current
/// display units
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatDistance(double fDistance)
{
	wxString strUnitTag;
	wxString strFormat;
	fDistance = (GetDisplayUnits() == SIUnits ? MilesToKilometers(fDistance) : fDistance);
	if ( fDistance < 0.1 )
	{
		fDistance *= (GetDisplayUnits() == SIUnits ? 1000 : 5280);
		strUnitTag = (GetDisplayUnits() == SIUnits ? wxT("meters") : wxT("feet"));
		strFormat  = (GetDisplayUnits() == SIUnits ? wxT("%.0f %s") : wxT("%.0f %s"));
	}
	else
	{
		strUnitTag = (GetDisplayUnits() == SIUnits ? wxT("km") : wxT("mi"));
		strFormat  = (GetDisplayUnits() == SIUnits ? wxT("%.1f %s") : wxT("%.2f %s"));
	}

	return wxString::Format(strFormat.c_str(), fDistance, strUnitTag.c_str());
}

//////////////////////////////////////////////////////////////////////////////
///
/// Convert a color into something suitable for storage in the registry
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatColor(const wxColour & c)
{
	return wxString::Format(wxT("#%02x%02x%02x"), c.Red(), c.Green(), c.Blue());
}

//////////////////////////////////////////////////////////////////////////////
///
/// Convert a color from the registry into a wxColour
///
//////////////////////////////////////////////////////////////////////////////
wxColour ParseColor(const wxString & strColor)
{
	long r, g, b;
	
	if (strColor.Length() != 7)
		return wxColour(0, 0, 0);
	
	strColor.Mid(1, 2).ToLong(&r, 16);
	strColor.Mid(3, 2).ToLong(&g, 16);
	strColor.Mid(5, 2).ToLong(&b, 16);
	
	return wxColour(r, g, b);
}

//////////////////////////////////////////////////////////////////////////////
///
/// Format a state/city name correctly (first letter of each word capitalized)
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatName(const wxString & strIn)
{
	wxString strOut;
	wxString strUpper;
	wxString strLower;
	unsigned int i;
	
	strUpper = strIn.Upper();
	strLower = strIn.Lower();

	strOut = strLower;
	
	if (strOut.size())
		strOut[(size_t) 0] = strUpper[(size_t) 0];

	for (i = 1; i < strOut.size(); i++)
		if (wxIsspace(strOut[i - 1]))
			strOut[i] = strUpper[i];

	strOut.Trim();

	return strOut;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Format a state abbreviation correctly (all upper case).
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatAbbreviation(const wxString & strIn)
{
	return strIn.Upper();
}


static EUnits g_eDisplayUnits = USUnits;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the current display units
///
//////////////////////////////////////////////////////////////////////////////
EUnits GetDisplayUnits()
{
	return g_eDisplayUnits;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Changes the current display units
///
//////////////////////////////////////////////////////////////////////////////
void SetDisplayUnits(EUnits eUnits)
{
	g_eDisplayUnits = eUnits;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Displays a heading textually
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatHeading(double fHeading)
{
	double fNormalizedHeading;
	wxString strRtn;
	
	fNormalizedHeading = fHeading;
	
	while (fHeading < 0)
		fHeading += 360;
	while (fHeading >= 360)
		fHeading -= 360;
		
	if (fHeading > 337.5 || fHeading <= 22.5)
		strRtn = wxT("N");
		
	if (fHeading > 22.5 && fHeading <= 67.5)
		strRtn = wxT("NE");
		
	if (fHeading > 67.5 && fHeading <= 112.5)
		strRtn = wxT("E");

	if (fHeading > 112.5 && fHeading <= 157.5)
		strRtn = wxT("SE");

	if (fHeading > 157.5 && fHeading <= 202.5)
		strRtn = wxT("S");

	if (fHeading > 202.5 && fHeading <= 247.5)
		strRtn = wxT("SW");

	if (fHeading > 247.5 && fHeading <= 292.5)
		strRtn = wxT("W");

	if (fHeading > 292.5 && fHeading <= 337.5)
		strRtn = wxT("NW");

	return strRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Formats a speed using the currently selected units
///
//////////////////////////////////////////////////////////////////////////////
wxString FormatSpeed(double fSpeed)
{
	switch (GetDisplayUnits())
	{
		case SIUnits:
			return wxString::Format(wxT("%.0fkph"), fSpeed * 1.609344);

		default:
			return wxString::Format(wxT("%.0fmph"), fSpeed);
	}
}
