/*
 *  LibRoadnav
 *  MapState.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Maps states to/from the FIPS code and abbreviation.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <vector>
#include <map>

#include <wx/wx.h>

#include "MapRepresentations.h"
#include "MapSupport.h"
#include "MapState.h"

using namespace std;

#define MAX_STATECODE			80

//////////////////////////////////////////////////////////////////////////////
///
/// State associations
///
//////////////////////////////////////////////////////////////////////////////
static map<wxString,wxUint32> mapStateNameToCode;
static wxString arStateCodeToName[MAX_STATECODE + 1];
static wxString arStateCodeToAbbreviation[MAX_STATECODE + 1];
static wxString arStateAbbreviationToName[26][26];
static map<wxString,wxString> mapStateNameToAbbreviation;
static vector<wxString> vStateAbbreviations;
static vector<wxString> vStateNames;

static wxString g_strEmpty = wxT("");

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the state information from a file.
///
/// Format should be:
///
/// Field #1: State FIPS code
/// Field #2: State two letter abbreviation
/// Field #3: State name
///
//////////////////////////////////////////////////////////////////////////////
bool LoadStateInformation(const wxString & strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (states)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		wxString strLine;
		long lCode;
		wxString strAbbr;
		wxString strName;
		
		strLine = wxString(szLine, *wxConvCurrent);
		strLine.Trim();
		
		strLine.BeforeFirst('\t').ToLong(&lCode);
		strLine = strLine.AfterFirst('\t');

		strAbbr = strLine.BeforeFirst('\t');
		strLine = strLine.AfterFirst('\t');

		strName = strLine.BeforeFirst('\t');
		strLine = strLine.AfterFirst('\t');

		wxASSERT(lCode <= MAX_STATECODE);

		mapStateNameToCode[strName] = lCode;
		arStateCodeToName[lCode] = strName;
		arStateCodeToAbbreviation[lCode] = strAbbr;
		
		wxASSERT(strAbbr.GetChar(0) >= 'A');
		wxASSERT(strAbbr.GetChar(0) <= 'Z');
		wxASSERT(strAbbr.GetChar(1) >= 'A');
		wxASSERT(strAbbr.GetChar(1) <= 'Z');
		arStateAbbreviationToName[strAbbr.GetChar(0) - 'A'][strAbbr.GetChar(1) - 'A'] = strName;
		
		mapStateNameToAbbreviation[strName] = strAbbr;
		vStateAbbreviations.push_back(strAbbr);
		vStateNames.push_back(strName);

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the array of state abbreviations
///
//////////////////////////////////////////////////////////////////////////////
vector<wxString> GetStateAbbreviations()
{
	return vStateAbbreviations;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the array of state names
///
//////////////////////////////////////////////////////////////////////////////
vector<wxString> GetStateNames()
{
	return vStateNames;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Returns the state abbreviation given the state name.
///
/// Returns "" for invalid abbreviations
///
//////////////////////////////////////////////////////////////////////////////
wxString StateAbbreviationByName(const wxString & strName)
{
	wxString strFormattedName = FormatName(strName);
	
	if (mapStateNameToAbbreviation.find(strFormattedName) != mapStateNameToAbbreviation.end())
		return mapStateNameToAbbreviation[strFormattedName];
	
	return wxT("");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the state abbreviation given the state's FIPS code.
///
/// Returns "" for invalid codes
///
//////////////////////////////////////////////////////////////////////////////
const wxString & StateAbbreviationByCode(int iCode)
{
	if (iCode > MAX_STATECODE)
		return g_strEmpty;
		
	return arStateCodeToAbbreviation[iCode];
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Returns the full name of a state given the state abbreviation.
///
/// Returns "" for invalid abbreviations
///
//////////////////////////////////////////////////////////////////////////////
const wxString & StateNameByAbbreviation(const wxString & strName)
{
	if (strName.Length() != 2)
		return g_strEmpty;
		
	char c1 = toupper(strName.GetChar(0));
	char c2 = toupper(strName.GetChar(1));
	
	if (c1 >= 'A' && c1 <= 'Z' && c2 >= 'A' && c2 <= 'Z')
		return arStateAbbreviationToName[c1 - 'A'][c2 - 'A'];
	
	return g_strEmpty;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the state code given either the state name or abbreviations.
///
//////////////////////////////////////////////////////////////////////////////
int StateCodeByName(const wxString & strStateName)
{
	wxString strName;

	strName = FormatName(strStateName);
	
	const wxString & strAbbrev = StateNameByAbbreviation(strStateName);

	if (strAbbrev != wxT(""))
		strName = strAbbrev;

	if (mapStateNameToCode.find(strName) != mapStateNameToCode.end())
		return mapStateNameToCode[strName];

	return -1;
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Returns the state name given the FIPS code.
///
//////////////////////////////////////////////////////////////////////////////
const wxString & StateNameByCode(int iStateCode)
{
	if (iStateCode > MAX_STATECODE)
		return g_strEmpty;
		
	return arStateCodeToName[iStateCode];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the closest state name to strState. May be in abbreviation
/// or written out form.
///
//////////////////////////////////////////////////////////////////////////////
wxString ClosestStateName(const wxString & strState)
{
	vector<wxString>::iterator i;
	wxString strBestMatch;
	double fBestMatch = -1;
	
	for (i = vStateAbbreviations.begin(); i != vStateAbbreviations.end(); i++)
	{
		double fThisSimilarity;
		
		fThisSimilarity = WordSimilarity(*i, strState);
		
		if (fThisSimilarity > fBestMatch)
		{
			fBestMatch = fThisSimilarity;
			strBestMatch = *i;
		}
	}
	
	for (i = vStateNames.begin(); i != vStateNames.end(); i++)
	{
		double fThisSimilarity;
		
		fThisSimilarity = WordSimilarity(*i, strState);
		
		if (fThisSimilarity > fBestMatch)
		{
			fBestMatch = fThisSimilarity;
			strBestMatch = *i;
		}
	}
	
	return strBestMatch;
}

AutoComplete_State::AutoComplete_State()
{
}

wxString AutoComplete_State::ClosestMatch(const wxString & strInput)
{
	return ClosestStateName(strInput);
}

int AutoComplete_State::GetUpdateDelay()
{
	return 1;
}

