/*
 *  LibRoadnav
 *  MapSupport.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Support functions for the mapping code.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/wx.h>
#include <wx/utils.h>
#include <wx/image.h>
#include <wx/filename.h>
#include <wx/filesys.h>
#include <wx/dir.h>

#include <vector>

#include "MapSupport.h"
#include "MapState.h"
#include "MapAbbreviations.h"
#include "DownloadThread.h"
#include "Direction.h"
#include "UnitConversion.h"
#include "DownloadFiles.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Like atoi, except for wxStrings instead of char *
///
//////////////////////////////////////////////////////////////////////////////
int wxtoi(const wxString & str)
{
	long lTmp;
	lTmp = 0;
	str.ToLong(&lTmp);
	return lTmp;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Implements "rm -rf" type functionality on strName.
///
//////////////////////////////////////////////////////////////////////////////
void RemoveDirectoryAndContents(wxString strName)
{
	wxDir * pcDir;
	wxString strEntry;
	bool bContinue;
	vector<wxString> vNames;
	vector<wxString>::iterator iNames;
	
	if (!wxDirExists(strName))
		return;

	pcDir = new wxDir(strName);

	bContinue = pcDir->GetFirst(&strEntry, wxEmptyString, wxDIR_FILES);

	while (bContinue)
	{
		wxFileName fnThis;

		fnThis.AssignDir(strName);
		fnThis.SetFullName(strEntry);
		vNames.push_back(fnThis.GetFullPath());

		bContinue = pcDir->GetNext(&strEntry);
	}
	
	for (iNames = vNames.begin(); iNames != vNames.end(); iNames++)
		wxRemoveFile(*iNames);
	vNames.clear();

	bContinue = pcDir->GetFirst(&strEntry, wxEmptyString, wxDIR_DIRS);

	while (bContinue)
	{
		wxFileName fnThis;

		fnThis.AssignDir(strName);
		fnThis.SetFullName(strEntry);
		vNames.push_back(fnThis.GetFullPath());


		bContinue = pcDir->GetNext(&strEntry);
	}
	for (iNames = vNames.begin(); iNames != vNames.end(); iNames++)
		RemoveDirectoryAndContents(*iNames);
	vNames.clear();

	delete pcDir;

	wxRmdir(strName);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the list of data that has been added to this tile-layer
/// set (checks the datalist.txt index)
///
//////////////////////////////////////////////////////////////////////////////
vector<string> GetIntegratedDataList()
{
	wxFileName fnList;
	FILE * hFile;
	char szBuffer[256];
	vector<string> vRtn;

	fnList.AssignDir(GetDataPath());
	fnList.SetFullName(wxT("datalist.txt"));

	hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "rb");

	if (!hFile)
		return vRtn;

	fgets(szBuffer, sizeof(szBuffer), hFile);
	while (!feof(hFile))
	{
		TrimWhitespace(szBuffer);
		
		vRtn.push_back(szBuffer);

		fgets(szBuffer, sizeof(szBuffer), hFile);
	}
	fclose(hFile);
	
	return vRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determines if the specified data has already been added to the tile-layer
/// set (checks the datalist.txt index)
///
//////////////////////////////////////////////////////////////////////////////
bool IsDataIntegrated(const wxString & strName)
{
	wxFileName fnList;
	FILE * hFile;
	char szBuffer[256];

	fnList.AssignDir(GetDataPath());
	fnList.SetFullName(wxT("datalist.txt"));

	hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "rb");

	if (!hFile)
		return false;

	fgets(szBuffer, sizeof(szBuffer), hFile);
	while (!feof(hFile))
	{
		TrimWhitespace(szBuffer);

		if (strName == wxString(szBuffer, *wxConvCurrent))
		{
			fclose(hFile);

			return true;
		}
		fgets(szBuffer, sizeof(szBuffer), hFile);
	}
	fclose(hFile);
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Marks the specified data has having been added to the tile-layer set
/// (appends to datalist.txt)
///
//////////////////////////////////////////////////////////////////////////////
void MarkDataIntegrated(const wxString & strName)
{
	wxFileName fnList;
	FILE * hFile;

	fnList.AssignDir(GetDataPath());
	fnList.SetFullName(wxT("datalist.txt"));

	hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "a+b");

	if (!hFile)
		return;

	fprintf(hFile, "%s\r\n", static_cast<const char *>(strName.mb_str(*wxConvCurrent)));

	fclose(hFile);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Copy one bitmap to another making a real copy, not just messing
/// with reference counters
///
//////////////////////////////////////////////////////////////////////////////
void CopyBitmap(wxBitmap & bmpSrc, wxBitmap & bmpDst)
{
	wxMemoryDC dcDst;
	
	bmpDst.Create(bmpSrc.GetWidth(), bmpSrc.GetHeight(), bmpSrc.GetDepth());
	
	dcDst.SelectObject(bmpDst);
	
	dcDst.DrawBitmap(bmpSrc, 0, 0);
}



/*wxBitmap ScaleDown(wxBitmap bmpOriginal, int iAntiAliasingLevel)
{
	// This is quite slow, but safe. Use wxImage::GetData() for speed instead.
	int iAAWidth;
	int iAAHeight;
	wxImage imgOriginal;
	wxImage imgAnti;
	
	iAAWidth = bmpOriginal.GetWidth() / iAntiAliasingLevel;
	iAAHeight =  bmpOriginal.GetHeight() / iAntiAliasingLevel;
	
	imgOriginal = bmpOriginal.ConvertToImage();
	imgAnti.Create(iAAWidth, iAAHeight);

	for (int y = 0; y < iAAHeight; y++)
	{
		for (int x = 0; x < iAAWidth; x++)
		{
			int r = 0;
			int g = 0;
			int b = 0;
			for (int x1 = 0; x1 < iAntiAliasingLevel; x1++)
			{
				for (int y1 = 0; y1 < iAntiAliasingLevel; y1++)
				{
					r += imgOriginal.GetRed(x * iAntiAliasingLevel + x1,
											y * iAntiAliasingLevel + y1);
					g += imgOriginal.GetGreen(x * iAntiAliasingLevel + x1,
											y * iAntiAliasingLevel + y1);
					b += imgOriginal.GetBlue(x * iAntiAliasingLevel + x1,
											y * iAntiAliasingLevel + y1);
				}
			}
			r /= iAntiAliasingLevel * iAntiAliasingLevel;
			g /= iAntiAliasingLevel * iAntiAliasingLevel;
			b /= iAntiAliasingLevel * iAntiAliasingLevel;
	
			imgAnti.SetRGB( x, y, r, g, b);
		}
	}
	
	return imgAnti.ConvertToBitmap();
}*/

/*
wxBitmap ScaleDown(wxBitmap & bmpOriginal, int iAntiAliasingLevel)
{
	// do it by hand
	// this version is optimized, although perhaps a bit less safe
	int iAAWidth;
	int iAAHeight;
	int iOrigWidth;
	int iOrigHeight;
	wxImage imgOriginal;
	wxImage imgAnti;
	unsigned char * pOriginal;
	unsigned char * pAnti;
	int iScalingFactor;
	
	if (iAntiAliasingLevel == 1)
		return bmpOriginal;
	
	iScalingFactor = iAntiAliasingLevel * iAntiAliasingLevel;
	
	iOrigWidth = bmpOriginal.GetWidth();
	iOrigHeight = bmpOriginal.GetHeight();

	iAAWidth = iOrigWidth / iAntiAliasingLevel;
	iAAHeight =  iOrigHeight / iAntiAliasingLevel;
	
	imgOriginal = bmpOriginal.ConvertToImage();
	imgAnti.Create(iAAWidth, iAAHeight);
	
	pOriginal = imgOriginal.GetData();
	pAnti = imgAnti.GetData();
	
	for (int y = 0; y < iAAHeight; y++)
	{
		for (int x = 0; x < iAAWidth; x++)
		{
			int r = 0;
			int g = 0;
			int b = 0;
			for (int y1 = 0; y1 < iAntiAliasingLevel; y1++)
			{
				for (int x1 = 0; x1 < iAntiAliasingLevel; x1++)
				{
					int iOffset;
					
					iOffset = GetImageDataOffset(x * iAntiAliasingLevel + x1,
												 y * iAntiAliasingLevel + y1,
												 iOrigWidth,
												 iOrigHeight);
					r += pOriginal[iOffset];
					g += pOriginal[iOffset + 1];
					b += pOriginal[iOffset + 2];
				}
			}

			int iOffset;
			iOffset = GetImageDataOffset(x, y, iAAWidth, iAAHeight);
			pAnti[iOffset] = r / iScalingFactor;
			pAnti[iOffset + 1] = g / iScalingFactor;
			pAnti[iOffset + 2] = b / iScalingFactor;
		}
	}
	
	return wxBitmap(imgAnti);
}*/


//////////////////////////////////////////////////////////////////////////////
///
/// Obtain the offset of a specific pixel
///
//////////////////////////////////////////////////////////////////////////////
inline int GetImageDataOffset(int x, int y, int w, int h)
{
	return x * 3 + y * w * 3;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Scale down a bitmap with antialiasing
///
//////////////////////////////////////////////////////////////////////////////
wxBitmap ScaleDown(wxBitmap & bmpOriginal, int iAntiAliasingLevel)
{
	return bmpOriginal.ConvertToImage().Rescale(bmpOriginal.GetWidth() / iAntiAliasingLevel, bmpOriginal.GetHeight() / iAntiAliasingLevel);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns distance in miles between two logitude and latitude coordinates 
/// on earth. 
///
//////////////////////////////////////////////////////////////////////////////
double Distance(Point a, Point b)
{
	a = a * (PI / 180);
	b = b * (PI / 180);
	double dLong = (a.GetLong() - b.GetLong());
	double dLat = (a.GetLat() - b.GetLat());

	// haversine distance formula
	
	double A = sin(dLat / 2) * sin(dLat / 2) + cos(a.GetLat()) * cos(b.GetLat()) * sin(dLong / 2) * sin(dLong / 2);
	double c = 2 * atan2(sqrt(A), sqrt(1-A));
	return EARTH_RADIUS * c;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the angle between two points
///
//////////////////////////////////////////////////////////////////////////////
Direction Bearing(const wxPoint& a,const wxPoint& b)
{
   wxSize size(b.x - a.x, b.y - a.y);
   double angle = atan2((double) size.GetWidth(), (double) -size.GetHeight() );
   if ( angle < 0 )
      angle += TWO_PI;

   angle *= 180./PI;

   return Direction(angle);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Performs a midstring extraction on a string, then trims off the whitespace
///
//////////////////////////////////////////////////////////////////////////////
char * ExtractField(char * pszText, int iStart, int iEnd)
{
	static char szField[1024];
	
	while (isspace(pszText[iStart - 1]) && iStart <= iEnd)
		iStart++;

	strncpy(szField, pszText + iStart - 1, iEnd - iStart + 1);
	szField[iEnd - iStart + 1] = 0;

	int iLength = strlen(szField);
	while (iLength && isspace(szField[iLength - 1]))
		iLength--;

	szField[iLength] = 0;

	return szField;
}



//////////////////////////////////////////////////////////////////////////////
///
/// \brief Trim the whitespace off the beginning and end of string p
///
//////////////////////////////////////////////////////////////////////////////
void TrimWhitespace(char * p)
{
	while (strlen(p) && isspace(p[strlen(p) - 1]))
		p[strlen(p) - 1] = 0;

	while (isspace(p[0]))
	{
		memmove(p, p + 1, strlen(p) * sizeof(char));
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Calculate sin(x), where x is in degrees
///
//////////////////////////////////////////////////////////////////////////////
double sin_deg(double x)
{
	return sin(DegreesToRadians(x));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Calculate cos(x), where x is in degrees
///
//////////////////////////////////////////////////////////////////////////////
double cos_deg(double x)
{
	return cos(DegreesToRadians(x));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Calculate tan(x), where x is in degrees
///
//////////////////////////////////////////////////////////////////////////////
double tan_deg(double x)
{
	return tan(DegreesToRadians(x));
}

static wxString g_strCommonDataPath;
static wxString g_strLocalDataPath = wxT("MapControlData");

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the system wide location for data files.
///
//////////////////////////////////////////////////////////////////////////////
void SetCommonDataPath(wxString strCommonDataPath)
{
	g_strCommonDataPath = strCommonDataPath;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Set the user's location for data files.
///
//////////////////////////////////////////////////////////////////////////////
void SetLocalDataPath(wxString strLocalDataPath)
{
	g_strLocalDataPath = strLocalDataPath;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Get the system wide location for data files.
///
//////////////////////////////////////////////////////////////////////////////
wxString GetCommonDataPath()
{
	return g_strCommonDataPath;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Get the user's location for data files.
///
//////////////////////////////////////////////////////////////////////////////
wxString GetLocalDataPath()
{
	return g_strLocalDataPath;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Get the path of the specified data file, downloading it if necessary.
///
/// This function tries both the common directory and the user directory.
///
//////////////////////////////////////////////////////////////////////////////
wxString GetDataPath(wxString strFilename, wxString strURL, bool bNonBlocking, bool bProgressDialog)
{
	wxString strRtn;
	wxFileName fnLocal;
	wxFileName fnCommon;
	
	fnCommon.AssignDir(g_strCommonDataPath);

	fnLocal.AssignDir(g_strLocalDataPath);

	if (!fnLocal.DirExists())
		fnLocal.Mkdir(0755);
		
	strRtn = fnLocal.GetFullPath();
	
	if (strFilename != wxT(""))
	{
		fnCommon.SetFullName(strFilename);
		fnLocal.SetFullName(strFilename);
		
		strRtn = fnCommon.GetFullPath();
		
		if (fnLocal.FileExists())
			strRtn = fnLocal.GetFullPath();
		else
		{
			if (!fnCommon.FileExists())
			{
				// file doesn't exist anywhere
				
				if (strURL != wxT(""))
				{
					if (bNonBlocking)
					{
						QueueDownload(strFilename, strURL, DL_ACTION_REFRESH);
						return wxT("");
					}

					SyncDownloadEntry sEntry;

					sEntry.m_strURL = strURL;
					sEntry.m_strOutputFilename = fnLocal.GetFullPath();
					sEntry.m_strProgressBarName = strURL;

					DownloadFile(sEntry, bProgressDialog);

					if (!sEntry.m_bDownloadedOkay)
						return wxT("");

					strRtn = fnLocal.GetFullPath();
				}
			}
		}
			
	}
	if (strRtn.Right(1) == '/' || strRtn.Right(1) == '\\')
		strRtn = strRtn.Left(strRtn.Length() - 1);
		
	return strRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Save the location of window pWnd to registry entry strName
///
//////////////////////////////////////////////////////////////////////////////
void SaveWindowPosition(wxConfigBase * pConfig, wxWindow * pWnd, wxString strName)
{
	wxRect rectWindow;

	rectWindow = pWnd->GetRect();
	pConfig->Write(strName + wxT(".x"), rectWindow.x);
	pConfig->Write(strName + wxT(".y"), rectWindow.y);
	pConfig->Write(strName + wxT(".w"), rectWindow.width);
	pConfig->Write(strName + wxT(".h"), rectWindow.height);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Restore the location of window pWnd from the registry entry strName
///
//////////////////////////////////////////////////////////////////////////////
void RestoreWindowPosition(wxConfigBase * pConfig, wxWindow * pWnd, wxString strName)
{
	wxRect rectWindow;
	
	pConfig->Read(strName + wxT(".x"), &rectWindow.x, -1);
	pConfig->Read(strName + wxT(".y"), &rectWindow.y, -1);
	pConfig->Read(strName + wxT(".w"), &rectWindow.width, 0);
	pConfig->Read(strName + wxT(".h"), &rectWindow.height, 0);
	
	if (rectWindow.x >= 0 && 
		rectWindow.y >= 0 &&
		rectWindow.width > 0 && 
		rectWindow.height > 0)
		pWnd->SetSize(rectWindow);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Rotate point ptTarget about origin ptCenter by fRotationAngle.
///
//////////////////////////////////////////////////////////////////////////////
wxPoint RotatePoint(const wxPoint & ptTarget, const wxPoint & ptCenter, double fRotationAngle)
{
	double fDX;
	double fDY;
	double fAngle;
	double fMagnitude;
	wxPoint ptRtn;
	
	fDX = ptTarget.x - ptCenter.x;
	fDY = ptTarget.y - ptCenter.y;

	fAngle = atan2(fDY, fDX);
	fMagnitude = sqrt(fDX * fDX + fDY * fDY);

	ptRtn.x = (int) (cos(fAngle + fRotationAngle / 180 * PI) * fMagnitude + ptCenter.x);
	ptRtn.y = (int) (sin(fAngle + fRotationAngle / 180 * PI) * fMagnitude + ptCenter.y);

	return ptRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Rotate point ptTarget about origin ptCenter by fRotationAngle.
///
//////////////////////////////////////////////////////////////////////////////
Point RotatePoint(const Point & ptTarget, const Point & ptCenter, double fRotationAngle)
{
	double fDX;
	double fDY;
	double fAngle;
	double fMagnitude;
	Point ptRtn;
	
	fDX = ptTarget.GetLong() - ptCenter.GetLong();
	fDY = ptTarget.GetLat() - ptCenter.GetLat();

	fAngle = atan2(fDY, fDX);
	fMagnitude = sqrt(fDX * fDX + fDY * fDY);

	ptRtn.SetLong(cos(fAngle + fRotationAngle / 180 * PI) * fMagnitude + ptCenter.GetLong());
	ptRtn.SetLat(sin(fAngle + fRotationAngle / 180 * PI) * fMagnitude + ptCenter.GetLat());

	return ptRtn;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Solves a 2 equation, 2 unknown set of linear equations.
///
/// a1 * x + a2 * y = a3
/// b1 * x + b2 * y = b3
///
//////////////////////////////////////////////////////////////////////////////
bool SolveLinearEquationSet(double a1, double a2, double a3, double b1, double b2, double b3, double & x, double & y)
{
	// check for equation independence
	if (a1 == 0 && a2 == 0)
		return true;
		
	if (b1 == 0 && b2 == 0)
		return true;
		
	if (a1 == 0 && b1 == 0)
		return true;
		
	if (a2 == 0 && b2 == 0)
		return true;
	
	// shortcuts for if there's a zero somewhere
	if (a1 == 0)
	{
		y = a3 / a2;
		x = (b3 - b2 * y) / b1;
		return false;
	}
	
	if (a2 == 0)
	{
		x = a3 / a1;
		y = (b3 - b1 * x) / b2;
		return false;
	}
	
	if (b1 == 0)
	{
		y = b3 / b2;
		x = (a3 - a2 * y) / a1;
		return false;
	}
	
	if (b2 == 0)
	{
		x = b3 / b1;
		y = (a3 - a1 * x) / a2;
		return false;
	}
	
	// a1, a2, b1, b2 are all non-zero
	// cancel out x in second equation
	
	// s = a1 / b1
	// (a1 - b1 * s) * x + (a2 - b2 * s) * y = a3 - b3 * s
	// (a2 - b2 * s) * y = a3 - b3 * s
	// y = (a3 - b3 * s) / (a2 - b2 * s)
	
	double s;
	s = a1 / b1;
	y = (a3 - b3 * s) / (a2 - b2 * s);

	// a1 * x + a2 * y = a3
	// a1 * x = a3 - a2 * y 
	// x = (a3 - a2 * y) / a1
		
	x = (a3 - a2 * y) / a1;
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the specified character is a vowel or not.
///
//////////////////////////////////////////////////////////////////////////////
bool IsVowel(wxChar c)
{
	switch (toupper(c))
	{
		case wxT('A'):
		case wxT('E'):
		case wxT('I'):
		case wxT('O'):
		case wxT('U'):
			return true;
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Internal function used by WordSimilarity
///
//////////////////////////////////////////////////////////////////////////////
double WordSimilarityOneWay(const wxString & strA, const wxString & strB)
{
	double * pScores;
	double * pScoresNew;
	double fScore;
	unsigned int iA;
	unsigned int iB;
	unsigned int nA;
	unsigned int nB;

	if (strA == strB)
		return 1;
	
	if (strA == wxT("") || strB == wxT(""))
		return 0;
	
	nA = strA.Length();
	nB = strB.Length();
	
	pScores = new double [nB];
	pScoresNew = new double [nB];
	
	for (iB = 0; iB < nB; iB++)
		pScores[iB] = ((double) (nB - iB)) * (nB - iB) / nB / nB;

	// step through each letter of A
	for (iA = 0; iA < nA; iA++)
	{
		// consider each possible state of B that A could be aligned to
		for (iB = 0; iB < nB; iB++)
		{
			unsigned int iPrev;
			double fBestScore = -1;
			double fLetterSimilarity;

			// how similar are the letters referenced by iA and iB?
			if (toupper(strA.GetChar(iA)) == toupper(strB.GetChar(iB)))
				fLetterSimilarity = 1;
			else if (IsVowel(strA.GetChar(iA)) && IsVowel(strB.GetChar(iB)))
				fLetterSimilarity = 0.5;
			else
				fLetterSimilarity = 0.1;
					
			
			// iterate through the ways we could get to this state
			for (iPrev = 0; iPrev < nB; iPrev++)
			{
				double fThisScore;
				double fDistancePenalty;
				
				// ideally, we should be advancing one char at a time
				if (iA == 0)
					fDistancePenalty = 1.0 / (1 << iB);
				else
					fDistancePenalty = 1.0 / (1 << abs((int) (iB - iPrev - 1)));
				
				// compute score
				fThisScore = pScores[iPrev] * fDistancePenalty * fLetterSimilarity;
				
				if (fThisScore > fBestScore)
					fBestScore = fThisScore;
			}
			
			pScoresNew[iB] = fBestScore;
		}

		// copy in new scores
		for (iB = 0; iB < nB; iB++)
			pScores[iB] = pScoresNew[iB];
	}

	// final weighting
	for (iB = 0; iB < nB; iB++)
		pScores[iB] *= ((double) (iB + 1)) * (iB + 1) / nB / nB;
	
	// pick best score
	fScore = 0;
	for (iB = 0; iB < nB; iB++)
	{
		if (pScores[iB] > fScore)
			fScore = pScores[iB];
	}
	
	// penalty for length difference
	//fScore *= 1.0 / (1 << abs((int) (nA - nB)));
	
	delete pScores;
	delete pScoresNew;
	return fScore;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a score between 0 and 1 indicating how similar two words
/// are.
///
//////////////////////////////////////////////////////////////////////////////
double WordSimilarity(const wxString & strA, const wxString & strB)
{
	double f1;
	double f2;
	
	f1 = WordSimilarityOneWay(strA, strB);
	f2 = WordSimilarityOneWay(strB, strA);
	
	if (f1 > f2)
		return f1;
	return f2;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a const bool * array containing all record types set
/// to true (visibile).
///
//////////////////////////////////////////////////////////////////////////////
const bool * AllVisibility()
{
	static bool bRtn[LAST_RECORD_TYPE + 1];
	static bool bInit = false;
	
	if (!bInit)
	{
		int i;
		
		for (i = 0; i < LAST_RECORD_TYPE + 1; i++)
			bRtn[i] = true;
			
		bInit = true;
	}
	
	return bRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a const bool * array containing all road types set
/// to true (visibile), and everything else false.
///
//////////////////////////////////////////////////////////////////////////////
const bool * OnlyRoadsVisibility()
{
	static bool bRtn[LAST_RECORD_TYPE + 1];
	static bool bInit = false;
	
	if (!bInit)
	{
		int i;

		wxASSERT(FIRST_RECORD_TYPE == 0);
		
		for (i = 0; i < LAST_RECORD_TYPE + 1; i++)
			bRtn[i] = (i >= FIRST_RECORD_TYPE_ROAD) && (i <= LAST_RECORD_TYPE_ROAD);

		bInit = true;
	}
	
	return bRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compares two visibility arrays
///
//////////////////////////////////////////////////////////////////////////////
bool VisibilityEqual(const bool * pb1, const bool * pb2)
{
	int i;
	
	wxASSERT(FIRST_RECORD_TYPE == 0);
	
	for (i = 0; i < LAST_RECORD_TYPE; i++)
	{
		if (pb1[i] != pb2[i])
			return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Copies visibility data from pbSrc to pbDest
///
//////////////////////////////////////////////////////////////////////////////
void CopyVisibility(bool * pbDest, const bool * pbSrc)
{
	int i;

	wxASSERT(FIRST_RECORD_TYPE == 0);
	
	for (i = 0; i < LAST_RECORD_TYPE; i++)
		pbDest[i] = pbSrc[i];
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the line formed by pt1 - pt2 intersects the rectangle
/// formed by corners ptTopLeft and ptBottomRight
///
//////////////////////////////////////////////////////////////////////////////
bool LineIntersectsRect(const wxPoint & ptTopLeft, const wxPoint & ptBottomRight, const wxPoint & pt1, const wxPoint & pt2, bool bCheckTLBR)
{
	int iSeg;
	int iLastVertical = 0;
	int iLastHorizontal = 0;
	wxPoint ptTopLeftChk(ptTopLeft);
	wxPoint ptBottomRightChk(ptBottomRight);

	if (bCheckTLBR)
	{	
		if (ptTopLeftChk.x > ptBottomRightChk.x)
		{
			int iTmp;
			
			iTmp = ptTopLeftChk.x;
			ptTopLeftChk.x = ptBottomRightChk.x;
			ptBottomRightChk.x = iTmp;
		}

		if (ptTopLeftChk.y > ptBottomRightChk.y)
		{
			int iTmp;
			
			iTmp = ptTopLeftChk.y;
			ptTopLeftChk.y = ptBottomRightChk.y;
			ptBottomRightChk.y = iTmp;
		}
	}

	for (iSeg = 0; iSeg < 2; iSeg++)
	{
		int iThisVertical;
		int iThisHorizontal;
		const wxPoint * pptThis;
		
		if (iSeg == 0)
			pptThis = &pt1;
		else
			pptThis = &pt2;
		
		if (pptThis->x < ptTopLeftChk.x)
			iThisHorizontal = -1;
		else if (pptThis->x > ptBottomRightChk.x)
			iThisHorizontal = 1;
		else
			iThisHorizontal = 0;
			
		
		if (pptThis->y < ptTopLeftChk.y)
			iThisVertical = -1;
		else if (pptThis->y > ptBottomRightChk.y)
			iThisVertical = 1;
		else
			iThisVertical = 0;
			
		if (iThisHorizontal == 0 && iThisVertical == 0)
			return true;
		
		if (iSeg != 0)
		{
			if (iThisHorizontal != iLastHorizontal &&
				iThisVertical != iLastVertical)
				return true;
			
			if (iThisHorizontal * iLastHorizontal == -1 &&
				iThisVertical == 0)
				return true;

			if (iThisVertical * iLastVertical == -1 &&
				iThisHorizontal == 0)
				return true;
			
		}
			
		iLastHorizontal = iThisHorizontal;
		iLastVertical = iThisVertical;
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Pack a Canadian zip code in "H0H-0H0" format into a 32 bit
/// unsigned integer.
///
/// Note, this function does not check if the input string is actually
/// in "H0H-0H0" format.
///
//////////////////////////////////////////////////////////////////////////////
wxUint32 EncodeCanadianZip(const wxString & strZip)
{
	wxUint32 uiRtn;
	int iPos;

	uiRtn = 0;
	iPos = 0;

	uiRtn = toupper(strZip.GetChar(iPos++)) - 'A'; // 5 bits
	uiRtn |= (strZip.GetChar(iPos++) - '0') << 5; // 4 bits
	uiRtn |= (toupper(strZip.GetChar(iPos++)) - 'A') << 9; // 5 bits

	if (strZip.GetChar(iPos) == '-')
		iPos++;   

	uiRtn |= (strZip.GetChar(iPos++) - '0') << 14; // 4 bits
	uiRtn |= (toupper(strZip.GetChar(iPos++)) - 'A') << 18; // 5 bits
	uiRtn |= (strZip.GetChar(iPos++) - '0') << 23; // 4 bits

	return uiRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Unpack a Canadian zip code from a 32 bit unsigned integer into
/// "H0H-0H0" format.
///
//////////////////////////////////////////////////////////////////////////////
wxString DecodeCanadianZip(wxUint32 uiZip)
{
	wxString strRtn;

	strRtn = wxT("");
	strRtn += (uiZip & 0x1f) + 'A';
	strRtn += ((uiZip >> 5) & 0x0f) + '0';
	strRtn += ((uiZip >> 9) & 0x1f) + 'A';
	strRtn += wxT('-');
	strRtn += ((uiZip >> 14) & 0x0f) + '0';
	strRtn += ((uiZip >> 18) & 0x1f) + 'A';
	strRtn += ((uiZip >> 23) & 0x0f) + '0';

	return strRtn;
}

void SerializePointArray(FILE * hFile, bool bSave, Point * pcPoints, int nPoints, const Point & ptReference)
{
	int iPoint;
	
	if (!nPoints)
		return;

	if (bSave)
	{
		for (iPoint = 0; iPoint < nPoints; iPoint++)
		{
			int iCoord;
			for (iCoord = 0; iCoord < 2; iCoord++)
			{
				wxInt32 lThis;
				wxInt32 lPrev;
				wxInt32 lDelta; 
				bool bForceNormal;
				
				bForceNormal = false;
				if (iCoord == 0)
				{
					lThis = pcPoints[iPoint].GetLongRaw();
					
					if (iPoint)
						lPrev = pcPoints[iPoint - 1].GetLongRaw();
					else
						lPrev = ptReference.GetLongRaw();
				}
				else
				{
					lThis = pcPoints[iPoint].GetLatRaw();
					
					if (iPoint)
						lPrev = pcPoints[iPoint - 1].GetLatRaw();
					else
						lPrev = ptReference.GetLatRaw();
				}
				
				if (iPoint == 0 && !ptReference.IsValid())
					bForceNormal = true;
				
				lDelta = lThis - lPrev;
				
				// free up LSB
				lDelta = lDelta << 1;
				
				// put sign in LSB, 1 = negative, 0 = positive
				if (lDelta < 0)
					lDelta = -lDelta | 1;
					
				// two MSB will indicate length and encoding
				// 0x = 1 byte (delta)
				// 10 = 2 bytes (delta)
				// 11 = 4 bytes (normal)
				if ((lDelta & 0xffffff80) == 0 && !bForceNormal)
				{
					// 1 byte
					wxUint8 iData = lDelta;
					fwrite(&iData, sizeof(iData), 1, hFile);
				}
				else if ((lDelta & 0xffffc000) == 0 && !bForceNormal)
				{
					// 2 bytes
					wxUint8 iData[2];
					
					lDelta |= 0x8000;
					
					iData[0] = (lDelta >> 8) & 0xff;
					iData[1] = (lDelta >> 0) & 0xff;
					fwrite(iData, sizeof(wxUint8), 2, hFile);
				}
				else
				{
					// 4 bytes
					wxUint8 iData[4];
					
					// free up LSB
					lThis = lThis << 1;
					
					// put sign in LSB, 1 = negative, 0 = positive
					if (lThis < 0)
						lThis = -lThis | 1;
						
					// set two MSB to 11
					lThis |= 0xc0000000;
						
					iData[0] = (lThis >> 24) & 0xff;
					iData[1] = (lThis >> 16) & 0xff;
					iData[2] = (lThis >> 8) & 0xff;
					iData[3] = (lThis >> 0) & 0xff;
					fwrite(iData, sizeof(wxUint8), 4, hFile);
				}
			}
		}
	}
	else
	{
		for (iPoint = 0; iPoint < nPoints; iPoint++)
		{
			int iCoord;
			for (iCoord = 0; iCoord < 2; iCoord++)
			{
				wxInt32 lThis;
				wxInt32 lPrev;
				wxUint8 iData[4];

				if (iCoord == 0)
				{
					if (iPoint)
						lPrev = pcPoints[iPoint - 1].GetLongRaw();
					else
						lPrev = ptReference.GetLongRaw();
				}
				else
				{
					if (iPoint)
						lPrev = pcPoints[iPoint - 1].GetLatRaw();
					else
						lPrev = ptReference.GetLatRaw();
				}
					
				fread(iData, sizeof(wxUint8), 1, hFile);
				
				if ((iData[0] & 0x80) == 0)
				{
					// 1 byte
					wxInt32 lDelta; 

					lDelta = iData[0];
					
					// get sign from LSB
					if (lDelta & 0x01)
						lDelta = -(lDelta >> 1);
					else
						lDelta = lDelta >> 1;
						
					lThis = lPrev + lDelta;
				}
				else if ((iData[0] & 0xc0) == 0x80)
				{
					// 2 bytes
					wxInt32 lDelta; 

					fread(iData + 1, sizeof(wxUint8), 1, hFile);
					
					lDelta = ((iData[0] & 0x3f) << 8) | (iData[1] << 0);

					// get sign from LSB
					if (lDelta & 0x01)
						lDelta = -(lDelta >> 1);
					else
						lDelta = lDelta >> 1;
						
					lThis = lPrev + lDelta;
				}
				else
				{
					// 4 bytes
					fread(iData + 1, sizeof(wxUint8), 3, hFile);
					
					lThis = (iData[0] << 24) | (iData[1] << 16) | (iData[2] << 8) | (iData[3] << 0);
					lThis &= 0x3fffffff;
					
					// get sign from LSB
					if (lThis & 0x01)
						lThis = -(lThis >> 1);
					else
						lThis = lThis >> 1;
				}

				if (iCoord == 0)
					pcPoints[iPoint].SetLongRaw(lThis);
				else
					pcPoints[iPoint].SetLatRaw(lThis);
			}			
		}
	}
}
