/*
 *  LibRoadnav
 *  MapZip.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Maps zip codes to/from their cooresponding states and cities.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#include <vector>
#include <map>

#include <wx/utils.h>
#include <wx/filename.h>

#include "MapState.h"
#include "MapRepresentations.h"
#include "MapSupport.h"
#include "MapZip.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// zip associations
///
//////////////////////////////////////////////////////////////////////////////
typedef vector<wxUint32> tZipCodeList;

static map<wxString,tZipCodeList> mapCityToZip;
static wxString arZipToCity[100000];
static wxString arZipToStateName[100000];
static wxString arZipToStateAbbreviation[100000];
static wxUint32 arZipToCountyCode[100000]; 

wxString g_strEmpty = wxT("");

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load zip code information from a file.
///
/// Format:
/// Field #1: Zip code
/// Field #2: City name
/// Field #3: FIPS state code
/// Field #4: FIPS county code
///
//////////////////////////////////////////////////////////////////////////////
bool LoadZipInformation(wxString strFilename)
{
	FILE * hFile;
	char szLine[2048];

	hFile = fopen(strFilename.mb_str(*wxConvCurrent), "rb");
	if (!hFile)
	{
		::wxMessageBox(wxString::Format(wxT("Error opening %s (zips)!"), strFilename.c_str()), wxT("Error"), wxOK);
		return true;
	}

	fgets(szLine, sizeof(szLine), hFile);
	while (!feof(hFile))
	{
		wxString strLine;
		long lZip;
		wxString strCity;
		long lStateCode;
		long lCountyCode;
		wxString strState;
		
		strLine = wxString(szLine, *wxConvCurrent);
		strLine.Trim();
		
		strLine.BeforeFirst('\t').ToLong(&lZip);
		strLine = strLine.AfterFirst('\t');
		
		strCity = strLine.BeforeFirst('\t');
		strLine = strLine.AfterFirst('\t');
		
		strLine.BeforeFirst('\t').ToLong(&lStateCode);
		strLine = strLine.AfterFirst('\t');

		strLine.BeforeFirst('\t').ToLong(&lCountyCode);
		strLine = strLine.AfterFirst('\t');
		
		strState = FormatName(StateNameByCode(lStateCode));
		
		wxASSERT(lZip < 100000);

		mapCityToZip[strCity + wxT("|") + strState].push_back(lZip);
		arZipToCity[lZip] = strCity;
		arZipToStateName[lZip] = strState;
		arZipToCountyCode[lZip] = lCountyCode;
		arZipToStateAbbreviation[lZip] = StateAbbreviationByCode(lStateCode);

		fgets(szLine, sizeof(szLine), hFile);
	}

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain all zip codes for a given city and/or state.
///
/// If pszCity is NULL, the zip codes for all cities in the state will be returned.
/// If pszState is NULL, the zip codes for all states with the appropriate 
/// city name will be returned.
///
//////////////////////////////////////////////////////////////////////////////
vector<wxUint32> ZipCodesByCityState(wxString strCity, wxString strState)
{
	wxString strSearchString;
	wxString strStateName;

	strStateName = strState;
	if (StateNameByAbbreviation(strState) != wxT(""))
		strStateName = StateNameByAbbreviation(strState);

	strSearchString = FormatName(strCity) + wxT("|") + FormatName(strStateName);

	if (mapCityToZip.find(strSearchString) != mapCityToZip.end())
		return mapCityToZip[strSearchString];

	vector<wxUint32> vZips;

	return vZips;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the state that a zip code belongs to.
///
//////////////////////////////////////////////////////////////////////////////
const wxString & StateByZip(wxUint32 iZip)
{
	if (iZip < 100000)
		return arZipToStateName[iZip];

	return g_strEmpty;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the state abbreviation that a zip code belongs to.
///
//////////////////////////////////////////////////////////////////////////////
const wxString & StateAbbreviationByZip(wxUint32 iZip)
{
	if (iZip < 100000)
		return arZipToStateAbbreviation[iZip];
		
	return g_strEmpty;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Find the city name that a zip code belongs to
///
//////////////////////////////////////////////////////////////////////////////
const wxString & CityByZip(wxUint32 iZip)
{
	if (iZip < 100000)
		return arZipToCity[iZip];

	return g_strEmpty;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the state + county code for a given zip code
///
//////////////////////////////////////////////////////////////////////////////
wxInt32 CountyCodeByZip(wxUint32 iZip)
{
	int iCounty = 0;
	int iState = 0;
	
	if (iZip < 100000)
		iCounty = arZipToCountyCode[iZip];

	iState = StateCodeByName(StateByZip(iZip));
	
	return iState * 1000 + iCounty;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Record the location of a US zip code to make looking addresses up easier
///
//////////////////////////////////////////////////////////////////////////////
void MarkZipCodeCoordinates(int iZip, Point pt)
{
	FILE * hFile;
	double fZero = 0;
	wxFileName fnList;
	static int iLastZip = 0;
	static Point ptLastTopLeft;
	static Point ptLastBottomRight;

	// if the zip code is the same as before, just update
	// the variables - don't actually save them yet
	if (iLastZip == iZip)
	{
		if (pt.GetLatRaw() < ptLastTopLeft.GetLatRaw())
			ptLastTopLeft.SetLatRaw(pt.GetLatRaw());
		if (pt.GetLongRaw() < ptLastTopLeft.GetLongRaw())
			ptLastTopLeft.SetLongRaw(pt.GetLongRaw());
		
		if (pt.GetLatRaw() > ptLastBottomRight.GetLatRaw())
			ptLastBottomRight.SetLatRaw(pt.GetLatRaw());
		if (pt.GetLongRaw() > ptLastBottomRight.GetLongRaw())
			ptLastBottomRight.SetLongRaw(pt.GetLongRaw());
		
		return;
	}

	// build filename	
	fnList.AssignDir(GetDataPath());
	fnList.SetFullName(wxT("ZipCodeCoords.dat"));

	// open file
	hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "r+b");

	// check for existence
	if (!hFile)
	{
		int iZip;

		hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "w+b");
		
		for (iZip = 0; iZip < 100000; iZip++)
		{
			fwrite(&fZero, sizeof(double), 1, hFile);
			fwrite(&fZero, sizeof(double), 1, hFile);

			fwrite(&fZero, sizeof(double), 1, hFile);
			fwrite(&fZero, sizeof(double), 1, hFile);
		}
	}

	// check for legitimate last zipcode
	if (iLastZip > 0)
	{
		double fTmp;
		
		// record last zipcode info
		fseek(hFile, sizeof(double) * 4 * iLastZip, SEEK_SET);

		fTmp = ptLastTopLeft.GetLong();
		fwrite(&fTmp, sizeof(double), 1, hFile);
		fTmp = ptLastTopLeft.GetLat();
		fwrite(&fTmp, sizeof(double), 1, hFile);

		fTmp = ptLastBottomRight.GetLong();
		fwrite(&fTmp, sizeof(double), 1, hFile);
		fTmp = ptLastBottomRight.GetLat();
		fwrite(&fTmp, sizeof(double), 1, hFile);
	}

	// switch over to the new zip code		
	iLastZip = iZip;

	fseek(hFile, sizeof(double) * 4 * iZip, SEEK_SET);

	// get previous bounds
	double fLong, fLat;
	fread(&fLong, sizeof(double), 1, hFile);
	fread(&fLat, sizeof(double), 1, hFile);
	ptLastTopLeft.Set(fLong, fLat);

	fread(&fLong, sizeof(double), 1, hFile);
	fread(&fLat, sizeof(double), 1, hFile);
	ptLastBottomRight.Set(fLong, fLat);

	fclose(hFile);

	// update bounds
	if (pt.GetLatRaw() < ptLastTopLeft.GetLatRaw() || fabs(ptLastTopLeft.GetLat()) < 0.001)
		ptLastTopLeft.SetLatRaw(pt.GetLatRaw());
	if (pt.GetLongRaw() < ptLastTopLeft.GetLongRaw() || fabs(ptLastTopLeft.GetLong()) < 0.001)
		ptLastTopLeft.SetLongRaw(pt.GetLongRaw());
	
	if (pt.GetLatRaw() > ptLastBottomRight.GetLatRaw() || fabs(ptLastBottomRight.GetLat()) < 0.001)
		ptLastBottomRight.SetLatRaw(pt.GetLatRaw());
	if (pt.GetLongRaw() > ptLastBottomRight.GetLongRaw() || fabs(ptLastBottomRight.GetLong()) < 0.001)
		ptLastBottomRight.SetLongRaw(pt.GetLongRaw());
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Obtain the long/lat coordinates of a zip code
///
//////////////////////////////////////////////////////////////////////////////
bool GetZipCodeCoordinates(int iZip, Point * pptTopLeft, Point * pptBottomRight)
{
	Point ptRtn;
	FILE * hFile;
	wxFileName fnList;
	
	wxASSERT(pptTopLeft);
	wxASSERT(pptBottomRight);

	if (iZip == 0)
	{
		*pptTopLeft = Point(0, 0);
		*pptBottomRight = Point(0, 0);
		return true;
	}

	fnList.AssignDir(GetDataPath());
	fnList.SetFullName(wxT("ZipCodeCoords.dat"));

	hFile = fopen(fnList.GetFullPath().mb_str(*wxConvCurrent), "r+b");

	if (!hFile)
	{
		*pptTopLeft = Point(0, 0);
		*pptBottomRight = Point(0, 0);
		return true;
	}

	fseek(hFile, sizeof(double) * 4 * iZip, SEEK_SET);

	double fLong, fLat;
	fread(&fLong, sizeof(double), 1, hFile);
	fread(&fLat, sizeof(double), 1, hFile);
	pptTopLeft->Set(fLong, fLat);

	fread(&fLong, sizeof(double), 1, hFile);
	fread(&fLat, sizeof(double), 1, hFile);
	pptBottomRight->Set(fLong, fLat);

	fclose(hFile);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the closest city name to strCity in state strState.
///
//////////////////////////////////////////////////////////////////////////////
wxString ClosestCityName(const wxString & strCity, const wxString & strState)
{
	wxString strSearchString;
	wxString strStateName;
	map<wxString,tZipCodeList>::iterator i;
	wxString strBestMatch;
	double fBestMatch = -1;
	
	strStateName = strState;
	if (StateNameByAbbreviation(strState) != wxT(""))
		strStateName = StateNameByAbbreviation(strState);
	strStateName = FormatName(strStateName);
	
	for (i = mapCityToZip.begin(); i != mapCityToZip.end(); i++)
	{
		wxString strThisCity;
		wxString strThisState;
		double fSimilarity;
		
		strThisCity = (*i).first.BeforeFirst(wxT('|'));
		strThisState = (*i).first.AfterFirst(wxT('|'));
		
		if (strThisState == strStateName || strStateName == wxT(""))
		{
			fSimilarity = WordSimilarity(strThisCity, strCity);
			
			if (fSimilarity > fBestMatch)
			{
				fBestMatch = fSimilarity;
				strBestMatch = strThisCity;
			}
		}
	}
	return strBestMatch;
}

AutoComplete_City::AutoComplete_City()
{
}

void AutoComplete_City::SetState(const wxString & strState)
{
	m_strState = strState;
}

wxString AutoComplete_City::ClosestMatch(const wxString & strInput)
{
	return ClosestCityName(strInput, m_strState);
}

int AutoComplete_City::GetUpdateDelay()
{
	return 1;
}

