/*
 *  LibRoadnav
 *  Point.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// This is rather similar to wxRealPoint.
///
/// Stores a 2D point. Can also calculate the angle from the origin to 
/// the point, distance to origin, and other goodies.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/wx.h>

#include "Point.h"

#define POINT_FACTOR		1000000.0
#define ONE_OVER_POINT_FACTOR 0.000001

wxString Point::m_strPointFormat;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Default constructor
///
//////////////////////////////////////////////////////////////////////////////
Point::Point()
{
	m_lLong = 0;
	m_lLat = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Construct a point given a couple floating point coordinates.
///
//////////////////////////////////////////////////////////////////////////////
Point::Point(double fLong, double fLat)
{
	m_lLong = (wxInt32) (fLong * POINT_FACTOR);
	m_lLat  = (wxInt32) (fLat  * POINT_FACTOR);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Construct a point given a couple fixed point coordinates.
///
//////////////////////////////////////////////////////////////////////////////
Point::Point(wxInt32 lLong, wxInt32 lLat, int iIgnore)
{
	SetRaw(lLong, lLat);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - does nothing
///
//////////////////////////////////////////////////////////////////////////////
Point::~Point()
{
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Assign some coordinates to this point
///
//////////////////////////////////////////////////////////////////////////////
void Point::Set(double fLong, double fLat)
{
	SetLong(fLong);
	SetLat(fLat);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert this point to a string suitable for storage in the registry
///
//////////////////////////////////////////////////////////////////////////////
wxString Point::ToString()
{
	return wxString::Format(wxT("%f, %f"), GetLong(), GetLat());
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Initialize this object from the provided string generated by
/// Point::ToString()
///
//////////////////////////////////////////////////////////////////////////////
int Point::FromString(wxString strDesc)
{
	double fTmp;
	
	strDesc.BeforeFirst(',').ToDouble(&fTmp);
	m_lLong = (wxInt32) (fTmp * POINT_FACTOR);
	
	strDesc = strDesc.AfterFirst(',');
	strDesc.Trim(FALSE);
	strDesc.BeforeFirst(',').ToDouble(&fTmp);
	m_lLat = (wxInt32) (fTmp * POINT_FACTOR);
	
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Parse a string created using FormatPoint. Tries to be as forgiving
/// as possible of unusual formatting.
///
//////////////////////////////////////////////////////////////////////////////
void Point::ParsePoint(wxString strDesc)
{
	double fValue;
	double fScale;

	int iPos;
	bool bFirst;

#if defined _MSC_VER
	strDesc.Replace(wxT(""), wxT(""));
#endif
	strDesc.Replace(wxT("'"), wxT(""));
	strDesc.Replace(wxT("\""), wxT(""));
	strDesc = strDesc.Trim();
	strDesc = strDesc.MakeUpper();

	bFirst = true;

	while (strDesc != wxT(""))
	{
		bool bCoordinateDone;
		bool bIsLongitude;
		bool bIsLatitude;

		bIsLongitude = false;
		bIsLatitude = false;
		bCoordinateDone = false;

		fValue = 0;
		fScale = 1;

		while (!bCoordinateDone)
		{
			bool bMatched = false;

			for (iPos = 0; isdigit(strDesc.GetChar(iPos)) || strDesc.GetChar(iPos) == '.' || strDesc.GetChar(iPos) == '-'; iPos++);

			if (iPos)
			{
				fValue += atof(strDesc.Left(iPos).mb_str(*wxConvCurrent)) * fScale;
				fScale /= 60;

				strDesc = strDesc.Mid(iPos);

				strDesc = strDesc.Trim(FALSE);

				bMatched = true;
			}

			if (strDesc.Left(1) == wxT("S"))
				fValue *= -1;
			if (strDesc.Left(1) == wxT("W"))
				fValue *= -1;

			if (strDesc.Left(1) == wxT("N") || strDesc.Left(1) == wxT("S"))
			{
				bIsLatitude = true;
				bCoordinateDone = true;

				strDesc = strDesc.Mid(1);
				strDesc = strDesc.Trim(FALSE);

				bMatched = true;
			}
			if (strDesc.Left(1) == wxT("E") || strDesc.Left(1) == wxT("W"))
			{
				bIsLongitude = true;
				bCoordinateDone = true;

				strDesc = strDesc.Mid(1);
				strDesc = strDesc.Trim(FALSE);

				bMatched = true;
			}
			if (strDesc.Left(1) == wxT(","))
			{
				bCoordinateDone = true;

				strDesc = strDesc.Mid(1);
				strDesc = strDesc.Trim(FALSE);

				bMatched = true;
			}

			if (!bMatched)
				strDesc = strDesc.Mid(1);

			if (strDesc == wxT(""))
				bCoordinateDone = true;
		}

		if (bIsLongitude)
			SetLong(fValue);
		else if (bIsLatitude)
			SetLat(fValue);
		else if (!bFirst)
			SetLong(fValue);
		else
			SetLat(fValue);

		bFirst = false;
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point less than another point (for sorting purposes)?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator <(const Point & x) const
{
	if (m_lLong < x.m_lLong)
		return true;

	if (m_lLong > x.m_lLong)
		return false;

	return m_lLat < x.m_lLat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point greater than another point (for sorting purposes)?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator >(const Point & x) const
{
	if (m_lLong > x.m_lLong)
		return true;

	if (m_lLong < x.m_lLong)
		return false;

	return m_lLat > x.m_lLat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point less than or equal to another point (for sorting purposes)?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator <=(const Point & x) const
{
	if (m_lLong < x.m_lLong)
		return true;

	if (m_lLong > x.m_lLong)
		return false;

	return m_lLat <= x.m_lLat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point greater than or equal to another point (for sorting purposes)?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator >=(const Point & x) const
{
	if (m_lLong > x.m_lLong)
		return true;

	if (m_lLong < x.m_lLong)
		return false;

	return m_lLat >= x.m_lLat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point equal to another point?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator == (const Point & x) const
{
	return (m_lLong == x.m_lLong) && (m_lLat == x.m_lLat);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Is this point not equal to another point?
///
//////////////////////////////////////////////////////////////////////////////
bool Point::operator != (const Point & x) const
{
	return (m_lLong != x.m_lLong) || (m_lLat != x.m_lLat);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Add this point to another point
///
//////////////////////////////////////////////////////////////////////////////
Point Point::operator + (const Point & x) const
{
	wxASSERT(((double) x.m_lLong + (double) m_lLong) <= 0x7fffffff);
	wxASSERT(((double) x.m_lLong + (double) m_lLong) >= -0x7fffffff);
	wxASSERT(((double) x.m_lLat + (double) m_lLat) <= 0x7fffffff);
	wxASSERT(((double) x.m_lLat + (double) m_lLat) >= -0x7fffffff);

	return Point(x.m_lLong + m_lLong, x.m_lLat + m_lLat, 0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Subtract another point from this one
///
//////////////////////////////////////////////////////////////////////////////
Point Point::operator - (const Point & x) const
{
	wxASSERT(((double) m_lLong - (double) x.m_lLong) <= 0x7fffffff);
	wxASSERT(((double) m_lLong - (double) x.m_lLong) >= -0x7fffffff);
	wxASSERT(((double) m_lLat - (double) x.m_lLat) <= 0x7fffffff);
	wxASSERT(((double) m_lLat - (double) x.m_lLat) >= -0x7fffffff);

	return Point(m_lLong - x.m_lLong, m_lLat - x.m_lLat, 0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Divide this point by a scalar
///
//////////////////////////////////////////////////////////////////////////////
Point Point::operator / (double x) const
{
	wxASSERT(((double) m_lLong / x ) <= 0x7fffffff);
	wxASSERT(((double) m_lLong / x) >= -0x7fffffff);
	wxASSERT(((double) m_lLat / x ) <= 0x7fffffff);
	wxASSERT(((double) m_lLat / x) >= -0x7fffffff);

	return Point((wxInt32) (m_lLong / x), (wxInt32) (m_lLat / x), 0);
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Multiply this point by a scalar
///
//////////////////////////////////////////////////////////////////////////////
Point Point::operator * (double x) const
{
	wxASSERT(((double) m_lLong * x ) <= 0x7fffffff);
	wxASSERT(((double) m_lLong * x) >= -0x7fffffff);
	wxASSERT(((double) m_lLat * x ) <= 0x7fffffff);
	wxASSERT(((double) m_lLat * x) >= -0x7fffffff);

	return Point((wxInt32) (m_lLong * x), (wxInt32) (m_lLat * x), 0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Distance from the origin to this point. Assumes cartesian projection.
///
//////////////////////////////////////////////////////////////////////////////
double Point::abs()
{
	return sqrt(GetLong() * GetLong() + GetLat() * GetLat());
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Angle from origin to this point, assuming cartesian projection.
///
//////////////////////////////////////////////////////////////////////////////
double Point::angle()
{
	return atan2((double) m_lLat, (double) m_lLong);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load/save this object from/to the supplied file handle
///
//////////////////////////////////////////////////////////////////////////////
void Point::Serialize(FILE * hFile, bool bSave, void * pvArg)
{
	SerializeItem(hFile, bSave, &m_lLong);
	SerializeItem(hFile, bSave, &m_lLat);
	return;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Format a point into a nice string
///
//////////////////////////////////////////////////////////////////////////////
wxString Point::FormatPoint()
{
	wxString strSel = m_strPointFormat;

	char hN = (m_lLat  > 0 ? 'N' : 'S');
	char hE = (m_lLong > 0 ? 'E' : 'W');

	double fLat  = fabs(GetLat());
	double fLong = fabs(GetLong());

	if ( strSel == POSITION_FORMAT_1 )
	{
		// ddd.dddddd h
#if defined _MSC_VER
		return wxString::Format(wxT("%.6f %c, %.6f %c"),
#else
		return wxString::Format(wxT("%.6f %c, %.6f %c"),
#endif
																fLat, 
																hN, 
																fLong, 
																hE
																);
   }
   else if ( strSel == POSITION_FORMAT_2 )
   {
      // ddd mm.mmm' h
      int dirN = (int) fLat;
      float minN = (fLat-dirN)*60;

      int dirE = (int) fLong;
      float minE = (fLong-dirE)*60;

#if defined _MSC_VER
      return wxString::Format(wxT("%d %06.3f' %c, %d %06.3f' %c"),
#else
      return wxString::Format(wxT("%d %06.3f' %c, %d %06.3f' %c"),
#endif
																dirN,
																minN,
																hN,
																dirE,
																minE,
																hE
																);
   }
   else
   {
      // ddd mm' ss.s" h
      int dirN = (int) fLat;
      int minN = (int) ((fLat-dirN)*60);
      float secN = ((fLat-dirN)*60 - minN)*60;

      int dirE = (int) fLong;
      int minE = (int) ((fLong-dirE)*60);
      float secE = ((fLong-dirE)*60 - minE)*60;

#if defined _MSC_VER
      return wxString::Format(wxT("%d %02d' %04.1f\" %c, %d %02d' %04.1f\" %c"),
#else
      return wxString::Format(wxT("%d %02d' %04.1f\" %c, %d %02d' %04.1f\" %c"),
#endif
																				dirN,
																				minN,
																				secN,
																				hN,
																				dirE,
																				minE,
																				secE,
																				hE
																				);
   }
}

void Point::SetPointFormat(wxString strFormat)
{
	m_strPointFormat = strFormat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if this object contains a valid set of longitude and
/// latitude coordinates.
///
//////////////////////////////////////////////////////////////////////////////
bool Point::IsValid() const
{
	return GetLong() >= -180 && GetLong() <= 180 && GetLat() >= -90 && GetLat() <= 90;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the coordinates for this object such that IsValid() returns
/// false.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetInvalid()
{
	Set(181, 181);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the longitude of this point in floating point form.
///
//////////////////////////////////////////////////////////////////////////////
double Point::GetLong() const
{
	return (double)m_lLong * ONE_OVER_POINT_FACTOR;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the latitude of this point in floating point form.
///
//////////////////////////////////////////////////////////////////////////////
double Point::GetLat() const
{
	return (double)m_lLat * ONE_OVER_POINT_FACTOR;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the longitude of this point in floating point form.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetLong(double fLong)
{
	wxASSERT((fLong * POINT_FACTOR) <= 0x7fffffff);
	wxASSERT((fLong * POINT_FACTOR) >= -0x7fffffff);
	
	m_lLong = (wxInt32) (fLong * POINT_FACTOR);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the latitude of this point in floating point form.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetLat(double fLat)
{
	wxASSERT((fLat * POINT_FACTOR) <= 0x7fffffff);
	wxASSERT((fLat * POINT_FACTOR) >= -0x7fffffff);

	m_lLat = (wxInt32) (fLat * POINT_FACTOR);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the longitude and latitude of this point in fixed point form.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetRaw(wxInt32 lLong, wxInt32 lLat)
{
	SetLongRaw(lLong);
	SetLatRaw(lLat);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the longitude of this point in fixed point form.
///
//////////////////////////////////////////////////////////////////////////////
wxInt32 Point::GetLongRaw() const
{
	return m_lLong;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Retrieves the latitude of this point in fixed point form.
///
//////////////////////////////////////////////////////////////////////////////
wxInt32 Point::GetLatRaw() const
{
	return m_lLat;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the longitude of this point in fixed point form.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetLongRaw(wxInt32 lLong)
{
	m_lLong = lLong;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Sets the latitude of this point in fixed point form.
///
//////////////////////////////////////////////////////////////////////////////
void Point::SetLatRaw(wxInt32 lLat)
{
	m_lLat = lLat;
}
