/*
 *  LibRoadnav
 *  StringTable.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains StringTable class
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>

#include "StringTable.h"

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - initialize object and load table into memory.
///
//////////////////////////////////////////////////////////////////////////////
StringTable::StringTable(wxString strFilename)
{
	m_strFilename = strFilename;
	Init();
	Load();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - call Clear()
///
//////////////////////////////////////////////////////////////////////////////
StringTable::~StringTable()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Initialize all member variables.
///
//////////////////////////////////////////////////////////////////////////////
void StringTable::Init()
{
	m_nGrowBy = 16;
	
	m_ppszTable = new char * [m_nGrowBy];
	m_puiSortKey = new wxUint32 [m_nGrowBy];
	m_nTableMemorySize = m_nGrowBy;
	m_nTableEntriesInUse = 0;
	m_bModified = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Frees all memory allocated
///
//////////////////////////////////////////////////////////////////////////////
void StringTable::Clear()
{
	wxUint32 iEntry;

	for (iEntry = 0; iEntry < m_nTableEntriesInUse; iEntry++)
		delete m_ppszTable[iEntry];
	delete m_ppszTable;
	delete m_puiSortKey;
	
	m_nTableEntriesInUse = 0;
	m_nTableMemorySize = 0;
	m_ppszTable = NULL;
	m_puiSortKey = NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Loads the table from m_strFilename
///
//////////////////////////////////////////////////////////////////////////////
void StringTable::Load()
{
	FILE * hFile;
	wxUint32 iEntry;
	
	hFile = fopen(m_strFilename.mb_str(*wxConvCurrent), "rb");
	
	if (!hFile)
		return;
		
	fread(&m_nTableEntriesInUse, sizeof(m_nTableEntriesInUse), 1, hFile);
	m_nTableMemorySize = m_nTableEntriesInUse;
	m_ppszTable = new char * [m_nTableMemorySize];
	m_puiSortKey = new wxUint32 [m_nTableMemorySize];

	fread(m_puiSortKey, sizeof(wxUint32), m_nTableEntriesInUse, hFile);
	
	for (iEntry = 0; iEntry < m_nTableEntriesInUse; iEntry++)
	{
		wxUint16 uiLength;
		
		fread(&uiLength, sizeof(uiLength), 1, hFile);

		m_ppszTable[iEntry] = new char [uiLength];
		fread(m_ppszTable[iEntry], sizeof(char), uiLength, hFile);
	}
	
	fclose(hFile);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Saves the table to m_strFilename
///
//////////////////////////////////////////////////////////////////////////////
void StringTable::Save()
{
	FILE * hFile;
	wxUint32 iEntry;
	
	if (!m_bModified)
		return;
	
	hFile = fopen(m_strFilename.mb_str(*wxConvCurrent), "wb");
	
	fwrite(&m_nTableEntriesInUse, sizeof(m_nTableEntriesInUse), 1, hFile);

	fwrite(m_puiSortKey, sizeof(wxUint32), m_nTableEntriesInUse, hFile);
	
	for (iEntry = 0; iEntry < m_nTableEntriesInUse; iEntry++)
	{
		wxUint16 uiLength;
		
		uiLength = strlen(m_ppszTable[iEntry]) + 1;
		fwrite(&uiLength, sizeof(uiLength), 1, hFile);
		fwrite(m_ppszTable[iEntry], sizeof(char), uiLength, hFile);
	}
	
	fclose(hFile);
	
	m_bModified = false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the index of the table entry with string pszName.
///
/// If it does not exist and bCreate is true, it will be created. If
/// it does not exist and bCreate is false, INVALID_STRING_ENTRY
/// will be returned.
///
//////////////////////////////////////////////////////////////////////////////
wxUint32 StringTable::GetEntryIndex(const char * pszName, bool bCreate)
{
	wxUint32 uiLower;
	wxUint32 uiUpper;
	wxUint32 uiCmpIndex = 0;

	if (m_nTableEntriesInUse)
	{
		uiLower = 0;
		uiUpper = m_nTableEntriesInUse - 1;

		while (1)
		{
			int iRet;
			
			uiCmpIndex = (uiLower + uiUpper) / 2;
			
			iRet = strcmp(pszName, m_ppszTable[m_puiSortKey[uiCmpIndex]]);
			
			if (iRet == 0)
			{
				return m_puiSortKey[uiCmpIndex];
			}
			else if (iRet < 0)
			{
				if (uiLower == uiCmpIndex)
					break;
					
				uiUpper = uiCmpIndex - 1;
			}
			else
			{
				if (uiUpper == uiCmpIndex)
				{
					uiCmpIndex++;
					break;
				}
					
				uiLower = uiCmpIndex + 1;
			}
		}
	}
	
	if (!bCreate)
		return INVALID_STRING_ENTRY;
		
	if (m_nTableEntriesInUse >= m_nTableMemorySize)
	{
		// grow the table
		char ** ppszOldTable;
		wxUint32 * piOldSortKey;
		
		ppszOldTable = m_ppszTable;
		piOldSortKey = m_puiSortKey;
		
		m_nTableMemorySize += m_nGrowBy;		
		
		m_ppszTable = new char * [m_nTableMemorySize];
		m_puiSortKey = new wxUint32 [m_nTableMemorySize];
		
		if (m_nTableEntriesInUse)
		{
			memcpy(m_ppszTable, ppszOldTable, sizeof(char *) * m_nTableEntriesInUse);
			memcpy(m_puiSortKey, piOldSortKey, sizeof(wxUint32) * m_nTableEntriesInUse);
		}
		
		delete ppszOldTable;
		delete piOldSortKey;
	}
	
	m_ppszTable[m_nTableEntriesInUse] = new char [strlen(pszName) + 1];
	strcpy(m_ppszTable[m_nTableEntriesInUse], pszName);
	
	if (uiCmpIndex != m_nTableEntriesInUse)
		memmove(&m_puiSortKey[uiCmpIndex + 1], &m_puiSortKey[uiCmpIndex], sizeof(wxUint32) * (m_nTableEntriesInUse - uiCmpIndex));
	m_puiSortKey[uiCmpIndex] = m_nTableEntriesInUse;
	
	m_nTableEntriesInUse++;
	m_bModified = true;
	
	return m_nTableEntriesInUse - 1;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a pointer to the entry with the supplied index.
///
//////////////////////////////////////////////////////////////////////////////
char * StringTable::GetEntryPointer(wxUint32 uiEntry)
{
	wxASSERT(uiEntry < m_nTableEntriesInUse);
	return m_ppszTable[uiEntry];
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns a pointer to the entry with the supplied name.
///
/// If it does not exist and bCreate is true, it will be created. If
/// it does not exist and bCreate is false, NULL
/// will be returned.
///
//////////////////////////////////////////////////////////////////////////////
char * StringTable::GetEntryPointer(const char * pszName, bool bCreate)
{
	wxUint32 uiEntry;
	
	uiEntry = GetEntryIndex(pszName, bCreate);
	
	if (uiEntry == INVALID_STRING_ENTRY)
		return NULL;
	
	return GetEntryPointer(uiEntry);
}
