/*
 *  LibRoadnav
 *  UTM.cpp
 *
 *  Copyright (c) 2004 - 2008 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// UTM conversion related functions
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#endif

#include <wx/wx.h>

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "Constants.h"
#include "MapSupport.h"
#include "UnitConversion.h"

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert from long/lat to NAD83 UTM coordinates
///
//////////////////////////////////////////////////////////////////////////////
void LongLatToUTM(double fLong, double fLat, double * pfEasting, double * pfNorthing, int * piZone)
{
	double fA; // equatorial radius in meters
	double fB; // polar radius in meters

	int iZone;
	double fLong0; // central meridian of zone (always positive)
	double fK0; // scale along long0
	double fE; // eccentricity of the earth's elliptical cross-section.
	double fE2;
	double fN;
	double fRho; // radius of curvature of the earth in the meridian plane.
	double fNu;
	double fP;
	double fPsec;

	double fS;
	double fA0;
	double fB0;
	double fC0;
	double fD0;
	double fE0;

	double fSin1;

	double fK1;
	double fK2;
	double fK3;
	double fK4;
	double fK5;

	// NAD83/WGS84
	fA = 6378137;
	fB = 6356752.3142;
	// end GRS 80

	if (fLong > 0)
		iZone = (int) (fLong / 6 + 1);
	else
		iZone = (int) (fLong / 6 + 31);

	fLong0 = iZone * 6 - 183;

	fP = fLong - fLong0;

	// make longitude and latitude positive
	fLong = fabs(fLong);
	fLat = fabs(fLat);

	fK0 = 0.9996;

	fE = sqrt(1 - (fB / fA) * (fB / fA));
	fE2 = fE * fE / (1 - fE * fE);

	fN = (fA - fB) / (fA + fB);

	fRho = fA * (1 - fE * fE) / pow((1 - fE * fE * sin_deg(fLat) * sin_deg(fLat)), 1.5);

	fNu = fA / sqrt((1 - fE * fE * sin_deg(fLat) * sin_deg(fLat)));

	fPsec = fP * 3600 / 10000;

	// meridional arc length
	fA0 = fA * (1 - fN + (5 * fN * fN / 4) * (1 - fN) + (81 * pow(fN, 4) / 64) * (1 - fN));
	fB0 = (3 * fA * fN / 2) * (1 - fN - (7 * fN * fN / 8) * (1 - fN) + 55 * fN * pow(fN, 4) / 64);
	fC0 = (15 * fA * fN * fN / 16) * (1 - fN + (3 * fN * fN / 4) * (1 - fN));
	fD0 = (35 * fA * pow(fN, 3) / 48) * (1 - fN + 11 * fN * fN / 16);
	fE0 = (315 * fA * pow(fN, 4) / 51) * (1 - fN);
	fS = fA0 * DegreesToRadians(fLat) - fB0 * sin_deg(2 * fLat) + fC0 * sin_deg(4 * fLat) + fD0 * sin_deg(6 * fLat) + fE0 * sin_deg(8 * fLat);

	// UTM constants
	fSin1 = PI / 180 / 3600;

	// UTM coefficients
	fK1 = fS * fK0;
	fK2 = fNu * sin_deg(fLat) * cos_deg(fLat) * fSin1 * fSin1 * fK0 * 100000000.0 / 2;
	fK3 = ((pow(fSin1, 4) * fNu * sin_deg(fLat) * pow(cos_deg(fLat), 3) / 24) * (5 - pow(tan_deg(fLat), 2) + 9 * fE2 * pow(cos_deg(fLat), 2) + 4 * fE2 * fE2 * pow(cos_deg(fLat), 4))) * fK0 * 10000000000000000.0;
	fK4 = fNu * cos_deg(fLat) * fSin1 * fK0 * 10000;
	fK5 = pow(fSin1 * cos_deg(fLat), 3) * (fNu / 6) * (1 - pow(tan_deg(fLat), 2) + fE2 * pow(cos_deg(fLat), 2)) * fK0 * 1000000000000.0;

	*pfNorthing = fK1 + fK2 * pow(fPsec, 2) + fK3 * pow(fPsec, 4);
	*pfEasting = 500000 + fK4 * fPsec + fK5 * pow(fPsec, 3);
	*piZone = iZone;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Convert from NAD83 UTM to long/lat
///
//////////////////////////////////////////////////////////////////////////////
void UTMToLongLat(double fEasting, double fNorthing, int iZone, double * pfLong, double * pfLat)
{
	double fA; // equatorial radius in meters
	double fB; // polar radius in meters

	double fE; // eccentricity of the earth's elliptical cross-section.
	double fE2;

	double fK0; // scale along long0

	double fArc; // arc length
	double fMu;
	double fE1;
	double fC1;
	double fC2;
	double fC3;
	double fC4;
	double fPhi1;

	double fSin1;
	double fC12;
	double fT1;
	double fN1;
	double fR1;
	double fH2;
	double fD;

	double fLatFact1;
	double fLatFact2;
	double fLatFact3;
	double fLatFact4;

	double fLongFact1;
	double fLongFact2;
	double fLongFact3;
	
	double fLong0;

	// NAD83/WGS84
	fA = 6378137;
	fB = 6356752.3142;
	// end GRS 80

	fE = sqrt(1 - (fB / fA) * (fB / fA));
	fE2 = fE * fE / (1 - fE * fE);
	
	fK0 = 0.9996;

	// footprint latitude
	fArc = fNorthing / fK0;
	fMu = fArc / (fA * (1 - pow(fE, 2) / 4 - 3 * pow(fE, 4) / 64 - 5 * pow(fE, 6) / 256));
	fE1 = (1 - sqrt(1 - fE * fE)) / (1 + sqrt(1 - fE * fE));
	fC1 = 3 * fE1 / 2 - 27 * pow(fE1, 3) / 32;
	fC2 = 21 * pow(fE1, 2) / 16 - 55 * pow(fE1, 4) / 32;
	fC3 = 151 * pow(fE1, 3) / 96;
	fC4 = 1097 * pow(fE1, 4) / 512;
	fPhi1 = fMu + fC1 * sin(2 * fMu) + fC2 * sin(4 * fMu) + fC3 * sin(6 * fMu) + fC4 * sin(8 * fMu);

	// constants
	fSin1 = PI / 180 / 3600;
	fC12 = fE2 * pow(cos(fPhi1), 2);
	fT1 = pow(tan(fPhi1), 2);
	fN1 = fA / sqrt(1 - pow(fE * sin(fPhi1), 2));
	fR1 = fA * (1 - fE * fE) / pow(1 - pow(fE * sin(fPhi1), 2), 1.5);
	fH2 = 500000 - fEasting;
	fD = fH2 / (fN1 * fK0);

	// latitude coefficients
	fLatFact1 = fN1 * tan(fPhi1) / fR1;
	fLatFact2 = fD * fD / 2;
	fLatFact3 = (5 + 3 * fT1 + 10 * fC12 - 4 * fC12 * fC12 - 9 * fE2) * pow(fD, 4) / 24;
	fLatFact4 = (61 + 90 * fT1 + 298 * fC12 + 45 * fT1 * fT1 - 252 * fE2 - 3 * fC12 * fC12) * pow(fD, 6) / 720;

	// longitude coefficients
	fLongFact1 = fD;
	fLongFact2 = (1 + 2 * fT1 + fC12) * pow(fD, 3) / 6;
	fLongFact3 = (5 - 2 * fC12 + 28 * fT1 - 3 * pow(fC12, 2) + 8 * fE2 + 24 * pow(fT1, 2)) * pow(fD, 5) / 120;

	fLong0 = 6 * iZone - 183;

	*pfLat = 180 * (fPhi1 - fLatFact1 * (fLatFact2 + fLatFact3 +fLatFact4)) / PI;
	*pfLong = fLong0 - (fLongFact1 - fLongFact2 + fLongFact3) / cos(fPhi1) * 180 / PI;
}
